/*******************************************************************************

              VV   VV    ii                    33333  DDDDDD
              VV   VV                         33   33  DD  DD
              VV   VV   iii    eeeee  ww   ww      33  DD  DD
              VV   VV    ii   ee   ee ww   ww   3333   DD  DD
              VV   VV    ii   eeeeeee ww w ww      33  DD  DD
               VV VV     ii   ee      wwwwwww 33   33  DD  DD
                VVV     iiii   eeeee  ww   ww  33333  DDDDDD
              -----------------------------------------------
                          3D Model Viewer Using OpenGL
                                 Developed by
                           dudley.benton@gmail.com

--- Description ----------------------------------------------------------------

  This program reads a 3D Studio, VMRL, or 3DV file identifies the textures,
  loads the texture, and displays the model using OpenGL. The elements and
  textures may be separate or packaged together in a ZIP file.

  Upon opening a ZIP archive, extract the files to the temporary folder and
  keep a list so that these can be deleted upon normal exit.

  On the first time rendering a model with textures, average the RGB over the
  element and then set the element color to this average value. Should the
  model then be saved to a non-textured format, the colors will be roughly
  preserved.

--- Revisions ------------------------------------------------------------------

  V1.00 (07/11/2016): initial release adapted from Knight's Tour in 3D
  V1.10 (02/21/2018): add JPEG and GIF textures
  V1.20 (06/15/2020): add WRL and 3DV file formats plus ZIP

*******************************************************************************/

#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <io.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <time.h>
#include <sys/utime.h>
#include <float.h>
#define _USE_MATH_DEFINES
#include <math.h>
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <commdlg.h>
#include <commctrl.h>
#include "..\\OpenGL\\gl.h"
#include "..\\utilities\\keys.h"
#include "..\\utilities\\colors.h"
#include "..\\utilities\\bitmap.h"
#include "..\\utilities\\gif89a.h"
#include "..\\utilities\\jpeg6b.h"
#include "..\\utilities\\ziplib.h"
#include "..\\utilities\\model.h"
#include "view3d.h"

WINGDIAPI void APIENTRY glGenTextures(GLsizei n,GLuint*textures);
WINGDIAPI void APIENTRY glBindTexture(GLenum target,GLuint texture);
MODEL Read3DS(char*filename);
MODEL Read3DV(char*filename);
MODEL ReadWRL(char*filename);
void Write3DS(MODEL*model,char*filename);
void Write3DV(MODEL*model,char*filename);
void WriteWRL(MODEL*model,char*filename);

/*******************************************************************************
  GLOBAL OBJECTS
*******************************************************************************/

HINSTANCE hInst; /* instance */
HWND hMain; /* application */
HWND hExit;
HWND hHelp;
HWND hZall;
HWND hPlot;
HWND hLogo;
HWND hMove;
HWND hRock;
HWND hStat;
HWND hTurn;
HWND hText;
HWND hDlay;
HMENU hMenu;

BITMAPINFOHEADER*bLogo;

HDC   pDC;   /* plot window device context */
HGLRC rDC;   /* OpenGL rendering context */
HPEN  pGray; /* gray pen */

MODEL model;

BITMAPINFOHEADER**bitmaps;
DWORD gu_color=-1;
DWORD*texture;
DWORD textures;
int delay_count;
BOOL delayed_painting=TRUE;
BOOL enable_lighting=TRUE;
BOOL fill_polygons=TRUE;
BOOL list_textures=FALSE;
BOOL mouse_clicked=FALSE;
BOOL pending_repaint=FALSE;
BOOL perspective=TRUE;

char file_name[FILENAME_MAX];

struct{float a,b,c;}Camera={3,5,100};
struct{int a,b;}Light={0,0};
struct{int a,b,c;float s,x,y,z;}View;
struct{int down,move,x1,x2,y1,y2;}Mouse;

/*******************************************************************************
  GENERIC PROCEDURES COMMON TO ANY WINDOWS PROGRAM
*******************************************************************************/

void WindowText(HWND hWnd,char*format,...)             /* works like printf() */
  {
  char bufr[1024];
  va_list arg_marker;
  va_start(arg_marker,format);
  if(vsprintf(bufr,format,arg_marker)>=sizeof(bufr))
    FatalAppExit(0,"WindowText(): buffer overflow");
  SetWindowText(hWnd,bufr);
  }

RECT*Rect(int left,int top,int right,int bottom)       /* fill RECT structure */
  {
  static RECT R;
  R.left=left;
  R.right=right;
  R.top=top;
  R.bottom=bottom;
  return(&R);
  }

SIZE GetTextSize(HDC hDC,char*text)   /* sT must be static due to Windows bug */
  {
  static SIZE sT;
  GetTextExtentPoint32(hDC,text,strlen(text),&sT);
  return(sT);
  }

void CenterWindowOnScreen(HWND hWnd)
  {
  int high,left,top,wide;
  RECT rw;
  GetWindowRect(hWnd,&rw);
  wide=rw.right-rw.left;
  high=rw.bottom-rw.top;
  left=(GetSystemMetrics(SM_CXSCREEN)-wide)/2;
  top=(GetSystemMetrics(SM_CYSCREEN)-high)/2;
  MoveWindow(hWnd,left,top,wide,high,TRUE);
  }

void PaintBitmap(HDC hDC,RECT*rc,BITMAPINFOHEADER*bHead,int stretch)
  {
  int colors,h,H,l,t,w,W;
  BYTE*bImage;
  l=rc->left;
  t=rc->top;
  w=rc->right-rc->left;
  h=rc->bottom-rc->top;
  W=bHead->biWidth;
  H=bHead->biHeight;
  if(w<8||h<8||W<8||H<8)
    return;
  if(bHead->biBitCount==24)
    colors=0;
  else
    {
    colors=bHead->biClrUsed;
    if(colors==0)
      colors=1<<bHead->biBitCount;
    }
  bImage=((BYTE*)bHead)+sizeof(BITMAPINFOHEADER)+colors*sizeof(DWORD);
  if(stretch<0)
    SetDIBitsToDevice(hDC,l,t,W,H,0,0,0,H,bImage,(BITMAPINFO*)bHead,DIB_RGB_COLORS);
  else if(stretch>0)
    StretchDIBits(hDC,l,t,w,h,0,0,W,H,bImage,(BITMAPINFO*)bHead,DIB_RGB_COLORS,SRCCOPY);
  else if(w*abs(H)<abs(W)*h)
    StretchDIBits(hDC,l,t+(h-w*abs(H)/abs(W))/2,w,w*abs(H)/abs(W),0,0,W,H,bImage,(BITMAPINFO*)bHead,DIB_RGB_COLORS,SRCCOPY);
  else
    StretchDIBits(hDC,l+(w-h*abs(W)/abs(H))/2,t,h*abs(W)/abs(H),h,0,0,W,H,bImage,(BITMAPINFO*)bHead,DIB_RGB_COLORS,SRCCOPY);
  }

int IsPowerOfTwo(int l)
  {
  int i;
  for(i=2;i<=1024;i*=2)
    if(l==i)
      return(1);
  return(0);
  }

/* OpenGL requires the height and width to be a power of two not exceeding 1024.
   The image must also be 24-bits/pixel. This will convert it if necessary. */

BITMAPINFOHEADER*BMPfixit(BITMAPINFOHEADER*bm1)
  {
  int height,wide,width;
  BITMAPINFO bi2;
  BITMAPINFOHEADER*bm2;
  BYTE*bits;
  HBITMAP bh1,bh2;
  HDC hDC;
  RECT rc;
  if(bm1->biBitCount==24)
    if(bm1->biWidth>=2)
      if(bm1->biWidth<=1024)
        if(IsPowerOfTwo(bm1->biWidth))
          if(bm1->biHeight>=2)
            if(bm1->biHeight<=1024)
              if(IsPowerOfTwo(bm1->biHeight))
                return(bm1);
  for(width=2;width!=1024;width*=2)
    if(width>=bm1->biWidth)
      break;
  for(height=2;height!=1024;height*=2)
    if(height>=bm1->biHeight)
      break;
  wide=4*((width*24+31)/32);
  memset(&bi2,0,sizeof(bi2));
  bi2.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
  bi2.bmiHeader.biPlanes=1;
  bi2.bmiHeader.biBitCount=24;
  bi2.bmiHeader.biWidth=width;
  bi2.bmiHeader.biHeight=height;
  bi2.bmiHeader.biSizeImage=wide*height;
  hDC=CreateCompatibleDC(NULL);
  bh2=CreateDIBSection(hDC,&bi2,DIB_RGB_COLORS,&bits,NULL,0);
  bh1=SelectObject(hDC,bh2);
  rc=*Rect(0,0,width,height);
  PaintBitmap(hDC,&rc,bm1,TRUE);
  SelectObject(hDC,bh1);
  DeleteDC(hDC);
  bm2=calloc(sizeof(BITMAPINFOHEADER)+wide*height,1);
  memcpy(bm2,&bi2.bmiHeader,sizeof(BITMAPINFOHEADER));
  memcpy(((BYTE*)bm2)+sizeof(BITMAPINFOHEADER),bits,wide*height);
  DeleteObject(bh2);
  return(bm2);
  }

int GetInputFileName(char*fname,char*filter,char*title)
  {
  OPENFILENAME OpenFileName;
  memset(&OpenFileName,0,sizeof(OPENFILENAME));
  OpenFileName.lStructSize    =sizeof(OPENFILENAME);
  OpenFileName.hwndOwner      =hMain;
  OpenFileName.hInstance      =hInst;
  OpenFileName.lpstrFilter    =filter;
  OpenFileName.nFilterIndex   =1;
  OpenFileName.lpstrFile      =fname;
  OpenFileName.nMaxFile       =FILENAME_MAX;
  OpenFileName.Flags          =OFN_PATHMUSTEXIST;
  OpenFileName.lpstrTitle     =title;
  if(!GetOpenFileName(&OpenFileName))
    fname[0]=0;
  return(fname[0]!=0);
  }

int GetOutputFileName(char*fname,char*filter,char*title)
  {
  OPENFILENAME SaveFileName;
  memset(&SaveFileName,0,sizeof(OPENFILENAME));
  SaveFileName.lStructSize =sizeof(OPENFILENAME);
  SaveFileName.hwndOwner   =hMain;
  SaveFileName.hInstance   =hInst;
  SaveFileName.lpstrFilter =filter;
  SaveFileName.nFilterIndex=1;
  SaveFileName.lpstrFile   =fname;
  SaveFileName.nMaxFile    =FILENAME_MAX;
  SaveFileName.Flags       =OFN_PATHMUSTEXIST|OFN_OVERWRITEPROMPT;
  SaveFileName.lpstrTitle  =title;
  if(!GetSaveFileName(&SaveFileName))
    fname[0]=0;
  return(fname[0]!=0);
  }

char*BuildFilter(char*title,char**types)
  {
  char*filter;
  int i,j,k,l;
  for(k=0;k<2;k++)
    {
    l=(int)strlen(title);
    if(k)
      memcpy(filter,title,l+1);
    i=l+1;
    for(j=0;types[j];j++)
      {
      if(j)
        {
        if(k)
          filter[i]=';';
        i++;
        }
      for(l=0;types[j][l];l++)
        if(types[j][l]==' ')
          break;
      if(k)
        memcpy(filter+i,types[j],l);
      i+=l;
      }
    if(k)
      filter[i]=0;
    i++;
    for(j=0;types[j];j++)
      {
      l=(int)strlen(types[j]);
      if(k)
        memcpy(filter+i,types[j],l+1);
      i+=l+1;
      for(l=0;types[j][l];l++)
        if(types[j][l]==' ')
          break;
      if(k)
        memcpy(filter+i,types[j],l);
      i+=l;
      if(k)
        filter[i]=0;
      i++;
      }
    if(k)
      filter[i]=0;
    else
      filter=calloc(i+1,sizeof(char));
    }
  return(filter);
  }

/*******************************************************************************
  AUXILIARY FUNCTIONS THAT SUPPLEMENT OPENGL
*******************************************************************************/

float*fltColor(DWORD c)
  {
  static float f[4];
  union{BYTE b[4];DWORD d;}u;
  u.d=c;
  f[0]=(float)(u.b[2]/255.);
  f[1]=(float)(u.b[1]/255.);
  f[2]=(float)(u.b[0]/255.);
  return(f);
  }

void glColor(DWORD color)
  {
  union{DWORD d;BYTE b[4];}u;
  if(color!=gu_color)
    {
    gu_color=color;
    u.d=color;
    glColor3ub(u.b[2],u.b[1],u.b[0]);
    }
  }

void CrossProduct(float*P,float*Q,float*R)
  {
  R[0]=P[1]*Q[2]-P[2]*Q[1];
  R[1]=P[2]*Q[0]-P[0]*Q[2];
  R[2]=P[0]*Q[1]-P[1]*Q[0];
  }

#define GL_BGR_EXT 0x80E0

DWORD InitializeTexture(BITMAPINFOHEADER*bm)
  {
  BYTE*bitmap;
  int i;
  DWORD t;
  if(bm->biBitCount!=24)
    {
    WindowText(hStat,"texture depth must be 24 bits/pixel (%li)",bm->biBitCount);
    return(0);
    }
  if(bm->biWidth>1024)
    {
    WindowText(hStat,"texture width must not exceed 1024 (%li)",bm->biWidth);
    return(0);
    }
  for(i=2;i<=1024;i*=2)
    if(i==bm->biWidth)
      break;
  if(i>1024)
    {
    WindowText(hStat,"texture width must be 2^n (%li)",bm->biWidth);
    return(0);
    }
  if(bm->biHeight>1024)
    {
    WindowText(hStat,"texture height must not exceed 1024 (%li)",bm->biHeight);
    return(0);
    }
  for(i=2;i<=1024;i*=2)
    if(i==bm->biHeight)
      break;
  if(i>1024)
    {
    WindowText(hStat,"texture height must be 2^n (%li)",bm->biHeight);
    return(0);
    }
  bitmap=((BYTE*)bm)+sizeof(BITMAPINFOHEADER);
  t=0;
  glGenTextures(1,&t);
  if(t==0)
    {
    WindowText(hStat,"can't generate texture: OpenGL error code=0x%lX",glGetError());
    return(0);
    }
  glBindTexture(GL_TEXTURE_2D,t);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
  glTexImage2D(GL_TEXTURE_2D,0,3,bm->biWidth,bm->biHeight,0,GL_BGR_EXT,GL_UNSIGNED_BYTE,bitmap);
  glDisable(GL_TEXTURE_2D);
  return(t);
  }

/*******************************************************************************
  FUNCTIONS UNIQUE TO LOADING AND RENDERING 3D STUDIO MODELS
*******************************************************************************/

BOOL exists(char*fname)
  {
  if(_access(fname,0))
    return FALSE;
  return TRUE;
  }

char*concat(char*str1,char*str2)
  {
  static char bufr[1024];
  strcpy(bufr,str1);
  strcat(bufr,str2);
  return(bufr);
  }

MODEL Load3DS(char*model_name)
  {
  char*ptr,texture_name[FILENAME_MAX];
  int i;
  BITMAPINFOHEADER*bm;
  FILE*fp;
  model=Read3DS(model_name);
  if(model.nodes<=0)
    return(model);
  textures=model.matls;
  if(list_textures)
    {
    GetModuleFileName(hInst,texture_name,sizeof(texture_name));
    strcpy(strrchr(texture_name,'.'),".lst");
    fp=fopen(texture_name,"wt");
    fprintf(fp,"%i textures\n",textures);
    }
  if(textures)
    {
    texture=calloc(textures,sizeof(DWORD));
    bitmaps=calloc(textures,sizeof(BITMAPINFOHEADER*));
    for(i=0;i<model.matls;i++)
      {
      strcpy(texture_name,model_name);
      if((ptr=strrchr(texture_name,'\\'))!=NULL)
        ptr[1]=0;
      else
        strcpy(texture_name,"");
      strcat(texture_name,model.matl[i].name);
      if(!exists(texture_name))
        {
        if(exists(concat(texture_name,".bmp")))
          strcat(texture_name,".bmp");
        else if(exists(concat(texture_name,".gif")))
          strcat(texture_name,".gif");
        else if(exists(concat(texture_name,".jpg")))
          strcat(texture_name,".jpg");
        else if(exists(concat(texture_name,".jpeg")))
          strcat(texture_name,".jpeg");
        }
      if(list_textures)
        fprintf(fp,"%s\n",texture_name);
      if(stristr(texture_name,".bmp"))
        {
        bm=BMPread(texture_name);
        bm=BMPfixit(bm);
        texture[i]=InitializeTexture(bm);
        bitmaps[i]=bm;
        }
      else if(stristr(texture_name,".gif"))
        {
        bm=ReadGIF1(texture_name);
        bm=BMPfixit(bm);
        texture[i]=InitializeTexture(bm);
        bitmaps[i]=bm;
        }
      else if(stristr(texture_name,".jpg")||stristr(texture_name,".jpeg"))
        {
        bm=ReadJPEG(texture_name);
        bm=BMPfixit(bm);
        texture[i]=InitializeTexture(bm);
        bitmaps[i]=bm;
        }
      else
        {
        texture[i]=InitializeTexture(bLogo);
        bitmaps[i]=bLogo;
        }
      }
    }
  if(list_textures)
    fclose(fp);
  return(model);
  }

char temp_path[FILENAME_MAX];
char temp_name[FILENAME_MAX];
char temp_modl[FILENAME_MAX];

int FileCreate(char*fname,void*data,size_t size,time_t time)
  {
  FILE*fp;
  strcpy(temp_name,temp_path);
  strcat(temp_name,fname);
  if((fp=fopen(temp_name,"wb"))==NULL)
    return(-1);
  if(stristr(fname,".3ds"))
    strcpy(temp_modl,temp_name);
  else if(stristr(fname,".3dv"))
    strcpy(temp_modl,temp_name);
  else if(stristr(fname,".wrl"))
    strcpy(temp_modl,temp_name);
  if(fwrite(data,1,size,fp)!=size)
    {
    fclose(fp);
    return(-1);
    }
  fclose(fp);
  return(1);
  }

void ExtractFiles(char*archive)
  {
  int i,n;
  size_t uncomp_size;
  void*bufr;
  mz_zip_archive zip;
  mz_zip_archive_file_stat stat;
  if(_access(archive,0))
    return;
  memset(&zip,0,sizeof(zip));
  if(mz_zip_reader_init_file(&zip,archive,0)==0)
    return;
  n=(int)mz_zip_reader_get_num_files(&zip);
  for(i=0;i<n;i++)
    {
    if(mz_zip_reader_is_file_a_directory(&zip,i))
      continue;
    if(!mz_zip_reader_file_stat(&zip,i,&stat))
      {
      mz_zip_reader_end(&zip);
      goto error;
      }
    if((bufr=mz_zip_reader_extract_file_to_heap(&zip,stat.m_filename,&uncomp_size,0))==NULL)
      goto error;
    if(uncomp_size!=stat.m_uncomp_size)
      goto error;
    this_file.comp_size=(size_t)stat.m_comp_size;
    if(FileCreate(stat.m_filename,bufr,(size_t)stat.m_uncomp_size,stat.m_time)<0)
      goto error;
    free(bufr);
    }
error:
  mz_zip_reader_end(&zip);
  }

MODEL LoadZIP(char*zname)
  {
  GetTempPath(sizeof(temp_path),temp_path);
  memset(temp_modl,0,sizeof(temp_modl));
  ExtractFiles(zname);
  if(stristr(temp_modl,".3ds"))
    model=Load3DS(temp_modl);
  else if(stristr(temp_modl,".3dv"))
    model=Read3DV(temp_modl);
  else if(stristr(temp_modl,".wrl"))
    model=ReadWRL(temp_modl);
  return(model);
  }

MODEL ImportModel(char*fname)
  {
  strcpy(file_name,fname);
  if(stristr(fname,".3ds"))
    model=Load3DS(fname);
  else if(stristr(fname,".3dv"))
    model=Read3DV(fname);
  else if(stristr(fname,".wrl"))
    model=ReadWRL(fname);
  else if(stristr(fname,".zip"))
    model=LoadZIP(fname);
  else
    Abort(__LINE__,"unexpected import file type");
  return(model);
  }

void ExportModel(char*fname)
  {
  if(stristr(fname,".3ds"))
    Write3DS(&model,fname);
  else if(stristr(fname,".3dv"))
    Write3DV(&model,fname);
  else if(stristr(fname,".wrl"))
    WriteWRL(&model,fname);
  else
    Abort(__LINE__,"unexpected export file type");
  }

DWORD RGBaverage(int t,float u1,float v1,float u2,float v2,float u3,float v3)
  {
  BYTE b,*bits,g,r;
  DWORD bb,gg,rr,rgb;
  int h,high,i,j,k,m=3,n,w,wide,width;
  float u,ul,ur,v,vb,vt;
  if(model.elems<10000)
    m=3;
  else if(model.elems<40000)
    m=2;
  else
    m=1;
  ul=u1;
  if(u2<ul)
    ul=u2;
  if(u3<ul)
    ul=u3;
  ur=u1;
  if(u2>ur)
    ur=u2;
  if(u3>ur)
    ur=u3;
  vb=v1;
  if(v2<vb)
    vb=v2;
  if(v3<vb)
    vb=v3;
  vt=v1;
  if(v2>vt)
    vt=v2;
  if(v3>vt)
    vt=v3;
  wide=bitmaps[t]->biWidth;
  high=bitmaps[t]->biHeight;
  if(wide<8||high<8)
    return(WHITE);
  width=BMPwidth(wide,24);
  bits=BMPbits(bitmaps[t]);
  bb=gg=rr=0;
  for(n=j=0;j<m;j++)
    {
    u=ul+(j+1)*(ur-ul)/(m+1);
    w=(int)(u*wide);
    while(w<0)
      w+=wide;
    while(w>wide)
      w-=wide;
    for(k=0;k<m;k++,n++)
      {
      v=vb+(k+1)*(vt-vb)/(m+1);
      h=(int)(v*high);
      while(h<0)
        h+=high;
      while(h>high)
        h-=high;
      i=width*h+w*3;
      b=bits[i++];
      g=bits[i++];
      r=bits[i++];
      bb+=b;
      gg+=g;
      rr+=r;
      }
    }
  b=(BYTE)(bb/n);
  g=(BYTE)(gg/n);
  r=(BYTE)(rr/n);
  rgb=r;
  rgb<<=8;
  rgb|=g;
  rgb<<=8;
  rgb|=b;
  return(rgb);
  }

void RenderModel()
  {
  int c,i,j,k,l,m;
  float n[3],p[3],q[3],s,v1[3],v2[3],v3[3],xc,xm,xx,yc,ym,yx,zc,zm,zx;
  if(model.nodes<3)
    {
    SetWindowText(hStat,"model has less than 3 nodes");
    return;
    }
  if(model.elems<1)
    {
    SetWindowText(hStat,"model has no elements");
    return;
    }
  xm=ym=zm=FLT_MAX;
  xx=yx=zx=-FLT_MAX;
  c=model.colors;
  for(i=0;i<model.nodes;i++)
    {
    xm=min(xm,model.node[i].x);
    xx=max(xx,model.node[i].x);
    ym=min(ym,model.node[i].y);
    yx=max(yx,model.node[i].y);
    zm=min(zm,model.node[i].z);
    zx=max(zx,model.node[i].z);
    }
  if(xx<=xm)
    {
    SetWindowText(hStat,"model is degenerate in x");
    return;
    }
  if(yx<=ym)
    {
    SetWindowText(hStat,"model is degenerate in y");
    return;
    }
  if(zx<=zm)
    {
    SetWindowText(hStat,"model is degenerate in z");
    return;
    }
  xc=(float)((xm+xx)/2.);
  yc=(float)((ym+yx)/2.);
  zc=(float)((zm+zx)/2.);
  s=(float)(max(xx-xm,max(yx-ym,zx-zm))/3.);
  for(l=0;l<model.elems;l++)
    {
    i=model.elem[l].i;
    j=model.elem[l].j;
    k=model.elem[l].k;
    m=model.elem[l].m;
    v1[0]=(model.node[i].x-xc)/s;
    v1[1]=(model.node[i].y-yc)/s;
    v1[2]=(model.node[i].z-zc)/s;
    v2[0]=(model.node[j].x-xc)/s;
    v2[1]=(model.node[j].y-yc)/s;
    v2[2]=(model.node[j].z-zc)/s;
    v3[0]=(model.node[k].x-xc)/s;
    v3[1]=(model.node[k].y-yc)/s;
    v3[2]=(model.node[k].z-zc)/s;
    p[0]=v2[0]-v1[0];
    p[1]=v2[1]-v1[1];
    p[2]=v2[2]-v1[2];
    q[0]=v3[0]-v1[0];
    q[1]=v3[1]-v1[1];
    q[2]=v3[2]-v1[2];
    CrossProduct(p,q,n);
    if(textures&&fill_polygons&&m>=0)
      {
      glColor(WHITE);
      glEnable(GL_TEXTURE_2D);
      glBindTexture(GL_TEXTURE_2D,texture[m]);
      glBegin(GL_TRIANGLES);
      glNormal3fv(n);
      glTexCoord2d(model.node[i].u,model.node[i].v);
      glVertex3fv(v1);
      glTexCoord2d(model.node[j].u,model.node[j].v);
      glVertex3fv(v2);
      glTexCoord2d(model.node[k].u,model.node[k].v);
      glVertex3fv(v3);
      glEnd();
      if(!model.colors)
        {
        model.elem[l].color=RGBaverage(m,model.node[i].u,model.node[i].v,model.node[j].u,model.node[j].v,model.node[k].u,model.node[k].v);
        c=textures;
        }
      }
    else
      {
      glDisable(GL_TEXTURE_2D);
      if(fill_polygons||Lightness(model.elem[l].color)>31)
        glColor(model.elem[l].color);
      else
        glColor(white);
      glBegin(GL_TRIANGLES);
      glNormal3fv(n);
      glVertex3fv(v1);
      glVertex3fv(v2);
      glVertex3fv(v3);
      glEnd();
      }
    }
  model.colors=c;
  }

/*******************************************************************************
  FUNCTIONS COMMON TO ANY OPENGL APPLICATION
*******************************************************************************/

void SetView(int a,int b,int c,float s,float x,float y,float z)
  {
  View.a=a;
  View.b=b;
  View.c=c;
  View.s=s;
  View.x=x;
  View.y=y;
  View.z=z;
  }

void RenderScene()
  {
  float lP[4];
  if(!pDC)
    return;
  glClearDepth(1);
  glClearColor(0,0,0,0);
  glClearStencil(0);
  glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT|GL_STENCIL_BUFFER_BIT);
  glEnable(GL_DEPTH_TEST);
  glDepthFunc(GL_LESS);
  glFrontFace(GL_CCW);
  glDisable(GL_CULL_FACE);
  glCullFace(GL_BACK);
  glEnable(GL_STENCIL_TEST);
  glStencilOp(GL_KEEP,GL_KEEP,GL_REPLACE);
  glStencilFunc(GL_ALWAYS,0,-1);
  glShadeModel(GL_FLAT);
  glDisable(GL_AUTO_NORMAL);
  glEnable(GL_NORMALIZE);
  glColorMaterial(GL_FRONT,GL_AMBIENT_AND_DIFFUSE);
  glEnable(GL_COLOR_MATERIAL);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  if(fill_polygons)
    glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
  else
    glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,FALSE);
  glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER,FALSE);
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT,fltColor(0x000000));
  glLightfv(GL_LIGHT0,GL_AMBIENT ,fltColor(0x777777));
  glLightfv(GL_LIGHT0,GL_DIFFUSE ,fltColor(0x999999));
  glLightfv(GL_LIGHT0,GL_SPECULAR,fltColor(0x000000));
    lP[0]=(float)(sin(Light.a*M_PI/180)*cos(Light.b*M_PI/180));
    lP[2]=(float)(cos(Light.a*M_PI/180)*cos(Light.b*M_PI/180));
    lP[1]=(float)sin(Light.b*M_PI/180);
    lP[3]=0;/* infinite distance to light source */
  glLightfv(GL_LIGHT0,GL_POSITION,lP);
  glEnable(GL_LIGHT0);
  if(enable_lighting)
    glEnable(GL_LIGHTING);
  else
    glDisable(GL_LIGHTING);
  glColor(white);
  glRotatef((float)View.a,1.F,0.F,0.F);
  glRotatef((float)View.b,0.F,1.F,0.F);
  glRotatef((float)View.c,0.F,0.F,1.F);
  glScalef(View.s,View.s,View.s);
  glTranslatef(View.x,View.y,View.z);
  if(model.elems)
    RenderModel();
  glFinish();
  SwapBuffers(pDC);
  }

HDC GetBestPixelFormat(HWND hWnd,int objects)     /* select best pixel format */
  {
  int i,j,pixel_formats,s,sx;
  HDC hDC;
  PIXELFORMATDESCRIPTOR pfd;
  if((hDC=GetDC(hWnd))==NULL)
    Abort(__LINE__,"can't get device context\nWindows error code %lu",GetLastError());
  pixel_formats=DescribePixelFormat(hDC,1,0,NULL);
  if(pixel_formats<1)
    Abort(__LINE__,"device doesn't support pixel formats");
  j=sx=-1;
  for(i=1;i<=pixel_formats;i++)
    {
    DescribePixelFormat(hDC,i,sizeof(PIXELFORMATDESCRIPTOR),&pfd);
    if(!(pfd.dwFlags&PFD_SUPPORT_OPENGL))
      continue;
    if(!(pfd.dwFlags&PFD_DOUBLEBUFFER))
      continue;
    if(pfd.iPixelType!=PFD_TYPE_RGBA)
      continue;
    if(pfd.dwFlags&PFD_NEED_PALETTE)
      continue;
    if(pfd.dwFlags&PFD_NEED_SYSTEM_PALETTE)
      continue;
    if(pfd.cColorBits<8)
      continue;
    if(pfd.cDepthBits<16)
      continue;
    if((1UL<<pfd.cStencilBits)<(unsigned)objects)
      continue;
    s=pfd.cDepthBits/16;
    if(pfd.dwFlags&PFD_GENERIC_FORMAT)
      s+=8;
    if(pfd.dwFlags&PFD_GENERIC_ACCELERATED)
      s+=16;
    if(pfd.cColorBits==GetDeviceCaps(hDC,BITSPIXEL))
      s+=32;
    if(s<=sx)
      continue;
    sx=s;
    j=i;
    }
  DescribePixelFormat(hDC,j,sizeof(PIXELFORMATDESCRIPTOR),&pfd);
  SetPixelFormat(hDC,j,&pfd);
  return(hDC);
  }

void WINAPI DelayedRePaint(HWND hWnd,WPARAM wParam,DWORD idEvent,long uTime)
  {
  if(delayed_painting)
    {
    if(!mouse_clicked)
      {
      if(pending_repaint)
        {
        if(++delay_count>3)
          {
          fill_polygons=TRUE;
          enable_lighting=TRUE;
          InvalidateRect(hPlot,NULL,FALSE);
          UpdateWindow(hPlot);
          pending_repaint=FALSE;
          }
        }
      }
    }
  }

void RePaint()
  {
  if(delayed_painting)
    {
    fill_polygons=FALSE;
    enable_lighting=FALSE;
    pending_repaint=TRUE;
    delay_count=0;
    }
  InvalidateRect(hPlot,NULL,FALSE);
  UpdateWindow(hPlot);
  }

/*******************************************************************************
  FUNCTIONS TO MANAGE THE WINDOWS GRAPHICAL USER INTERFACE
*******************************************************************************/

void PositionWindows(int wide,int high)
  {
  int b=36,f=1,h,l,p=20,w,x,y;
  RECT rc;
  if(IsIconic(hMain))
    return;
  if(!hPlot)
    return;
  if(!bLogo)
    return;
  GetClientRect(hMain,&rc);
  w=bLogo->biWidth;
  x=rc.right-f-w;
  y=rc.bottom-2*p;
  MoveWindow(hExit,x,y,w,p,TRUE);y-=f+p;
  MoveWindow(hHelp,x,y,w,p,TRUE);y-=f+p;
  MoveWindow(hZall,x,y,w,p,TRUE);
  b=(w-2*f)/3;
  y-=b+4;
  MoveWindow(hMove,x,y,b,b,TRUE);
  MoveWindow(hRock,x+b+(w-3*b)/2,y,b,b,TRUE);
  MoveWindow(hTurn,x+w-b,y,b,b,TRUE);
  y-=b-p;
  l=(17*p)/20;
  MoveWindow(hText,x,y,w,l,TRUE);
  y-=f+p;
  MoveWindow(hDlay,x,y,w,p,TRUE);
  y=f;
  l=(w*bLogo->biHeight)/bLogo->biWidth;
  MoveWindow(hLogo,x,y,w,l,TRUE);
  y=f;
  w=x-5*f;
  w=4*(w/4);
  h=rc.bottom-f-p;
  x=2*f;
  y=f;
  MoveWindow(hPlot,x,y,w,h,TRUE);
  y+=f+p;
  MoveWindow(hStat,0,rc.bottom-(p+2*f)-1,rc.right,p+2*f,TRUE);
  glViewport(0,0,w,h);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  if(w>h)
    {
    float a=(float)(w/(h+0.01));
    if(perspective)
      glFrustum(-a,a,-1,1,Camera.a,Camera.c);
    else
      glOrtho(-a,a,-1,1,Camera.a,Camera.c);
    }
  else
    {
    float a=(float)(h/(w+0.01));
    if(perspective)
      glFrustum(-1,1,-a,a,Camera.a,Camera.c);
    else
      glOrtho(-1,1,-a,a,Camera.a,Camera.c);
    }
  glTranslatef(0,0,-Camera.b);
  }

void InvertLine(HDC hDC,int x1,int y1,int x2,int y2)
  {
  int ROP2;
  POINT P;
  ROP2=GetROP2(hDC);
  SetROP2(hDC,R2_NOT);
  MoveToEx(hDC,x1,y1,&P);
  LineTo(  hDC,x2,y2);
  SetROP2(hDC,ROP2);
  }

void InvertFrame(HDC hDC,RECT*rc)
  {
  int ROP2;
  POINT P;
  ROP2=GetROP2(hDC);
  SetROP2(hDC,R2_NOT);
  MoveToEx(hDC,rc->left ,rc->top   ,&P);
  LineTo(  hDC,rc->right,rc->top   );
  LineTo(  hDC,rc->right,rc->bottom);
  LineTo(  hDC,rc->left ,rc->bottom);
  LineTo(  hDC,rc->left ,rc->top   );
  SetROP2(hDC,ROP2);
  }

LRESULT WINAPI PlotProc(HWND hWnd,DWORD msg,WPARAM wParam,LPARAM lParam)
  {
  if(msg==WM_CLOSE)
    return(FALSE);
  if(msg==WM_CREATE)
    return(FALSE);
  if(msg==WM_DESTROY)
    return(FALSE);
  if(msg==WM_MOUSEMOVE)
    {
    if((wParam&MK_RBUTTON)&&(Mouse.down&MK_RBUTTON))
      {
      RECT rc;
      HDC hDC=GetDC(hWnd);
      if(Mouse.move)
        {
        if(Mouse.down&MK_CONTROL)
          InvertLine(hDC,Mouse.x1,Mouse.y1,Mouse.x2,Mouse.y2);
        else
          {
          rc.left  =min(Mouse.x1,Mouse.x2);
          rc.right =max(Mouse.x1,Mouse.x2);
          rc.top   =min(Mouse.y1,Mouse.y2);
          rc.bottom=max(Mouse.y1,Mouse.y2);
          InvertFrame(hDC,&rc);
          }
        }
      Mouse.down=wParam;
      Mouse.x2=LOWORD(lParam);
      Mouse.y2=HIWORD(lParam);
      rc.left  =min(Mouse.x1,Mouse.x2);
      rc.right =max(Mouse.x1,Mouse.x2);
      rc.top   =min(Mouse.y1,Mouse.y2);
      rc.bottom=max(Mouse.y1,Mouse.y2);
      if(Mouse.down&MK_CONTROL)
        InvertLine(hDC,Mouse.x1,Mouse.y1,Mouse.x2,Mouse.y2);
      else
        {
        rc.left  =min(Mouse.x1,Mouse.x2);
        rc.right =max(Mouse.x1,Mouse.x2);
        rc.top   =min(Mouse.y1,Mouse.y2);
        rc.bottom=max(Mouse.y1,Mouse.y2);
        InvertFrame(hDC,&rc);
        }
      Mouse.move++;
      ReleaseDC(hWnd,hDC);
      }
    }
  if(msg==WM_PAINT)
    {
    PAINTSTRUCT pS;
    HDC hDC=BeginPaint(hWnd,&pS);
    if(pDC)
      {
      if(hDC!=pDC)
        Abort(__LINE__,"OpenGL device context not preserved");
      RenderScene();
      }
    EndPaint(hWnd,&pS);
    return(FALSE);
    }
  if(msg==WM_RBUTTONDOWN)
    {
    mouse_clicked=TRUE;
    Mouse.down=wParam;
    Mouse.x1=LOWORD(lParam);
    Mouse.y1=HIWORD(lParam);
    Mouse.move=0;
    SetCapture(hWnd);
    return(FALSE);
    }
  if(msg==WM_RBUTTONUP)
    {
    if(Mouse.down&MK_RBUTTON)
      {
      float cx,cy,ds,dx,dy,s1,s2;
      HDC hDC;
      RECT rc;
      if(Mouse.move)
        {
        hDC=GetDC(hWnd);
        if(Mouse.down&MK_CONTROL)
          InvertLine(hDC,Mouse.x1,Mouse.y1,Mouse.x2,Mouse.y2);
        else
          {
          rc.left  =min(Mouse.x1,Mouse.x2);
          rc.right =max(Mouse.x1,Mouse.x2);
          rc.top   =min(Mouse.y1,Mouse.y2);
          rc.bottom=max(Mouse.y1,Mouse.y2);
          InvertFrame(hDC,&rc);
          }
        ReleaseDC(hWnd,hDC);
        }
      Mouse.x2=LOWORD(lParam);
      Mouse.y2=HIWORD(lParam);
      GetClientRect(hWnd,&rc);
      dx=(float)(Mouse.x2-Mouse.x1);
      dy=(float)(Mouse.y2-Mouse.y1);
      ds=(float)(hypot(rc.right,rc.bottom)/sqrt(2));
      if(Mouse.down&MK_CONTROL)
        {
        View.x+=2*dx/ds;
        View.y-=2*dy/ds;
        RePaint();
        }
      else
        {
        s1=View.s;
        if(fabs(dx)*rc.bottom>fabs(dy)*rc.right)
          s2=(float)(s1*rc.right/fabs(dx));
        else
          s2=(float)(s1*rc.bottom/fabs(dy));
        s2=(float)min(s2,2*s1);
        if(hypot(dx,dy)<32)
          s2=s1;
        cx=(float)(rc.right /2-(Mouse.x1+Mouse.x2)/2);
        cy=(float)(rc.bottom/2-(Mouse.y1+Mouse.y2)/2);
        View.x=(View.x/s1+(cx/ds))*s2;
        View.y=(View.y/s1-(cx/ds))*s2;
        View.s=s2;
        RePaint();
        }
      }
    Mouse.down=0;
    Mouse.move=0;
    ReleaseCapture();
    mouse_clicked=FALSE;
    return(FALSE);
    }
  if(msg==WM_SIZE)
    {
    PositionWindows(0,0);
    return(FALSE);
    }
  return(DefWindowProc(hWnd,msg,wParam,lParam));
  }

LRESULT WINAPI AboutProc(HWND hWnd,DWORD msg,DWORD wParam,LPARAM lParam)
  {
  if(msg==WM_COMMAND)
    {
    if(HIWORD(wParam)==BN_CLICKED||HIWORD(wParam)==BN_DOUBLECLICKED)
      if(LOWORD(wParam)==IDOK)
        return(EndDialog(hWnd,LOWORD(wParam)));
    return(FALSE);
    }
  if(msg==WM_CTLCOLORDLG)
    return((LRESULT)GetStockObject(LTGRAY_BRUSH));
  if(msg==WM_CTLCOLORSTATIC)
    {
    SelectObject((HDC)wParam,GetStockObject(ANSI_VAR_FONT));
    SetTextColor((HDC)wParam,BLACK);
    SetBkMode((HDC)wParam,TRANSPARENT);
    return((DWORD)GetStockObject(LTGRAY_BRUSH));
    }
  if(msg==WM_DESTROY)
    return(FALSE);
  if(msg==WM_INITDIALOG)
    {
    CenterWindowOnScreen(hWnd);
    return(TRUE);
    }
  return(FALSE);
  }

int IsCommand(WPARAM wParam,LPARAM lParam)
  {
  if(wParam==PUSH_EXIT||wParam==PUSH_ALT_F4)
    {
    PostQuitMessage(0);
    return(TRUE);
    }
  if(wParam==PUSH_ABOUT)
    {
    DialogBox(hInst,"ABOUT",hMain,(DLGPROC)AboutProc);
    SetFocus(hMain);
    return(TRUE);
    }
  if(wParam==PUSH_HELP||wParam==PUSH_F1)
    {
    WinHelp(hMain,"View3D.hlp",HELP_CONTENTS,0);
    SetFocus(hMain);
    return(TRUE);
    }
  if(wParam==PUSH_F2)
    {
    char cbuf[256];
    sprintf(cbuf,"Camera=%lG,%lG,%lG\nLight=%i,%i\nView=%i,%i,%i,%lG,%lG,%lG,%lG",Camera.a,Camera.b,Camera.c,Light.a,Light.b,View.a,View.b,View.c,View.s,View.x,View.y,View.z);
    MessageBox(NULL,cbuf,"current settings",MB_OK);
    return(TRUE);
    }
  if(wParam==PUSH_DELAY)
    {
    delayed_painting^=1;
    SendMessage(hDlay,BM_SETCHECK,delayed_painting,0);
    RePaint();
    return(TRUE);
    }
  if(wParam==PUSH_IMPORT)
    {
    char*filter;
    char*Files[]={"*.3ds (3D Studio)","*.wrl (Virtual Reality Markup Language)",
      "*.3dv (3D View)","*.zip (model + texture package)",NULL};
    FreeModel(&model);
    GetCurrentDirectory(sizeof(file_name),file_name);
    strcat(file_name,"\\*.3ds");
    filter=BuildFilter("3D Models",Files);
    if(GetInputFileName(file_name,filter,"Import 3D Model"))
      model=ImportModel(file_name);
    SetView(0,0,0,1.,0,0,0);
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  if(wParam==PUSH_EXPORT)
    {
    char*filter;
    char*Files[]={"*.3ds (3D Studio)","*.wrl (Virtual Reality Markup Language)",
      "*.3dv (3D View)","*.zip (model + texture package)",NULL};
    GetCurrentDirectory(sizeof(file_name),file_name);
    strcat(file_name,"\\*.3ds");
    filter=BuildFilter("3D Models",Files);
    if(GetOutputFileName(file_name,filter,"Export 3D Model"))
      ExportModel(file_name);
    return(TRUE);
    }
  return(FALSE);
  }

int IsZoom(WPARAM wParam,LPARAM lParam)
  {
  if(wParam==PUSH_ZALL||wParam==PUSH_CTL_Z)
    {
    SetView(0,0,0,1.F,0.F,0.F,0.F);
    RePaint();
    return(TRUE);
    }
  if(wParam==PUSH_ZOOM||wParam==PUSH_PGUP)
    {
    View.s*=1.05F;
    RePaint();
    return(TRUE);
    }
  if(wParam==PUSH_ZOUT||wParam==PUSH_PGDN)
    {
    View.s/=1.05F;
    RePaint();
    return(TRUE);
    }
  return(FALSE);
  }

int IsRotation(WPARAM wParam,LPARAM lParam)
  {
  if(wParam==PUSH_CTL_A)
    {
    View.a= 90;
    View.b=270;
    View.c=  0;
    RePaint();
    return(TRUE);
    }
  if(wParam==PUSH_UP)
    {
    View.a-=10;
    if(View.a<0)
      View.a=350;
    RePaint();
    return(TRUE);
    }
  if(wParam==PUSH_DOWN)
    {
    View.a+=10;
    if(View.a>350)
      View.a=0;
    RePaint();
    return(TRUE);
    }
  if(wParam==PUSH_END)
    {
    View.c+=10;
    if(View.c>350)
      View.c=0;
    RePaint();
    return(TRUE);
    }
  if(wParam==PUSH_HOME)
    {
    View.c-=10;
    if(View.c<0)
      View.c=350;
    RePaint();
    return(TRUE);
    }
  if(wParam==PUSH_RIGHT)
    {
    View.b+=10;
    if(View.b>350)
      View.b=0;
    RePaint();
    return(TRUE);
    }
  if(wParam==PUSH_LEFT)
    {
    View.b-=10;
    if(View.b<0)
      View.b=350;
    RePaint();
    return(TRUE);
    }
  return(FALSE);
  }

int IsTranslation(WPARAM wParam,LPARAM lParam)
  {
  if(wParam==PUSH_CTL_LEFT)
    {
    View.x+=0.2F;
    RePaint();
    return(TRUE);
    }
  if(wParam==PUSH_CTL_RIGHT)
    {
    View.x-=0.2F;
    RePaint();
    return(TRUE);
    }
  if(wParam==PUSH_CTL_HOME)
    {
    View.y+=0.2F;
    RePaint();
    return(TRUE);
    }
  if(wParam==PUSH_CTL_END)
    {
    View.y-=0.2F;
    RePaint();
    return(TRUE);
    }
  if(wParam==PUSH_CTL_UP)
    {
    View.z+=0.2F;
    RePaint();
    return(TRUE);
    }
  if(wParam==PUSH_CTL_DOWN)
    {
    View.z-=0.2F;
    RePaint();
    return(TRUE);
    }
  return(FALSE);
  }

void Position(float x,float y,float z)
  {
  float M[16];
  glPushMatrix();
  glLoadIdentity();
  glRotatef((float)View.a,1.F,0.F,0.F);
  glRotatef((float)View.b,0.F,1.F,0.F);
  glRotatef((float)View.c,0.F,0.F,1.F);
  glGetFloatv(GL_MODELVIEW_MATRIX,M);
  View.x+=M[0]*x+M[1]*y+M[ 2]*z+M[ 3];
  View.y+=M[4]*x+M[5]*y+M[ 6]*z+M[ 7];
  View.z+=M[8]*x+M[9]*y+M[10]*z+M[11];
  glPopMatrix();
  }

int IsMovement(WPARAM wParam,LPARAM lParam)
  {
  if(wParam==PUSH_ALT_UP)
    {
    Position(0.F,-0.2F,0.F);
    RePaint();
    return(TRUE);
    }
  if(wParam==PUSH_ALT_DOWN)
    {
    Position(0.F,0.2F,0.F);
    RePaint();
    return(TRUE);
    }
  if(wParam==PUSH_ALT_LEFT)
    {
    Position(0.2F,0.F,0.F);
    RePaint();
    return(TRUE);
    }
  if(wParam==PUSH_ALT_RIGHT)
    {
    Position(-0.2F,0.F,0.F);
    RePaint();
    return(TRUE);
    }
  if(wParam==PUSH_ALT_HOME)
    {
    Position(0.F,0.F,-0.2F);
    RePaint();
    return(TRUE);
    }
  if(wParam==PUSH_ALT_END)
    {
    Position(0.F,0.F,0.2F);
    RePaint();
    return(TRUE);
    }
  return(FALSE);
  }

LRESULT WINAPI MainProc(HWND hWnd,DWORD msg,WPARAM wParam,LPARAM lParam)
  {
  if(msg==WM_CLOSE)
    {
    PostQuitMessage(0);
    return(FALSE);
    }
  if(msg==WM_COMMAND)
    {
    if(IsCommand(LOWORD(wParam),lParam))
      ;
    else if(IsZoom(LOWORD(wParam),lParam))
      ;
    else if(IsRotation(LOWORD(wParam),lParam))
      ;
    else if(IsTranslation(LOWORD(wParam),lParam))
      ;
    else if(IsMovement(LOWORD(wParam),lParam))
      ;
    SetFocus(hMain);
    return(FALSE);
    }
  if(msg==WM_CREATE)
    return(FALSE);
  if(msg==WM_CTLCOLORBTN)
    {
    SetTextColor((HDC)wParam,BLACK);
    SetBkColor((HDC)wParam,LTGRAY);
    return((DWORD)GetStockObject(LTGRAY_BRUSH));
    }
  if(msg==WM_CTLCOLOREDIT)
    {
    SetTextColor((HDC)wParam,BLACK);
    SetBkColor((HDC)wParam,WHITE);
    return((DWORD)GetStockObject(WHITE_BRUSH));
    }
  if(msg==WM_CTLCOLORSTATIC)
    {
    SetTextColor((HDC)wParam,BLACK);
    SetBkColor((HDC)wParam,LTGRAY);
    return((DWORD)GetStockObject(LTGRAY_BRUSH));
    }
  if(msg==WM_GETMINMAXINFO)
    {
    MINMAXINFO*mInfo=(MINMAXINFO*)lParam;
    mInfo->ptMinTrackSize.x=560;
    mInfo->ptMinTrackSize.y=480;
    return(FALSE);
    }
  if(msg==WM_PAINT)
    {
    HDC hDC;
    PAINTSTRUCT pS;
    hDC=BeginPaint(hWnd,&pS);
    EndPaint(hWnd,&pS);
    return(FALSE);
    }
  if(msg==WM_SIZE)
    {
    if(wParam==SIZE_MAXIMIZED||wParam==SIZE_RESTORED)
      PositionWindows(LOWORD(lParam),HIWORD(lParam));
    return(FALSE);
    }
  return(DefWindowProc(hWnd,msg,wParam,lParam));
  }

LRESULT WINAPI LogoProc(HWND hWnd,UINT msg,WPARAM wParam,LPARAM lParam)
  {
  if(msg==WM_CLOSE)
    return(FALSE);
  if(msg==WM_CREATE)
    return(FALSE);
  if(msg==WM_DESTROY)
    return(FALSE);
  if(msg==WM_PAINT)
    {
    HDC hDC;
    PAINTSTRUCT ps;
    RECT rc;
    if((hDC=BeginPaint(hWnd,&ps))==NULL)
      Abort(__LINE__,"can't get device context to begin painting\nWin Err Code %i",GetLastError());
    GetClientRect(hWnd,&rc);
    if(bLogo)
      PaintBitmap(hDC,&rc,bLogo,TRUE);
    EndPaint(hWnd,&ps);
    return(FALSE);
    }
  return(DefWindowProc(hWnd,msg,wParam,lParam));
  }

/*******************************************************************************
  FUNCTIONS TO CREATE 3D CONTROL BUTTONS
*******************************************************************************/

void DrawLine(HDC hDC,DWORD rgb,int x,...)       /* variable number of points */
  {
  int y;
  HPEN hPen;
  POINT P;
  va_list arg_marker;
  hPen=CreatePen(PS_SOLID,1,rgb);
  hPen=SelectObject(hDC,hPen);
  va_start(arg_marker,x);
  y=va_arg(arg_marker,int);
  MoveToEx(hDC,x,y,&P);
  while(1)
    {
    x=va_arg(arg_marker,int);
    if(x==INT_MIN||x==INT_MAX)
      break;
    y=va_arg(arg_marker,int);
    LineTo(hDC,x,y);
    }
  hPen=SelectObject(hDC,hPen);
  DeleteObject(hPen);
  }

void Arrow(HDC hDC,DWORD fill,DWORD outline,int a,int x,int y,int b)
  {
  int i;
  float C,S;
  HBRUSH hBrush;
  HPEN hPen;
  POINT P[8],Q[8];
  hBrush=CreateSolidBrush(fill);
  hBrush=SelectObject(hDC,hBrush);
  hPen=CreatePen(PS_SOLID,1,outline);
  hPen=SelectObject(hDC,hPen);
  C=(float)cos(a*M_PI/180);
  S=(float)sin(a*M_PI/180);
  P[0].x=P[7].x=-b;
  P[1].x=P[2].x=P[5].x=P[6].x=0;
  P[3].x=P[4].x=b;
  P[5].y=-b;
  P[4].y=P[6].y=P[7].y=0;
  P[0].y=P[1].y=P[3].y=1;
  P[2].y=b+1;
  for(i=0;i<8;i++)
    {
    Q[i].x=nint( C*P[i].x+S*P[i].y);
    Q[i].y=nint(-S*P[i].x+C*P[i].y);
    }
  for(i=0;i<8;i++)
    {
    P[i].x=Q[i].x+x;
    P[i].y=Q[i].y+y;
    }
  Polygon(hDC,P,8);
  hBrush=SelectObject(hDC,hBrush);
  DeleteObject(hBrush);
  hPen=SelectObject(hDC,hPen);
  DeleteObject(hPen);
  }

LRESULT WINAPI MoveProc(HWND hWnd,DWORD msg,WPARAM wParam,LPARAM lParam)
  {                     /* this draws a group of four custom position buttons */
  if(msg==WM_CLOSE)                   /* it's supposed to look like 4 arrows */
    return(FALSE);
  if(msg==WM_CREATE)
    return(FALSE);
  if(msg==WM_DESTROY)
    return(FALSE);
  if(msg==WM_LBUTTONDOWN)
    {
    int state,x,y;
    RECT rc;
    GetClientRect(hWnd,&rc);
    x=LOWORD(lParam);
    y=HIWORD(lParam);
    if(y>x)
      {
      if(y>rc.bottom-x)
        state=2;
      else
        state=4;
      }
    else
      {
      if(y>rc.bottom-x)
        state=8;
      else
        state=1;
      }
    SetWindowLong(hWnd,GWL_USERDATA,state);
    InvalidateRect(hWnd,NULL,TRUE);
    return(FALSE);
    }
  if(msg==WM_LBUTTONUP)
    {
    int state;
    state=GetWindowLong(hWnd,GWL_USERDATA);
    if(state&1)
      PostMessage(GetParent(hWnd),WM_COMMAND,PUSH_ALT_UP,0);
    if(state&2)
      PostMessage(GetParent(hWnd),WM_COMMAND,PUSH_ALT_DOWN,0);
    if(state&4)
      PostMessage(GetParent(hWnd),WM_COMMAND,PUSH_ALT_LEFT,0);
    if(state&8)
      PostMessage(GetParent(hWnd),WM_COMMAND,PUSH_ALT_RIGHT,0);
    SetWindowLong(hWnd,GWL_USERDATA,0);
    InvalidateRect(hWnd,NULL,TRUE);
    return(FALSE);
    }
  if(msg==WM_PAINT)
    {
    int l,state;
    HDC hDC;
    PAINTSTRUCT pS;
    RECT rc;
    hDC=BeginPaint(hWnd,&pS);
    state=GetWindowLong(hWnd,GWL_USERDATA);
    GetClientRect(hWnd,&rc);
    FillRect(hDC,&rc,GetStockObject(LTGRAY_BRUSH));
    l=(rc.right+rc.bottom)/4;
    DrawLine(hDC,BLACK,0,1,0,2*l-1,INT_MIN);
    DrawLine(hDC,BLACK,1,2*l-1,2*l-1,2*l-1,INT_MIN);
    DrawLine(hDC,BLACK,2*l-1,2*l-2,2*l-1,0,INT_MIN);
    DrawLine(hDC,BLACK,2*l-2,0,0,0,INT_MIN);
    DrawLine(hDC,BLACK,1,1,2*l-1,2*l-1,INT_MIN);
    DrawLine(hDC,BLACK,1,2*l-2,2*l-1,0,INT_MIN);
    DrawLine(hDC,state&1?WHITE:DKGRAY,2,1,l,l-1,INT_MIN); /* up */
    DrawLine(hDC,state&1?WHITE:DKGRAY,3,1,l,l-2,INT_MIN);
    DrawLine(hDC,state&1?WHITE:DKGRAY,2*l-3,1,l-1,l-1,INT_MIN);
    DrawLine(hDC,state&1?WHITE:DKGRAY,2*l-4,1,l-1,l-2,INT_MIN);
    DrawLine(hDC,state&1?DKGRAY:WHITE,4,1,2*l-4,1,INT_MIN);
    DrawLine(hDC,state&2?WHITE:DKGRAY,4,2*l-2,2*l-4,2*l-2,INT_MIN); /* down */
    DrawLine(hDC,state&2?DKGRAY:WHITE,2,2*l-2,l,l,INT_MIN);
    DrawLine(hDC,state&2?DKGRAY:WHITE,3,2*l-2,l,l+1,INT_MIN);
    DrawLine(hDC,state&2?DKGRAY:WHITE,2*l-3,2*l-2,l-1,l,INT_MIN);
    DrawLine(hDC,state&2?DKGRAY:WHITE,2*l-4,2*l-2,l-1,l+1,INT_MIN);
    DrawLine(hDC,state&4?WHITE:DKGRAY,1,2*l-3,l-1,l-1,INT_MIN); /* left */
    DrawLine(hDC,state&4?WHITE:DKGRAY,1,2*l-4,l-2,l-1,INT_MIN);
    DrawLine(hDC,state&4?DKGRAY:WHITE,1,2,l-1,l,INT_MIN);
    DrawLine(hDC,state&4?DKGRAY:WHITE,1,3,l-2,l,INT_MIN);
    DrawLine(hDC,state&4?DKGRAY:WHITE,1,4,1,2*l-4,INT_MIN);
    DrawLine(hDC,state&8?WHITE:DKGRAY,2*l-2,2*l-3,l,l-1,INT_MIN); /* right */
    DrawLine(hDC,state&8?WHITE:DKGRAY,2*l-2,2*l-4,l+1,l-1,INT_MIN);
    DrawLine(hDC,state&8?DKGRAY:WHITE,2*l-2,4,2*l-2,2*l-4,INT_MIN);
    DrawLine(hDC,state&8?DKGRAY:WHITE,2*l-2,2,l,l,INT_MIN);
    DrawLine(hDC,state&8?DKGRAY:WHITE,2*l-2,3,l+1,l,INT_MIN);
    Arrow(hDC,BLACK,BLACK,  0,9*l/5-3,  l  -1,l/5);
    Arrow(hDC,BLACK,BLACK, 90,  l-1  ,2*l/5  ,l/5);
    Arrow(hDC,BLACK,BLACK,180,2*l/5  ,  l    ,l/5);
    Arrow(hDC,BLACK,BLACK,270,  l    ,9*l/5-3,l/5);
    EndPaint(hWnd,&pS);
    return(FALSE);
    }
  return(DefWindowProc(hWnd,msg,wParam,lParam));
  }

LRESULT WINAPI RockProc(HWND hWnd,DWORD msg,WPARAM wParam,LPARAM lParam)
  {            /* this draws a group of four custom position/rotation buttons */
  if(msg==WM_CLOSE)      /* it's supposed to look like a 4-way rocker switch */
    return(FALSE);                          /* the middle does something too! */
  if(msg==WM_CREATE)
    return(FALSE);
  if(msg==WM_DESTROY)
    return(FALSE);
  if(msg==WM_LBUTTONDOWN)
    {
    int state,x,y;
    RECT rc;
    GetClientRect(hWnd,&rc);
    x=LOWORD(lParam);
    y=HIWORD(lParam);
    if(hypot(x-rc.right/2,y-rc.bottom/2)<max(2,(rc.bottom+rc.right)/12))
      state=16;
    else if(y>x)
      {
      if(y>rc.bottom-x)
        state=2;
      else
        state=4;
      }
    else
      {
      if(y>rc.bottom-x)
        state=8;
      else
        state=1;
      }
    SetWindowLong(hWnd,GWL_USERDATA,state);
    InvalidateRect(hWnd,NULL,TRUE);
    return(FALSE);
    }
  if(msg==WM_LBUTTONUP)
    {
    int state;
    state=GetWindowLong(hWnd,GWL_USERDATA);
    if(state&1)
      PostMessage(GetParent(hWnd),WM_COMMAND,PUSH_ALT_HOME,0);
    if(state&2)
      PostMessage(GetParent(hWnd),WM_COMMAND,PUSH_ALT_END,0);
    if(state&4)
      PostMessage(GetParent(hWnd),WM_COMMAND,PUSH_HOME,0);
    if(state&8)
      PostMessage(GetParent(hWnd),WM_COMMAND,PUSH_END,0);
    if(state&16)
      PostMessage(GetParent(hWnd),WM_COMMAND,PUSH_PGDN,0);
    SetWindowLong(hWnd,GWL_USERDATA,0);
    InvalidateRect(hWnd,NULL,TRUE);
    return(FALSE);
    }
  if(msg==WM_PAINT)
    {
    int a,b,c,l,state;
    HDC hDC;
    PAINTSTRUCT pS;
    RECT rc;
    hDC=BeginPaint(hWnd,&pS);
    state=GetWindowLong(hWnd,GWL_USERDATA);
    GetClientRect(hWnd,&rc);
    FillRect(hDC,&rc,GetStockObject(LTGRAY_BRUSH));
    l=(rc.right+rc.bottom)/4;
    a=2*l/3-1;
    b=4*l/3;
    c=2*l-1;
    DrawLine(hDC,BLACK,a,0,b,0,b,a,c,a,c,b,b,b,b,c,a,c,a,b,0,b,0,a,a,a,a,0,INT_MIN);
    DrawLine(hDC,state&1?DKGRAY:WHITE,a+1,a+1,a+1,  1,b-1,  1,INT_MIN);
    DrawLine(hDC,state&1?WHITE:DKGRAY,b-1,  1,b-1,a+1,a+1,a+1,INT_MIN);
    DrawLine(hDC,state&2?DKGRAY:WHITE,a+1,c-1,a+1,b-1,b-1,b-1,INT_MIN);
    DrawLine(hDC,state&2?WHITE:DKGRAY,b-1,b-1,b-1,c-1,a+1,c-1,INT_MIN);
    DrawLine(hDC,state&4?DKGRAY:WHITE,  1,b-1,  1,a+1,a+1,a+1,INT_MIN);
    DrawLine(hDC,state&4?WHITE:DKGRAY,a+1,a+1,a+1,b-1,  1,b-1,INT_MIN);
    DrawLine(hDC,state&8?DKGRAY:WHITE,b-1,b-1,b-1,a+1,c-1,a+1,INT_MIN);
    DrawLine(hDC,state&8?WHITE:DKGRAY,c-1,a+1,c-1,b-1,b-1,b-1,INT_MIN);
    Arrow(hDC,BLACK,BLACK,  0,9*l/5-3,  l  -1,l/5);
    Arrow(hDC,BLACK,BLACK, 90,  l-1  ,2*l/5  ,l/5);
    Arrow(hDC,BLACK,BLACK,180,2*l/5  ,  l    ,l/5);
    Arrow(hDC,BLACK,BLACK,270,  l    ,9*l/5-3,l/5);
    if(state&16)
      {
      SelectObject(hDC,GetStockObject(BLACK_PEN));
      SelectObject(hDC,GetStockObject(GRAY_BRUSH));
      Rectangle(hDC,a,a,b,b);
      }
    EndPaint(hWnd,&pS);
    return(FALSE);
    }
  return(DefWindowProc(hWnd,msg,wParam,lParam));
  }

LRESULT WINAPI TurnProc(HWND hWnd,DWORD msg,WPARAM wParam,LPARAM lParam)
  {                     /* this draws a group of four custom rotation buttons */
  if(msg==WM_CLOSE)     /* it looks sort of like a segmented life-saver ring */
    return(FALSE);                          /* the middle does something too! */
  if(msg==WM_CREATE)
    return(FALSE);
  if(msg==WM_DESTROY)
    return(FALSE);
  if(msg==WM_LBUTTONDOWN)
    {
    int state,x,y;
    RECT rc;
    GetClientRect(hWnd,&rc);
    x=LOWORD(lParam);
    y=HIWORD(lParam);
    if(hypot(x-rc.right/2,y-rc.bottom/2)<max(2,(rc.bottom+rc.right)/12))
      state=16;
    else if(x>rc.right/2)
      {
      if(y>rc.bottom/2)
        state=4;
      else
        state=2;
      }
    else
      {
      if(y>rc.bottom/2)
        state=8;
      else
        state=1;
      }
    SetWindowLong(hWnd,GWL_USERDATA,state);
    InvalidateRect(hWnd,NULL,TRUE);
    return(FALSE);
    }
  if(msg==WM_LBUTTONUP)
    {
    int state;
    state=GetWindowLong(hWnd,GWL_USERDATA);
    if(state&1)
      PostMessage(GetParent(hWnd),WM_COMMAND,PUSH_UP,0);
    if(state&2)
      PostMessage(GetParent(hWnd),WM_COMMAND,PUSH_DOWN,0);
    if(state&4)
      PostMessage(GetParent(hWnd),WM_COMMAND,PUSH_LEFT,0);
    if(state&8)
      PostMessage(GetParent(hWnd),WM_COMMAND,PUSH_RIGHT,0);
    if(state&16)
      PostMessage(GetParent(hWnd),WM_COMMAND,PUSH_PGUP,0);
    SetWindowLong(hWnd,GWL_USERDATA,0);
    InvalidateRect(hWnd,NULL,TRUE);
    return(FALSE);
    }
  if(msg==WM_PAINT)
    {
    int l,r,state;
    HDC hDC;
    HRGN hClip;
    PAINTSTRUCT pS;
    RECT rc;
    hDC=BeginPaint(hWnd,&pS);
    state=GetWindowLong(hWnd,GWL_USERDATA);
    GetClientRect(hWnd,&rc);
    FillRect(hDC,&rc,GetStockObject(LTGRAY_BRUSH));
    l=(rc.right+rc.bottom)/4;
    hClip=CreateRectRgn(0,0,l,l);
    hClip=SelectObject(hDC,hClip);
    SelectObject(hDC,GetStockObject(BLACK_PEN));
    if(state&1)
      SelectObject(hDC,GetStockObject(GRAY_BRUSH));
    else
      SelectObject(hDC,GetStockObject(WHITE_BRUSH));
    Ellipse(hDC,0,0,2*l,2*l);
    if(state&1)
      SelectObject(hDC,pGray);
    else
      SelectObject(hDC,GetStockObject(WHITE_PEN));
    SelectObject(hDC,GetStockObject(LTGRAY_BRUSH));
    r=2;
    Ellipse(hDC,r,r,2*l-r,2*l-r);
    if(state&1)
      {
      SelectObject(hDC,GetStockObject(WHITE_PEN));
      SelectObject(hDC,GetStockObject(WHITE_BRUSH));
      }
    else
      {
      SelectObject(hDC,pGray);
      SelectObject(hDC,GetStockObject(GRAY_BRUSH));
      }
    r=l/2;
    Ellipse(hDC,r,r,2*l-r,2*l-r);
    SelectObject(hDC,GetStockObject(BLACK_PEN));
    SelectObject(hDC,GetStockObject(LTGRAY_BRUSH));
    r+=2;
    Ellipse(hDC,r,r,2*l-r,2*l-r);
    hClip=SelectObject(hDC,hClip);
    DeleteObject(hClip);
    hClip=CreateRectRgn(l,0,2*l,l);
    hClip=SelectObject(hDC,hClip);
    SelectObject(hDC,GetStockObject(BLACK_PEN));
    if(state&2)
      SelectObject(hDC,GetStockObject(GRAY_BRUSH));
    else
      SelectObject(hDC,GetStockObject(WHITE_BRUSH));
    Ellipse(hDC,0,0,2*l,2*l);
    if(state&2)
      SelectObject(hDC,pGray);
    else
      SelectObject(hDC,GetStockObject(WHITE_PEN));
    SelectObject(hDC,GetStockObject(LTGRAY_BRUSH));
    r=2;
    Ellipse(hDC,r,r,2*l-r,2*l-r);
    if(state&2)
      {
      SelectObject(hDC,GetStockObject(WHITE_PEN));
      SelectObject(hDC,GetStockObject(WHITE_BRUSH));
      }
    else
      {
      SelectObject(hDC,pGray);
      SelectObject(hDC,GetStockObject(GRAY_BRUSH));
      }
    r=l/2;
    Ellipse(hDC,r,r,2*l-r,2*l-r);
    SelectObject(hDC,GetStockObject(BLACK_PEN));
    SelectObject(hDC,GetStockObject(LTGRAY_BRUSH));
    r+=2;
    Ellipse(hDC,r,r,2*l-r,2*l-r);
    hClip=SelectObject(hDC,hClip);
    DeleteObject(hClip);
    hClip=CreateRectRgn(l,l,2*l,2*l);
    hClip=SelectObject(hDC,hClip);
    SelectObject(hDC,GetStockObject(BLACK_PEN));
    if(state&4)
      SelectObject(hDC,GetStockObject(WHITE_BRUSH));
    else
      SelectObject(hDC,GetStockObject(GRAY_BRUSH));
    Ellipse(hDC,0,0,2*l,2*l);
    if(state&4)
      SelectObject(hDC,GetStockObject(WHITE_PEN));
    else
      SelectObject(hDC,pGray);
    SelectObject(hDC,GetStockObject(LTGRAY_BRUSH));
    r=2;
    Ellipse(hDC,r,r,2*l-r,2*l-r);
    if(state&4)
      {
      SelectObject(hDC,pGray);
      SelectObject(hDC,GetStockObject(GRAY_BRUSH));
      }
    else
      {
      SelectObject(hDC,GetStockObject(WHITE_PEN));
      SelectObject(hDC,GetStockObject(WHITE_BRUSH));
      }
    r=l/2;
    Ellipse(hDC,r,r,2*l-r,2*l-r);
    SelectObject(hDC,GetStockObject(BLACK_PEN));
    SelectObject(hDC,GetStockObject(LTGRAY_BRUSH));
    r+=2;
    Ellipse(hDC,r,r,2*l-r,2*l-r);
    hClip=SelectObject(hDC,hClip);
    DeleteObject(hClip);
    hClip=CreateRectRgn(0,l,l,2*l);
    hClip=SelectObject(hDC,hClip);
    SelectObject(hDC,GetStockObject(BLACK_PEN));
    if(state&8)
      SelectObject(hDC,GetStockObject(WHITE_BRUSH));
    else
      SelectObject(hDC,GetStockObject(GRAY_BRUSH));
    Ellipse(hDC,0,0,2*l,2*l);
    if(state&8)
      SelectObject(hDC,GetStockObject(WHITE_PEN));
    else
      SelectObject(hDC,pGray);
    SelectObject(hDC,GetStockObject(LTGRAY_BRUSH));
    r=2;
    Ellipse(hDC,r,r,2*l-r,2*l-r);
    if(state&8)
      {
      SelectObject(hDC,pGray);
      SelectObject(hDC,GetStockObject(GRAY_BRUSH));
      }
    else
      {
      SelectObject(hDC,GetStockObject(WHITE_PEN));
      SelectObject(hDC,GetStockObject(WHITE_BRUSH));
      }
    r=l/2;
    Ellipse(hDC,r,r,2*l-r,2*l-r);
    SelectObject(hDC,GetStockObject(BLACK_PEN));
    SelectObject(hDC,GetStockObject(LTGRAY_BRUSH));
    r+=2;
    Ellipse(hDC,r,r,2*l-r,2*l-r);
    hClip=SelectObject(hDC,hClip);
    DeleteObject(hClip);
    hClip=CreateRectRgn(0,0,2*l,2*l);
    hClip=SelectObject(hDC,hClip);
    DrawLine(hDC,BLACK,l,1,l,l/2+2,INT_MIN);
    DrawLine(hDC,BLACK,l,3*l/2-2,l,2*l-1,INT_MIN);
    DrawLine(hDC,BLACK,1,l,l/2+2,l,INT_MIN);
    DrawLine(hDC,BLACK,3*l/2-2,l,2*l-1,l,INT_MIN);
    DrawLine(hDC,DKGRAY,l-1,1,l-1,l/2+2,INT_MIN);
    DrawLine(hDC,DKGRAY,l-1,3*l/2-2,l-1,2*l-1,INT_MIN);
    DrawLine(hDC,DKGRAY,1,l-1,l/2+2,l-1,INT_MIN);
    DrawLine(hDC,DKGRAY,3*l/2-2,l-1,2*l-1,l-1,INT_MIN);
    if(state&16)
      {
      r=l/2+2;
      SelectObject(hDC,GetStockObject(BLACK_PEN));
      SelectObject(hDC,GetStockObject(GRAY_BRUSH));
      Ellipse(hDC,r,r,2*l-r,2*l-r);
      }
    Arrow(hDC,BLACK,BLACK, 45,3*l/2-2,  l/2  ,l/5);
    Arrow(hDC,BLACK,BLACK,135,  l/2  ,  l/2+1,l/5);
    Arrow(hDC,BLACK,BLACK,225,  l/2+1,3*l/2-1,l/5);
    Arrow(hDC,BLACK,BLACK,315,3*l/2-1,3*l/2-2,l/5);
    hClip=SelectObject(hDC,hClip);
    DeleteObject(hClip);
    EndPaint(hWnd,&pS);
    return(FALSE);
    }
  return(DefWindowProc(hWnd,msg,wParam,lParam));
  }

/*******************************************************************************
  FUNCTIONS TO CREATE APPLICATION WINDOWS AND LOAD RESOURCES
*******************************************************************************/

HWND CreatePushButton(DWORD ID,char*text)
  {
  static HWND hWnd;
  if((hWnd=CreateWindow("BUTTON",text,
      WS_CHILD|WS_CLIPSIBLINGS|WS_VISIBLE|BS_PUSHBUTTON,
      0,0,0,0,hMain,(void*)ID,hInst,NULL))==NULL)
    Abort(__LINE__,"can't create pushbutton \"%s\"\nWin error code %i",text,GetLastError());
  return(hWnd);
  }

HWND CreateRadioButton(DWORD ID,char*text,int check)
  {
  static HWND hWnd;
  if((hWnd=CreateWindow("BUTTON",text,
      WS_CHILD|WS_CLIPSIBLINGS|WS_VISIBLE|BS_RADIOBUTTON,
      0,0,0,0,hMain,(void*)ID,hInst,NULL))==NULL)
    Abort(__LINE__,"can't create radio button \"%s\"\nWin error code %i",text,GetLastError());
  SendMessage(hWnd,BM_SETCHECK,check,0);
  return(hWnd);
  }

void CreateWindows()
  {
  int h,high,w,wide;
  wide=GetSystemMetrics(SM_CXSCREEN);
  high=GetSystemMetrics(SM_CYSCREEN);
  w=(3*wide)/4;
  h=(7*high)/8;
  if((hMain=CreateWindow("MAIN","3D Viewer - use arrow keys (+Alt/Ctl) also PgUp/Dn End/Home",WS_OVERLAPPEDWINDOW|WS_CLIPCHILDREN|WS_VISIBLE,
      (wide-w)/2,(high-h)/2,w,h,NULL,LoadMenu(hInst,"MENU"),hInst,NULL))==NULL)
    Abort(__LINE__,"can't create main window\nWin error code %i",GetLastError());
  if((hMenu=GetMenu(hMain))==NULL)
    Abort(__LINE__,"can't get handle to menu\nerror code %i",GetLastError());
  EnableMenuItem(hMenu,PUSH_HELP  ,MF_BYCOMMAND|MF_GRAYED);
  if((hLogo=CreateWindow("LOGO",NULL,
      WS_CHILD|WS_CLIPSIBLINGS|WS_VISIBLE|SS_WHITERECT|WS_BORDER,
      0,0,0,0,hMain,NULL,hInst,NULL))==NULL)
    Abort(__LINE__,"can't create logo window\nWin error code %i",GetLastError());
  if((hMove=CreateWindow("MOVE",NULL,
      WS_CHILD|WS_CLIPSIBLINGS|WS_VISIBLE|SS_GRAYRECT,
      0,0,0,0,hMain,NULL,hInst,NULL))==NULL)
    Abort(__LINE__,"can't create movement control\nWin error code %i",GetLastError());
  if((hRock=CreateWindow("ROCK",NULL,
      WS_CHILD|WS_CLIPSIBLINGS|WS_VISIBLE|SS_GRAYRECT,
      0,0,0,0,hMain,NULL,hInst,NULL))==NULL)
    Abort(__LINE__,"can't create rocker control\nWin error code %i",GetLastError());
  if((hTurn=CreateWindow("TURN",NULL,
      WS_CHILD|WS_CLIPSIBLINGS|WS_VISIBLE|SS_GRAYRECT,
      0,0,0,0,hMain,NULL,hInst,NULL))==NULL)
    Abort(__LINE__,"can't create rotation control\nWin error code %i",GetLastError());
  if((hText=CreateWindow("STATIC","3D Controls",
      WS_CHILD|WS_CLIPSIBLINGS|WS_VISIBLE|SS_CENTER,
      0,0,0,0,hMain,NULL,hInst,NULL))==NULL)
    Abort(__LINE__,"can't create text\nWin error code %i",GetLastError());
  if((hPlot=CreateWindow("PLOT",NULL,
      WS_CHILD|WS_CLIPSIBLINGS|WS_VISIBLE|WS_BORDER,
      0,0,0,0,hMain,0,hInst,NULL))==NULL)
    Abort(__LINE__,"can't create plotting window\nWin error code %i",GetLastError());
  if((hStat=CreateStatusWindow(WS_CHILD|WS_CLIPSIBLINGS|WS_VISIBLE,
      "ready",hMain,0x7777))==NULL)
    Abort(__LINE__,"can't create status window\nWin error code %i",GetLastError());
  hExit=CreatePushButton(PUSH_EXIT,"Exit");
  hHelp=CreatePushButton(PUSH_HELP,"Help");
  hZall=CreatePushButton(PUSH_ZALL,"Reset");
  EnableWindow(hHelp,FALSE);
  hDlay=CreateRadioButton(PUSH_DELAY,"delayed paint",delayed_painting);
  SendMessage(hDlay,BM_SETCHECK,delayed_painting,0);
  }

typedef struct{void*data;DWORD size;}RESOURCE;

RESOURCE GetResource(char*name,char*type)
  {
  HGLOBAL rl;
  HRSRC rf;
  static RESOURCE res;
  if((rf=FindResource(hInst,name,type))==NULL)
    Abort(__LINE__,"can't find resource %s\nWin Err Code %i",name,GetLastError());
  if((rl=LoadResource(hInst,rf))==NULL)
    Abort(__LINE__,"can't load resource %s\nWin Err Code %i",name,GetLastError());
  if((res.data=LockResource(rl))==NULL)
    Abort(__LINE__,"can't lock resource %s\nWin Err Code %i",name,GetLastError());
  res.size=SizeofResource(hInst,rf);
  return(res);
  }

BITMAPINFOHEADER*LoadBitmapResource(char*name)
  {
  RESOURCE res;
  res=GetResource(name,(char*)RT_BITMAP);
  return((BITMAPINFOHEADER*)res.data);
  }

void LoadResources()
  {
  bLogo=LoadBitmapResource("LOGO");
  }

void Register(char*cName,void*wProc,HBRUSH brush,char*cursor,DWORD style)
  {
  WNDCLASS wc;
  memset(&wc,0,sizeof(WNDCLASS));
  wc.hInstance    =hInst;
  wc.hIcon        =LoadIcon(hInst,"ICON");
  wc.style        =CS_HREDRAW|CS_VREDRAW|style;
  wc.lpszClassName=cName;
  wc.lpfnWndProc  =(WNDPROC)wProc;
  wc.hbrBackground=brush;
  wc.hCursor      =LoadCursor(NULL,cursor);
  if(!RegisterClass(&wc))
    Abort(__LINE__,"can't register %s class\nWin error code %li",cName,GetLastError());
  }

void RegisterClasses()
  {
  Register("MAIN",MainProc,GetStockObject(LTGRAY_BRUSH),IDC_ARROW,0);
  Register("LOGO",LogoProc,GetStockObject(LTGRAY_BRUSH),IDC_ARROW,0);
  Register("MOVE",MoveProc,GetStockObject(LTGRAY_BRUSH),IDC_ARROW,0);
  Register("ROCK",RockProc,GetStockObject(LTGRAY_BRUSH),IDC_ARROW,0);
  Register("TURN",TurnProc,GetStockObject(LTGRAY_BRUSH),IDC_ARROW,0);
  Register("PLOT",PlotProc,GetStockObject(  NULL_BRUSH),IDC_CROSS,CS_OWNDC|CS_DBLCLKS);
  }

extern int __argc;
extern char**__argv;

int WINAPI WinMain(HINSTANCE hCurrent,HINSTANCE hPrevious,char*lCommand,int nShow)
  {
  HACCEL acc;
  MSG msg;
  hInst=hCurrent;
  pGray=CreatePen(PS_SOLID,1,DKGRAY);
  InitCommonControls();
  LoadResources();
  RegisterClasses();
  CreateWindows();
  if((pDC=GetBestPixelFormat(hPlot,0))==0)
    Abort(__LINE__,"can't find best pixel context\nWin error code %li",GetLastError());
  if((rDC=wglCreateContext(pDC))==0)
    Abort(__LINE__,"can't create OpenGL context\nWin error code %li",GetLastError());
  if(!wglMakeCurrent(pDC,rDC))
    Abort(__LINE__,"can't make OpenGL context current\nWin error code %li",GetLastError());
  PositionWindows(0,0);
  InvalidateRect(hPlot,NULL,FALSE);
  if((acc=LoadAccelerators(hInst,"FAST"))==NULL)
    Abort(__LINE__,"can't load accelerator table\nWin error code %li",GetLastError());
  SetTimer(hMain,1,250,(TIMERPROC)DelayedRePaint);
  if(__argc>1)
    {
    model=ImportModel(__argv[1]);
    SetView(0,0,0,1.,0,0,0);
    InvalidateRect(hPlot,NULL,FALSE);
    }
  if(model.nodes<=0)
    PostMessage(hMain,WM_COMMAND,PUSH_IMPORT,0);
  while(GetMessage(&msg,NULL,0,0))
    {
    if(msg.message==WM_MOUSEMOVE||msg.message==WM_NCMOUSEMOVE)
      {
      if(msg.hwnd==hExit)
        SetWindowText(hStat,"press this button to exit (alt-F4 will also work)");
      else if(msg.hwnd==hZall)
        SetWindowText(hStat,"press this button to reset view");
      else if(msg.hwnd==hLogo)
        SetWindowText(hStat,"if you find this helpful Email me at dudley.benton@gmail.com");
      else if(msg.hwnd==hMove)
        SetWindowText(hStat,"move scene left, right, up, or down (or use alt-arrows)");
      else if(msg.hwnd==hRock)
        SetWindowText(hStat,"this controls pitch and proximity (or use PgUp/Dn alt-PgUp/Dn)");
      else if(msg.hwnd==hTurn)
        SetWindowText(hStat,"this controls yaw and roll (or use arrow keys)");
      else if(msg.hwnd==hPlot)
        {
        if(model.nodes)
          WindowText(hStat,"%s, %i nodes, %i elements, %u textures",file_name,model.nodes,model.elems,textures);
        else
          SetWindowText(hStat,"right click and drag to zoom or shift right click and drag to pan");
        }
      }
    TranslateAccelerator(msg.hwnd,acc,&msg);
    TranslateMessage(&msg);
    DispatchMessage(&msg);
    }
  return(0);
  }
