/* gifread.c - Functions to read GIFs.
   Copyright (C) 1997-2019 Eddie Kohler, ekohler@gmail.com
   This file is part of the LCDF GIF library.
   The LCDF GIF library is free software. It is distributed under the GNU
   General Public License, version 2; you can copy, distribute, or alter it at
   will, as long as this notice is kept intact and this source code is made
   available. There is no warranty, express or implied. */

#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include "bitmap.h"
#include "gif89a.h"

#define GIF_MAJOR_VERSION 1
#define GIF_MINOR_VERSION 5
#define GIF_VERSION "1.5"

typedef struct
  {
  BYTE haspixel;
  BYTE r;
  BYTE g;
  BYTE b;
  DWORD pixel;
  }GIF_COLOR;

typedef struct GIF_COLORMAP
  {
  int ncol;
  int capacity;
  DWORD userflags;
  int refcount;
  GIF_COLOR*col;
  }GIF_COLORMAP;

typedef struct GIF_COMMENT
  {
  char**str;
  int*len;
  int count;
  int cap;
  }GIF_COMMENT;

typedef struct GIF_IMAGE
  {
  char*identifier;
  GIF_COMMENT*comment;
  GIF_COLORMAP*local;
  short transparent;
  WORD delay;
  BYTE disposal;
  WORD left;
  WORD top;
  WORD width;
  WORD height;
  BYTE interlace;
  BYTE**img; /* img[y][x]==image byte (x,y) */
  BYTE*image_data;
  DWORD compressed_len;
  BYTE*compressed;
  void*user_data;
  int refcount;
  }GIF_IMAGE;

typedef struct GIF_EXTENSION GIF_EXTENSION;

typedef struct GIF_STREAM
  {
  GIF_COLORMAP*global;
  BYTE background;
  WORD screen_width;
  WORD screen_height;
  long loopcount; /* -1 means no loop count */
  GIF_COMMENT*comment;
  GIF_IMAGE**images;
  int nimages;
  int imagescap;
  GIF_EXTENSION*extensions;
  unsigned errors;
  int userflags;
  int refcount;
  }GIF_STREAM;

typedef struct GIF_EXTENSION
  {
  int kind; /* negative kinds are reserved */
  char*application;
  BYTE*data;
  DWORD length;
  int position;
  GIF_STREAM*stream;
  GIF_EXTENSION*next;
  }GIF_EXTENSION;

#define GifScreenWidth(gfs) ((gfs)->screen_width)
#define GifScreenHeight(gfs) ((gfs)->screen_height)
#define GifImageCount(gfs) ((gfs)->nimages)

#define GIF_DISPOSAL_NONE 0
#define GIF_DISPOSAL_ASIS 1
#define GIF_DISPOSAL_BACKGROUND 2
#define GIF_DISPOSAL_PREVIOUS 3

#define GifImageWidth(gfi) ((gfi)->width)
#define GifImageHeight(gfi) ((gfi)->height)
#define GifImageDelay(gfi) ((gfi)->delay)
#define GifImageUserData(gfi) ((gfi)->userdata)
#define GifSetImageUserData(gfi,v) ((gfi)->userdata=v)

typedef void (*GifReadErrorHandler) (char*,int,void*);

#define GIF_COLOREQ(c1,c2) ((c1)->r==(c2)->r&&(c1)->g==(c2)->g&&(c1)->b==(c2)->b)

#define GIF_READ_COMPRESSED 1
#define GIF_READ_UNCOMPRESSED 2
#define GIF_READ_CONST_RECORD 4
#define GIF_READ_TRAILING_GARBAGE_OK 8
#define GIF_WRITE_CAREFUL_MIN_CODE_SIZE 1

#define GifReadFile(f)GifFullReadFile((f),GIF_READ_UNCOMPRESSED,0,0)
#define GifReadRecord(r)GifFullReadRecord((r),GIF_READ_UNCOMPRESSED,0,0)
#define GifCompressImage(s,i)GifFullCompressImage((s),(i),0)
#define GifWriteFile(s,f)GifFullWriteFile((s),0,(f))

#define GIF_T_STREAM 0
#define GIF_T_IMAGE 1
#define GIF_T_COLORMAP 2

typedef void (*GIF_DELETIONHOOKFUNC) (int,void*,void*);

typedef struct GIF_DELETIONHOOK GIF_DELETIONHOOK;

typedef struct GIF_DELETIONHOOK
  {
  int kind;
  GIF_DELETIONHOOKFUNC func;
  void*callback_data;
  struct GIF_DELETIONHOOK*next;
  }GIF_DELETIONHOOK;

GIF_DELETIONHOOK*all_hooks;

typedef WORD GifCode;
#define GIF_MAX_CODE_BITS 12
#define GIF_MAX_CODE 0x1000
#define GIF_MAX_BLOCK 255

#define GifNew(t)         ((t*)malloc(sizeof(t)))
#define GifNewArray(t,n)  ((t*)malloc(sizeof(t)*(n)))
#define GifReArray(p,t,n) ((p)=((t*)realloc((void*)(p),sizeof(t)*(n))))

GIF_STREAM*GifNewStream(void)
  {
  GIF_STREAM*gfs=GifNew(GIF_STREAM);
  if(!gfs)
    return(0);
  gfs->global=0;
  gfs->background=0;
  gfs->screen_width=gfs->screen_height=0;
  gfs->loopcount=-1;
  gfs->comment=0;
  gfs->images=0;
  gfs->nimages=gfs->imagescap=0;
  gfs->extensions=0;
  gfs->errors=0;
  gfs->userflags=0;
  gfs->refcount=0;
  return(gfs);
  }

GIF_IMAGE*GifNewImage(void)
  {
  GIF_IMAGE*gfi=GifNew(GIF_IMAGE);
  if(!gfi)
    return(0);
  gfi->identifier=0;
  gfi->comment=0;
  gfi->local=0;
  gfi->transparent=-1;
  gfi->disposal=GIF_DISPOSAL_NONE;
  gfi->delay=0;
  gfi->left=gfi->top=gfi->width=gfi->height=0;
  gfi->interlace=0;
  gfi->img=0;
  gfi->image_data=0;
  gfi->compressed_len=0;
  gfi->compressed=0;
  gfi->user_data=0;
  gfi->refcount=0;
  return(gfi);
  }

GIF_COLORMAP*GifNewColormap(void)
  {
  GIF_COLORMAP*gfcm=GifNew(GIF_COLORMAP);
  if(!gfcm)
    return(0);
  gfcm->ncol=0;
  gfcm->capacity=0;
  gfcm->col=0;
  gfcm->refcount=0;
  gfcm->userflags=0;
  return(gfcm);
  }

GIF_COLORMAP*GifNewFullColormap(int count,int capacity)
  {
  GIF_COLORMAP*gfcm=GifNew(GIF_COLORMAP);
  if(!gfcm||capacity<=0||count<0)
    return(0);
  if(count>capacity)
    capacity=count;
  gfcm->ncol=count;
  gfcm->capacity=capacity;
  gfcm->col=GifNewArray(GIF_COLOR,capacity);
  gfcm->refcount=0;
  gfcm->userflags=0;
  if(!gfcm->col)
    {
    free(gfcm);
    return(0);
    }
  else
    return(gfcm);
  }

GIF_COMMENT*GifNewComment(void)
  {
  GIF_COMMENT*gfcom=GifNew(GIF_COMMENT);
  if(!gfcom)
    return(0);
  gfcom->str=0;
  gfcom->len=0;
  gfcom->count=gfcom->cap=0;
  return(gfcom);
  }

char*GifCopyString(char*s)
  {
  int l;
  char*copy;
  if(!s)
    return(0);
  l=(int)strlen(s);
  copy=GifNewArray(char,l+1);
  if(!copy)
    return(0);
  memcpy(copy,s,l+1);
  return(copy);
  }

void GifDeleteExtension(GIF_EXTENSION*gfex)
  {
  if(!gfex)
    return;
  free(gfex->data);
  free(gfex->application);
  if(gfex->stream)
    {
    GIF_STREAM*gfs=gfex->stream;
    GIF_EXTENSION*prev,*trav;
    for(prev=0,trav=gfs->extensions;
         trav&&trav!=gfex;
         prev=trav,trav=trav->next)
      ;
    if(trav)
      {
      if(prev)
        prev->next=trav->next;
      else
        gfs->extensions=trav->next;
      }
    }
  free(gfex);
  }

GIF_EXTENSION*GifNewExtension(int kind,char*app_name)
  {
  GIF_EXTENSION*gfex=GifNew(GIF_EXTENSION);
  if(!gfex)
    return(0);
  gfex->kind=app_name?255:kind;
  gfex->application=GifCopyString(app_name);
  gfex->data=0;
  gfex->position=0;
  gfex->stream=0;
  gfex->next=0;
  if(!gfex->application&&app_name)
    {
    GifDeleteExtension(gfex);
    return(0);
    }
  return(gfex);
  }

int GifAddImage(GIF_STREAM*gfs,GIF_IMAGE*gfi)
  {
  if(gfs->nimages>=gfs->imagescap)
    {
    if(gfs->imagescap)
      gfs->imagescap*=2;
    else
      gfs->imagescap=2;
    GifReArray(gfs->images,GIF_IMAGE*,gfs->imagescap);
    if(!gfs->images)
      return(0);
    }
  gfs->images[gfs->nimages]=gfi;
  gfs->nimages++;
  gfi->refcount++;
  return(1);
  }

void GifDeleteComment(GIF_COMMENT*gfcom)
  {
  int i;
  if(!gfcom)
    return;
  for(i=0;i<gfcom->count;i++)
    free(gfcom->str[i]);
  free(gfcom->str);
  free(gfcom->len);
  free(gfcom);
  }

void GifDeleteColormap(GIF_COLORMAP*gfcm)
  {
  GIF_DELETIONHOOK*hook;
  if(!gfcm)
    return;
  if(--gfcm->refcount>0)
    return;
  for(hook=all_hooks;hook;hook=hook->next)
    if(hook->kind==GIF_T_COLORMAP)
      (*hook->func)(GIF_T_COLORMAP,gfcm,hook->callback_data);
  free(gfcm->col);
  free(gfcm);
  }

void GifDeleteImage(GIF_IMAGE*gfi)
  {
  GIF_DELETIONHOOK*hook;
  if(!gfi)
    return;
  if(--gfi->refcount>0)
    return;
  for(hook=all_hooks;hook;hook=hook->next)
    if(hook->kind==GIF_T_IMAGE)
      (*hook->func)(GIF_T_IMAGE,gfi,hook->callback_data);
  if(gfi->identifier)
    free(gfi->identifier);
  GifDeleteComment(gfi->comment);
  GifDeleteColormap(gfi->local);
  if(gfi->image_data)
    free((void*)gfi->image_data);
  if(gfi->img)
    free(gfi->img);
  if(gfi->compressed)
    free((void*)gfi->compressed);
  if(gfi->user_data)
    free(gfi->user_data);
  free(gfi);
  }

void GifRemoveImage(GIF_STREAM*gfs,int inum)
  {
  int j;
  if(inum<0||inum>=gfs->nimages)
    return;
  GifDeleteImage(gfs->images[inum]);
  for(j=inum;j<gfs->nimages-1;j++)
    gfs->images[j]=gfs->images[j+1];
  gfs->nimages--;
  }

int GifAddCommentTake(GIF_COMMENT*gfcom,char*x,int xlen)
  {
  if(gfcom->count>=gfcom->cap)
    {
    if(gfcom->cap)
      gfcom->cap*=2;
    else
      gfcom->cap=2;
    GifReArray(gfcom->str,char*,gfcom->cap);
    GifReArray(gfcom->len,int,gfcom->cap);
    if(!gfcom->str||!gfcom->len)
      return(0);
    }
  if(xlen<0)
    xlen=(int)strlen(x);
  gfcom->str[gfcom->count]=x;
  gfcom->len[gfcom->count]=xlen;
  gfcom->count++;
  return(1);
  }

int GifAddComment(GIF_COMMENT*gfcom,char*x,int xlen)
  {
  char*new_x;
  if(xlen<0)
    xlen=(int)strlen(x);
  new_x=GifNewArray(char,xlen);
  if(!new_x)
    return(0);
  memcpy(new_x,x,xlen);
  if(GifAddCommentTake(gfcom,new_x,xlen)==0)
    {
    free(new_x);
    return(0);
    }
  else
    return(1);
  }

int GifAddExtension(GIF_STREAM*gfs,GIF_EXTENSION*gfex,int pos)
  {
  GIF_EXTENSION*prev,*trav;
  if(gfex->stream)
    return(0);
  for(prev=0,trav=gfs->extensions;trav&&trav->position<=pos;prev=trav,trav=trav->next)
    ;
  if(prev)
    prev->next=gfex;
  else
    gfs->extensions=gfex;
  gfex->next=trav;
  return(1);
  }

int GifImageNumber(GIF_STREAM*gfs,GIF_IMAGE*gfi)
  {
  int i;
  for(i=0;i<gfs->nimages;i++)
    if(gfs->images[i]==gfi)
      return(i);
  return(-1);
  }

void GifCalculateScreenSize(GIF_STREAM*gfs,int force)
  {
  int i;
  int screen_width=0;
  int screen_height=0;
  for(i=0;i<gfs->nimages;i++)
    {
    GIF_IMAGE*gfi=gfs->images[i];
    if(screen_width<gfi->left+gfi->width)
      screen_width=gfi->left+gfi->width;
    if(screen_height<gfi->top+gfi->height)
      screen_height=gfi->top+gfi->height;
    }
  if(screen_width==0&&(gfs->screen_width==0||force))
    screen_width=640;
  if(screen_height==0&&(gfs->screen_height==0||force))
    screen_height=480;
  if(gfs->screen_width<screen_width||force)
    gfs->screen_width=screen_width;
  if(gfs->screen_height<screen_height||force)
    gfs->screen_height=screen_height;
  }

GIF_COLORMAP*GifCopyColormap(GIF_COLORMAP*src)
  {
  int i;
  GIF_COLORMAP*dest;
  if(!src)
    return(0);
  dest=GifNewFullColormap(src->ncol,src->capacity);
  if(!dest)
    return(0);
  for(i=0;i<src->ncol;i++)
    {
    dest->col[i]=src->col[i];
    dest->col[i].haspixel=0;
    }
  return(dest);
  }

void GifDeleteStream(GIF_STREAM*gfs)
  {
  GIF_EXTENSION*gfex;
  GIF_DELETIONHOOK*hook;
  int i;
  if(!gfs)
    return;
  if(--gfs->refcount>0)
    return;
  GifDeleteColormap(gfs->global);
  GifDeleteComment(gfs->comment);
  for(i=0;i<gfs->nimages;i++)
    GifDeleteImage(gfs->images[i]);
  free(gfs->images);
  gfex=gfs->extensions;
  while(gfex)
    {
    GIF_EXTENSION*next=gfex->next;
    gfex->stream=0;
    GifDeleteExtension(gfex);
    gfex=next;
    }
  for(hook=all_hooks;hook;hook=hook->next)
    if(hook->kind==GIF_T_STREAM)
      (*hook->func)(GIF_T_STREAM,gfs,hook->callback_data);
  free(gfs);
  }

GIF_STREAM*GifCopyStreamSkeleton(GIF_STREAM*gfs)
  {
  GIF_STREAM*ngfs=GifNewStream();
  ngfs->global=GifCopyColormap(gfs->global);
  ngfs->background=gfs->background;
  ngfs->screen_width=gfs->screen_width;
  ngfs->screen_height=gfs->screen_height;
  ngfs->loopcount=gfs->loopcount;
  if(gfs->global &&!ngfs->global)
    {
    GifDeleteStream(ngfs);
    return(0);
    }
  else
    return(ngfs);
  }

GIF_IMAGE*GifCopyImage(GIF_IMAGE*src)
  {
  GIF_IMAGE*dest;
  BYTE*data;
  int i;
  if(!src)
    return(0);
  dest=GifNewImage();
  if(!dest)
    return(0);
  dest->identifier=GifCopyString(src->identifier);
  if(!dest->identifier&&src->identifier)
    goto failure;
  if(src->comment)
    {
    dest->comment=GifNewComment();
    if(!dest->comment)
      goto failure;
    for(i=0;i<src->comment->count;i++)
      if(!GifAddComment(dest->comment,src->comment->str[i],src->comment->len[i]))
        goto failure;
    }
  dest->local=GifCopyColormap(src->local);
  if(!dest->local&&src->local)
    goto failure;
  dest->transparent=src->transparent;
  dest->delay=src->delay;
  dest->disposal=src->disposal;
  dest->left=src->left;
  dest->top=src->top;
  dest->width=src->width;
  dest->height=src->height;
  dest->interlace=src->interlace;
  if(src->img)
    {
    dest->img=GifNewArray(BYTE*,dest->height+1);
    dest->image_data=GifNewArray(BYTE,dest->width*dest->height);
    if(!dest->img||!dest->image_data)
      goto failure;
    for(i=0,data=dest->image_data;i<dest->height;i++)
      {
      memcpy(data,src->img[i],dest->width);
      dest->img[i]=data;
      data+=dest->width;
      }
    dest->img[dest->height]=0;
    }
  if(src->compressed)
    {
    dest->compressed=GifNewArray(BYTE,src->compressed_len);
    memcpy(dest->compressed,src->compressed,src->compressed_len);
    dest->compressed_len=src->compressed_len;
    }
  return(dest);
failure:
  GifDeleteImage(dest);
  return(0);
  }

GIF_STREAM*GifCopyStreamImages(GIF_STREAM*gfs)
  {
  GIF_STREAM*ngfs=GifCopyStreamSkeleton(gfs);
  int i;
  if(!ngfs)
    return(0);
  for(i=0;i<gfs->nimages;i++)
    {
    GIF_IMAGE*gfi=GifCopyImage(gfs->images[i]);
    if(!gfi||!GifAddImage(ngfs,gfi))
      {
      GifDeleteStream(ngfs);
      return(0);
      }
    }
  return(ngfs);
  }

void GifRemoveDeletionHook(int kind,void (*func)(int,void*,void*),void*cb)
  {
  GIF_DELETIONHOOK*hook=all_hooks,*prev=0;
  while(hook)
    {
    if(hook->kind==kind&&hook->func==func
       &&hook->callback_data==cb)
      {
      if(prev)
        prev->next=hook->next;
      else
        all_hooks=hook->next;
      free(hook);
      return;
      }
    prev=hook;
    hook=hook->next;
    }
  }

int GifAddDeletionHook(int kind,void (*func)(int,void*,void*),void*cb)
  {
  GIF_DELETIONHOOK*hook=GifNew(GIF_DELETIONHOOK);
  if(!hook)
    return(0);
  GifRemoveDeletionHook(kind,func,cb);
  hook->kind=kind;
  hook->func=func;
  hook->callback_data=cb;
  hook->next=all_hooks;
  all_hooks=hook;
  return(1);
  }

int GifColorEq(GIF_COLOR*c1,GIF_COLOR*c2)
  {
  return(GIF_COLOREQ(c1,c2));
  }

int GifFindColor(GIF_COLORMAP*gfcm,GIF_COLOR*c)
  {
  int i;
  for(i=0;i<gfcm->ncol;i++)
    if(GIF_COLOREQ(&gfcm->col[i],c))
      return(i);
  return(-1);
  }

int GifAddColor(GIF_COLORMAP*gfcm,GIF_COLOR*c,int look_from)
  {
  int i;
  if(look_from>=0)
    for(i=look_from;i<gfcm->ncol;i++)
      if(GIF_COLOREQ(&gfcm->col[i],c))
        return(i);
  if(gfcm->ncol>=gfcm->capacity)
    {
    gfcm->capacity*=2;
    GifReArray(gfcm->col,GIF_COLOR,gfcm->capacity);
    if(gfcm->col==0)
      return(-1);
    }
  i=gfcm->ncol;
  gfcm->ncol++;
  gfcm->col[i]=*c;
  return(i);
  }

GIF_IMAGE*GifGetImage(GIF_STREAM*gfs,int imagenumber)
  {
  if(imagenumber>=0&&imagenumber<gfs->nimages)
    return(gfs->images[imagenumber]);
  else
    return(0);
  }

GIF_IMAGE*GifGetNamedImage(GIF_STREAM*gfs,char*name)
  {
  int i;
  if(!name)
    return(gfs->nimages?gfs->images[0]:0);
  for(i=0;i<gfs->nimages;i++)
    if(gfs->images[i]->identifier &&
        strcmp(gfs->images[i]->identifier,name)==0)
      return(gfs->images[i]);
  return(0);
  }

GIF_EXTENSION*GifGetExtension(GIF_STREAM*gfs,int id,GIF_EXTENSION*search_from)
  {
  if(!search_from)
    search_from=gfs->extensions;
  while(search_from)
    {
    if(search_from->kind==id)
      return(search_from);
    search_from=search_from->next;
    }
  return(0);
  }

void GifReleaseCompressedImage(GIF_IMAGE*gfi)
  {
  if(gfi->compressed)
    free(gfi->compressed);
  gfi->compressed=0;
  gfi->compressed_len=0;
  }

void GifReleaseUncompressedImage(GIF_IMAGE*gfi)
  {
  if(gfi->img)
    free(gfi->img);
  if(gfi->image_data)
    free(gfi->image_data);
  gfi->img=0;
  gfi->image_data=0;
  }

int GifClipImage(GIF_IMAGE*gfi,int left,int top,int width,int height)
  {
  int new_width=gfi->width,new_height=gfi->height;
  int y;
  if(!gfi->img)
    return(0);
  if(gfi->left<left)
    {
    int shift=left-gfi->left;
    for(y=0;y<gfi->height;y++)
      gfi->img[y]+=shift;
    gfi->left+=shift;
    new_width-=shift;
    }
  if(gfi->top<top)
    {
    int shift=top-gfi->top;
    for(y=gfi->height-1;y>=shift;y++)
      gfi->img[y-shift]=gfi->img[y];
    gfi->top+=shift;
    new_height-=shift;
    }
  if(gfi->left+new_width>=width)
    new_width=width-gfi->left;
  if(gfi->top+new_height>=height)
    new_height=height-gfi->top;
  if(new_width<0)
    new_width=0;
  if(new_height<0)
    new_height=0;
  gfi->width=new_width;
  gfi->height=new_height;
  return(1);
  }

int GifInterlaceLine(int line,int height)
  {
  height--;
  if(line>height/2)
    return(line*2-(height|1));
  else if(line>height/4)
    return(line*4-((height&~1)|2));
  else if(line>height/8)
    return(line*8-((height&~3)|4));
  else
    return(line*8);
  }

int GifSetUncompressedImage(GIF_IMAGE*gfi,BYTE*image_data,int data_interlaced)
  {
  int i;
  int width=gfi->width;
  int height=gfi->height;
  BYTE**img;
  GifReleaseUncompressedImage(gfi);
  if(!image_data)
    return(0);
  img=GifNewArray(BYTE*,height+1);
  if(!img)
    return(0);
  if(data_interlaced)
    for(i=0;i<height;i++)
      img[GifInterlaceLine(i,height)]=image_data+width*i;
  else
    for(i=0;i<height;i++)
      img[i]=image_data+width*i;
  img[height]=0;
  gfi->img=img;
  gfi->image_data=image_data;
  return(1);
  }

int GifCreateUncompressedImage(GIF_IMAGE*gfi)
  {
  BYTE*data=GifNewArray(BYTE,gfi->width*gfi->height);
  return(GifSetUncompressedImage(gfi,data,gfi->interlace));
  }

typedef struct
  {
  GIF_STREAM*stream;
  GifCode*prefix;
  BYTE*suffix;
  WORD*length;
  WORD width;
  WORD height;
  BYTE*image;
  BYTE*maximage;
  unsigned decodepos;
  GifReadErrorHandler handler;
  void*handler_thunk;
  }GIF_CONTEXT1;

typedef struct GIF_READER
  {
  FILE*f;
  BYTE*v;
  DWORD w;
  DWORD length;
  int is_record;
  int is_eoi;
  BYTE(*byte_getter)(struct GIF_READER*);
  void (*block_getter)(BYTE*,DWORD,struct GIF_READER*);
  DWORD(*offseter)(struct GIF_READER*);
  int (*eofer)(struct GIF_READER*);
  }GIF_READER;

#define gifgetc(grr)((char)(*grr->byte_getter)(grr))
#define gifgetbyte(grr)((*grr->byte_getter)(grr))
#define gifgetblock(ptr,size,grr)((*grr->block_getter)(ptr,size,grr))
#define gifgetoffset(grr)((*grr->offseter)(grr))
#define gifeof(grr)((*grr->eofer)(grr))

WORD GifGetWord(GIF_READER*grr)
  {
  BYTE one=gifgetbyte(grr);
  BYTE two=gifgetbyte(grr);
  return(one|(two<<8));
  }

BYTE GifFileByteGetter(GIF_READER*grr)
  {
  int i=getc(grr->f);
  return(i==EOF?0:(BYTE)i);
  }

void GifFileBlockGetter(BYTE*p,DWORD s,GIF_READER*grr)
  {
  fread(p,1,s,grr->f);
  }

DWORD GifFileOffseter(GIF_READER*grr)
  {
  return(ftell(grr->f));
  }

int GifFileWriteEof(GIF_READER*grr)
  {
  int c=getc(grr->f);
  if(c==EOF)
    return(1);
  else
    {
    ungetc(c,grr->f);
    return(0);
    }
  }

BYTE GifRecordByteGetter(GIF_READER*grr)
  {
  return(grr->w?(grr->w--,*grr->v++):0);
  }

void GifRecordBlockGetter(BYTE*p,DWORD s,GIF_READER*grr)
  {
  if(s>grr->w)
    s=grr->w;
  memcpy(p,grr->v,s);
  grr->w-=s,grr->v+=s;
  }

DWORD GifRecordOffseter(GIF_READER*grr)
  {
  return(grr->length-grr->w);
  }

int GifRecordEofer(GIF_READER*grr)
  {
  return(grr->w==0);
  }

void GifMakeDataReader(GIF_READER*grr,BYTE*data,DWORD length)
  {
  grr->v=data;
  grr->length=length;
  grr->w=length;
  grr->is_record=1;
  grr->byte_getter=GifRecordByteGetter;
  grr->block_getter=GifRecordBlockGetter;
  grr->offseter=GifRecordOffseter;
  grr->eofer=GifRecordEofer;
  }

void GifReadError(GIF_CONTEXT1*gfc,char*error)
  {
  gfc->stream->errors++;
  if(gfc->handler)
    gfc->handler(error,gfc->stream->nimages,gfc->handler_thunk);
  }

BYTE GifGetCode(GIF_CONTEXT1*gfc,GifCode code)
  {
  BYTE*suffixes=gfc->suffix;
  GifCode*prefixes=gfc->prefix;
  BYTE*ptr;
  int lastsuffix;
  WORD codelength=gfc->length[code];
  gfc->decodepos+=codelength;
  ptr=gfc->image+gfc->decodepos;
  if(ptr>gfc->maximage||!codelength)
    {
    GifReadError(gfc,(!codelength?"bad code":"too much image data"));
    gfc->decodepos=(unsigned int)(gfc->maximage-gfc->image);
    return(0);
    }
  do
    {
    lastsuffix=suffixes[code];
    *--ptr=lastsuffix;
    code=prefixes[code];
    }
  while(--codelength>0);
  return(lastsuffix);
  }

int GifReadImageBlock(GIF_READER*grr,BYTE*buffer,int*bit_pos_store,int*bit_len_store,int bits_needed)
  {
  int bit_position=*bit_pos_store;
  int bit_length=*bit_len_store;
  BYTE block_len;
  while(bit_position+bits_needed>bit_length)
    {
    if(bit_position>=8)
      {
      int i=bit_position/8;
      buffer[0]=buffer[i];
      buffer[1]=buffer[i+1];
      bit_position-=i*8;
      bit_length-=i*8;
      }
    block_len=gifgetbyte(grr);
    if(block_len==0)
      return(0);
    gifgetblock(buffer+bit_length/8,block_len,grr);
    bit_length+=block_len*8;
    }
  *bit_pos_store=bit_position;
  *bit_len_store=bit_length;
  return(1);
  }

void GifReadImageData(GIF_CONTEXT1*gfc,GIF_READER*grr)
  {
 /* we need a bit more than GIF_MAX_BLOCK in case a single code is split across blocks */
  BYTE buffer[GIF_MAX_BLOCK+5];
  int i;
  DWORD accum;

  int bit_position;
  int bit_length;

  GifCode code;
  GifCode old_code;
  GifCode clear_code;
  GifCode eoi_code;
  GifCode next_code;
#define CUR_BUMP_CODE (1<<bits_needed)
#define CUR_CODE_MASK ((1<<bits_needed)-1)

  int min_code_size;
  int bits_needed;
  gfc->decodepos=0;

  min_code_size=gifgetbyte(grr);
  if(min_code_size>=GIF_MAX_CODE_BITS)
    {
    GifReadError(gfc,"min_code_size too big");
    min_code_size=GIF_MAX_CODE_BITS-1;
    }
  else if(min_code_size<2)
    {
    GifReadError(gfc,"min_code_size too small");
    min_code_size=2;
    }
  clear_code=1<<min_code_size;
  for(code=0;code<clear_code;code++)
    {
    gfc->prefix[code]=49428;
    gfc->suffix[code]=(BYTE)code;
    gfc->length[code]=1;
    }
  eoi_code=clear_code+1;

  next_code=eoi_code;
  bits_needed=min_code_size+1;

  code=clear_code;

  bit_length=bit_position=0;
 /* Thus the `Read in the next data block.' code below will be invoked on the first time through: exactly right! */

  while(1)
    {

    old_code=code;

   /* GET A CODE INTO THE `code' VARIABLE.
*
*9.Dec.1998-Rather than maintain a byte pointer and a bit offset into the current byte (and the processing associated with
*that),we maintain one number: the offset,in bits,from the beginning of `buffer'. This much cleaner choice was
*inspired by Patrick J. Naughton <naughton@wind.sun.com>'s GIF-reading code,which does the same thing. His code
*distributed as part of XV in xvgif.c. */

    if(bit_position+bits_needed>bit_length)
     /* Read in the next data block. */
      if(!GifReadImageBlock(grr,buffer,&bit_position,&bit_length,bits_needed))
        goto zero_length_block;

    i=bit_position/8;
    accum=buffer[i]+(buffer[i+1]<<8);
    if(bits_needed>=8)
      accum|=(buffer[i+2])<<16;
    code=(GifCode)((accum>>(bit_position%8))&CUR_CODE_MASK);
    bit_position+=bits_needed;

   /* CHECK FOR SPECIAL OR BAD CODES: clear_code,eoi_code,or a code that is too large. */
    if(code==clear_code)
      {
      bits_needed=min_code_size+1;
      next_code=eoi_code;
      continue;

      }
    else if(code==eoi_code)
      break;

    else if(code>next_code&&next_code)
      {
     /* code>next_code: a (hopefully recoverable)error.
*
*Bug fix,5/27: Do this even if old_code==clear_code,and set code to 0 to prevent errors later. (If we didn't zero
*code,we'd later set old_code=code;then we had old_code>=next_code;so the prefixes array got all screwed up!) */
      GifReadError(gfc,"unexpected code");
      code=0;
      }

   /* PROCESS THE CURRENT CODE and define the next code. If no meaningful next code should be defined,then we have set
*next_code to either `eoi_code' or `clear_code'--so we'll store useless prefix/suffix data in a useless place. */

   /* *First,* set up the prefix and length for the next code (in case code==next_code). */
    gfc->prefix[next_code]=old_code;
    gfc->length[next_code]=gfc->length[old_code]+1;

   /* Use GifGetCode to process code. It's nice that it returns the first pixel in code: that's what we need. */
    gfc->suffix[next_code]=GifGetCode(gfc,code);

   /* Special processing if code==next_code: we didn't know code's final suffix when we called GifGetCode,but we do now. */
    if(code==next_code)
      gfc->image[gfc->decodepos-1]=gfc->suffix[next_code];

   /* Increment next_code except for the `clear_code' special case (that's when we're reading at the end of a GIF) */
    if(next_code!=clear_code)
      {
      next_code++;
      if(next_code==CUR_BUMP_CODE)
        {
        if(bits_needed<GIF_MAX_CODE_BITS)
          bits_needed++;
        else
          next_code=clear_code;
        }
      }

    }

 /* read blocks until zero-length reached. */
  i=gifgetbyte(grr);
  while(i>0)
    {
    gifgetblock(buffer,i,grr);
    i=gifgetbyte(grr);
    }

 /* zero-length block reached. */
zero_length_block:

  if(gfc->image+gfc->decodepos<gfc->maximage)
    GifReadError(gfc,"not enough image data for image size");
  else if(gfc->image+gfc->decodepos>gfc->maximage)
    GifReadError(gfc,"too much image data for image size");
  }

GIF_COLORMAP*GifReadColorTable(int size,GIF_READER*grr)
  {
  GIF_COLORMAP*gfcm=GifNewFullColormap(size,size);
  GIF_COLOR*c;
  if(!gfcm)
    return(0);

  for(c=gfcm->col;size;size--,c++)
    {
    c->r=gifgetbyte(grr);
    c->g=gifgetbyte(grr);
    c->b=gifgetbyte(grr);
    c->haspixel=0;
    }

  return(gfcm);
  }

int GifReadLogicalScreenDescriptor(GIF_STREAM*gfs,GIF_READER*grr)
  {
  BYTE packed;
  gfs->screen_width=GifGetWord(grr);
  gfs->screen_height=GifGetWord(grr);
  packed=gifgetbyte(grr);
  gfs->background=gifgetbyte(grr);
  gifgetbyte(grr);
  if(packed&0x80)
    {
    int ncol=1<<((packed&0x07)+1);
    gfs->global=GifReadColorTable(ncol,grr);
    if(!gfs->global)
      return(0);
    gfs->global->refcount=1;
    }
  return(1);
  }

int GifReadCompressedImage(GIF_IMAGE*gfi,GIF_READER*grr,int read_flags)
  {
  if(grr->is_record)
    {
    BYTE*first=grr->v;
    DWORD pos;
    pos=1;
    while(pos<grr->w)
      {
      int amt=grr->v[pos];
      pos+=amt+1;
      if(amt==0)
        break;
      }
    if(pos>grr->w)
      pos=grr->w;
    gfi->compressed_len=pos;
    if(read_flags&GIF_READ_CONST_RECORD)
      {
      gfi->compressed=(BYTE*)first;
      }
    else
      {
      gfi->compressed=GifNewArray(BYTE,gfi->compressed_len);
      if(!gfi->compressed)
        return(0);
      memcpy(gfi->compressed,first,gfi->compressed_len);
      }
    grr->v+=pos;
    grr->w-=pos;
    }
  else
    {
    DWORD comp_cap=1024;
    DWORD comp_len;
    BYTE*comp=GifNewArray(BYTE,comp_cap);
    int i;
    if(!comp)
      return(0);
    i=gifgetbyte(grr);
    comp[0]=i;
    comp_len=1;
    i=gifgetbyte(grr);
    while(i>0)
      {
      if(comp_len+i+2>comp_cap)
        {
        comp_cap*=2;
        GifReArray(comp,BYTE,comp_cap);
        if(!comp)
          return(0);
        }
      comp[comp_len]=i;
      gifgetblock(comp+comp_len+1,i,grr);
      comp_len+=i+1;
      i=gifgetbyte(grr);
      }
    comp[comp_len++]=0;
    gfi->compressed=comp;
    gfi->compressed_len=comp_len;
    }
  return(1);
  }

int GifUncompressImage(GIF_CONTEXT1*gfc,GIF_IMAGE*gfi,GIF_READER*grr)
  {
  if(!GifCreateUncompressedImage(gfi))
    return(0);
  gfc->width=gfi->width;
  gfc->height=gfi->height;
  gfc->image=gfi->image_data;
  gfc->maximage=gfi->image_data+gfi->width*gfi->height;
  GifReadImageData(gfc,grr);
  return(1);
  }

int GifFullUncompressImage(GIF_IMAGE*gfi,GifReadErrorHandler h,void*hthunk)
  {
  GIF_CONTEXT1 gfc;
  GIF_STREAM fake_gfs;
  GIF_READER grr;
  int ok=0;
  if(gfi->img)
    return(1);
  if(gfi->image_data)
    return(0);
  fake_gfs.errors=0;
  gfc.stream=&fake_gfs;
  gfc.prefix=GifNewArray(GifCode,GIF_MAX_CODE);
  gfc.suffix=GifNewArray(BYTE,GIF_MAX_CODE);
  gfc.length=GifNewArray(WORD,GIF_MAX_CODE);
  gfc.handler=h;
  gfc.handler_thunk=hthunk;
  if(gfi&&gfc.prefix&&gfc.suffix&&gfc.length&&gfi->compressed)
    {
    GifMakeDataReader(&grr,gfi->compressed,gfi->compressed_len);
    ok=GifUncompressImage(&gfc,gfi,&grr);
    }
  free(gfc.prefix);
  free(gfc.suffix);
  free(gfc.length);
  return(ok&&!fake_gfs.errors);
  }

int GifReadImage(GIF_READER*grr,GIF_CONTEXT1*gfc,GIF_IMAGE*gfi,int read_flags)
  {
  BYTE packed;
  gfi->left=GifGetWord(grr);
  gfi->top=GifGetWord(grr);
  gfi->width=GifGetWord(grr);
  gfi->height=GifGetWord(grr);
  packed=gifgetbyte(grr);

  if(packed&0x80)
   { /* have a local color table */
    int ncol=1<<((packed&0x07)+1);
    gfi->local=GifReadColorTable(ncol,grr);
    if(!gfi->local)
      return(0);
    gfi->local->refcount=1;
    }

  gfi->interlace=(packed&0x40)!=0;

 /* Keep the compressed data if asked */
  if(read_flags&GIF_READ_COMPRESSED)
    {
    if(!GifReadCompressedImage(gfi,grr,read_flags))
      return(0);
    if(read_flags&GIF_READ_UNCOMPRESSED)
      {
      GIF_READER new_grr;
      GifMakeDataReader(&new_grr,gfi->compressed,gfi->compressed_len);
      if(!GifUncompressImage(gfc,gfi,&new_grr))
        return(0);
      }
    }
  else if(read_flags&GIF_READ_UNCOMPRESSED)
    {
    if(!GifUncompressImage(gfc,gfi,grr))
      return(0);
    }
  else
    {
   /* skip over the image */
    BYTE buffer[GIF_MAX_BLOCK];
    int i=gifgetbyte(grr);
    while(i>0)
      {
      gifgetblock(buffer,i,grr);
      i=gifgetbyte(grr);
      }
    }

  return(1);
  }

void GifReadGraphicControlExtension(GIF_CONTEXT1*gfc,GIF_IMAGE*gfi,GIF_READER*grr)
  {
  BYTE len;
  BYTE crap[GIF_MAX_BLOCK];
  len=gifgetbyte(grr);

  if(len==4)
    {
    BYTE packed=gifgetbyte(grr);
    gfi->disposal=(packed>>2)&0x07;
    gfi->delay=GifGetWord(grr);
    gfi->transparent=gifgetbyte(grr);
    if(!(packed&0x01))
      gfi->transparent=-1;
    len-=4;
    }

  if(len>0)
    {
    GifReadError(gfc,"odd graphic extension format");
    gifgetblock(crap,len,grr);
    }

  len=gifgetbyte(grr);
  while(len>0)
    {
    GifReadError(gfc,"odd graphic extension format");
    gifgetblock(crap,len,grr);
    len=gifgetbyte(grr);
    }
  }

char*last_name;

char*GifSkipData(char*data,int*store_len,GIF_READER*grr)
  {
  BYTE len=gifgetbyte(grr);
  int total_len=0;
  while(len>0)
    {
    GifReArray(data,char,total_len+len+1);
    if(!data)
      return(0);
    gifgetblock((BYTE*)data,len,grr);
    total_len+=len;
    data[total_len]=0;
    len=gifgetbyte(grr);
    }
  if(store_len)
    *store_len=total_len;
  return(data);
  }

int GifReadUnknownExtension(GIF_STREAM*gfs,int kind,char*app_name,int position,GIF_READER*grr)
  {
  BYTE block_len=gifgetbyte(grr);
  BYTE*data=0;
  BYTE data_len=0;
  GIF_EXTENSION*gfex=0;
  while(block_len>0)
    {
    if(data)
      GifReArray(data,BYTE,data_len+block_len+1);
    else
      data=GifNewArray(BYTE,block_len+1);
    if(!data)
      goto done;
    gifgetblock(data+data_len,block_len,grr);
    data_len+=block_len;
    block_len=gifgetbyte(grr);
    }

  if(data)
    gfex=GifNewExtension(kind,app_name);
  if(gfex)
    {
    gfex->data=data;
    gfex->length=data_len;
    data[data_len]=0;
    GifAddExtension(gfs,gfex,position);
    }

done:
  if(!gfex)
    free(data);
  while(block_len>0)
    {
    BYTE buffer[GIF_MAX_BLOCK];
    gifgetblock(buffer,block_len,grr);
    block_len=gifgetbyte(grr);
    }
  return(gfex!=0);
  }

int GifReadApplicationExtension(GIF_CONTEXT1*gfc,int position,GIF_READER*grr)
  {
  GIF_STREAM*gfs=gfc->stream;
  BYTE buffer[GIF_MAX_BLOCK+1];
  BYTE len=gifgetbyte(grr);
  gifgetblock(buffer,len,grr);

 /* Read the Netscape loop extension. */
  if(len==11&&memcmp(buffer,"NETSCAPE2.0",11)==0)
    {

    len=gifgetbyte(grr);
    if(len==3)
      {
      gifgetbyte(grr); /* throw away the 1 */
      gfs->loopcount=GifGetWord(grr);
      len=gifgetbyte(grr);
      if(len)
        GifReadError(gfc,"bad loop extension");
      }
    else
      GifReadError(gfc,"bad loop extension");

    while(len>0)
      {
      gifgetblock(buffer,len,grr);
      len=gifgetbyte(grr);
      }
    return(1);

    }
  else
    {
    buffer[len]=0;
    return(GifReadUnknownExtension(gfs,0xFF,(char*)buffer,position,grr));
    }
  }

int GifReadCommentExtension(GIF_IMAGE*gfi,GIF_READER*grr)
  {
  int len;
  GIF_COMMENT*gfcom=gfi->comment;
  char*m=GifSkipData(0,&len,grr);
  if(m)
    {
    if(!gfcom)
      gfcom=gfi->comment=GifNewComment();
    if(!gfcom||!GifAddCommentTake(gfcom,m,len))
      return(0);
    }
  return(1);
  }

GIF_STREAM*GifReadStream(GIF_READER*grr,int read_flags,GifReadErrorHandler handler,void*handler_thunk)
  {
  GIF_STREAM*gfs;
  GIF_IMAGE*gfi;
  GIF_IMAGE*new_gfi;
  GIF_CONTEXT1 gfc;
  int extension_position=0;
  int unknown_block_type=0;

  if(gifgetc(grr)!='G'||gifgetc(grr)!='I'||gifgetc(grr)!='F')
    return(0);

  gifgetc(grr);
  gifgetc(grr);
  gifgetc(grr);

  gfs=GifNewStream();
  gfi=GifNewImage();

  gfc.stream=gfs;
  gfc.prefix=GifNewArray(GifCode,GIF_MAX_CODE);
  gfc.suffix=GifNewArray(BYTE,GIF_MAX_CODE);
  gfc.length=GifNewArray(WORD,GIF_MAX_CODE);
  gfc.handler=handler;
  gfc.handler_thunk=handler_thunk;

  if(!gfs||!gfi||!gfc.prefix||!gfc.suffix||!gfc.length)
    goto done;

  if(!GifReadLogicalScreenDescriptor(gfs,grr))
    goto done;

  while(!gifeof(grr))
    {
    BYTE block=gifgetbyte(grr);
    switch (block)
      {
      case',': /* image block */
        gfi->identifier=last_name;
        last_name=0;
        if(!GifReadImage(grr,&gfc,gfi,read_flags)||!GifAddImage(gfs,gfi))
          {
          GifDeleteImage(gfi);
          goto done;
          }
        new_gfi=GifNewImage();
        if(!new_gfi)
          goto done;
        gfi=new_gfi;
        extension_position++;
        break;
      case';': /* terminator */
        goto done;
      case'!': /* extension */
        block=gifgetbyte(grr);
        switch (block)
          {
          case 0xF9:
            GifReadGraphicControlExtension(&gfc,gfi,grr);
            break;
          case 0xCE:
            last_name=GifSkipData(last_name,0,grr);
            break;
          case 0xFE:
            if(!GifReadCommentExtension(gfi,grr))
              goto done;
            break;
          case 0xFF:
            GifReadApplicationExtension(&gfc,extension_position,grr);
            break;
          default:
            GifReadUnknownExtension(gfs,block,0,extension_position,grr);
            break;
            }
        break;
      default:
        if(!unknown_block_type)
          {
          char buf[256];
          sprintf(buf,"unknown block type %d at file offset %d",block,gifgetoffset(grr)-1);
          GifReadError(&gfc,buf);
          unknown_block_type=1;
          }
        break;
      }
    }

done:
 /* Move comments after last image into stream. */
  gfs->comment=gfi->comment;
  gfi->comment=0;

  GifDeleteImage(gfi);
  if(last_name)
    free(last_name);
  if(gfc.prefix)
    free(gfc.prefix);
  if(gfc.suffix)
    free(gfc.suffix);
  if(gfc.length)
    free(gfc.length);

  if(gfs->errors==0&&!(read_flags&GIF_READ_TRAILING_GARBAGE_OK)&&!grr->eofer(grr))
    {
    GifReadError(&gfc,"trailing garbage after GIF ignored");
   /* but clear error count,since the GIF itself was all right */
    gfs->errors=0;
    }

  return(gfs);
  }

GIF_STREAM*GifFullReadFile(FILE*f,int read_flags,GifReadErrorHandler h,void*hthunk)
  {
  GIF_READER grr;
  if(!f)
    return(0);
  grr.f=f;
  grr.is_record=0;
  grr.byte_getter=GifFileByteGetter;
  grr.block_getter=GifFileBlockGetter;
  grr.offseter=GifFileOffseter;
  grr.eofer=GifFileWriteEof;
  return(GifReadStream(&grr,read_flags,h,hthunk));
  }

GIF_STREAM*GifFullReadRecord(GIF_RECORD*gifrec,int read_flags,GifReadErrorHandler h,void*hthunk)
  {
  GIF_READER grr;
  if(!gifrec)
    return(0);
  GifMakeDataReader(&grr,gifrec->data,gifrec->length);
  if(read_flags&GIF_READ_CONST_RECORD)
    read_flags|=GIF_READ_COMPRESSED;
  return(GifReadStream(&grr,read_flags,h,hthunk));
  }
#undef GifReadFile
#undef GifReadRecord

GIF_STREAM*GifReadFile(FILE*f)
  {
  return(GifFullReadFile(f,GIF_READ_UNCOMPRESSED,0,0));
  }

GIF_STREAM*GifReadRecord(GIF_RECORD*gifrec)
  {
  return(GifFullReadRecord(gifrec,GIF_READ_UNCOMPRESSED,0,0));
  }

int gif_write_flags;
#define WRITE_BUFFER_SIZE 255
#define NODES_SIZE GIF_MAX_CODE
#define LINKS_SIZE GIF_MAX_CODE

/* 1.Aug.1999-Removed code hashing in favor of an adaptive tree strategy
   based on Whirlgif-3.04,written by Hans Dinsen-Hansen <dino@danbbs.dk>. Mr.
   Dinsen-Hansen brought the adaptive tree strategy to my attention and argued
   at length that it was better than hashing. In fact,he was right: it runs a
   lot faster. However,it does NOT create "better" results in any way.

   Each code is represented by a Node. The Nodes form a tree with variable
   fan-out--up to `clear_code' children per Node. There are two kinds of
   Node,TABLE and LINKS. In a TABLE node,the children are stored in a table
   indexed by suffix--thus,it's very efficient to determine a given child.
   In a LINKS node,the existent children are stored in a linked list. This is
   slightly slower to access. When a LINKS node gets more than
   `MAX_LINKS_TYPE-1' children,it is converted to a TABLE node. (This is why
   it's an adaptive tree.)

   Problems with this implementation: MAX_LINKS_TYPE is fixed,so GIFs with
   very small numbers of colors (2,4,8)won't get the speed benefits of
   TABLE nodes. */

#define TABLE_TYPE 0
#define LINKS_TYPE 1
#define MAX_LINKS_TYPE 5

typedef struct GIF_NODE
  {
  GifCode code;
  BYTE type;
  BYTE suffix;
  struct GIF_NODE*sibling;
  union
    {
    struct GIF_NODE*s;
    struct GIF_NODE**m;
    }child;
  }GIF_NODE;

typedef struct GIF_CONTEXT2
  {
  GIF_NODE*nodes;
  int nodes_pos;
  GIF_NODE**links;
  int links_pos;
  }GIF_CONTEXT2;

typedef struct GIF_WRITER
  {
  FILE*f;
  BYTE*v;
  DWORD pos;
  DWORD cap;
  int flags;
  int global_size;
  int local_size;
  void (*byte_putter)(BYTE,struct GIF_WRITER*);
  void (*block_putter)(BYTE*,WORD,struct GIF_WRITER*);
  }GIF_WRITER;

#define gifputbyte(b,grr)((*grr->byte_putter)(b,grr))
#define gifputblock(b,l,grr)((*grr->block_putter)(b,l,grr))

void GifPutWord(WORD uns,GIF_WRITER*grr)
  {
  gifputbyte(uns&0xFF,grr);
  gifputbyte(uns>>8,grr);
  }

void GifFileBytePutter(BYTE b,GIF_WRITER*grr)
  {
  fputc(b,grr->f);
  }

void GifFileBlockPutter(BYTE*block,WORD size,GIF_WRITER*grr)
  {
  fwrite(block,size,1,grr->f);
  }

void GifMemoryBytePutter(BYTE b,GIF_WRITER*grr)
  {
  if(grr->pos>=grr->cap)
    {
    grr->cap*=2;
    GifReArray(grr->v,BYTE,grr->cap);
    }
  if(grr->v)
    {
    grr->v[grr->pos]=b;
    grr->pos++;
    }
  }

void GifMemoryBlockPutter(BYTE*data,WORD len,GIF_WRITER*grr)
  {
  if(grr->pos+len>=grr->cap)
    {
    grr->cap*=2;
    GifReArray(grr->v,BYTE,grr->cap);
    }
  if(grr->v)
    {
    memcpy(grr->v+grr->pos,data,len);
    grr->pos+=len;
    }
  }

void GifChangeNodeToTable(GIF_CONTEXT2*gfc,GIF_NODE*work_node,GIF_NODE*next_node,GifCode clear_code)
  {
  GifCode c;
  GIF_NODE**table=&gfc->links[gfc->links_pos];
  GIF_NODE*n;
  gfc->links_pos+=clear_code;
  for(c=0;c<clear_code;c++)
    table[c]=0;
  table[next_node->suffix]=next_node;
  for(n=work_node->child.s;n;n=n->sibling)
    table[n->suffix]=n;
  work_node->type=TABLE_TYPE;
  work_node->child.m=table;
  }

void GifWriteCompressedData(BYTE**img,WORD width,WORD height,BYTE min_code_bits,GIF_CONTEXT2*gfc,GIF_WRITER*grr)
  {
  BYTE buffer[WRITE_BUFFER_SIZE];
  BYTE*buf;
  WORD xleft;
  BYTE*imageline;
  DWORD leftover;
  BYTE bits_left_over;
  GIF_NODE*work_node;
  GIF_NODE*next_node;
  GifCode next_code;
  GifCode output_code;
  GifCode clear_code;
  GifCode eoi_code;
#undef CUR_BUMP_CODE
#define CUR_BUMP_CODE (1<<cur_code_bits)
  BYTE suffix;
  BYTE cur_code_bits;
 /* Here we go! */
  gifputbyte(min_code_bits,grr);
  clear_code=1<<min_code_bits;
  eoi_code=clear_code+1;

  cur_code_bits=min_code_bits+1;
 /* next_code set by first runthrough of output clear_code */

  work_node=0;
  output_code=clear_code;
 /* Because output_code is clear_code,we'll initialize next_code,et al. below. */

  bits_left_over=0;
  leftover=0;
  buf=buffer;
  xleft=width;
  imageline=img[0];

  while(1)
    {
    leftover|=output_code<<bits_left_over;
    bits_left_over+=cur_code_bits;
    while(bits_left_over>=8)
      {
      *buf++=(BYTE)leftover&0xFF;
      leftover=(leftover>>8)&0x00FFFFFF;
      bits_left_over-=8;
      if(buf==buffer+WRITE_BUFFER_SIZE)
        {
        gifputbyte(WRITE_BUFFER_SIZE,grr);
        gifputblock(buffer,WRITE_BUFFER_SIZE,grr);
        buf=buffer;
        }
      }

    if(output_code==clear_code)
      {
      GifCode c;
      cur_code_bits=min_code_bits+1;
      next_code=eoi_code+1;
      gfc->nodes_pos=clear_code;
      gfc->links_pos=0;
      for(c=0;c<clear_code;c++)
        {
        gfc->nodes[c].code=c;
        gfc->nodes[c].type=LINKS_TYPE;
        gfc->nodes[c].suffix=(BYTE)c;
        gfc->nodes[c].child.s=0;
        }
      }
    else if(next_code>CUR_BUMP_CODE)
      {
     /* bump up compression size */
      if(cur_code_bits==GIF_MAX_CODE_BITS)
        {
        output_code=clear_code;
        continue;
        }
      else
        cur_code_bits++;

      }
    else if(output_code==eoi_code)
      break;

   /* If height is 0--no more pixels to write--we output work_node next time around. */
    while(height!=0)
      {
      suffix=*imageline;
      if(suffix>=clear_code)
       /* should not happen unless GIF_WRITE_CAREFUL_MIN_CODE_BITS */
        suffix=0;
      if(!work_node)
        next_node=&gfc->nodes[suffix];
      else if(work_node->type==TABLE_TYPE)
        next_node=work_node->child.m[suffix];
      else
        for(next_node=work_node->child.s;next_node;
             next_node=next_node->sibling)
          if(next_node->suffix==suffix)
            break;

      imageline++;
      xleft--;
      if(xleft==0)
        {
        xleft=width;
        height--;
        img++;
        imageline=img[0];
        }

      if(!next_node)
        {
       /* We need to output the current code and add a new one to our dictionary. First reserve a node for the added code.
*It's LINKS_TYPE at first. */
        next_node=&gfc->nodes[gfc->nodes_pos];
        gfc->nodes_pos++;
        next_node->code=next_code;
        next_code++;
        next_node->type=LINKS_TYPE;
        next_node->suffix=suffix;
        next_node->child.s=0;

       /* link next_node into work_node's set of children */
        if(work_node->type==TABLE_TYPE)
          work_node->child.m[suffix]=next_node;
        else if(work_node->type<MAX_LINKS_TYPE
                ||gfc->links_pos+clear_code>LINKS_SIZE)
          {
          next_node->sibling=work_node->child.s;
          work_node->child.s=next_node;
          work_node->type++;
          }
        else
          GifChangeNodeToTable(gfc,work_node,next_node,clear_code);

       /* Output the current code. */
        output_code=work_node->code;
        work_node=&gfc->nodes[suffix];
        goto found_output_code;
        }

      work_node=next_node;
      }

   /* Ran out of data if we get here. */
    output_code=(work_node?work_node->code:eoi_code);
    work_node=0;

found_output_code:;
    }

  if(bits_left_over>0)
    *buf++=(BYTE)leftover;

  if(buf!=buffer)
    {
    gifputbyte((BYTE)(buf-buffer),grr);
    gifputblock(buffer,(WORD)(buf-buffer),grr);
    }

  gifputbyte(0,grr);
  }

int GifGetColorTableSize(GIF_STREAM*gfs,GIF_IMAGE*gfi,GIF_WRITER*grr)
  {
  GIF_COLORMAP*gfcm=(gfi?gfi->local:gfs->global);
  int ncol,totalcol,i;
  if(!gfcm||gfcm->ncol<=0)
    return(0);

 /* Make sure ncol is reasonable */
  ncol=gfcm->ncol;

 /* Possibly bump up 'ncol' based on 'transparent' values,if careful_min_code_bits */
  if(grr->flags&GIF_WRITE_CAREFUL_MIN_CODE_SIZE)
    {
    if(gfi&&gfi->transparent>=ncol)
      ncol=gfi->transparent+1;
    else if(!gfi)
      for(i=0;i<gfs->nimages;i++)
        if(gfs->images[i]->transparent>=ncol)
          ncol=gfs->images[i]->transparent+1;
    }

 /* Make sure the colormap is a power of two entries! */
 /* GIF format doesn't allow a colormap with only 1 entry. */
  if(ncol>256)
    ncol=256;
  for(totalcol=2;totalcol<ncol;totalcol*=2)
    /* nada */;

  return(totalcol);
  }

int GifWriteImageData(GIF_IMAGE*gfi,BYTE min_code_bits,GIF_CONTEXT2*gfc,GIF_WRITER*grr)
  {
  BYTE**img=gfi->img;
  WORD width=gfi->width,height=gfi->height;
  if(gfi->interlace)
    {
    WORD y;
    BYTE**nimg=GifNewArray(BYTE*,height+1);
    if(!nimg)
      return(0);

    for(y=0;y<height;y++)
      nimg[y]=img[GifInterlaceLine(y,height)];
    nimg[height]=0;

    GifWriteCompressedData(nimg,width,height,min_code_bits,gfc,grr);

    free(nimg);
    }
  else
    GifWriteCompressedData(img,width,height,min_code_bits,gfc,grr);

  return(1);
  }

void GifWriteColorTable(GIF_COLORMAP*gfcm,int totalcol,GIF_WRITER*grr)
  {
  GIF_COLOR*c=gfcm->col;
  int i,ncol=gfcm->ncol;
  for(i=0;i<ncol&&i<totalcol;i++,c++)
    {
    gifputbyte(c->r,grr);
    gifputbyte(c->g,grr);
    gifputbyte(c->b,grr);
    }

 /* Pad out colormap with black. */
  for(;i<totalcol;i++)
    {
    gifputbyte(0,grr);
    gifputbyte(0,grr);
    gifputbyte(0,grr);
    }
  }

int GifCalculateMinCodeBits(GIF_STREAM*gfs,GIF_IMAGE*gfi,GIF_WRITER*grr);

int GifWriteImage(GIF_STREAM*gfs,GIF_IMAGE*gfi,GIF_CONTEXT2*gfc,GIF_WRITER*grr)
  {
  BYTE min_code_bits,packed=0;
  grr->local_size=GifGetColorTableSize(gfs,gfi,grr);

  gifputbyte(',',grr);
  GifPutWord(gfi->left,grr);
  GifPutWord(gfi->top,grr);
  GifPutWord(gfi->width,grr);
  GifPutWord(gfi->height,grr);

  if(grr->local_size>0)
    {
    int size=2;
    packed|=0x80;
    while(size<grr->local_size)
      size*=2,packed++;
    }

  if(gfi->interlace)
    packed|=0x40;
  gifputbyte(packed,grr);

  if(grr->local_size>0)
    GifWriteColorTable(gfi->local,grr->local_size,grr);

 /* calculate min_code_bits here (because calculation may involve recompression,if GIF_WRITE_CAREFUL_MIN_CODE_BITS is true) */
  min_code_bits=GifCalculateMinCodeBits(gfs,gfi,grr);

 /* use existing compressed data if it exists. This will tend to whip people's asses who uncompress an image,keep the
*compressed data around,but modify the uncompressed data anyway. That sucks. */
  if(gfi->compressed)
    {
    BYTE*compressed=gfi->compressed;
    DWORD compressed_len=gfi->compressed_len;
    while(compressed_len>0)
      {
      WORD amt=(WORD)(compressed_len>0x7000?0x7000:compressed_len);
      gifputblock(compressed,amt,grr);
      compressed+=amt;
      compressed_len-=amt;
      }

    }
  else
    GifWriteImageData(gfi,min_code_bits,gfc,grr);

  return(1);
  }

void GifWriteLogicalScreenDescriptor(GIF_STREAM*gfs,GIF_WRITER*grr)
  {
  BYTE packed=0x70; /* high resolution colors */
  grr->global_size=GifGetColorTableSize(gfs,0,grr);

  GifCalculateScreenSize(gfs,0);
  GifPutWord(gfs->screen_width,grr);
  GifPutWord(gfs->screen_height,grr);

  if(grr->global_size>0)
    {
    WORD size=2;
    packed|=0x80;
    while(size<grr->global_size)
      size*=2,packed++;
    }

  gifputbyte(packed,grr);
  gifputbyte(gfs->background,grr);
  gifputbyte(0,grr); /* no aspect ratio information */

  if(grr->global_size>0)
    GifWriteColorTable(gfs->global,grr->global_size,grr);
  }

int GifFullCompressImage(GIF_STREAM*gfs,GIF_IMAGE*gfi,int flags)
  {
  int ok=0;
  BYTE min_code_bits;
  GIF_WRITER grr;
  GIF_CONTEXT2 gfc;
  gfc.nodes=GifNewArray(GIF_NODE,NODES_SIZE);
  gfc.links=GifNewArray(GIF_NODE*,LINKS_SIZE);

  grr.v=GifNewArray(BYTE,1024);
  grr.pos=0;
  grr.cap=1024;
  grr.byte_putter=GifMemoryBytePutter;
  grr.block_putter=GifMemoryBlockPutter;
  grr.flags=flags;
  grr.global_size=GifGetColorTableSize(gfs,0,&grr);
  grr.local_size=GifGetColorTableSize(gfs,gfi,&grr);

  if(!gfc.nodes||!gfc.links||!grr.v)
    goto done;

  min_code_bits=GifCalculateMinCodeBits(gfs,gfi,&grr);
  ok=GifWriteImageData(gfi,min_code_bits,&gfc,&grr);

done:
  if(!ok)
    {
    free(grr.v);
    grr.v=0;
    }
  gfi->compressed=grr.v;
  gfi->compressed_len=grr.pos;
  free(gfc.nodes);
  free(gfc.links);
  return(grr.v!=0);
  }

int GifCalculateMinCodeBits(GIF_STREAM*gfs,GIF_IMAGE*gfi,GIF_WRITER*grr)
  {
  int colors_used=-1,min_code_bits,i;
  if(grr->flags&GIF_WRITE_CAREFUL_MIN_CODE_SIZE)
    {
   /* calculate m_c_b based on colormap */
    if(grr->local_size>0)
      colors_used=grr->local_size;
    else if(grr->global_size>0)
      colors_used=grr->global_size;

    }
  else if(gfi->compressed)
    {
   /* take m_c_b from compressed image */
    colors_used=1<<gfi->compressed[0];

    }
  else if(gfi->img)
    {
   /* calculate m_c_b from uncompressed data */
    int x,y,width=gfi->width,height=gfi->height;
    colors_used=0;
    for(y=0;y<height&&colors_used<128;y++)
      {
      BYTE*data=gfi->img[y];
      for(x=width;x>0;x--,data++)
        if(*data>colors_used)
          colors_used=*data;
      }
    colors_used++;

    }
  else
    {
   /* should never happen */
    colors_used=256;
    }

  min_code_bits=2; /* min_code_bits of 1 isn't allowed */
  i=4;
  while(i<colors_used)
    {
    min_code_bits++;
    i*=2;
    }

  if((grr->flags&GIF_WRITE_CAREFUL_MIN_CODE_SIZE)
     &&gfi->compressed&&gfi->compressed[0]!=min_code_bits)
    {
    if(GifFullUncompressImage(gfi,0,0))
      GifFullCompressImage(gfs,gfi,grr->flags);
    }

  return(min_code_bits);
  }
/* extension byte table:
   0x01 plain text extension
   0xCE name*
   0xF9 graphic control extension
   0xFE comment extension
   0xFF application extension
 */

void GifWriteGraphicControlExtension(GIF_IMAGE*gfi,GIF_WRITER*grr)
  {
  BYTE packed=0;
  gifputbyte('!',grr);
  gifputbyte(0xF9,grr);
  gifputbyte(4,grr);
  if(gfi->transparent>=0)
    packed|=0x01;
  packed|=(gfi->disposal&0x07)<<2;
  gifputbyte(packed,grr);
  GifPutWord(gfi->delay,grr);
  gifputbyte((BYTE)gfi->transparent,grr);
  gifputbyte(0,grr);
  }

void GifPutData(BYTE*data,int len,GIF_WRITER*grr)
  {
  while(len>0)
    {
    int s=len>255?255:len;
    gifputbyte(s,grr);
    gifputblock(data,s,grr);
    data+=s;
    len-=s;
    }
  gifputbyte(0,grr);
  }

void GifWriteNameExtension(char*id,GIF_WRITER*grr)
  {
  gifputbyte('!',grr);
  gifputbyte(0xCE,grr);
  GifPutData((BYTE*)id,(int)strlen(id),grr);
  }

void GifWriteCommentExtensions(GIF_COMMENT*gfcom,GIF_WRITER*grr)
  {
  int i;
  for(i=0;i<gfcom->count;i++)
    {
    gifputbyte('!',grr);
    gifputbyte(0xFE,grr);
    GifPutData((BYTE*)gfcom->str[i],gfcom->len[i],grr);
    }
  }

void GifWriteNetscapeLoopExtension(WORD value,GIF_WRITER*grr)
  {
  gifputblock((BYTE*)"!\xFF\x0BNETSCAPE2.0\x03\x01",16,grr);
  GifPutWord(value,grr);
  gifputbyte(0,grr);
  }

void GifWriteGenericExtension(GIF_EXTENSION*gfex,GIF_WRITER*grr)
  {
  DWORD pos=0;
  if(gfex->kind<0)
    return; /* ignore our private extensions */

  gifputbyte('!',grr);
  gifputbyte(gfex->kind,grr);
  if(gfex->kind==255)
   { /* an application extension */
    int len=gfex->application?(int)strlen(gfex->application):0;
    if(len)
      {
      gifputbyte(len,grr);
      gifputblock((BYTE*)gfex->application,len,grr);
      }
    }
  while(pos+255<gfex->length)
    {
    gifputbyte(255,grr);
    gifputblock(gfex->data+pos,255,grr);
    pos+=255;
    }
  if(pos<gfex->length)
    {
    DWORD len=gfex->length-pos;
    gifputbyte((BYTE)len,grr);
    gifputblock(gfex->data+pos,(WORD)len,grr);
    }
  gifputbyte(0,grr);
  }

int GifWriteStream(GIF_STREAM*gfs,GIF_WRITER*grr)
  {
  int ok=0;
  int i;
  GIF_IMAGE*gfi;
  GIF_EXTENSION*gfex=gfs->extensions;
  GIF_CONTEXT2 gfc;
  gfc.nodes=GifNewArray(GIF_NODE,NODES_SIZE);
  gfc.links=GifNewArray(GIF_NODE*,LINKS_SIZE);
  if(!gfc.nodes||!gfc.links)
    goto done;

    {
    BYTE isgif89a=0;
    if(gfs->comment||gfs->loopcount>-1)
      isgif89a=1;
    for(i=0;i<gfs->nimages&&!isgif89a;i++)
      {
      gfi=gfs->images[i];
      if(gfi->identifier||gfi->transparent!=-1||gfi->disposal||gfi->delay||gfi->comment)
        isgif89a=1;
      }
    if(isgif89a)
      gifputblock((BYTE*)"GIF89a",6,grr);
    else
      gifputblock((BYTE*)"GIF87a",6,grr);
    }

  GifWriteLogicalScreenDescriptor(gfs,grr);

  if(gfs->loopcount>-1)
    GifWriteNetscapeLoopExtension((WORD)gfs->loopcount,grr);

  for(i=0;i<gfs->nimages;i++)
    {
    GIF_IMAGE*gfi=gfs->images[i];
    while(gfex&&gfex->position==i)
      {
      GifWriteGenericExtension(gfex,grr);
      gfex=gfex->next;
      }
    if(gfi->comment)
      GifWriteCommentExtensions(gfi->comment,grr);
    if(gfi->identifier)
      GifWriteNameExtension(gfi->identifier,grr);
    if(gfi->transparent!=-1||gfi->disposal||gfi->delay)
      GifWriteGraphicControlExtension(gfi,grr);
    if(!GifWriteImage(gfs,gfi,&gfc,grr))
      goto done;
    }

  while(gfex)
    {
    GifWriteGenericExtension(gfex,grr);
    gfex=gfex->next;
    }
  if(gfs->comment)
    GifWriteCommentExtensions(gfs->comment,grr);

  gifputbyte(';',grr);
  ok=1;

done:
  free(gfc.nodes);
  free(gfc.links);
  return(ok);
  }

int GifFullWriteFile(GIF_STREAM*gfs,int flags,FILE*f)
  {
  GIF_WRITER grr;
  grr.f=f;
  grr.byte_putter=GifFileBytePutter;
  grr.block_putter=GifFileBlockPutter;
  grr.flags=flags;
  return(GifWriteStream(gfs,&grr));
  }
#undef GifCompressImage
#undef GifWriteFile

int GifCompressImage(GIF_STREAM*gfs,GIF_IMAGE*gfi)
  {
  return(GifFullCompressImage(gfs,gfi,0));
  }

int GifWriteFile(GIF_STREAM*gfs,FILE*f)
  {
  return(GifFullWriteFile(gfs,0,f));
  }

GIFX ReadGIF(char*filename)
  {

/* read a GIF sequence comprised of one or more images and convert to bitmaps */

  int colors,f,h,i,w,wide;
  BYTE b,g,*last_image,*this_image,r;
  DWORD*palette;
  FILE*fp;
  GIFX gx;
  GIF_IMAGE*gi;
  GIF_STREAM*gs;

/* initialize GIFX structure */

  memset(&gx,0,sizeof(gx));

/* open file */

  if((fp=fopen(filename,"rb"))==NULL)
    return(gx);

/* call library function to read GIF_STREAM */

  if((gs=GifFullReadFile(fp,GIF_READ_UNCOMPRESSED,0,0))==NULL)
    {
    fclose(fp);
    return(gx);
    }
  fclose(fp);

/* allocate memory for array of bitmap pointers */

  if((gx.bm=calloc(gs->nimages,sizeof(BITMAPINFOHEADER*)))==NULL)
    return(gx);
  gx.frames=gs->nimages;
  this_image=NULL;
  wide=BMPwidth(gs->screen_width,8);

/* convert images from GIF_STREAM to bitmap frames in GIFX */

  for(f=0;f<gx.frames;f++)
    {
    gi=gs->images[f];
    fprintf(fp,"%i %i\n",f,gi->transparent);
    gx.delay=gi->delay;
    if(gi->local)
      colors=gi->local->ncol;
    else
      colors=gs->global->ncol;

/* allocate memory for bitmap frame */

    if((gx.bm[f]=calloc(sizeof(BITMAPINFOHEADER)+colors*sizeof(DWORD)+wide*gs->screen_height,1))==NULL)
      {
      while(f>0)
        free(gx.bm[--f]);
      free(gx.bm);
      memset(&gx,0,sizeof(gx));
      return(gx);
      }
    palette=(DWORD*)(((BYTE*)gx.bm[f])+sizeof(BITMAPINFOHEADER));
    gx.bm[f]->biSize=sizeof(BITMAPINFOHEADER);
    gx.bm[f]->biPlanes=1;
    gx.bm[f]->biBitCount=8;
    gx.bm[f]->biWidth=gs->screen_width;
    gx.bm[f]->biHeight=gs->screen_height;
    gx.bm[f]->biClrUsed=colors;
    gx.bm[f]->biSizeImage=wide*gs->screen_height;

/* copy palette from GIF_STREAM to bitmap frame */

    last_image=this_image;
    this_image=(BYTE*)(palette+colors);
    if(gi->local)
      {
      for(i=0;i<colors;i++)
        {
        r=gi->local->col[i].r;
        g=gi->local->col[i].g;
        b=gi->local->col[i].b;
        palette[i]=RGB(b,g,r);
        }
      }
    else
      {
      for(i=0;i<colors;i++)
        {
        r=gs->global->col[i].r;
        g=gs->global->col[i].g;
        b=gs->global->col[i].b;
        palette[i]=RGB(b,g,r);
        }
      }

/* copy image from GIF_STREAM to bitmap frame */

    memset(this_image,gs->background,gx.bm[f]->biSizeImage);
    for(h=0;h<gi->height;h++)
      {
      i=(gs->screen_height-1-(gi->top+h))*wide+gi->left;
      for(w=0;w<gi->width;w++,i++)
        {
        b=gi->img[h][w];
        this_image[i]=b;
        if(last_image)
          if(b==(BYTE)gi->transparent)
            this_image[i]=last_image[i];
        }
      }
    }

/* free memory associated with GIF_STREAM */

  GifDeleteStream(gs);

/* return bitmap frames */

  return(gx);
  }

BITMAPINFOHEADER*ReadGIF1(char*filename)
  {

/* read the first frame from a GIF sequence and convert to a bitmap */

  int i;
  BITMAPINFOHEADER*bh;
  GIFX gx;
/* read the entire GIF sequence */

  gx=ReadGIF(filename);
  if(gx.frames<1)
    return(NULL);
  bh=gx.bm[0];

/* free memory associated with all but the first frame */

  for(i=1;i<gx.frames;i++)
    free(gx.bm[i]);
  free(gx.bm);

/* return the first frame as a bitmap */

  return(bh);
  }

GIFX LoadGIF(GIF_RECORD gr)
  {
/* load a GIF sequence comprised of one or more images and convert to bitmaps */

  int colors,f,h,i,w,wide;
  BYTE b,g,*last_image,*this_image,r;
  DWORD*palette;
  GIFX gx;
  GIF_IMAGE*gi;
  GIF_STREAM*gs;

/* initialize GIFX structure */

  memset(&gx,0,sizeof(gx));

/* load data */

  gs=GifReadRecord(&gr);

/* allocate memory for array of bitmap pointers */

  if((gx.bm=calloc(gs->nimages,sizeof(BITMAPINFOHEADER*)))==NULL)
    return(gx);
  gx.frames=gs->nimages;
  this_image=NULL;
  wide=BMPwidth(gs->screen_width,8);

/* convert images from GIF_STREAM to bitmap frames in GIFX */

  for(f=0;f<gx.frames;f++)
    {
    gi=gs->images[f];
    gx.delay=gi->delay;
    if(gi->local)
      colors=gi->local->ncol;
    else
      colors=gs->global->ncol;

/* allocate memory for bitmap frame */

    if((gx.bm[f]=calloc(sizeof(BITMAPINFOHEADER)+colors*sizeof(DWORD)+wide*gs->screen_height,1))==NULL)
      {
      while(f>0)
        free(gx.bm[--f]);
      free(gx.bm);
      memset(&gx,0,sizeof(gx));
      return(gx);
      }
    palette=(DWORD*)(((BYTE*)gx.bm[f])+sizeof(BITMAPINFOHEADER));
    gx.bm[f]->biSize=sizeof(BITMAPINFOHEADER);
    gx.bm[f]->biPlanes=1;
    gx.bm[f]->biBitCount=8;
    gx.bm[f]->biWidth=gs->screen_width;
    gx.bm[f]->biHeight=gs->screen_height;
    gx.bm[f]->biClrUsed=colors;
    gx.bm[f]->biSizeImage=wide*gs->screen_height;

/* copy palette from GIF_STREAM to bitmap frame */

    this_image=(BYTE*)(palette+colors);
    if(gi->local)
      {
      for(i=0;i<colors;i++)
        {
        r=gi->local->col[i].r;
        g=gi->local->col[i].g;
        b=gi->local->col[i].b;
        palette[i]=RGB(b,g,r);
        }
      }
    else
      {
      for(i=0;i<colors;i++)
        {
        r=gs->global->col[i].r;
        g=gs->global->col[i].g;
        b=gs->global->col[i].b;
        palette[i]=RGB(b,g,r);
        }
      }

/* copy image from GIF_STREAM to bitmap frame */

    last_image=this_image;
    memset(this_image,gs->background,gx.bm[f]->biSizeImage);
    for(h=0;h<gi->height;h++)
      {
      i=(gs->screen_height-1-(gi->top+h))*wide+gi->left;
      for(w=0;w<gi->width;w++,i++)
        {
        b=gi->img[h][w];
        this_image[i]=b;
        }
      }
    }

/* free memory associated with GIF_STREAM */

  GifDeleteStream(gs);

/* return bitmap frames */

  return(gx);
  }

BITMAPINFOHEADER*LoadGIF1(GIF_RECORD gr)
  {
  int i;
  BITMAPINFOHEADER*bh;
  GIFX gx;
  gx=LoadGIF(gr);
  if(gx.frames<1)
    return(NULL);
  bh=gx.bm[0];
  for(i=1;i<gx.frames;i++)
    free(gx.bm[i]);
  free(gx.bm);
  return(bh);
  }

void FreeGIFs(GIF_STREAM*gs)
  {
  int n;
  if(gs)
    {
    if(gs->images)
      {
      for(n=0;n<gs->nimages;n++)
        {
        if(gs->images[n])
          {
          if(gs->images[n]->img)
            free(gs->images[n]->img);
          if(gs->images[n]->image_data)
            free(gs->images[n]->image_data);
          if(gs->images[n]->compressed)
            free(gs->images[n]->compressed);
          if(gs->images[n]->local)
            {
            if(gs->images[n]->local->col)
              free(gs->images[n]->local->col);
            free(gs->images[n]->local);
            }
          free(gs->images[n]);
          }
        }
      free(gs->images);
      }
    if(gs->global)
      {
      if(gs->global->col)
        free(gs->global->col);
      free(gs->global);
      }
    free(gs);
    }
  }

/* This function will write a sequence of bitmaps to a GIF file.
   It can only handle 8 bits/pixel, all the frames having the
   same size, and only one global palette. */

int WriteGifFile(GIFX gx,char*filename)
  {
  int colors,depth,f,h,height,i,success,w,wide,width;
  BITMAPINFOHEADER*bm;
  BYTE*image;
  DWORD*palette;
  FILE*fp;
  GIF_STREAM*gs;
/* assume 8 bits/pixel */

  depth=gx.bm[0]->biBitCount;
  if(depth!=8)
    return(__LINE__);
  colors=gx.bm[0]->biClrUsed;
  if(colors<2||colors>256)
    return(__LINE__);

/* assume all the images are the same size and color depth */

  width=gx.bm[0]->biWidth;
  wide=BMPwidth(width,depth);
  height=gx.bm[0]->biHeight;
  bm=gx.bm[0];
  for(f=1;f<gx.frames;f++)
    if(memcmp(gx.bm[f],bm,sizeof(BITMAPINFOHEADER)))
      return(__LINE__);

/* assume there is only one global palette */

  palette=(DWORD*)(((BYTE*)bm)+sizeof(BITMAPINFOHEADER));
  for(f=1;f<gx.frames;f++)
    if(memcmp(((BYTE*)gx.bm[f])+sizeof(BITMAPINFOHEADER),palette,colors*sizeof(DWORD)))
      return(__LINE__);

/* allocate and begin to populate GIF_STREAM structure */

  if((gs=calloc(1,sizeof(GIF_STREAM)))==NULL)
    return(__LINE__);
  gs->screen_height=width;
  gs->screen_height=height;

/* allocate populate global palette structure */

  if((gs->global=calloc(1,sizeof(GIF_COLORMAP)))==NULL)
    {
    FreeGIFs(gs);
    return(__LINE__);
    }
  gs->global->ncol=colors;
  gs->global->capacity=colors;

/* allocate and populate global palette entries */

  if((gs->global->col=calloc(colors,sizeof(GIF_COLOR)))==NULL)
    {
    FreeGIFs(gs);
    return(__LINE__);
    }
  palette=(DWORD*)(((BYTE*)gx.bm[0])+sizeof(BITMAPINFOHEADER));
  for(i=0;i<colors;i++)
    {
    gs->global->col[i].r=RedRGB(palette[i]);
    gs->global->col[i].g=GreenRGB(palette[i]);
    gs->global->col[i].b=BlueRGB(palette[i]);
    }

/* allocate array of pointers to the image structures */

  gs->nimages=gx.frames;
  if((gs->images=calloc(gx.frames,sizeof(GIF_IMAGE*)))==NULL)
    {
    FreeGIFs(gs);
    return(__LINE__);
    }

/* allocate image structures */

  for(f=0;f<gx.frames;f++)
    {
    if((gs->images[f]=calloc(sizeof(GIF_IMAGE),1))==NULL)
      {
      FreeGIFs(gs);
      return(__LINE__);
      }
    }

/* create images */

  for(f=0;f<gx.frames;f++)
    {
    gs->images[f]->delay=gx.delay;
    gs->images[f]->width=width;
    gs->images[f]->height=height;
    gs->images[f]->transparent=gx.transparency;

/* allocate image data array */

    if((gs->images[f]->image_data=calloc(width*height,sizeof(BYTE)))==NULL)
      {
      FreeGIFs(gs);
      return(__LINE__);
      }

/* allocate image index array */

    if((gs->images[f]->img=calloc(height,sizeof(BYTE*)))==NULL)
      {
      FreeGIFs(gs);
      return(__LINE__);
      }

/* set index values */

    for(h=0;h<height;h++)
      gs->images[f]->img[h]=gs->images[f]->image_data+h*width;

/* copy image pixels */

    palette=(DWORD*)(((BYTE*)gx.bm[f])+sizeof(BITMAPINFOHEADER));
    image=(BYTE*)(palette+colors);
    for(h=0;h<height;h++)
      for(w=0;w<width;w++)
        gs->images[f]->img[h][w]=image[(height-1-h)*wide+w];
    }

/* open file */

  if((fp=fopen(filename,"wb"))==NULL)
    return(__LINE__);

/* write GIF_STREAM */

  success=GifFullWriteFile(gs,0,fp);

/* close file */

  fclose(fp);

/* delete GIF_STREAM */

  FreeGIFs(gs);

/* return success */

  if(success)
    return(0);
  return(__LINE__);
  }

GIFX FreeGIFX(GIFX gx)
  {
  int i;
  static GIFX gy;
  if(gx.bm)
    {
    for(i=0;i<gx.frames;i++)
      if(gx.bm[i])
        free(gx.bm[i]);
    free(gx.bm);
    }
  return(gy);
  }

/* This function will write a sequence of bitmaps to a GIF file.
   It can handle 1, 4, 8,or 24 bits/pixel, frames having differing
   sizes, and multiple palettes. */

int WriteGIF(GIFX gx,char*filename)
  {
  int f,h,height=0,i,used[256],w,wide=0,width=0;
  BITMAPINFOHEADER*bm;
  BYTE*image1,*image2,t;

/* step 1: convert all images to 8 bits/pixel */

  for(f=0;f<gx.frames;f++)
    {
    if(gx.bm[f]->biBitCount==1)
      {
      bm=BMP1to8(gx.bm[f]);
      free(gx.bm[f]);
      gx.bm[f]=bm;
      }
    else if(gx.bm[f]->biBitCount==4)
      {
      bm=BMP4to8(gx.bm[f]);
      free(gx.bm[f]);
      gx.bm[f]=bm;
      }
    else if(gx.bm[f]->biBitCount==24)
      {
      bm=BMPdiffuse24(gx.bm[f]);
      free(gx.bm[f]);
      gx.bm[f]=bm;
      }
    }

/* step 2: make all images the same size */

  if(gx.frames>1)
    {
    for(width=height=f=0;f<gx.frames;f++)
      {
      if(gx.bm[f]->biWidth>width)
        width=gx.bm[f]->biWidth;
      if(gx.bm[f]->biHeight>height)
        height=gx.bm[f]->biHeight;
      }
    for(f=0;f<gx.frames;f++)
      {
      if(gx.bm[f]->biWidth!=width||gx.bm[f]->biHeight!=height)
        {
        bm=BMPresize(gx.bm[f],width,height);
        free(gx.bm[f]);
        gx.bm[f]=bm;
        }
      }
    }

/* step 3: check for unused palette index available for transparency */

  gx.transparency=-1;
  if(gx.frames>1)
    {
    memset(used,0,sizeof(used));
    wide=BMPwidth(width,8);
    for(f=0;f<gx.frames;f++)
      {
      image1=BMPbits(gx.bm[f]);
      for(h=0;h<height;h++)
        for(w=0;w<width;w++)
          used[image1[wide*h+w]]++;
      }
    for(i=0;i<256;i++)
      if(used[i]==0)
        break;
    if(i<256)
      {
      gx.transparency=i;
      t=(BYTE)i;
      }
    }

/* step 4: implement transparency */

  if(gx.transparency)
    {
    for(f=gx.frames-1;f>0;f--)
      {
      image1=BMPbits(gx.bm[f-1]);
      image2=BMPbits(gx.bm[f]);
      for(h=0;h<height;h++)
        {
        i=wide*h;
        for(w=0;w<width;w++,i++)
          if(image2[i]==image1[i])
            image2[i]=t;
        }
      }
    }

  return(WriteGifFile(gx,filename));
  }

int WriteGIF1(BITMAPINFOHEADER*bm,char*filename)
  {

/* write a single bitmap to a GIF file */

  static BITMAPINFOHEADER*bh[1];
  static GIFX gx;
/* create a GIFX structure containing only one bitmap frame */

  bh[0]=bm;
  gx.bm=(BITMAPINFOHEADER**)&bh;
  gx.frames=1;

/* write the bitmap to a GIF file */

  return(WriteGIF(gx,filename));
  }
