/*******************************************************************************

  This program has been adapted from miniz.c v1.15, a public domain
  deflate/inflate, zlib-subset, ZIP reading/writing/appending program
  written by Rich Geldreich <richgel99@gmail.com> and last updated 10/13/2013.

  Besides cleaning up and removing things specific to operating systems that
  don't exist (or are no longer used by anyone), this adaptation includes
  a file and folder search which can be used just like the WIN32 calls.

  One of the main purposes of this adaptation is to allow entire ZIP files
  to be embedded in an executable as static data or as a resource in order
  to create a much more manageable implementation rather than requiring that
  an entire folder structure and multiple files be created for an
  application to function. This way such an application can be completely
  self contained.

  While there may have been a necessity to share every possible resource
  in the days of 640 kB memory. Those days are long gone. What has
  happened is that many applications have become massive sprawling
  complexes of bug-infested bloatware that present endless hassles.
  This trend, of course, has been led by Micro$lop like a crusade. It
  arises from sloth and greed like anerobic bacteria in a cesspool.

  This adaptation was written by Dudley J. Benton, Knoxville, Tennessee,
  dudley.benton@gmail.com.

*******************************************************************************/

#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <io.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/utime.h>
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include "ziplib.h"

THIS_FILE this_file;

enum{
  MZ_NO_FLUSH=0,
  MZ_PARTIAL_FLUSH=1,
  MZ_SYNC_FLUSH=2,
  MZ_FULL_FLUSH=3,
  MZ_FINISH=4,
  MZ_BLOCK=5
  };

enum{
  MZ_OK=0,
  MZ_STREAM_END=1,
  MZ_NEED_DICT=2,
  MZ_ERRNO=-1,
  MZ_STREAM_ERROR=-2,
  MZ_DATA_ERROR=-3,
  MZ_MEM_ERROR=-4,
  MZ_BUF_ERROR=-5,
  MZ_VERSION_ERROR=-6,
  MZ_PARAM_ERROR=-10000
  };

enum{
  TINFL_FLAG_PARSE_ZLIB_HEADER=1,
  TINFL_FLAG_HAS_MORE_INPUT=2,
  TINFL_FLAG_USING_NON_WRAPPING_OUTPUT_BUF=4,
  TINFL_FLAG_COMPUTE_ADLER32=8
  };

enum{
  TINFL_MAX_HUFF_TABLES=3,
  TINFL_MAX_HUFF_SYMBOLS_0=288,
  TINFL_MAX_HUFF_SYMBOLS_1=32,
  TINFL_MAX_HUFF_SYMBOLS_2=19,
  TINFL_FAST_LOOKUP_BITS=10,
  TINFL_FAST_LOOKUP_SIZE=1<<TINFL_FAST_LOOKUP_BITS
  };

enum{
  TDEFL_HUFFMAN_ONLY=0,
  TDEFL_DEFAULT_MAX_PROBES=128,
  TDEFL_MAX_PROBES_MASK=0xFFF
  };

enum{
  TDEFL_WRITE_ZLIB_HEADER            =0x01000,
  TDEFL_COMPUTE_ADLER32              =0x02000,
  TDEFL_GREEDY_PARSING_FLAG          =0x04000,
  TDEFL_NONDETERMINISTIC_PARSING_FLAG=0x08000,
  TDEFL_RLE_MATCHES                  =0x10000,
  TDEFL_FILTER_MATCHES               =0x20000,
  TDEFL_FORCE_ALL_STATIC_BLOCKS      =0x40000,
  TDEFL_FORCE_ALL_RAW_BLOCKS         =0x80000
  };

enum{
  TDEFL_MAX_HUFF_TABLES=3,
  TDEFL_MAX_HUFF_SYMBOLS_0=288,
  TDEFL_MAX_HUFF_SYMBOLS_1=32,
  TDEFL_MAX_HUFF_SYMBOLS_2=19,
  TDEFL_LZ_DICT_SIZE=32768,
  TDEFL_LZ_DICT_SIZE_MASK=TDEFL_LZ_DICT_SIZE-1,
  TDEFL_MIN_MATCH_LEN=3,
  TDEFL_MAX_MATCH_LEN=258
  };

enum{
  TDEFL_LZ_CODE_BUF_SIZE=64 * 1024,
  TDEFL_OUT_BUF_SIZE=(TDEFL_LZ_CODE_BUF_SIZE * 13 ) / 10,
  TDEFL_MAX_HUFF_SYMBOLS=288,
  TDEFL_LZ_HASH_BITS=15,
  TDEFL_LEVEL1_HASH_SIZE_MASK=4095,
  TDEFL_LZ_HASH_SHIFT=(TDEFL_LZ_HASH_BITS + 2) / 3,
  TDEFL_LZ_HASH_SIZE=1 << TDEFL_LZ_HASH_BITS
  };

enum{
  TDEFL_MAX_SUPPORTED_HUFF_CODESIZE=32
  };

enum{
  MZ_ZIP_END_OF_CENTRAL_DIR_HEADER_SIG=0x06054b50,
  MZ_ZIP_CENTRAL_DIR_HEADER_SIG=0x02014b50,
  MZ_ZIP_LOCAL_DIR_HEADER_SIG=0x04034b50,
  MZ_ZIP_LOCAL_DIR_HEADER_SIZE=30,
  MZ_ZIP_CENTRAL_DIR_HEADER_SIZE=46,
  MZ_ZIP_END_OF_CENTRAL_DIR_HEADER_SIZE=22,
  MZ_ZIP_CDH_SIG_OFS=0,
  MZ_ZIP_CDH_VERSION_MADE_BY_OFS=4,
  MZ_ZIP_CDH_VERSION_NEEDED_OFS=6,
  MZ_ZIP_CDH_BIT_FLAG_OFS=8,
  MZ_ZIP_CDH_METHOD_OFS=10,
  MZ_ZIP_CDH_FILE_TIME_OFS=12,
  MZ_ZIP_CDH_FILE_DATE_OFS=14,
  MZ_ZIP_CDH_CRC32_OFS=16,
  MZ_ZIP_CDH_COMPRESSED_SIZE_OFS=20,
  MZ_ZIP_CDH_DECOMPRESSED_SIZE_OFS=24,
  MZ_ZIP_CDH_FILENAME_LEN_OFS=28,
  MZ_ZIP_CDH_EXTRA_LEN_OFS=30,
  MZ_ZIP_CDH_COMMENT_LEN_OFS=32,
  MZ_ZIP_CDH_DISK_START_OFS=34,
  MZ_ZIP_CDH_INTERNAL_ATTR_OFS=36,
  MZ_ZIP_CDH_EXTERNAL_ATTR_OFS=38,
  MZ_ZIP_CDH_LOCAL_HEADER_OFS=42,
  MZ_ZIP_LDH_SIG_OFS=0,
  MZ_ZIP_LDH_VERSION_NEEDED_OFS=4,
  MZ_ZIP_LDH_BIT_FLAG_OFS=6,
  MZ_ZIP_LDH_METHOD_OFS=8,
  MZ_ZIP_LDH_FILE_TIME_OFS=10,
  MZ_ZIP_LDH_FILE_DATE_OFS=12,
  MZ_ZIP_LDH_CRC32_OFS=14,
  MZ_ZIP_LDH_COMPRESSED_SIZE_OFS=18,
  MZ_ZIP_LDH_DECOMPRESSED_SIZE_OFS=22,
  MZ_ZIP_LDH_FILENAME_LEN_OFS=26,
  MZ_ZIP_LDH_EXTRA_LEN_OFS=28,
  MZ_ZIP_ECDH_SIG_OFS=0,
  MZ_ZIP_ECDH_NUM_THIS_DISK_OFS=4,
  MZ_ZIP_ECDH_NUM_DISK_CDIR_OFS=6,
  MZ_ZIP_ECDH_CDIR_NUM_ENTRIES_ON_DISK_OFS=8,
  MZ_ZIP_ECDH_CDIR_TOTAL_ENTRIES_OFS=10,
  MZ_ZIP_ECDH_CDIR_SIZE_OFS=12,
  MZ_ZIP_ECDH_CDIR_OFS_OFS=16,
  MZ_ZIP_ECDH_COMMENT_SIZE_OFS=20,
  };

typedef enum{
  MZ_ZIP_FLAG_CASE_SENSITIVE               =0x0100,
  MZ_ZIP_FLAG_IGNORE_PATH                  =0x0200,
  MZ_ZIP_FLAG_COMPRESSED_DATA              =0x0400,
  MZ_ZIP_FLAG_DO_NOT_SORT_CENTRAL_DIRECTORY=0x0800
  }mz_zip_flags;

typedef enum{
  TINFL_STATUS_BAD_PARAM=-3,
  TINFL_STATUS_ADLER32_MISMATCH=-2,
  TINFL_STATUS_FAILED=-1,
  TINFL_STATUS_DONE=0,
  TINFL_STATUS_NEEDS_MORE_INPUT=1,
  TINFL_STATUS_HAS_MORE_OUTPUT=2
  }tinfl_status;

typedef enum{
  TDEFL_STATUS_BAD_PARAM=-2,
  TDEFL_STATUS_PUT_BUF_FAILED=-1,
  TDEFL_STATUS_OKAY=0,
  TDEFL_STATUS_DONE=1,
  }tdefl_status;

typedef enum{
  TDEFL_NO_FLUSH=0,
  TDEFL_SYNC_FLUSH=2,
  TDEFL_FULL_FLUSH=3,
  TDEFL_FINISH=4
  }tdefl_flush;

struct tinfl_decompressor_tag;
typedef struct tinfl_decompressor_tag tinfl_decompressor;

typedef struct
  {
  BYTE m_code_size[TINFL_MAX_HUFF_SYMBOLS_0];
  __int16 m_look_up[TINFL_FAST_LOOKUP_SIZE],m_tree[TINFL_MAX_HUFF_SYMBOLS_0 * 2];
  } tinfl_huff_table;

typedef DWORD tinfl_bit_buf_t;

struct tinfl_decompressor_tag
  {
  DWORD m_state,m_num_bits,m_zhdr0,m_zhdr1,m_z_adler32,m_final,m_type,m_check_adler32,m_dist,m_counter,m_num_extra,m_table_sizes[TINFL_MAX_HUFF_TABLES];
  tinfl_bit_buf_t m_bit_buf;
  size_t m_dist_from_out_buf_start;
  tinfl_huff_table m_tables[TINFL_MAX_HUFF_TABLES];
  BYTE m_raw_header[4],m_len_codes[TINFL_MAX_HUFF_SYMBOLS_0 + TINFL_MAX_HUFF_SYMBOLS_1 + 137];
  };

typedef int (*tdefl_put_buf_func_ptr)(void* pBuf,int len,void*pUser);

typedef struct
  {
  tdefl_put_buf_func_ptr m_pPut_buf_func;
  void*m_pPut_buf_user;
  DWORD m_flags,m_max_probes[2];
  int m_greedy_parsing;
  DWORD m_adler32,m_lookahead_pos,m_lookahead_size,m_dict_size;
  BYTE *m_pLZ_code_buf,*m_pLZ_flags,*m_pOutput_buf,*m_pOutput_buf_end;
  DWORD m_num_flags_left,m_total_lz_bytes,m_lz_code_buf_dict_pos,m_bits_in,m_bit_buffer;
  DWORD m_saved_match_dist,m_saved_match_len,m_saved_lit,m_output_flush_ofs,m_output_flush_remaining,m_finished,m_block_index,m_wants_to_finish;
  tdefl_status m_prev_return_status;
  void*m_pIn_buf;
  void*m_pOut_buf;
  size_t *m_pIn_buf_size,*m_pOut_buf_size;
  tdefl_flush m_flush;
  BYTE *m_pSrc;
  size_t m_src_buf_left,m_out_buf_ofs;
  BYTE m_dict[TDEFL_LZ_DICT_SIZE + TDEFL_MAX_MATCH_LEN - 1];
  WORD m_huff_count[TDEFL_MAX_HUFF_TABLES][TDEFL_MAX_HUFF_SYMBOLS];
  WORD m_huff_codes[TDEFL_MAX_HUFF_TABLES][TDEFL_MAX_HUFF_SYMBOLS];
  BYTE m_huff_code_sizes[TDEFL_MAX_HUFF_TABLES][TDEFL_MAX_HUFF_SYMBOLS];
  BYTE m_lz_code_buf[TDEFL_LZ_CODE_BUF_SIZE];
  WORD m_next[TDEFL_LZ_DICT_SIZE];
  WORD m_hash[TDEFL_LZ_HASH_SIZE];
  BYTE m_output_buf[TDEFL_OUT_BUF_SIZE];
  } tdefl_compressor;

typedef unsigned char mz_validate_uint16[sizeof(WORD)==2 ? 1 : -1];
typedef unsigned char mz_validate_uint32[sizeof(DWORD)==4 ? 1 : -1];
typedef unsigned char mz_validate_uint64[sizeof(QWORD)==8 ? 1 : -1];

DWORD mz_adler32(DWORD adler,unsigned char*ptr,size_t buf_len)
  {
  DWORD i,s1=(DWORD)(adler & 0xffff),s2=(DWORD)(adler >> 16);
  size_t block_len=buf_len % 5552;
  if(!ptr) return (1);
  while(buf_len)   {
    for(i=0; i + 7 < block_len; i+=8,ptr+=8)
      {
      s1+=ptr[0],s2+=s1;
      s1+=ptr[1],s2+=s1;
      s1+=ptr[2],s2+=s1;
      s1+=ptr[3],s2+=s1;
      s1+=ptr[4],s2+=s1;
      s1+=ptr[5],s2+=s1;
      s1+=ptr[6],s2+=s1;
      s1+=ptr[7],s2+=s1;
      }
    for( ; i < block_len; ++i)
      s1+=*ptr++,s2+=s1;
    s1 %= 65521U,s2 %= 65521U; buf_len-=block_len; block_len=5552;
    }
  return (s2 << 16) + s1;
  }

DWORD mz_crc32(DWORD crc,BYTE *ptr,size_t buf_len)
  {
  static DWORD s_crc32[16]=  { 0,0x1db71064,0x3b6e20c8,0x26d930ac,0x76dc4190,0x6b6b51f4,0x4db26158,0x5005713c,
    0xedb88320,0xf00f9344,0xd6d6a3e8,0xcb61b38c,0x9b64c2b0,0x86d3d2d4,0xa00ae278,0xbdbdf21c   };
  DWORD crcu32=(DWORD)crc;
  if(!ptr) return (0);
  crcu32=~crcu32; while(buf_len--)   { BYTE b=*ptr++; crcu32=(crcu32 >> 4) ^ s_crc32[(crcu32 & 0xF) ^ (b & 0xF)]; crcu32=(crcu32 >> 4) ^ s_crc32[(crcu32 & 0xF) ^ (b >> 4)];   }
  return ~crcu32;
  }

void*def_alloc_func(void*opaque,size_t items,size_t size)   { (void)opaque,(void)items,(void)size; return malloc(items * size);   }
void def_free_func(void*opaque,void*address)   { (void)opaque,(void)address; free(address);   }
void*def_realloc_func(void*opaque,void*address,size_t items,size_t size)   { (void)opaque,(void)address,(void)items,(void)size; return realloc(address,items * size);   }

int mz_deflateEnd(mz_stream* pStream)
  {
  if(!pStream) return MZ_STREAM_ERROR;
  if(pStream->state)
    {
    pStream->zfree(pStream->opaque,pStream->state);
    pStream->state=NULL;
    }
  return MZ_OK;
  }

DWORD s_tdefl_num_probes[11]=  { 0,1,6,32,16,32,128,256,512,768,1500   };

DWORD tdefl_create_comp_flags_from_zip_params(int level,int window_bits,int strategy)
  {
  DWORD comp_flags=s_tdefl_num_probes[(level >= 0) ? min(10,level) : MZ_DEFAULT_LEVEL] | ((level <= 3) ? TDEFL_GREEDY_PARSING_FLAG : 0);
  if(window_bits > 0) comp_flags |= TDEFL_WRITE_ZLIB_HEADER;
  if(!level) comp_flags |= TDEFL_FORCE_ALL_RAW_BLOCKS;
  else if(strategy == MZ_FILTERED) comp_flags |= TDEFL_FILTER_MATCHES;
  else if(strategy == MZ_HUFFMAN_ONLY) comp_flags &= ~TDEFL_MAX_PROBES_MASK;
  else if(strategy == MZ_FIXED) comp_flags |= TDEFL_FORCE_ALL_STATIC_BLOCKS;
  else if(strategy == MZ_RLE) comp_flags |= TDEFL_RLE_MATCHES;
  return comp_flags;
  }

tdefl_status tdefl_flush_output_buffer(tdefl_compressor *d)
  {
  if(d->m_pIn_buf_size)
    {
    *d->m_pIn_buf_size=d->m_pSrc - (BYTE *)d->m_pIn_buf;
    }
  if(d->m_pOut_buf_size)
    {
    size_t n=min(*d->m_pOut_buf_size - d->m_out_buf_ofs,d->m_output_flush_remaining);
    memcpy((BYTE *)d->m_pOut_buf + d->m_out_buf_ofs,d->m_output_buf + d->m_output_flush_ofs,n);
    d->m_output_flush_ofs+=(DWORD)n;
    d->m_output_flush_remaining-=(DWORD)n;
    d->m_out_buf_ofs+=n;
    *d->m_pOut_buf_size=d->m_out_buf_ofs;
    }
  return (d->m_finished && !d->m_output_flush_remaining) ? TDEFL_STATUS_DONE : TDEFL_STATUS_OKAY;
  }

void tdefl_find_match(tdefl_compressor *d,DWORD lookahead_pos,DWORD max_dist,DWORD max_match_len,DWORD *pMatch_dist,DWORD *pMatch_len)
  {
  DWORD dist,pos=lookahead_pos & TDEFL_LZ_DICT_SIZE_MASK,match_len=*pMatch_len,probe_pos=pos,next_probe_pos,probe_len;
  DWORD num_probes_left=d->m_max_probes[match_len >= 32];
  WORD *s=(WORD*)(d->m_dict + pos),*p,*q;
  WORD c01=*(WORD*)(&d->m_dict[pos + match_len - 1]),s01=*(WORD*)(s);
  (void)( (!!(max_match_len <= TDEFL_MAX_MATCH_LEN)) || (assert("max_match_len <= TDEFL_MAX_MATCH_LEN"),0) ); if(max_match_len <= match_len) return;
  for( ; ; )
    {
    for( ; ; )
      {
      if(--num_probes_left == 0) return;
      next_probe_pos=d->m_next[probe_pos]; if((!next_probe_pos) || ((dist=(WORD)(lookahead_pos - next_probe_pos)) > max_dist)) return; probe_pos=next_probe_pos & TDEFL_LZ_DICT_SIZE_MASK; if(*(WORD*)(&d->m_dict[probe_pos + match_len - 1]) == c01) break;; next_probe_pos=d->m_next[probe_pos]; if((!next_probe_pos) || ((dist=(WORD)(lookahead_pos - next_probe_pos)) > max_dist)) return; probe_pos=next_probe_pos & TDEFL_LZ_DICT_SIZE_MASK; if(*(WORD*)(&d->m_dict[probe_pos + match_len - 1]) == c01) break;; next_probe_pos=d->m_next[probe_pos]; if((!next_probe_pos) || ((dist=(WORD)(lookahead_pos - next_probe_pos)) > max_dist)) return; probe_pos=next_probe_pos & TDEFL_LZ_DICT_SIZE_MASK; if(*(WORD*)(&d->m_dict[probe_pos + match_len - 1]) == c01) break;;
      }
    if(!dist) break; q=(WORD*)(d->m_dict + probe_pos); if(*(WORD*)(q)!=s01) continue; p=s; probe_len=32;
    do{
      }while( (*(WORD*)(++p) == *(WORD*)(++q)) && (*(WORD*)(++p) == *(WORD*)(++q)) &&
                   (*(WORD*)(++p) == *(WORD*)(++q)) && (*(WORD*)(++p) == *(WORD*)(++q)) && (--probe_len > 0) );
    if(!probe_len)
      {
      *pMatch_dist=dist; *pMatch_len=min(max_match_len,TDEFL_MAX_MATCH_LEN); break;
      }
    else if((probe_len=((DWORD)(p - s) * 2) + (DWORD)(*(BYTE*)p == *(BYTE*)q)) > match_len)
      {
      *pMatch_dist=dist; if((*pMatch_len=match_len=min(max_match_len,probe_len)) == max_match_len) break;
      c01=*(WORD*)(&d->m_dict[pos + match_len - 1]);
      }
    }
  }

WORD s_tdefl_len_sym[256]=  {
  257,258,259,260,261,262,263,264,265,265,266,266,267,267,268,268,269,269,269,269,270,270,270,270,271,271,271,271,272,272,272,272,
  273,273,273,273,273,273,273,273,274,274,274,274,274,274,274,274,275,275,275,275,275,275,275,275,276,276,276,276,276,276,276,276,
  277,277,277,277,277,277,277,277,277,277,277,277,277,277,277,277,278,278,278,278,278,278,278,278,278,278,278,278,278,278,278,278,
  279,279,279,279,279,279,279,279,279,279,279,279,279,279,279,279,280,280,280,280,280,280,280,280,280,280,280,280,280,280,280,280,
  281,281,281,281,281,281,281,281,281,281,281,281,281,281,281,281,281,281,281,281,281,281,281,281,281,281,281,281,281,281,281,281,
  282,282,282,282,282,282,282,282,282,282,282,282,282,282,282,282,282,282,282,282,282,282,282,282,282,282,282,282,282,282,282,282,
  283,283,283,283,283,283,283,283,283,283,283,283,283,283,283,283,283,283,283,283,283,283,283,283,283,283,283,283,283,283,283,283,
  284,284,284,284,284,284,284,284,284,284,284,284,284,284,284,284,284,284,284,284,284,284,284,284,284,284,284,284,284,284,284,285   };

BYTE s_tdefl_len_extra[256]=  {
  0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
  4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,
  5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
  5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,0   };

BYTE s_tdefl_small_dist_sym[512]=  {
  0,1,2,3,4,4,5,5,6,6,6,6,7,7,7,7,8,8,8,8,8,8,8,8,9,9,9,9,9,9,9,9,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,11,11,11,11,11,11,
  11,11,11,11,11,11,11,11,11,11,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,13,
  13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,14,14,14,14,14,14,14,14,14,14,14,14,
  14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,
  14,14,14,14,14,14,14,14,14,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,
  15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,16,16,16,16,16,16,16,16,16,16,16,16,16,
  16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,
  16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,
  16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,17,17,17,17,17,17,17,17,17,17,17,17,17,17,
  17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,
  17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,
  17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17   };

BYTE s_tdefl_small_dist_extra[512]=  {
  0,0,0,0,1,1,1,1,2,2,2,2,2,2,2,2,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,5,5,5,5,5,5,5,5,
  5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
  6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
  6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
  7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
  7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
  7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
  7,7,7,7,7,7,7,7   };

BYTE s_tdefl_large_dist_sym[128]=  {
  0,0,18,19,20,20,21,21,22,22,22,22,23,23,23,23,24,24,24,24,24,24,24,24,25,25,25,25,25,25,25,25,26,26,26,26,26,26,26,26,26,26,26,26,
  26,26,26,26,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,
  28,28,28,28,28,28,28,28,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29   };

BYTE s_tdefl_large_dist_extra[128]=  {
  0,0,8,8,9,9,9,9,10,10,10,10,10,10,10,10,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,
  12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,
  13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13   };

BYTE s_tdefl_packed_code_size_syms_swizzle[]=  { 16,17,18,0,8,7,9,6,10,5,11,4,12,3,13,2,14,1,15   };

void tdefl_huffman_enforce_max_code_size(int *pNum_codes,int code_list_len,int max_code_size)
  {
  int i; DWORD total=0; if(code_list_len <= 1) return;
  for(i=max_code_size + 1; i <= TDEFL_MAX_SUPPORTED_HUFF_CODESIZE; i++) pNum_codes[max_code_size]+=pNum_codes[i];
  for(i=max_code_size; i > 0; i--) total+=(((DWORD)pNum_codes[i]) << (max_code_size - i));
  while(total!=(1UL << max_code_size))
    {
    pNum_codes[max_code_size]--;
    for(i=max_code_size - 1; i > 0; i--) if(pNum_codes[i])   { pNum_codes[i]--; pNum_codes[i + 1]+=2; break;   }
    total--;
    }
  }

typedef struct   { WORD m_key,m_sym_index;   } tdefl_sym_freq;

tdefl_sym_freq* tdefl_radix_sort_syms(DWORD num_syms,tdefl_sym_freq* pSyms0,tdefl_sym_freq* pSyms1)
  {
  DWORD total_passes=2,pass_shift,pass,i,hist[256 * 2]; tdefl_sym_freq* pCur_syms=pSyms0,*pNew_syms=pSyms1;
  memset(&hist,0,sizeof(hist));
  for(i=0; i < num_syms; i++)   { DWORD freq=pSyms0[i].m_key; hist[freq & 0xFF]++; hist[256 + ((freq >> 8) & 0xFF)]++;   }
  while((total_passes > 1) && (num_syms == hist[(total_passes - 1) * 256])) total_passes--;
  for(pass_shift=0,pass=0; pass < total_passes; pass++,pass_shift+=8)
    {
    DWORD* pHist=&hist[pass << 8];
    DWORD offsets[256],cur_ofs=0;
    for(i=0; i < 256; i++)   { offsets[i]=cur_ofs; cur_ofs+=pHist[i];   }
    for(i=0; i < num_syms; i++) pNew_syms[offsets[(pCur_syms[i].m_key >> pass_shift) & 0xFF]++]=pCur_syms[i];
      { tdefl_sym_freq* t=pCur_syms; pCur_syms=pNew_syms; pNew_syms=t;   }
    }
  return pCur_syms;
  }

void tdefl_calculate_minimum_redundancy(tdefl_sym_freq *A,int n)
  {
  int root,leaf,next,avbl,used,dpth;
  if(n==0) return; else if(n==1)   { A[0].m_key=1; return;   }
  A[0].m_key+=A[1].m_key; root=0; leaf=2;
  for(next=1; next < n-1; next++)
    {
    if(leaf>=n || A[root].m_key<A[leaf].m_key)   { A[next].m_key=A[root].m_key; A[root++].m_key=(WORD)next;   } else A[next].m_key=A[leaf++].m_key;
    if(leaf>=n || (root<next && A[root].m_key<A[leaf].m_key))   { A[next].m_key=(WORD)(A[next].m_key + A[root].m_key); A[root++].m_key=(WORD)next;   } else A[next].m_key=(WORD)(A[next].m_key + A[leaf++].m_key);
    }
  A[n-2].m_key=0; for(next=n-3; next>=0; next--) A[next].m_key=A[A[next].m_key].m_key+1;
  avbl=1; used=dpth=0; root=n-2; next=n-1;
  while(avbl>0)
    {
    while(root>=0 && (int)A[root].m_key==dpth)   { used++; root--;   }
    while(avbl>used)   { A[next--].m_key=(WORD)(dpth); avbl--;   }
    avbl=2*used; dpth++; used=0;
    }
  }

void tdefl_optimize_huffman_table(tdefl_compressor *d,int table_num,int table_len,int code_size_limit,int static_table)
  {
  int i,j,l,num_codes[1 + TDEFL_MAX_SUPPORTED_HUFF_CODESIZE]; DWORD next_code[TDEFL_MAX_SUPPORTED_HUFF_CODESIZE + 1]; memset(&num_codes,0,sizeof(num_codes));
  if(static_table)
    {
    for(i=0; i < table_len; i++) num_codes[d->m_huff_code_sizes[table_num][i]]++;
    }
  else
    {
    tdefl_sym_freq syms0[TDEFL_MAX_HUFF_SYMBOLS],syms1[TDEFL_MAX_HUFF_SYMBOLS],*pSyms;
    int num_used_syms=0;
    WORD *pSym_count=&d->m_huff_count[table_num][0];
    for(i=0; i < table_len; i++) if(pSym_count[i])   { syms0[num_used_syms].m_key=(WORD)pSym_count[i]; syms0[num_used_syms++].m_sym_index=(WORD)i;   }
    pSyms=tdefl_radix_sort_syms(num_used_syms,syms0,syms1); tdefl_calculate_minimum_redundancy(pSyms,num_used_syms);
    for(i=0; i < num_used_syms; i++) num_codes[pSyms[i].m_key]++;
    tdefl_huffman_enforce_max_code_size(num_codes,num_used_syms,code_size_limit);
    memset(&d->m_huff_code_sizes[table_num],0,sizeof(d->m_huff_code_sizes[table_num])); memset(&d->m_huff_codes[table_num],0,sizeof(d->m_huff_codes[table_num]));
    for(i=1,j=num_used_syms; i <= code_size_limit; i++)
      for(l=num_codes[i]; l > 0; l--) d->m_huff_code_sizes[table_num][pSyms[--j].m_sym_index]=(BYTE)(i);
    }
  next_code[1]=0; for(j=0,i=2; i <= code_size_limit; i++) next_code[i]=j=((j + num_codes[i - 1]) << 1);
  for(i=0; i < table_len; i++)
    {
    DWORD rev_code=0,code,code_size; if((code_size=d->m_huff_code_sizes[table_num][i]) == 0) continue;
    code=next_code[code_size]++; for(l=code_size; l > 0; l--,code >>= 1) rev_code=(rev_code << 1) | (code & 1);
    d->m_huff_codes[table_num][i]=(WORD)rev_code;
    }
  }

void tdefl_start_static_block(tdefl_compressor *d)
  {
  DWORD i;
  BYTE *p=&d->m_huff_code_sizes[0][0];
  for(i=0; i <= 143; ++i) *p++=8;
  for( ; i <= 255; ++i) *p++=9;
  for( ; i <= 279; ++i) *p++=7;
  for( ; i <= 287; ++i) *p++=8;
  memset(d->m_huff_code_sizes[1],5,32);
  tdefl_optimize_huffman_table(d,0,288,15,1);
  tdefl_optimize_huffman_table(d,1,32,15,1);
  do{
    DWORD bits=1; DWORD len=2; (void)( (!!(bits <= ((1U << len) - 1U))) || (assert("bits <= ((1U << len) - 1U)"),0) ); d->m_bit_buffer |= (bits << d->m_bits_in); d->m_bits_in+=len; while(d->m_bits_in >= 8)   { if(d->m_pOutput_buf < d->m_pOutput_buf_end) *d->m_pOutput_buf++=(BYTE)(d->m_bit_buffer); d->m_bit_buffer >>= 8; d->m_bits_in-=8;   }
    }while(0);
  }

DWORD mz_bitmasks[17]=  { 0x0000,0x0001,0x0003,0x0007,0x000F,0x001F,0x003F,0x007F,0x00FF,0x01FF,0x03FF,0x07FF,0x0FFF,0x1FFF,0x3FFF,0x7FFF,0xFFFF   };

int tdefl_compress_lz_codes(tdefl_compressor *d)
  {
  DWORD flags;
  BYTE *pLZ_codes;
  flags=1;
  for(pLZ_codes=d->m_lz_code_buf; pLZ_codes < d->m_pLZ_code_buf; flags >>= 1)
    {
    if(flags == 1)
      flags=*pLZ_codes++ | 0x100;
    if(flags & 1)
      {
      DWORD sym,num_extra_bits;
      DWORD match_len=pLZ_codes[0],match_dist=(pLZ_codes[1] | (pLZ_codes[2] << 8)); pLZ_codes+=3;
      (void)( (!!(d->m_huff_code_sizes[0][s_tdefl_len_sym[match_len]])) || (assert("d->m_huff_code_sizes[0][s_tdefl_len_sym[match_len]]"),0) );
      do{
        DWORD bits=d->m_huff_codes[0][s_tdefl_len_sym[match_len]]; DWORD len=d->m_huff_code_sizes[0][s_tdefl_len_sym[match_len]]; (void)( (!!(bits <= ((1U << len) - 1U))) || (assert("bits <= ((1U << len) - 1U)"),0) ); d->m_bit_buffer |= (bits << d->m_bits_in); d->m_bits_in+=len; while(d->m_bits_in >= 8)   { if(d->m_pOutput_buf < d->m_pOutput_buf_end) *d->m_pOutput_buf++=(BYTE)(d->m_bit_buffer); d->m_bit_buffer >>= 8; d->m_bits_in-=8;   }
        }while(0);
      do{
        DWORD bits=match_len & mz_bitmasks[s_tdefl_len_extra[match_len]]; DWORD len=s_tdefl_len_extra[match_len]; (void)( (!!(bits <= ((1U << len) - 1U))) || (assert("bits <= ((1U << len) - 1U)"),0) ); d->m_bit_buffer |= (bits << d->m_bits_in); d->m_bits_in+=len; while(d->m_bits_in >= 8)   { if(d->m_pOutput_buf < d->m_pOutput_buf_end) *d->m_pOutput_buf++=(BYTE)(d->m_bit_buffer); d->m_bit_buffer >>= 8; d->m_bits_in-=8;   }
        }while(0);
      if(match_dist < 512)
        {
        sym=s_tdefl_small_dist_sym[match_dist]; num_extra_bits=s_tdefl_small_dist_extra[match_dist];
        }
      else
        {
        sym=s_tdefl_large_dist_sym[match_dist >> 8]; num_extra_bits=s_tdefl_large_dist_extra[match_dist >> 8];
        }
      (void)( (!!(d->m_huff_code_sizes[1][sym])) || (assert("d->m_huff_code_sizes[1][sym]"),0) );
      do{
        DWORD bits=d->m_huff_codes[1][sym]; DWORD len=d->m_huff_code_sizes[1][sym]; (void)( (!!(bits <= ((1U << len) - 1U))) || (assert("bits <= ((1U << len) - 1U)"),0) ); d->m_bit_buffer |= (bits << d->m_bits_in); d->m_bits_in+=len; while(d->m_bits_in >= 8)   { if(d->m_pOutput_buf < d->m_pOutput_buf_end) *d->m_pOutput_buf++=(BYTE)(d->m_bit_buffer); d->m_bit_buffer >>= 8; d->m_bits_in-=8;   }
        }while(0);
      do{
        DWORD bits=match_dist & mz_bitmasks[num_extra_bits]; DWORD len=num_extra_bits; (void)( (!!(bits <= ((1U << len) - 1U))) || (assert("bits <= ((1U << len) - 1U)"),0) ); d->m_bit_buffer |= (bits << d->m_bits_in); d->m_bits_in+=len; while(d->m_bits_in >= 8)   { if(d->m_pOutput_buf < d->m_pOutput_buf_end) *d->m_pOutput_buf++=(BYTE)(d->m_bit_buffer); d->m_bit_buffer >>= 8; d->m_bits_in-=8;   }
        }while(0);
      }
    else
      {
      DWORD lit=*pLZ_codes++;
      (void)( (!!(d->m_huff_code_sizes[0][lit])) || (assert("d->m_huff_code_sizes[0][lit]"),0) );
      do{
        DWORD bits=d->m_huff_codes[0][lit]; DWORD len=d->m_huff_code_sizes[0][lit]; (void)( (!!(bits <= ((1U << len) - 1U))) || (assert("bits <= ((1U << len) - 1U)"),0) ); d->m_bit_buffer |= (bits << d->m_bits_in); d->m_bits_in+=len; while(d->m_bits_in >= 8)   { if(d->m_pOutput_buf < d->m_pOutput_buf_end) *d->m_pOutput_buf++=(BYTE)(d->m_bit_buffer); d->m_bit_buffer >>= 8; d->m_bits_in-=8;   }
        }while(0);
      }
    }
  do{
    DWORD bits=d->m_huff_codes[0][256]; DWORD len=d->m_huff_code_sizes[0][256]; (void)( (!!(bits <= ((1U << len) - 1U))) || (assert("bits <= ((1U << len) - 1U)"),0) ); d->m_bit_buffer |= (bits << d->m_bits_in); d->m_bits_in+=len; while(d->m_bits_in >= 8)   { if(d->m_pOutput_buf < d->m_pOutput_buf_end) *d->m_pOutput_buf++=(BYTE)(d->m_bit_buffer); d->m_bit_buffer >>= 8; d->m_bits_in-=8;   }
    }while(0);
  return (d->m_pOutput_buf < d->m_pOutput_buf_end);
  }

void tdefl_start_dynamic_block(tdefl_compressor *d)
  {
  int num_lit_codes,num_dist_codes,num_bit_lengths; DWORD i,total_code_sizes_to_pack,num_packed_code_sizes,rle_z_count,rle_repeat_count,packed_code_sizes_index;
  BYTE code_sizes_to_pack[TDEFL_MAX_HUFF_SYMBOLS_0 + TDEFL_MAX_HUFF_SYMBOLS_1],packed_code_sizes[TDEFL_MAX_HUFF_SYMBOLS_0 + TDEFL_MAX_HUFF_SYMBOLS_1],prev_code_size=0xFF;
  d->m_huff_count[0][256]=1;
  tdefl_optimize_huffman_table(d,0,TDEFL_MAX_HUFF_SYMBOLS_0,15,0);
  tdefl_optimize_huffman_table(d,1,TDEFL_MAX_HUFF_SYMBOLS_1,15,0);
  for(num_lit_codes=286; num_lit_codes > 257; num_lit_codes--) if(d->m_huff_code_sizes[0][num_lit_codes - 1]) break;
  for(num_dist_codes=30; num_dist_codes > 1; num_dist_codes--) if(d->m_huff_code_sizes[1][num_dist_codes - 1]) break;
  memcpy(code_sizes_to_pack,&d->m_huff_code_sizes[0][0],num_lit_codes);
  memcpy(code_sizes_to_pack + num_lit_codes,&d->m_huff_code_sizes[1][0],num_dist_codes);
  total_code_sizes_to_pack=num_lit_codes + num_dist_codes; num_packed_code_sizes=0; rle_z_count=0; rle_repeat_count=0;
  memset(&d->m_huff_count[2][0],0,sizeof(d->m_huff_count[2][0]) * TDEFL_MAX_HUFF_SYMBOLS_2);
  for(i=0; i < total_code_sizes_to_pack; i++)
    {
    BYTE code_size=code_sizes_to_pack[i];
    if(!code_size)
      {
        { if(rle_repeat_count)   { if(rle_repeat_count < 3)   { d->m_huff_count[2][prev_code_size]=(WORD)(d->m_huff_count[2][prev_code_size] + rle_repeat_count); while(rle_repeat_count--) packed_code_sizes[num_packed_code_sizes++]=prev_code_size;   } else   { d->m_huff_count[2][16]=(WORD)(d->m_huff_count[2][16] + 1); packed_code_sizes[num_packed_code_sizes++]=16; packed_code_sizes[num_packed_code_sizes++]=(BYTE)(rle_repeat_count - 3);   } rle_repeat_count=0;   }   };
      if(++rle_z_count == 138)   {   { if(rle_z_count)   { if(rle_z_count < 3)   { d->m_huff_count[2][0]=(WORD)(d->m_huff_count[2][0] + rle_z_count); while(rle_z_count--) packed_code_sizes[num_packed_code_sizes++]=0;   } else if(rle_z_count <= 10)   { d->m_huff_count[2][17]=(WORD)(d->m_huff_count[2][17] + 1); packed_code_sizes[num_packed_code_sizes++]=17; packed_code_sizes[num_packed_code_sizes++]=(BYTE)(rle_z_count - 3);   } else   { d->m_huff_count[2][18]=(WORD)(d->m_huff_count[2][18] + 1); packed_code_sizes[num_packed_code_sizes++]=18; packed_code_sizes[num_packed_code_sizes++]=(BYTE)(rle_z_count - 11);   } rle_z_count=0;   }   };   }
      }
    else
      {
        { if(rle_z_count)   { if(rle_z_count < 3)   { d->m_huff_count[2][0]=(WORD)(d->m_huff_count[2][0] + rle_z_count); while(rle_z_count--) packed_code_sizes[num_packed_code_sizes++]=0;   } else if(rle_z_count <= 10)   { d->m_huff_count[2][17]=(WORD)(d->m_huff_count[2][17] + 1); packed_code_sizes[num_packed_code_sizes++]=17; packed_code_sizes[num_packed_code_sizes++]=(BYTE)(rle_z_count - 3);   } else   { d->m_huff_count[2][18]=(WORD)(d->m_huff_count[2][18] + 1); packed_code_sizes[num_packed_code_sizes++]=18; packed_code_sizes[num_packed_code_sizes++]=(BYTE)(rle_z_count - 11);   } rle_z_count=0;   }   };
      if(code_size!=prev_code_size)
        {
          { if(rle_repeat_count)   { if(rle_repeat_count < 3)   { d->m_huff_count[2][prev_code_size]=(WORD)(d->m_huff_count[2][prev_code_size] + rle_repeat_count); while(rle_repeat_count--) packed_code_sizes[num_packed_code_sizes++]=prev_code_size;   } else   { d->m_huff_count[2][16]=(WORD)(d->m_huff_count[2][16] + 1); packed_code_sizes[num_packed_code_sizes++]=16; packed_code_sizes[num_packed_code_sizes++]=(BYTE)(rle_repeat_count - 3);   } rle_repeat_count=0;   }   };
        d->m_huff_count[2][code_size]=(WORD)(d->m_huff_count[2][code_size] + 1); packed_code_sizes[num_packed_code_sizes++]=code_size;
        }
      else if(++rle_repeat_count == 6)
        {
          { if(rle_repeat_count)   { if(rle_repeat_count < 3)   { d->m_huff_count[2][prev_code_size]=(WORD)(d->m_huff_count[2][prev_code_size] + rle_repeat_count); while(rle_repeat_count--) packed_code_sizes[num_packed_code_sizes++]=prev_code_size;   } else   { d->m_huff_count[2][16]=(WORD)(d->m_huff_count[2][16] + 1); packed_code_sizes[num_packed_code_sizes++]=16; packed_code_sizes[num_packed_code_sizes++]=(BYTE)(rle_repeat_count - 3);   } rle_repeat_count=0;   }   };
        }
      }
    prev_code_size=code_size;
    }
  if(rle_repeat_count)   {   { if(rle_repeat_count)   { if(rle_repeat_count < 3)   { d->m_huff_count[2][prev_code_size]=(WORD)(d->m_huff_count[2][prev_code_size] + rle_repeat_count); while(rle_repeat_count--) packed_code_sizes[num_packed_code_sizes++]=prev_code_size;   } else   { d->m_huff_count[2][16]=(WORD)(d->m_huff_count[2][16] + 1); packed_code_sizes[num_packed_code_sizes++]=16; packed_code_sizes[num_packed_code_sizes++]=(BYTE)(rle_repeat_count - 3);   } rle_repeat_count=0;   }   };   } else   {   { if(rle_z_count)   { if(rle_z_count < 3)   { d->m_huff_count[2][0]=(WORD)(d->m_huff_count[2][0] + rle_z_count); while(rle_z_count--) packed_code_sizes[num_packed_code_sizes++]=0;   } else if(rle_z_count <= 10)   { d->m_huff_count[2][17]=(WORD)(d->m_huff_count[2][17] + 1); packed_code_sizes[num_packed_code_sizes++]=17; packed_code_sizes[num_packed_code_sizes++]=(BYTE)(rle_z_count - 3);   } else   { d->m_huff_count[2][18]=(WORD)(d->m_huff_count[2][18] + 1); packed_code_sizes[num_packed_code_sizes++]=18; packed_code_sizes[num_packed_code_sizes++]=(BYTE)(rle_z_count - 11);   } rle_z_count=0;   }   };   }
  tdefl_optimize_huffman_table(d,2,TDEFL_MAX_HUFF_SYMBOLS_2,7,0);
  do{
    DWORD bits=2; DWORD len=2; (void)( (!!(bits <= ((1U << len) - 1U))) || (assert("bits <= ((1U << len) - 1U)"),0) ); d->m_bit_buffer |= (bits << d->m_bits_in); d->m_bits_in+=len; while(d->m_bits_in >= 8)   { if(d->m_pOutput_buf < d->m_pOutput_buf_end) *d->m_pOutput_buf++=(BYTE)(d->m_bit_buffer); d->m_bit_buffer >>= 8; d->m_bits_in-=8;   }
    }while(0);
  do{
    DWORD bits=num_lit_codes - 257; DWORD len=5; (void)( (!!(bits <= ((1U << len) - 1U))) || (assert("bits <= ((1U << len) - 1U)"),0) ); d->m_bit_buffer |= (bits << d->m_bits_in); d->m_bits_in+=len; while(d->m_bits_in >= 8)   { if(d->m_pOutput_buf < d->m_pOutput_buf_end) *d->m_pOutput_buf++=(BYTE)(d->m_bit_buffer); d->m_bit_buffer >>= 8; d->m_bits_in-=8;   }
    }while(0);
  do{
    DWORD bits=num_dist_codes - 1; DWORD len=5; (void)( (!!(bits <= ((1U << len) - 1U))) || (assert("bits <= ((1U << len) - 1U)"),0) ); d->m_bit_buffer |= (bits << d->m_bits_in); d->m_bits_in+=len; while(d->m_bits_in >= 8)   { if(d->m_pOutput_buf < d->m_pOutput_buf_end) *d->m_pOutput_buf++=(BYTE)(d->m_bit_buffer); d->m_bit_buffer >>= 8; d->m_bits_in-=8;   }
    }while(0);
  for(num_bit_lengths=18; num_bit_lengths >= 0; num_bit_lengths--) if(d->m_huff_code_sizes[2][s_tdefl_packed_code_size_syms_swizzle[num_bit_lengths]]) break;
  num_bit_lengths=max(4,(num_bit_lengths + 1)); do{
    DWORD bits=num_bit_lengths - 4; DWORD len=4; (void)( (!!(bits <= ((1U << len) - 1U))) || (assert("bits <= ((1U << len) - 1U)"),0) ); d->m_bit_buffer |= (bits << d->m_bits_in); d->m_bits_in+=len; while(d->m_bits_in >= 8)   { if(d->m_pOutput_buf < d->m_pOutput_buf_end) *d->m_pOutput_buf++=(BYTE)(d->m_bit_buffer); d->m_bit_buffer >>= 8; d->m_bits_in-=8;   }
    }while(0);
  for(i=0; (int)i < num_bit_lengths; i++) do{
    DWORD bits=d->m_huff_code_sizes[2][s_tdefl_packed_code_size_syms_swizzle[i]]; DWORD len=3; (void)( (!!(bits <= ((1U << len) - 1U))) || (assert("bits <= ((1U << len) - 1U)"),0) ); d->m_bit_buffer |= (bits << d->m_bits_in); d->m_bits_in+=len; while(d->m_bits_in >= 8)   { if(d->m_pOutput_buf < d->m_pOutput_buf_end) *d->m_pOutput_buf++=(BYTE)(d->m_bit_buffer); d->m_bit_buffer >>= 8; d->m_bits_in-=8;   }
    }while(0);
  for(packed_code_sizes_index=0; packed_code_sizes_index < num_packed_code_sizes; )
    {
    DWORD code=packed_code_sizes[packed_code_sizes_index++]; (void)( (!!(code < TDEFL_MAX_HUFF_SYMBOLS_2)) || (assert("code < TDEFL_MAX_HUFF_SYMBOLS_2"),0) );
    do{
      DWORD bits=d->m_huff_codes[2][code]; DWORD len=d->m_huff_code_sizes[2][code]; (void)( (!!(bits <= ((1U << len) - 1U))) || (assert("bits <= ((1U << len) - 1U)"),0) ); d->m_bit_buffer |= (bits << d->m_bits_in); d->m_bits_in+=len; while(d->m_bits_in >= 8)   { if(d->m_pOutput_buf < d->m_pOutput_buf_end) *d->m_pOutput_buf++=(BYTE)(d->m_bit_buffer); d->m_bit_buffer >>= 8; d->m_bits_in-=8;   }
      }while(0);
    if(code >= 16) do{
      DWORD bits=packed_code_sizes[packed_code_sizes_index++]; DWORD len="\02\03\07"[code - 16]; (void)( (!!(bits <= ((1U << len) - 1U))) || (assert("bits <= ((1U << len) - 1U)"),0) ); d->m_bit_buffer |= (bits << d->m_bits_in); d->m_bits_in+=len; while(d->m_bits_in >= 8)   { if(d->m_pOutput_buf < d->m_pOutput_buf_end) *d->m_pOutput_buf++=(BYTE)(d->m_bit_buffer); d->m_bit_buffer >>= 8; d->m_bits_in-=8;   }
      }while(0);
    }
  }

int tdefl_compress_block(tdefl_compressor *d,int static_block)
  {
  if(static_block)
    tdefl_start_static_block(d);
  else
    tdefl_start_dynamic_block(d);
  return tdefl_compress_lz_codes(d);
  }

int tdefl_flush_block(tdefl_compressor *d,int flush)
  {
  DWORD saved_bit_buf,saved_bits_in;
  BYTE *pSaved_output_buf;
  int comp_block_succeeded=0;
  int n,use_raw_block=((d->m_flags & TDEFL_FORCE_ALL_RAW_BLOCKS)!=0) && (d->m_lookahead_pos - d->m_lz_code_buf_dict_pos) <= d->m_dict_size;
  BYTE *pOutput_buf_start=((d->m_pPut_buf_func == NULL) && ((*d->m_pOut_buf_size - d->m_out_buf_ofs) >= TDEFL_OUT_BUF_SIZE)) ? ((BYTE *)d->m_pOut_buf + d->m_out_buf_ofs) : d->m_output_buf;
  d->m_pOutput_buf=pOutput_buf_start;
  d->m_pOutput_buf_end=d->m_pOutput_buf + TDEFL_OUT_BUF_SIZE - 16;
  (void)( (!!(!d->m_output_flush_remaining)) || (assert("!d->m_output_flush_remaining"),0) );
  d->m_output_flush_ofs=0;
  d->m_output_flush_remaining=0;
  *d->m_pLZ_flags=(BYTE)(*d->m_pLZ_flags >> d->m_num_flags_left);
  d->m_pLZ_code_buf-=(d->m_num_flags_left == 8);
  if((d->m_flags & TDEFL_WRITE_ZLIB_HEADER) && (!d->m_block_index))
    {
    do{
      DWORD bits=0x78; DWORD len=8; (void)( (!!(bits <= ((1U << len) - 1U))) || (assert("bits <= ((1U << len) - 1U)"),0) ); d->m_bit_buffer |= (bits << d->m_bits_in); d->m_bits_in+=len; while(d->m_bits_in >= 8)   { if(d->m_pOutput_buf < d->m_pOutput_buf_end) *d->m_pOutput_buf++=(BYTE)(d->m_bit_buffer); d->m_bit_buffer >>= 8; d->m_bits_in-=8;   }
      }while(0);
    do{
        DWORD bits=0x01; DWORD len=8; (void)( (!!(bits <= ((1U << len) - 1U))) || (assert("bits <= ((1U << len) - 1U)"),0) ); d->m_bit_buffer |= (bits << d->m_bits_in); d->m_bits_in+=len; while(d->m_bits_in >= 8)   { if(d->m_pOutput_buf < d->m_pOutput_buf_end) *d->m_pOutput_buf++=(BYTE)(d->m_bit_buffer); d->m_bit_buffer >>= 8; d->m_bits_in-=8;   }
        }while(0);
    }
  do{
    DWORD bits=flush == TDEFL_FINISH; DWORD len=1; (void)( (!!(bits <= ((1U << len) - 1U))) || (assert("bits <= ((1U << len) - 1U)"),0) ); d->m_bit_buffer |= (bits << d->m_bits_in); d->m_bits_in+=len; while(d->m_bits_in >= 8)   { if(d->m_pOutput_buf < d->m_pOutput_buf_end) *d->m_pOutput_buf++=(BYTE)(d->m_bit_buffer); d->m_bit_buffer >>= 8; d->m_bits_in-=8;   }
    }while(0);
  pSaved_output_buf=d->m_pOutput_buf; saved_bit_buf=d->m_bit_buffer; saved_bits_in=d->m_bits_in;
  if(!use_raw_block)
    comp_block_succeeded=tdefl_compress_block(d,(d->m_flags & TDEFL_FORCE_ALL_STATIC_BLOCKS) || (d->m_total_lz_bytes < 48));
  if( ((use_raw_block) || ((d->m_total_lz_bytes) && ((d->m_pOutput_buf - pSaved_output_buf + 1U) >= d->m_total_lz_bytes))) &&
       ((d->m_lookahead_pos - d->m_lz_code_buf_dict_pos) <= d->m_dict_size) )
    {
    DWORD i; d->m_pOutput_buf=pSaved_output_buf; d->m_bit_buffer=saved_bit_buf,d->m_bits_in=saved_bits_in;
    do{
      DWORD bits=0; DWORD len=2; (void)( (!!(bits <= ((1U << len) - 1U))) || (assert("bits <= ((1U << len) - 1U)"),0) ); d->m_bit_buffer |= (bits << d->m_bits_in); d->m_bits_in+=len; while(d->m_bits_in >= 8)   { if(d->m_pOutput_buf < d->m_pOutput_buf_end) *d->m_pOutput_buf++=(BYTE)(d->m_bit_buffer); d->m_bit_buffer >>= 8; d->m_bits_in-=8;   }
      }while(0);
    if(d->m_bits_in)   { do{
      DWORD bits=0; DWORD len=8 - d->m_bits_in; (void)( (!!(bits <= ((1U << len) - 1U))) || (assert("bits <= ((1U << len) - 1U)"),0) ); d->m_bit_buffer |= (bits << d->m_bits_in); d->m_bits_in+=len; while(d->m_bits_in >= 8)   { if(d->m_pOutput_buf < d->m_pOutput_buf_end) *d->m_pOutput_buf++=(BYTE)(d->m_bit_buffer); d->m_bit_buffer >>= 8; d->m_bits_in-=8;   }
      }while(0);
      }
    for(i=2; i; --i,d->m_total_lz_bytes ^= 0xFFFF)
      {
      do{
        DWORD bits=d->m_total_lz_bytes & 0xFFFF; DWORD len=16; (void)( (!!(bits <= ((1U << len) - 1U))) || (assert("bits <= ((1U << len) - 1U)"),0) ); d->m_bit_buffer |= (bits << d->m_bits_in); d->m_bits_in+=len; while(d->m_bits_in >= 8)   { if(d->m_pOutput_buf < d->m_pOutput_buf_end) *d->m_pOutput_buf++=(BYTE)(d->m_bit_buffer); d->m_bit_buffer >>= 8; d->m_bits_in-=8;   }
        }while(0);
      }
    for(i=0; i < d->m_total_lz_bytes; ++i)
      {
      do{
        DWORD bits=d->m_dict[(d->m_lz_code_buf_dict_pos + i) & TDEFL_LZ_DICT_SIZE_MASK]; DWORD len=8; (void)( (!!(bits <= ((1U << len) - 1U))) || (assert("bits <= ((1U << len) - 1U)"),0) ); d->m_bit_buffer |= (bits << d->m_bits_in); d->m_bits_in+=len; while(d->m_bits_in >= 8)   { if(d->m_pOutput_buf < d->m_pOutput_buf_end) *d->m_pOutput_buf++=(BYTE)(d->m_bit_buffer); d->m_bit_buffer >>= 8; d->m_bits_in-=8;   }
        }while(0);
      }
    }
  else if(!comp_block_succeeded)
    {
    d->m_pOutput_buf=pSaved_output_buf; d->m_bit_buffer=saved_bit_buf,d->m_bits_in=saved_bits_in;
    tdefl_compress_block(d,1);
    }
  if(flush)
    {
    if(flush == TDEFL_FINISH)
      {
      if(d->m_bits_in)   { do{
        DWORD bits=0; DWORD len=8 - d->m_bits_in; (void)( (!!(bits <= ((1U << len) - 1U))) || (assert("bits <= ((1U << len) - 1U)"),0) ); d->m_bit_buffer |= (bits << d->m_bits_in); d->m_bits_in+=len; while(d->m_bits_in >= 8)   { if(d->m_pOutput_buf < d->m_pOutput_buf_end) *d->m_pOutput_buf++=(BYTE)(d->m_bit_buffer); d->m_bit_buffer >>= 8; d->m_bits_in-=8;   }
        }while(0);
        }
      if(d->m_flags & TDEFL_WRITE_ZLIB_HEADER)   { DWORD i,a=d->m_adler32; for(i=0; i < 4; i++)   { do{
        DWORD bits=(a >> 24) & 0xFF; DWORD len=8; (void)( (!!(bits <= ((1U << len) - 1U))) || (assert("bits <= ((1U << len) - 1U)"),0) ); d->m_bit_buffer |= (bits << d->m_bits_in); d->m_bits_in+=len; while(d->m_bits_in >= 8)   { if(d->m_pOutput_buf < d->m_pOutput_buf_end) *d->m_pOutput_buf++=(BYTE)(d->m_bit_buffer); d->m_bit_buffer >>= 8; d->m_bits_in-=8;   }
        }while(0);
      a <<= 8;   }   }
      }
    else
      {
      DWORD i,z=0; do{
        DWORD bits=0; DWORD len=3; (void)( (!!(bits <= ((1U << len) - 1U))) || (assert("bits <= ((1U << len) - 1U)"),0) ); d->m_bit_buffer |= (bits << d->m_bits_in); d->m_bits_in+=len; while(d->m_bits_in >= 8)   { if(d->m_pOutput_buf < d->m_pOutput_buf_end) *d->m_pOutput_buf++=(BYTE)(d->m_bit_buffer); d->m_bit_buffer >>= 8; d->m_bits_in-=8;   }
        }while(0);
      if(d->m_bits_in)   { do{
          DWORD bits=0; DWORD len=8 - d->m_bits_in; (void)( (!!(bits <= ((1U << len) - 1U))) || (assert("bits <= ((1U << len) - 1U)"),0) ); d->m_bit_buffer |= (bits << d->m_bits_in); d->m_bits_in+=len; while(d->m_bits_in >= 8)   { if(d->m_pOutput_buf < d->m_pOutput_buf_end) *d->m_pOutput_buf++=(BYTE)(d->m_bit_buffer); d->m_bit_buffer >>= 8; d->m_bits_in-=8;   }
          }while(0);
          } for(i=2; i; --i,z ^= 0xFFFF)   { do{
            DWORD bits=z & 0xFFFF; DWORD len=16; (void)( (!!(bits <= ((1U << len) - 1U))) || (assert("bits <= ((1U << len) - 1U)"),0) ); d->m_bit_buffer |= (bits << d->m_bits_in); d->m_bits_in+=len; while(d->m_bits_in >= 8)   { if(d->m_pOutput_buf < d->m_pOutput_buf_end) *d->m_pOutput_buf++=(BYTE)(d->m_bit_buffer); d->m_bit_buffer >>= 8; d->m_bits_in-=8;   }
            }while(0);
            }
      }
    }
  (void)( (!!(d->m_pOutput_buf < d->m_pOutput_buf_end)) || (assert("d->m_pOutput_buf < d->m_pOutput_buf_end"),0) );
  memset(&d->m_huff_count[0][0],0,sizeof(d->m_huff_count[0][0]) * TDEFL_MAX_HUFF_SYMBOLS_0);
  memset(&d->m_huff_count[1][0],0,sizeof(d->m_huff_count[1][0]) * TDEFL_MAX_HUFF_SYMBOLS_1);
  d->m_pLZ_code_buf=d->m_lz_code_buf + 1; d->m_pLZ_flags=d->m_lz_code_buf; d->m_num_flags_left=8; d->m_lz_code_buf_dict_pos+=d->m_total_lz_bytes; d->m_total_lz_bytes=0; d->m_block_index++;
  if((n=(int)(d->m_pOutput_buf - pOutput_buf_start))!=0)
    {
    if(d->m_pPut_buf_func)
      {
      *d->m_pIn_buf_size=d->m_pSrc - (BYTE *)d->m_pIn_buf;
      if(!(*d->m_pPut_buf_func)(d->m_output_buf,n,d->m_pPut_buf_user))
        return (d->m_prev_return_status=TDEFL_STATUS_PUT_BUF_FAILED);
      }
    else if(pOutput_buf_start == d->m_output_buf)
      {
      int bytes_to_copy=(int)min((size_t)n,(size_t)(*d->m_pOut_buf_size - d->m_out_buf_ofs));
      memcpy((BYTE *)d->m_pOut_buf + d->m_out_buf_ofs,d->m_output_buf,bytes_to_copy);
      d->m_out_buf_ofs+=bytes_to_copy;
      if((n-=bytes_to_copy)!=0)
        {
        d->m_output_flush_ofs=bytes_to_copy;
        d->m_output_flush_remaining=n;
        }
      }
    else
      {
      d->m_out_buf_ofs+=n;
      }
    }
  return d->m_output_flush_remaining;
  }

int tdefl_compress_fast(tdefl_compressor *d)
  {
  DWORD lookahead_pos=d->m_lookahead_pos,lookahead_size=d->m_lookahead_size,dict_size=d->m_dict_size,total_lz_bytes=d->m_total_lz_bytes,num_flags_left=d->m_num_flags_left;
  BYTE *pLZ_code_buf=d->m_pLZ_code_buf,*pLZ_flags=d->m_pLZ_flags;
  DWORD cur_pos=lookahead_pos & TDEFL_LZ_DICT_SIZE_MASK;
  while((d->m_src_buf_left) || ((d->m_flush) && (lookahead_size)))
    {
    DWORD TDEFL_COMP_FAST_LOOKAHEAD_SIZE=4096;
    DWORD dst_pos=(lookahead_pos + lookahead_size) & TDEFL_LZ_DICT_SIZE_MASK;
    DWORD num_bytes_to_process=(DWORD)min(d->m_src_buf_left,TDEFL_COMP_FAST_LOOKAHEAD_SIZE - lookahead_size);
    d->m_src_buf_left-=num_bytes_to_process;
    lookahead_size+=num_bytes_to_process;
    while(num_bytes_to_process)
      {
      DWORD n=min(TDEFL_LZ_DICT_SIZE - dst_pos,num_bytes_to_process);
      memcpy(d->m_dict + dst_pos,d->m_pSrc,n);
      if(dst_pos < (TDEFL_MAX_MATCH_LEN - 1))
        memcpy(d->m_dict + TDEFL_LZ_DICT_SIZE + dst_pos,d->m_pSrc,min(n,(TDEFL_MAX_MATCH_LEN - 1) - dst_pos));
      d->m_pSrc+=n;
      dst_pos=(dst_pos + n) & TDEFL_LZ_DICT_SIZE_MASK;
      num_bytes_to_process-=n;
      }
    dict_size=min(TDEFL_LZ_DICT_SIZE - lookahead_size,dict_size);
    if((!d->m_flush) && (lookahead_size < TDEFL_COMP_FAST_LOOKAHEAD_SIZE)) break;
    while(lookahead_size >= 4)
      {
      DWORD cur_match_dist,cur_match_len=1;
      BYTE *pCur_dict=d->m_dict + cur_pos;
      DWORD first_trigram=(*(DWORD *)pCur_dict) & 0xFFFFFF;
      DWORD hash=(first_trigram ^ (first_trigram >> (24 - (TDEFL_LZ_HASH_BITS - 8)))) & TDEFL_LEVEL1_HASH_SIZE_MASK;
      DWORD probe_pos=d->m_hash[hash];
      d->m_hash[hash]=(WORD)lookahead_pos;
      if(((cur_match_dist=(WORD)(lookahead_pos - probe_pos)) <= dict_size) && ((*(DWORD *)(d->m_dict + (probe_pos &= TDEFL_LZ_DICT_SIZE_MASK)) & 0xFFFFFF) == first_trigram))
        {
        WORD *p=(WORD *)pCur_dict;
        WORD *q=(WORD *)(d->m_dict + probe_pos);
        DWORD probe_len=32;
        do{
          }while( (*(WORD*)(++p) == *(WORD*)(++q)) && (*(WORD*)(++p) == *(WORD*)(++q)) &&
          (*(WORD*)(++p) == *(WORD*)(++q)) && (*(WORD*)(++p) == *(WORD*)(++q)) && (--probe_len > 0) );
        cur_match_len=((DWORD)(p - (WORD *)pCur_dict) * 2) + (DWORD)(*(BYTE *)p == *(BYTE *)q);
        if(!probe_len)
          cur_match_len=cur_match_dist ? TDEFL_MAX_MATCH_LEN : 0;
        if((cur_match_len < TDEFL_MIN_MATCH_LEN) || ((cur_match_len == TDEFL_MIN_MATCH_LEN) && (cur_match_dist >= 8U*1024U)))
          {
          cur_match_len=1;
          *pLZ_code_buf++=(BYTE)first_trigram;
          *pLZ_flags=(BYTE)(*pLZ_flags >> 1);
          d->m_huff_count[0][(BYTE)first_trigram]++;
          }
        else
          {
          DWORD s0,s1;
          cur_match_len=min(cur_match_len,lookahead_size);
          (void)( (!!((cur_match_len >= TDEFL_MIN_MATCH_LEN) && (cur_match_dist >= 1) && (cur_match_dist <= TDEFL_LZ_DICT_SIZE))) || (assert("(cur_match_len >= TDEFL_MIN_MATCH_LEN) && (cur_match_dist >= 1) && (cur_match_dist <= TDEFL_LZ_DICT_SIZE)"),0) );
          cur_match_dist--;
          pLZ_code_buf[0]=(BYTE)(cur_match_len - TDEFL_MIN_MATCH_LEN);
          *(WORD *)(&pLZ_code_buf[1])=(WORD)cur_match_dist;
          pLZ_code_buf+=3;
          *pLZ_flags=(BYTE)((*pLZ_flags >> 1) | 0x80);
          s0=s_tdefl_small_dist_sym[cur_match_dist & 511];
          s1=s_tdefl_large_dist_sym[cur_match_dist >> 8];
          d->m_huff_count[1][(cur_match_dist < 512) ? s0 : s1]++;
          d->m_huff_count[0][s_tdefl_len_sym[cur_match_len - TDEFL_MIN_MATCH_LEN]]++;
          }
        }
      else
        {
        *pLZ_code_buf++=(BYTE)first_trigram;
        *pLZ_flags=(BYTE)(*pLZ_flags >> 1);
        d->m_huff_count[0][(BYTE)first_trigram]++;
        }
      if(--num_flags_left == 0)   { num_flags_left=8; pLZ_flags=pLZ_code_buf++;   }
      total_lz_bytes+=cur_match_len;
      lookahead_pos+=cur_match_len;
      dict_size=min(dict_size + cur_match_len,TDEFL_LZ_DICT_SIZE);
      cur_pos=(cur_pos + cur_match_len) & TDEFL_LZ_DICT_SIZE_MASK;
      (void)( (!!(lookahead_size >= cur_match_len)) || (assert("lookahead_size >= cur_match_len"),0) );
      lookahead_size-=cur_match_len;
      if(pLZ_code_buf > &d->m_lz_code_buf[TDEFL_LZ_CODE_BUF_SIZE - 8])
        {
        int n;
        d->m_lookahead_pos=lookahead_pos; d->m_lookahead_size=lookahead_size; d->m_dict_size=dict_size;
        d->m_total_lz_bytes=total_lz_bytes; d->m_pLZ_code_buf=pLZ_code_buf; d->m_pLZ_flags=pLZ_flags; d->m_num_flags_left=num_flags_left;
        if((n=tdefl_flush_block(d,0))!=0)
          return (n < 0) ? 0 : 1;
        total_lz_bytes=d->m_total_lz_bytes; pLZ_code_buf=d->m_pLZ_code_buf; pLZ_flags=d->m_pLZ_flags; num_flags_left=d->m_num_flags_left;
        }
      }
    while(lookahead_size)
      {
      BYTE lit=d->m_dict[cur_pos];
      total_lz_bytes++;
      *pLZ_code_buf++=lit;
      *pLZ_flags=(BYTE)(*pLZ_flags >> 1);
      if(--num_flags_left == 0)   { num_flags_left=8; pLZ_flags=pLZ_code_buf++;   }
      d->m_huff_count[0][lit]++;
      lookahead_pos++;
      dict_size=min(dict_size + 1,TDEFL_LZ_DICT_SIZE);
      cur_pos=(cur_pos + 1) & TDEFL_LZ_DICT_SIZE_MASK;
      lookahead_size--;
      if(pLZ_code_buf > &d->m_lz_code_buf[TDEFL_LZ_CODE_BUF_SIZE - 8])
        {
        int n;
        d->m_lookahead_pos=lookahead_pos; d->m_lookahead_size=lookahead_size; d->m_dict_size=dict_size;
        d->m_total_lz_bytes=total_lz_bytes; d->m_pLZ_code_buf=pLZ_code_buf; d->m_pLZ_flags=pLZ_flags; d->m_num_flags_left=num_flags_left;
        if((n=tdefl_flush_block(d,0))!=0)
          return (n < 0) ? 0 : 1;
        total_lz_bytes=d->m_total_lz_bytes; pLZ_code_buf=d->m_pLZ_code_buf; pLZ_flags=d->m_pLZ_flags; num_flags_left=d->m_num_flags_left;
        }
      }
    }
  d->m_lookahead_pos=lookahead_pos; d->m_lookahead_size=lookahead_size; d->m_dict_size=dict_size;
  d->m_total_lz_bytes=total_lz_bytes; d->m_pLZ_code_buf=pLZ_code_buf; d->m_pLZ_flags=pLZ_flags; d->m_num_flags_left=num_flags_left;
  return 1;
  }

void tdefl_record_literal(tdefl_compressor *d,BYTE lit)
  {
  d->m_total_lz_bytes++;
  *d->m_pLZ_code_buf++=lit;
  *d->m_pLZ_flags=(BYTE)(*d->m_pLZ_flags >> 1); if(--d->m_num_flags_left == 0)   { d->m_num_flags_left=8; d->m_pLZ_flags=d->m_pLZ_code_buf++;   }
  d->m_huff_count[0][lit]++;
  }

void tdefl_record_match(tdefl_compressor *d,DWORD match_len,DWORD match_dist)
  {
  DWORD s0,s1;
  (void)( (!!((match_len >= TDEFL_MIN_MATCH_LEN) && (match_dist >= 1) && (match_dist <= TDEFL_LZ_DICT_SIZE))) || (assert("(match_len >= TDEFL_MIN_MATCH_LEN) && (match_dist >= 1) && (match_dist <= TDEFL_LZ_DICT_SIZE)"),0) );
  d->m_total_lz_bytes+=match_len;
  d->m_pLZ_code_buf[0]=(BYTE)(match_len - TDEFL_MIN_MATCH_LEN);
  match_dist-=1;
  d->m_pLZ_code_buf[1]=(BYTE)(match_dist & 0xFF);
  d->m_pLZ_code_buf[2]=(BYTE)(match_dist >> 8); d->m_pLZ_code_buf+=3;
  *d->m_pLZ_flags=(BYTE)((*d->m_pLZ_flags >> 1) | 0x80); if(--d->m_num_flags_left == 0)   { d->m_num_flags_left=8; d->m_pLZ_flags=d->m_pLZ_code_buf++;   }
  s0=s_tdefl_small_dist_sym[match_dist & 511]; s1=s_tdefl_large_dist_sym[(match_dist >> 8) & 127];
  d->m_huff_count[1][(match_dist < 512) ? s0 : s1]++;
  if(match_len >= TDEFL_MIN_MATCH_LEN) d->m_huff_count[0][s_tdefl_len_sym[match_len - TDEFL_MIN_MATCH_LEN]]++;
  }

int tdefl_compress_normal(tdefl_compressor *d)
  {
  BYTE *pSrc=d->m_pSrc; size_t src_buf_left=d->m_src_buf_left;
  tdefl_flush flush=d->m_flush;
  while((src_buf_left) || ((flush) && (d->m_lookahead_size)))
    {
    DWORD len_to_move,cur_match_dist,cur_match_len,cur_pos;
    if((d->m_lookahead_size + d->m_dict_size) >= (TDEFL_MIN_MATCH_LEN - 1))
      {
      DWORD dst_pos=(d->m_lookahead_pos + d->m_lookahead_size) & TDEFL_LZ_DICT_SIZE_MASK,ins_pos=d->m_lookahead_pos + d->m_lookahead_size - 2;
      DWORD hash=(d->m_dict[ins_pos & TDEFL_LZ_DICT_SIZE_MASK] << TDEFL_LZ_HASH_SHIFT) ^ d->m_dict[(ins_pos + 1) & TDEFL_LZ_DICT_SIZE_MASK];
      DWORD num_bytes_to_process=(DWORD)min(src_buf_left,TDEFL_MAX_MATCH_LEN - d->m_lookahead_size);
      BYTE *pSrc_end=pSrc + num_bytes_to_process;
      src_buf_left-=num_bytes_to_process;
      d->m_lookahead_size+=num_bytes_to_process;
      while(pSrc!=pSrc_end)
        {
        BYTE c=*pSrc++; d->m_dict[dst_pos]=c; if(dst_pos < (TDEFL_MAX_MATCH_LEN - 1)) d->m_dict[TDEFL_LZ_DICT_SIZE + dst_pos]=c;
        hash=((hash << TDEFL_LZ_HASH_SHIFT) ^ c) & (TDEFL_LZ_HASH_SIZE - 1);
        d->m_next[ins_pos & TDEFL_LZ_DICT_SIZE_MASK]=d->m_hash[hash]; d->m_hash[hash]=(WORD)(ins_pos);
        dst_pos=(dst_pos + 1) & TDEFL_LZ_DICT_SIZE_MASK; ins_pos++;
        }
      }
    else
      {
      while((src_buf_left) && (d->m_lookahead_size < TDEFL_MAX_MATCH_LEN))
        {
        BYTE c=*pSrc++;
        DWORD dst_pos=(d->m_lookahead_pos + d->m_lookahead_size) & TDEFL_LZ_DICT_SIZE_MASK;
        src_buf_left--;
        d->m_dict[dst_pos]=c;
        if(dst_pos < (TDEFL_MAX_MATCH_LEN - 1))
          d->m_dict[TDEFL_LZ_DICT_SIZE + dst_pos]=c;
        if((++d->m_lookahead_size + d->m_dict_size) >= TDEFL_MIN_MATCH_LEN)
          {
          DWORD ins_pos=d->m_lookahead_pos + (d->m_lookahead_size - 1) - 2;
          DWORD hash=((d->m_dict[ins_pos & TDEFL_LZ_DICT_SIZE_MASK] << (TDEFL_LZ_HASH_SHIFT * 2)) ^ (d->m_dict[(ins_pos + 1) & TDEFL_LZ_DICT_SIZE_MASK] << TDEFL_LZ_HASH_SHIFT) ^ c) & (TDEFL_LZ_HASH_SIZE - 1);
          d->m_next[ins_pos & TDEFL_LZ_DICT_SIZE_MASK]=d->m_hash[hash]; d->m_hash[hash]=(WORD)(ins_pos);
          }
        }
      }
    d->m_dict_size=min(TDEFL_LZ_DICT_SIZE - d->m_lookahead_size,d->m_dict_size);
    if((!flush) && (d->m_lookahead_size < TDEFL_MAX_MATCH_LEN))
      break;
    len_to_move=1; cur_match_dist=0; cur_match_len=d->m_saved_match_len ? d->m_saved_match_len : (TDEFL_MIN_MATCH_LEN - 1); cur_pos=d->m_lookahead_pos & TDEFL_LZ_DICT_SIZE_MASK;
    if(d->m_flags & (TDEFL_RLE_MATCHES | TDEFL_FORCE_ALL_RAW_BLOCKS))
      {
      if((d->m_dict_size) && (!(d->m_flags & TDEFL_FORCE_ALL_RAW_BLOCKS)))
        {
        BYTE c=d->m_dict[(cur_pos - 1) & TDEFL_LZ_DICT_SIZE_MASK];
        cur_match_len=0; while(cur_match_len < d->m_lookahead_size)   { if(d->m_dict[cur_pos + cur_match_len]!=c) break; cur_match_len++;   }
        if(cur_match_len < TDEFL_MIN_MATCH_LEN) cur_match_len=0; else cur_match_dist=1;
        }
      }
    else
      {
      tdefl_find_match(d,d->m_lookahead_pos,d->m_dict_size,d->m_lookahead_size,&cur_match_dist,&cur_match_len);
      }
    if(((cur_match_len == TDEFL_MIN_MATCH_LEN) && (cur_match_dist >= 8U*1024U)) || (cur_pos == cur_match_dist) || ((d->m_flags & TDEFL_FILTER_MATCHES) && (cur_match_len <= 5)))
      {
      cur_match_dist=cur_match_len=0;
      }
    if(d->m_saved_match_len)
      {
      if(cur_match_len > d->m_saved_match_len)
        {
        tdefl_record_literal(d,(BYTE)d->m_saved_lit);
        if(cur_match_len >= 128)
          {
          tdefl_record_match(d,cur_match_len,cur_match_dist);
          d->m_saved_match_len=0; len_to_move=cur_match_len;
          }
        else
          {
          d->m_saved_lit=d->m_dict[cur_pos]; d->m_saved_match_dist=cur_match_dist; d->m_saved_match_len=cur_match_len;
          }
        }
      else
        {
        tdefl_record_match(d,d->m_saved_match_len,d->m_saved_match_dist);
        len_to_move=d->m_saved_match_len - 1; d->m_saved_match_len=0;
        }
      }
    else if(!cur_match_dist)
      tdefl_record_literal(d,d->m_dict[min(cur_pos,sizeof(d->m_dict) - 1)]);
    else if((d->m_greedy_parsing) || (d->m_flags & TDEFL_RLE_MATCHES) || (cur_match_len >= 128))
      {
      tdefl_record_match(d,cur_match_len,cur_match_dist);
      len_to_move=cur_match_len;
      }
    else
      {
      d->m_saved_lit=d->m_dict[min(cur_pos,sizeof(d->m_dict) - 1)]; d->m_saved_match_dist=cur_match_dist; d->m_saved_match_len=cur_match_len;
      }
    d->m_lookahead_pos+=len_to_move;
    (void)( (!!(d->m_lookahead_size >= len_to_move)) || (assert("d->m_lookahead_size >= len_to_move"),0) );
    d->m_lookahead_size-=len_to_move;
    d->m_dict_size=min(d->m_dict_size + len_to_move,TDEFL_LZ_DICT_SIZE);
    if( (d->m_pLZ_code_buf > &d->m_lz_code_buf[TDEFL_LZ_CODE_BUF_SIZE - 8]) ||
         ( (d->m_total_lz_bytes > 31*1024) && (((((DWORD)(d->m_pLZ_code_buf - d->m_lz_code_buf) * 115) >> 7) >= d->m_total_lz_bytes) || (d->m_flags & TDEFL_FORCE_ALL_RAW_BLOCKS))) )
      {
      int n;
      d->m_pSrc=pSrc; d->m_src_buf_left=src_buf_left;
      if((n=tdefl_flush_block(d,0))!=0)
        return (n < 0) ? 0 : 1;
      }
    }
  d->m_pSrc=pSrc; d->m_src_buf_left=src_buf_left;
  return 1;
  }

tdefl_status tdefl_compress(tdefl_compressor *d,void*pIn_buf,size_t *pIn_buf_size,void*pOut_buf,size_t *pOut_buf_size,tdefl_flush flush)
  {
  if(!d)
    {
    if(pIn_buf_size) *pIn_buf_size=0;
    if(pOut_buf_size) *pOut_buf_size=0;
    return TDEFL_STATUS_BAD_PARAM;
    }
  d->m_pIn_buf=pIn_buf; d->m_pIn_buf_size=pIn_buf_size;
  d->m_pOut_buf=pOut_buf; d->m_pOut_buf_size=pOut_buf_size;
  d->m_pSrc=(BYTE *)(pIn_buf); d->m_src_buf_left=pIn_buf_size ? *pIn_buf_size : 0;
  d->m_out_buf_ofs=0;
  d->m_flush=flush;
  if( ((d->m_pPut_buf_func!=NULL) == ((pOut_buf!=NULL) || (pOut_buf_size!=NULL))) || (d->m_prev_return_status!=TDEFL_STATUS_OKAY) ||
        (d->m_wants_to_finish && (flush!=TDEFL_FINISH)) || (pIn_buf_size && *pIn_buf_size && !pIn_buf) || (pOut_buf_size && *pOut_buf_size && !pOut_buf) )
    {
    if(pIn_buf_size) *pIn_buf_size=0;
    if(pOut_buf_size) *pOut_buf_size=0;
    return (d->m_prev_return_status=TDEFL_STATUS_BAD_PARAM);
    }
  d->m_wants_to_finish |= (flush == TDEFL_FINISH);
  if((d->m_output_flush_remaining) || (d->m_finished))
    return (d->m_prev_return_status=tdefl_flush_output_buffer(d));
  if(((d->m_flags & TDEFL_MAX_PROBES_MASK) == 1) &&
      ((d->m_flags & TDEFL_GREEDY_PARSING_FLAG)!=0) &&
      ((d->m_flags & (TDEFL_FILTER_MATCHES | TDEFL_FORCE_ALL_RAW_BLOCKS | TDEFL_RLE_MATCHES)) == 0))
    {
    if(!tdefl_compress_fast(d))
      return d->m_prev_return_status;
    }
  else
    {
    if(!tdefl_compress_normal(d))
      return d->m_prev_return_status;
    }
  if((d->m_flags & (TDEFL_WRITE_ZLIB_HEADER | TDEFL_COMPUTE_ADLER32)) && (pIn_buf))
    d->m_adler32=(DWORD)mz_adler32(d->m_adler32,(BYTE *)pIn_buf,d->m_pSrc - (BYTE *)pIn_buf);
  if((flush) && (!d->m_lookahead_size) && (!d->m_src_buf_left) && (!d->m_output_flush_remaining))
    {
    if(tdefl_flush_block(d,flush) < 0)
      return d->m_prev_return_status;
    d->m_finished=(flush == TDEFL_FINISH);
    if(flush == TDEFL_FULL_FLUSH)   { memset(&d->m_hash,0,sizeof(d->m_hash)); memset(&d->m_next,0,sizeof(d->m_next)); d->m_dict_size=0;   }
    }
  return (d->m_prev_return_status=tdefl_flush_output_buffer(d));
  }

tdefl_status tdefl_compress_buffer(tdefl_compressor *d,void*pIn_buf,size_t in_buf_size,tdefl_flush flush)
  {
  (void)( (!!(d->m_pPut_buf_func)) || (assert("d->m_pPut_buf_func"),0) ); return tdefl_compress(d,pIn_buf,&in_buf_size,NULL,NULL,flush);
  }

tdefl_status tdefl_init(tdefl_compressor *d,tdefl_put_buf_func_ptr pPut_buf_func,void*pPut_buf_user,int flags)
  {
  d->m_pPut_buf_func=pPut_buf_func; d->m_pPut_buf_user=pPut_buf_user;
  d->m_flags=(DWORD)(flags); d->m_max_probes[0]=1 + ((flags & 0xFFF) + 2) / 3; d->m_greedy_parsing=(flags & TDEFL_GREEDY_PARSING_FLAG)!=0;
  d->m_max_probes[1]=1 + (((flags & 0xFFF) >> 2) + 2) / 3;
  if(!(flags & TDEFL_NONDETERMINISTIC_PARSING_FLAG)) memset(&d->m_hash,0,sizeof(d->m_hash));
  d->m_lookahead_pos=d->m_lookahead_size=d->m_dict_size=d->m_total_lz_bytes=d->m_lz_code_buf_dict_pos=d->m_bits_in=0;
  d->m_output_flush_ofs=d->m_output_flush_remaining=d->m_finished=d->m_block_index=d->m_bit_buffer=d->m_wants_to_finish=0;
  d->m_pLZ_code_buf=d->m_lz_code_buf + 1; d->m_pLZ_flags=d->m_lz_code_buf; d->m_num_flags_left=8;
  d->m_pOutput_buf=d->m_output_buf; d->m_pOutput_buf_end=d->m_output_buf; d->m_prev_return_status=TDEFL_STATUS_OKAY;
  d->m_saved_match_dist=d->m_saved_match_len=d->m_saved_lit=0; d->m_adler32=1;
  d->m_pIn_buf=NULL; d->m_pOut_buf=NULL;
  d->m_pIn_buf_size=NULL; d->m_pOut_buf_size=NULL;
  d->m_flush=TDEFL_NO_FLUSH; d->m_pSrc=NULL; d->m_src_buf_left=0; d->m_out_buf_ofs=0;
  memset(&d->m_huff_count[0][0],0,sizeof(d->m_huff_count[0][0]) * TDEFL_MAX_HUFF_SYMBOLS_0);
  memset(&d->m_huff_count[1][0],0,sizeof(d->m_huff_count[1][0]) * TDEFL_MAX_HUFF_SYMBOLS_1);
  return TDEFL_STATUS_OKAY;
  }

tdefl_status tdefl_get_prev_return_status(tdefl_compressor *d)
  {
  return d->m_prev_return_status;
  }

DWORD tdefl_get_adler32(tdefl_compressor *d)
  {
  return d->m_adler32;
  }

int mz_deflateInit2(mz_stream* pStream,int level,int method,int window_bits,int mem_level,int strategy)
  {
  tdefl_compressor *pComp;
  DWORD comp_flags=TDEFL_COMPUTE_ADLER32 | tdefl_create_comp_flags_from_zip_params(level,window_bits,strategy);
  if(!pStream) return MZ_STREAM_ERROR;
  if((method!=8) || ((mem_level < 1) || (mem_level > 9)) || ((window_bits!=15) && (-window_bits!=15))) return MZ_PARAM_ERROR;
  pStream->data_type=0;
  pStream->adler=(1);
  pStream->msg=NULL;
  pStream->reserved=0;
  pStream->total_in=0;
  pStream->total_out=0;
  if(!pStream->zalloc) pStream->zalloc=def_alloc_func;
  if(!pStream->zfree) pStream->zfree=def_free_func;
  pComp=(tdefl_compressor *)pStream->zalloc(pStream->opaque,1,sizeof(tdefl_compressor));
  if(!pComp)
    return MZ_MEM_ERROR;
  pStream->state=(struct mz_internal_state *)pComp;
  if(tdefl_init(pComp,NULL,NULL,comp_flags)!=TDEFL_STATUS_OKAY)
    {
    mz_deflateEnd(pStream);
    return MZ_PARAM_ERROR;
    }
  return MZ_OK;
  }

int mz_deflateInit(mz_stream* pStream,int level)
  {
  return mz_deflateInit2(pStream,level,8,15,9,MZ_DEFAULT_STRATEGY);
  }

int mz_deflateReset(mz_stream* pStream)
  {
  if((!pStream) || (!pStream->state) || (!pStream->zalloc) || (!pStream->zfree)) return MZ_STREAM_ERROR;
  pStream->total_in=pStream->total_out=0;
  tdefl_init((tdefl_compressor*)pStream->state,NULL,NULL,((tdefl_compressor*)pStream->state)->m_flags);
  return MZ_OK;
  }

int mz_deflate(mz_stream* pStream,int flush)
  {
  size_t in_bytes,out_bytes;
  DWORD orig_total_in,orig_total_out;
  int mz_status=MZ_OK;
  if((!pStream) || (!pStream->state) || (flush < 0) || (flush > MZ_FINISH) || (!pStream->next_out)) return MZ_STREAM_ERROR;
  if(!pStream->avail_out) return MZ_BUF_ERROR;
  if(flush == MZ_PARTIAL_FLUSH) flush=MZ_SYNC_FLUSH;
  if(((tdefl_compressor*)pStream->state)->m_prev_return_status == TDEFL_STATUS_DONE)
    return (flush == MZ_FINISH) ? MZ_STREAM_END : MZ_BUF_ERROR;
  orig_total_in=pStream->total_in; orig_total_out=pStream->total_out;
  for( ; ; )
    {
    tdefl_status defl_status;
    in_bytes=pStream->avail_in; out_bytes=pStream->avail_out;
    defl_status=tdefl_compress((tdefl_compressor*)pStream->state,pStream->next_in,&in_bytes,pStream->next_out,&out_bytes,(tdefl_flush)flush);
    pStream->next_in+=(DWORD)in_bytes; pStream->avail_in-=(DWORD)in_bytes;
    pStream->total_in+=(DWORD)in_bytes; pStream->adler=tdefl_get_adler32((tdefl_compressor*)pStream->state);
    pStream->next_out+=(DWORD)out_bytes; pStream->avail_out-=(DWORD)out_bytes;
    pStream->total_out+=(DWORD)out_bytes;
    if(defl_status < 0)
      {
      mz_status=MZ_STREAM_ERROR;
      break;
      }
    else if(defl_status == TDEFL_STATUS_DONE)
      {
      mz_status=MZ_STREAM_END;
      break;
      }
    else if(!pStream->avail_out)
      break;
    else if((!pStream->avail_in) && (flush!=MZ_FINISH))
      {
      if((flush) || (pStream->total_in!=orig_total_in) || (pStream->total_out!=orig_total_out))
        break;
      return MZ_BUF_ERROR;
      }
    }
  return mz_status;
  }

DWORD mz_deflateBound(mz_stream* pStream,DWORD source_len)
  {
  (void)pStream;
  return max(128 + (source_len * 110) / 100,128 + source_len + ((source_len / (31 * 1024)) + 1) * 5);
  }

int mz_compress2(unsigned char*pDest,DWORD *pDest_len,unsigned char*pSource,DWORD source_len,int level)
  {
  int status;
  mz_stream stream;
  memset(&stream,0,sizeof(stream));
  if((source_len | *pDest_len) > 0xFFFFFFFFU) return MZ_PARAM_ERROR;
  stream.next_in=pSource;
  stream.avail_in=(DWORD)source_len;
  stream.next_out=pDest;
  stream.avail_out=(DWORD)*pDest_len;
  status=mz_deflateInit(&stream,level);
  if(status!=MZ_OK) return status;
  status=mz_deflate(&stream,MZ_FINISH);
  if(status!=MZ_STREAM_END)
    {
    mz_deflateEnd(&stream);
    return (status == MZ_OK) ? MZ_BUF_ERROR : status;
    }
  *pDest_len=stream.total_out;
  return mz_deflateEnd(&stream);
  }

int mz_compress(unsigned char*pDest,DWORD *pDest_len,unsigned char*pSource,DWORD source_len)
  {
  return mz_compress2(pDest,pDest_len,pSource,source_len,MZ_DEFAULT_COMPRESSION);
  }

DWORD mz_compressBound(DWORD source_len)
  {
  return mz_deflateBound(NULL,source_len);
  }

typedef struct
  {
  tinfl_decompressor m_decomp;
  DWORD m_dict_ofs,m_dict_avail,m_first_call,m_has_flushed; int m_window_bits;
  BYTE m_dict[32768];
  tinfl_status m_last_status;
  } inflate_state;

int mz_inflateInit2(mz_stream* pStream,int window_bits)
  {
  inflate_state *pDecomp;
  if(!pStream) return MZ_STREAM_ERROR;
  if((window_bits!=15) && (-window_bits!=15)) return MZ_PARAM_ERROR;
  pStream->data_type=0;
  pStream->adler=0;
  pStream->msg=NULL;
  pStream->total_in=0;
  pStream->total_out=0;
  pStream->reserved=0;
  if(!pStream->zalloc) pStream->zalloc=def_alloc_func;
  if(!pStream->zfree) pStream->zfree=def_free_func;
  pDecomp=(inflate_state*)pStream->zalloc(pStream->opaque,1,sizeof(inflate_state));
  if(!pDecomp) return MZ_MEM_ERROR;
  pStream->state=(struct mz_internal_state *)pDecomp;
  do{
    (&pDecomp->m_decomp)->m_state=0;
    }while(0);
  pDecomp->m_dict_ofs=0;
  pDecomp->m_dict_avail=0;
  pDecomp->m_last_status=TINFL_STATUS_NEEDS_MORE_INPUT;
  pDecomp->m_first_call=1;
  pDecomp->m_has_flushed=0;
  pDecomp->m_window_bits=window_bits;
  return MZ_OK;
  }

int mz_inflateInit(mz_stream* pStream)
  {
   return mz_inflateInit2(pStream,15);
  }

tinfl_status tinfl_decompress(tinfl_decompressor *r,BYTE *pIn_buf_next,size_t *pIn_buf_size,BYTE *pOut_buf_start,BYTE *pOut_buf_next,size_t *pOut_buf_size,DWORD decomp_flags)
  {
  static int s_length_base[31]=  { 3,4,5,6,7,8,9,10,11,13,15,17,19,23,27,31,35,43,51,59,67,83,99,115,131,163,195,227,258,0,0   };
  static int s_length_extra[31]=   { 0,0,0,0,0,0,0,0,1,1,1,1,2,2,2,2,3,3,3,3,4,4,4,4,5,5,5,5,0,0,0   };
  static int s_dist_base[32]=  { 1,2,3,4,5,7,9,13,17,25,33,49,65,97,129,193,257,385,513,769,1025,1537,2049,3073,4097,6145,8193,12289,16385,24577,0,0  };
  static int s_dist_extra[32]=  { 0,0,0,0,1,1,2,2,3,3,4,4,5,5,6,6,7,7,8,8,9,9,10,10,11,11,12,12,13,13  };
  static BYTE s_length_dezigzag[19]=  { 16,17,18,0,8,7,9,6,10,5,11,4,12,3,13,2,14,1,15   };
  static int s_min_table_sizes[3]=  { 257,1,4   };
  tinfl_status status=TINFL_STATUS_FAILED; DWORD num_bits,dist,counter,num_extra; tinfl_bit_buf_t bit_buf;
  BYTE *pIn_buf_cur=pIn_buf_next,*pIn_buf_end=pIn_buf_next + *pIn_buf_size;
  BYTE *pOut_buf_cur=pOut_buf_next,*pOut_buf_end=pOut_buf_next + *pOut_buf_size;
  size_t out_buf_size_mask=(decomp_flags & TINFL_FLAG_USING_NON_WRAPPING_OUTPUT_BUF) ? (size_t)-1 : ((pOut_buf_next - pOut_buf_start) + *pOut_buf_size) - 1,dist_from_out_buf_start;
  if(((out_buf_size_mask + 1) & out_buf_size_mask) || (pOut_buf_next < pOut_buf_start))   { *pIn_buf_size=*pOut_buf_size=0; return TINFL_STATUS_BAD_PARAM;   }
  num_bits=r->m_num_bits; bit_buf=r->m_bit_buf; dist=r->m_dist; counter=r->m_counter; num_extra=r->m_num_extra; dist_from_out_buf_start=r->m_dist_from_out_buf_start;
  switch(r->m_state)   { case 0:
  bit_buf=num_bits=dist=counter=num_extra=r->m_zhdr0=r->m_zhdr1=0; r->m_z_adler32=r->m_check_adler32=1;
  if(decomp_flags & TINFL_FLAG_PARSE_ZLIB_HEADER)
    {
    do{
      if(pIn_buf_cur >= pIn_buf_end)   { for( ; ; )   { if(decomp_flags & TINFL_FLAG_HAS_MORE_INPUT)   { do{
        status=TINFL_STATUS_NEEDS_MORE_INPUT; r->m_state=1; goto common_exit; case 1:;
        }while(0);
      if(pIn_buf_cur < pIn_buf_end)   { r->m_zhdr0=*pIn_buf_cur++; break;   }   } else   { r->m_zhdr0=0; break;   }   }   } else r->m_zhdr0=*pIn_buf_cur++;   }while(0);
    do{
          if(pIn_buf_cur >= pIn_buf_end)   { for( ; ; )   { if(decomp_flags & TINFL_FLAG_HAS_MORE_INPUT)   { do{
            status=TINFL_STATUS_NEEDS_MORE_INPUT; r->m_state=2; goto common_exit; case 2:;
            }while(0);
          if(pIn_buf_cur < pIn_buf_end)   { r->m_zhdr1=*pIn_buf_cur++; break;   }   } else   { r->m_zhdr1=0; break;   }   }   } else r->m_zhdr1=*pIn_buf_cur++;   }while(0);
    counter=(((r->m_zhdr0 * 256 + r->m_zhdr1) % 31!=0) || (r->m_zhdr1 & 32) || ((r->m_zhdr0 & 15)!=8));
    if(!(decomp_flags & TINFL_FLAG_USING_NON_WRAPPING_OUTPUT_BUF)) counter |= (((1U << (8U + (r->m_zhdr0 >> 4))) > 32768U) || ((out_buf_size_mask + 1) < (size_t)(1U << (8U + (r->m_zhdr0 >> 4)))));
    if(counter)   { do{
      for( ; ; )   { do{
        status=TINFL_STATUS_FAILED; r->m_state=36; goto common_exit; case 36:;
        }while(0);
        }   }while(0);
        }
    }
  do
    {
    do{
      if(num_bits < (DWORD)(3))   { do{
        DWORD c; do{
          if(pIn_buf_cur >= pIn_buf_end)   { for( ; ; )   { if(decomp_flags & TINFL_FLAG_HAS_MORE_INPUT)   { do{
            status=TINFL_STATUS_NEEDS_MORE_INPUT; r->m_state=3; goto common_exit; case 3:;
            }while(0);
          if(pIn_buf_cur < pIn_buf_end)   { c=*pIn_buf_cur++; break;   }   } else   { c=0; break;   }   }   } else c=*pIn_buf_cur++;   }while(0);
        bit_buf |= (((tinfl_bit_buf_t)c) << num_bits); num_bits+=8;   }while(num_bits < (DWORD)(3));   } r->m_final=bit_buf & ((1 << (3)) - 1); bit_buf >>= (3); num_bits-=(3);   }while(0);
      r->m_type=r->m_final >> 1;
    if(r->m_type == 0)
      {
      do{
        if(num_bits < (DWORD)(num_bits & 7))   { do{
          DWORD c; do{
            if(pIn_buf_cur >= pIn_buf_end)   { for( ; ; )   { if(decomp_flags & TINFL_FLAG_HAS_MORE_INPUT)   { do{
              status=TINFL_STATUS_NEEDS_MORE_INPUT; r->m_state=5; goto common_exit; case 5:;
              }while(0);
            if(pIn_buf_cur < pIn_buf_end)   { c=*pIn_buf_cur++; break;   }   } else   { c=0; break;   }   }   } else c=*pIn_buf_cur++;   }while(0);
          bit_buf |= (((tinfl_bit_buf_t)c) << num_bits); num_bits+=8;   }while(num_bits < (DWORD)(num_bits & 7));   } bit_buf >>= (num_bits & 7); num_bits-=(num_bits & 7);   }while(0);
      for(counter=0; counter < 4; ++counter)   { if(num_bits) do{
        if(num_bits < (DWORD)(8))   { do{
          DWORD c; do{
            if(pIn_buf_cur >= pIn_buf_end)   { for( ; ; )   { if(decomp_flags & TINFL_FLAG_HAS_MORE_INPUT)   { do{
              status=TINFL_STATUS_NEEDS_MORE_INPUT; r->m_state=6; goto common_exit; case 6:;
              }while(0);
            if(pIn_buf_cur < pIn_buf_end)   { c=*pIn_buf_cur++; break;   }   } else   { c=0; break;   }   }   } else c=*pIn_buf_cur++;   }while(0);
          bit_buf |= (((tinfl_bit_buf_t)c) << num_bits); num_bits+=8;   }while(num_bits < (DWORD)(8));   } r->m_raw_header[counter]=bit_buf & ((1 << (8)) - 1); bit_buf >>= (8); num_bits-=(8);   }while(0);
        else do{
                if(pIn_buf_cur >= pIn_buf_end)   { for( ; ; )   { if(decomp_flags & TINFL_FLAG_HAS_MORE_INPUT)   { do{
                  status=TINFL_STATUS_NEEDS_MORE_INPUT; r->m_state=7; goto common_exit; case 7:;
                  }while(0);
                if(pIn_buf_cur < pIn_buf_end)   { r->m_raw_header[counter]=*pIn_buf_cur++; break;   }   } else   { r->m_raw_header[counter]=0; break;   }   }   } else r->m_raw_header[counter]=*pIn_buf_cur++;   }while(0);
                }
      if((counter=(r->m_raw_header[0] | (r->m_raw_header[1] << 8)))!=(DWORD)(0xFFFF ^ (r->m_raw_header[2] | (r->m_raw_header[3] << 8))))   { do{
        for( ; ; )   { do{
          status=TINFL_STATUS_FAILED; r->m_state=39; goto common_exit; case 39:;
          }while(0);
          }   }while(0);
          }
      while((counter) && (num_bits))
        {
        do{
          if(num_bits < (DWORD)(8))   { do{
            DWORD c; do{
              if(pIn_buf_cur >= pIn_buf_end)   { for( ; ; )   { if(decomp_flags & TINFL_FLAG_HAS_MORE_INPUT)   { do{
                status=TINFL_STATUS_NEEDS_MORE_INPUT; r->m_state=51; goto common_exit; case 51:;
                }while(0);
              if(pIn_buf_cur < pIn_buf_end)   { c=*pIn_buf_cur++; break;   }   } else   { c=0; break;   }   }   } else c=*pIn_buf_cur++;   }while(0);
            bit_buf |= (((tinfl_bit_buf_t)c) << num_bits); num_bits+=8;   }while(num_bits < (DWORD)(8));   } dist=bit_buf & ((1 << (8)) - 1); bit_buf >>= (8); num_bits-=(8);   }while(0);
        while(pOut_buf_cur >= pOut_buf_end)   { do{
          status=TINFL_STATUS_HAS_MORE_OUTPUT; r->m_state=52; goto common_exit; case 52:;
          }while(0);
          }
        *pOut_buf_cur++=(BYTE)dist;
        counter--;
        }
      while(counter)
        {
        size_t n; while(pOut_buf_cur >= pOut_buf_end)   { do{
          status=TINFL_STATUS_HAS_MORE_OUTPUT; r->m_state=9; goto common_exit; case 9:;
          }while(0);
          }
        while(pIn_buf_cur >= pIn_buf_end)
          {
          if(decomp_flags & TINFL_FLAG_HAS_MORE_INPUT)
            {
            do{
              status=TINFL_STATUS_NEEDS_MORE_INPUT; r->m_state=38; goto common_exit; case 38:;
              }while(0);
            }
          else
            {
            do{
              for( ; ; )   { do{
                status=TINFL_STATUS_FAILED; r->m_state=40; goto common_exit; case 40:;
                }while(0);
                }   }while(0);
            }
          }
        n=min(min((size_t)(pOut_buf_end - pOut_buf_cur),(size_t)(pIn_buf_end - pIn_buf_cur)),counter);
        memcpy(pOut_buf_cur,pIn_buf_cur,n); pIn_buf_cur+=n; pOut_buf_cur+=n; counter-=(DWORD)n;
        }
      }
    else if(r->m_type == 3)
      {
      do{
        for( ; ; )   { do{
          status=TINFL_STATUS_FAILED; r->m_state=10; goto common_exit; case 10:;
          }while(0);
          }   }while(0);
      }
    else
      {
      if(r->m_type == 1)
        {
        BYTE *p=r->m_tables[0].m_code_size; DWORD i;
        r->m_table_sizes[0]=288; r->m_table_sizes[1]=32; memset(r->m_tables[1].m_code_size,5,32);
        for( i=0; i <= 143; ++i) *p++=8; for( ; i <= 255; ++i) *p++=9; for( ; i <= 279; ++i) *p++=7; for( ; i <= 287; ++i) *p++=8;
        }
      else
        {
        for(counter=0; counter < 3; counter++)   { do{
          if(num_bits < (DWORD)("\05\05\04"[counter]))   { do{
            DWORD c; do{
              if(pIn_buf_cur >= pIn_buf_end)   { for( ; ; )   { if(decomp_flags & TINFL_FLAG_HAS_MORE_INPUT)   { do{
                status=TINFL_STATUS_NEEDS_MORE_INPUT; r->m_state=11; goto common_exit; case 11:;
                }while(0);
              if(pIn_buf_cur < pIn_buf_end)   { c=*pIn_buf_cur++; break;   }   } else   { c=0; break;   }   }   } else c=*pIn_buf_cur++;   }while(0);
            bit_buf |= (((tinfl_bit_buf_t)c) << num_bits); num_bits+=8;   }while(num_bits < (DWORD)("\05\05\04"[counter]));   } r->m_table_sizes[counter]=bit_buf & ((1 << ("\05\05\04"[counter])) - 1); bit_buf >>= ("\05\05\04"[counter]); num_bits-=("\05\05\04"[counter]);   }while(0);
          r->m_table_sizes[counter]+=s_min_table_sizes[counter];   }
        memset(&r->m_tables[2].m_code_size,0,sizeof(r->m_tables[2].m_code_size)); for(counter=0; counter < r->m_table_sizes[2]; counter++)   { DWORD s; do{
          if(num_bits < (DWORD)(3))   { do{
            DWORD c; do{
              if(pIn_buf_cur >= pIn_buf_end)   { for( ; ; )   { if(decomp_flags & TINFL_FLAG_HAS_MORE_INPUT)   { do{
                status=TINFL_STATUS_NEEDS_MORE_INPUT; r->m_state=14; goto common_exit; case 14:;
                }while(0);
              if(pIn_buf_cur < pIn_buf_end)   { c=*pIn_buf_cur++; break;   }   } else   { c=0; break;   }   }   } else c=*pIn_buf_cur++;   }while(0);
            bit_buf |= (((tinfl_bit_buf_t)c) << num_bits); num_bits+=8;   }while(num_bits < (DWORD)(3));   } s=bit_buf & ((1 << (3)) - 1); bit_buf >>= (3); num_bits-=(3);   }while(0);
          r->m_tables[2].m_code_size[s_length_dezigzag[counter]]=(BYTE)s;   }
        r->m_table_sizes[2]=19;
        }
      for( ; (int)r->m_type >= 0; r->m_type--)
        {
        int tree_next,tree_cur; tinfl_huff_table *pTable;
        DWORD i,j,used_syms,total,sym_index,next_code[17],total_syms[16]; pTable=&r->m_tables[r->m_type]; memset(&total_syms,0,sizeof(total_syms)); memset(&pTable->m_look_up,0,sizeof(pTable->m_look_up)); memset(&pTable->m_tree,0,sizeof(pTable->m_tree));
        for(i=0; i < r->m_table_sizes[r->m_type]; ++i) total_syms[pTable->m_code_size[i]]++;
        used_syms=0,total=0; next_code[0]=next_code[1]=0;
        for(i=1; i <= 15; ++i)   { used_syms+=total_syms[i]; next_code[i + 1]=(total=((total + total_syms[i]) << 1));   }
        if((65536!=total) && (used_syms > 1))
          {
          do{
            for( ; ; )   { do{
              status=TINFL_STATUS_FAILED; r->m_state=35; goto common_exit; case 35:;
              }while(0);
              }   }while(0);
          }
        for(tree_next=-1,sym_index=0; sym_index < r->m_table_sizes[r->m_type]; ++sym_index)
          {
          DWORD rev_code=0,l,cur_code,code_size=pTable->m_code_size[sym_index]; if(!code_size) continue;
          cur_code=next_code[code_size]++; for(l=code_size; l > 0; l--,cur_code >>= 1) rev_code=(rev_code << 1) | (cur_code & 1);
          if(code_size <= TINFL_FAST_LOOKUP_BITS)   { __int16 k=(__int16)((code_size << 9) | sym_index); while(rev_code < TINFL_FAST_LOOKUP_SIZE)   { pTable->m_look_up[rev_code]=k; rev_code+=(1 << code_size);   } continue;   }
          if(0 == (tree_cur=pTable->m_look_up[rev_code & (TINFL_FAST_LOOKUP_SIZE - 1)]))   { pTable->m_look_up[rev_code & (TINFL_FAST_LOOKUP_SIZE - 1)]=(__int16)tree_next; tree_cur=tree_next; tree_next-=2;   }
          rev_code >>= (TINFL_FAST_LOOKUP_BITS - 1);
          for(j=code_size; j > (TINFL_FAST_LOOKUP_BITS + 1); j--)
            {
            tree_cur-=((rev_code >>= 1) & 1);
            if(!pTable->m_tree[-tree_cur - 1])   { pTable->m_tree[-tree_cur - 1]=(__int16)tree_next; tree_cur=tree_next; tree_next-=2;   } else tree_cur=pTable->m_tree[-tree_cur - 1];
            }
          tree_cur-=((rev_code >>= 1) & 1); pTable->m_tree[-tree_cur - 1]=(__int16)sym_index;
          }
        if(r->m_type == 2)
          {
          for(counter=0; counter < (r->m_table_sizes[0] + r->m_table_sizes[1]); )
            {
            DWORD s; do{
              int temp; DWORD code_len,c; if(num_bits < 15)   { if((pIn_buf_end - pIn_buf_cur) < 2)   { do{
                temp=(&r->m_tables[2])->m_look_up[bit_buf & (TINFL_FAST_LOOKUP_SIZE - 1)]; if(temp >= 0)   { code_len=temp >> 9; if((code_len) && (num_bits >= code_len)) break;   } else if(num_bits > TINFL_FAST_LOOKUP_BITS)   { code_len=TINFL_FAST_LOOKUP_BITS; do{
                  temp=(&r->m_tables[2])->m_tree[~temp + ((bit_buf >> code_len++) & 1)];   }while((temp < 0) && (num_bits >= (code_len + 1))); if(temp >= 0) break;   } do{
                    if(pIn_buf_cur >= pIn_buf_end)   { for( ; ; )   { if(decomp_flags & TINFL_FLAG_HAS_MORE_INPUT)   { do{
                      status=TINFL_STATUS_NEEDS_MORE_INPUT; r->m_state=16; goto common_exit; case 16:;
                      }while(0);
                    if(pIn_buf_cur < pIn_buf_end)   { c=*pIn_buf_cur++; break;   }   } else   { c=0; break;   }   }   } else c=*pIn_buf_cur++;   }while(0);
                  bit_buf |= (((tinfl_bit_buf_t)c) << num_bits); num_bits+=8;   }while(num_bits < 15);;   } else   { bit_buf |= (((tinfl_bit_buf_t)pIn_buf_cur[0]) << num_bits) | (((tinfl_bit_buf_t)pIn_buf_cur[1]) << (num_bits + 8)); pIn_buf_cur+=2; num_bits+=16;   }   } if((temp=(&r->m_tables[2])->m_look_up[bit_buf & (TINFL_FAST_LOOKUP_SIZE - 1)]) >= 0) code_len=temp >> 9,temp &= 511; else   { code_len=TINFL_FAST_LOOKUP_BITS; do{
                        temp=(&r->m_tables[2])->m_tree[~temp + ((bit_buf >> code_len++) & 1)];   }while(temp < 0);   } dist=temp; bit_buf >>= code_len; num_bits-=code_len;
                        }while(0);
                      if(dist < 16)   { r->m_len_codes[counter++]=(BYTE)dist; continue;   }
            if((dist == 16) && (!counter))
              {
              do{
                for( ; ; )   { do{
                  status=TINFL_STATUS_FAILED; r->m_state=17; goto common_exit; case 17:;
                  }while(0);
                  }   }while(0);
              }
            num_extra="\02\03\07"[dist - 16]; do{
              if(num_bits < (DWORD)(num_extra))   { do{
                DWORD c; do{
                  if(pIn_buf_cur >= pIn_buf_end)   { for( ; ; )   { if(decomp_flags & TINFL_FLAG_HAS_MORE_INPUT)   { do{
                    status=TINFL_STATUS_NEEDS_MORE_INPUT; r->m_state=18; goto common_exit; case 18:;
                    }while(0);
                  if(pIn_buf_cur < pIn_buf_end)   { c=*pIn_buf_cur++; break;   }   } else   { c=0; break;   }   }   } else c=*pIn_buf_cur++;   }while(0);
                bit_buf |= (((tinfl_bit_buf_t)c) << num_bits); num_bits+=8;   }while(num_bits < (DWORD)(num_extra));   } s=bit_buf & ((1 << (num_extra)) - 1); bit_buf >>= (num_extra); num_bits-=(num_extra);   }while(0);
              s+="\03\03\013"[dist - 16];
            memset(r->m_len_codes + counter,(dist == 16) ? r->m_len_codes[counter - 1] : 0,s); counter+=s;
            }
          if((r->m_table_sizes[0] + r->m_table_sizes[1])!=counter)
            {
            do{
              for( ; ; )   { do{
                status=TINFL_STATUS_FAILED; r->m_state=21; goto common_exit; case 21:;
                }while(0);
                }   }while(0);
            }
          memcpy(r->m_tables[0].m_code_size,r->m_len_codes,r->m_table_sizes[0]); memcpy(r->m_tables[1].m_code_size,r->m_len_codes + r->m_table_sizes[0],r->m_table_sizes[1]);
          }
        }
      for( ; ; )
        {
        BYTE *pSrc;
        for( ; ; )
          {
          if(((pIn_buf_end - pIn_buf_cur) < 4) || ((pOut_buf_end - pOut_buf_cur) < 2))
            {
            do{
              int temp; DWORD code_len,c; if(num_bits < 15)   { if((pIn_buf_end - pIn_buf_cur) < 2)   { do{
                temp=(&r->m_tables[0])->m_look_up[bit_buf & (TINFL_FAST_LOOKUP_SIZE - 1)]; if(temp >= 0)   { code_len=temp >> 9; if((code_len) && (num_bits >= code_len)) break;   } else if(num_bits > TINFL_FAST_LOOKUP_BITS)   { code_len=TINFL_FAST_LOOKUP_BITS; do{
                  temp=(&r->m_tables[0])->m_tree[~temp + ((bit_buf >> code_len++) & 1)];   }while((temp < 0) && (num_bits >= (code_len + 1))); if(temp >= 0) break;   } do{
                    if(pIn_buf_cur >= pIn_buf_end)   { for( ; ; )   { if(decomp_flags & TINFL_FLAG_HAS_MORE_INPUT)   { do{
                      status=TINFL_STATUS_NEEDS_MORE_INPUT; r->m_state=23; goto common_exit; case 23:;
                      }while(0);
                    if(pIn_buf_cur < pIn_buf_end)   { c=*pIn_buf_cur++; break;   }   } else   { c=0; break;   }   }   } else c=*pIn_buf_cur++;   }while(0);
                  bit_buf |= (((tinfl_bit_buf_t)c) << num_bits); num_bits+=8;   }while(num_bits < 15);;   } else   { bit_buf |= (((tinfl_bit_buf_t)pIn_buf_cur[0]) << num_bits) | (((tinfl_bit_buf_t)pIn_buf_cur[1]) << (num_bits + 8)); pIn_buf_cur+=2; num_bits+=16;   }   } if((temp=(&r->m_tables[0])->m_look_up[bit_buf & (TINFL_FAST_LOOKUP_SIZE - 1)]) >= 0) code_len=temp >> 9,temp &= 511; else   { code_len=TINFL_FAST_LOOKUP_BITS; do{
                        temp=(&r->m_tables[0])->m_tree[~temp + ((bit_buf >> code_len++) & 1)];   }while(temp < 0);   } counter=temp; bit_buf >>= code_len; num_bits-=code_len;
                        }while(0);
            if(counter >= 256)
              break;
            while(pOut_buf_cur >= pOut_buf_end)   { do{
              status=TINFL_STATUS_HAS_MORE_OUTPUT; r->m_state=24; goto common_exit; case 24:;
              }while(0);
              }
            *pOut_buf_cur++=(BYTE)counter;
            }
          else
            {
            int sym2; DWORD code_len;
            if(num_bits < 15)   { bit_buf |= (((tinfl_bit_buf_t)*((WORD *)(pIn_buf_cur))) << num_bits); pIn_buf_cur+=2; num_bits+=16;   }
            if((sym2=r->m_tables[0].m_look_up[bit_buf & (TINFL_FAST_LOOKUP_SIZE - 1)]) >= 0)
              code_len=sym2 >> 9;
            else
              {
              code_len=TINFL_FAST_LOOKUP_BITS; do{
                sym2=r->m_tables[0].m_tree[~sym2 + ((bit_buf >> code_len++) & 1)];   }while(sym2 < 0);
              }
            counter=sym2; bit_buf >>= code_len; num_bits-=code_len;
            if(counter & 256)
              break;
            if(num_bits < 15)   { bit_buf |= (((tinfl_bit_buf_t)*((WORD *)(pIn_buf_cur))) << num_bits); pIn_buf_cur+=2; num_bits+=16;   }
            if((sym2=r->m_tables[0].m_look_up[bit_buf & (TINFL_FAST_LOOKUP_SIZE - 1)]) >= 0)
              code_len=sym2 >> 9;
            else
              {
              code_len=TINFL_FAST_LOOKUP_BITS; do{
                sym2=r->m_tables[0].m_tree[~sym2 + ((bit_buf >> code_len++) & 1)];   }while(sym2 < 0);
              }
            bit_buf >>= code_len; num_bits-=code_len;
            pOut_buf_cur[0]=(BYTE)counter;
            if(sym2 & 256)
              {
              pOut_buf_cur++;
              counter=sym2;
              break;
              }
            pOut_buf_cur[1]=(BYTE)sym2;
            pOut_buf_cur+=2;
            }
          }
        if((counter &= 511) == 256) break;
        num_extra=s_length_extra[counter - 257]; counter=s_length_base[counter - 257];
        if(num_extra)   { DWORD extra_bits; do{
          if(num_bits < (DWORD)(num_extra))   { do{
            DWORD c; do{
              if(pIn_buf_cur >= pIn_buf_end)   { for( ; ; )   { if(decomp_flags & TINFL_FLAG_HAS_MORE_INPUT)   { do{
                status=TINFL_STATUS_NEEDS_MORE_INPUT; r->m_state=25; goto common_exit; case 25:;
                }while(0);
              if(pIn_buf_cur < pIn_buf_end)   { c=*pIn_buf_cur++; break;   }   } else   { c=0; break;   }   }   } else c=*pIn_buf_cur++;   }while(0);
            bit_buf |= (((tinfl_bit_buf_t)c) << num_bits); num_bits+=8;   }while(num_bits < (DWORD)(num_extra));   } extra_bits=bit_buf & ((1 << (num_extra)) - 1); bit_buf >>= (num_extra); num_bits-=(num_extra);   }while(0);
          counter+=extra_bits;   }
        do{
          int temp; DWORD code_len,c; if(num_bits < 15)   { if((pIn_buf_end - pIn_buf_cur) < 2)   { do{
            temp=(&r->m_tables[1])->m_look_up[bit_buf & (TINFL_FAST_LOOKUP_SIZE - 1)]; if(temp >= 0)   { code_len=temp >> 9; if((code_len) && (num_bits >= code_len)) break;   } else if(num_bits > TINFL_FAST_LOOKUP_BITS)   { code_len=TINFL_FAST_LOOKUP_BITS; do{
              temp=(&r->m_tables[1])->m_tree[~temp + ((bit_buf >> code_len++) & 1)];   }while((temp < 0) && (num_bits >= (code_len + 1))); if(temp >= 0) break;   } do{
                if(pIn_buf_cur >= pIn_buf_end)   { for( ; ; )   { if(decomp_flags & TINFL_FLAG_HAS_MORE_INPUT)   { do{
                  status=TINFL_STATUS_NEEDS_MORE_INPUT; r->m_state=26; goto common_exit; case 26:;
                  }while(0);
                if(pIn_buf_cur < pIn_buf_end)   { c=*pIn_buf_cur++; break;   }   } else   { c=0; break;   }   }   } else c=*pIn_buf_cur++;   }while(0);
              bit_buf |= (((tinfl_bit_buf_t)c) << num_bits); num_bits+=8;   }while(num_bits < 15);;   } else   { bit_buf |= (((tinfl_bit_buf_t)pIn_buf_cur[0]) << num_bits) | (((tinfl_bit_buf_t)pIn_buf_cur[1]) << (num_bits + 8)); pIn_buf_cur+=2; num_bits+=16;   }   } if((temp=(&r->m_tables[1])->m_look_up[bit_buf & (TINFL_FAST_LOOKUP_SIZE - 1)]) >= 0) code_len=temp >> 9,temp &= 511; else   { code_len=TINFL_FAST_LOOKUP_BITS; do{
                    temp=(&r->m_tables[1])->m_tree[~temp + ((bit_buf >> code_len++) & 1)];   }while(temp < 0);   } dist=temp; bit_buf >>= code_len; num_bits-=code_len;
                    }while(0);
        num_extra=s_dist_extra[dist]; dist=s_dist_base[dist];
        if(num_extra)   { DWORD extra_bits; do{
          if(num_bits < (DWORD)(num_extra))   { do{
            DWORD c; do{
              if(pIn_buf_cur >= pIn_buf_end)   { for( ; ; )   { if(decomp_flags & TINFL_FLAG_HAS_MORE_INPUT)   { do{
                status=TINFL_STATUS_NEEDS_MORE_INPUT; r->m_state=27; goto common_exit; case 27:;
                }while(0);
              if(pIn_buf_cur < pIn_buf_end)   { c=*pIn_buf_cur++; break;   }   } else   { c=0; break;   }   }   } else c=*pIn_buf_cur++;   }while(0);
            bit_buf |= (((tinfl_bit_buf_t)c) << num_bits); num_bits+=8;   }while(num_bits < (DWORD)(num_extra));   } extra_bits=bit_buf & ((1 << (num_extra)) - 1); bit_buf >>= (num_extra); num_bits-=(num_extra);   }while(0);
          dist+=extra_bits;   }
        dist_from_out_buf_start=pOut_buf_cur - pOut_buf_start;
        if((dist > dist_from_out_buf_start) && (decomp_flags & TINFL_FLAG_USING_NON_WRAPPING_OUTPUT_BUF))
          {
          do{
            for( ; ; )   { do{
              status=TINFL_STATUS_FAILED; r->m_state=37; goto common_exit; case 37:;
              }while(0);
              }   }while(0);
          }
        pSrc=pOut_buf_start + ((dist_from_out_buf_start - dist) & out_buf_size_mask);
        if((max(pOut_buf_cur,pSrc) + counter) > pOut_buf_end)
          {
          while(counter--)
            {
            while(pOut_buf_cur >= pOut_buf_end)   { do{
              status=TINFL_STATUS_HAS_MORE_OUTPUT; r->m_state=53; goto common_exit; case 53:;
              }while(0);
              }
            *pOut_buf_cur++=pOut_buf_start[(dist_from_out_buf_start++ - dist) & out_buf_size_mask];
            }
          continue;
          }
        else if((counter >= 9) && (counter <= dist))
          {
          BYTE *pSrc_end=pSrc + (counter & ~7);
          do
            {
            ((DWORD *)pOut_buf_cur)[0]=((DWORD *)pSrc)[0];
            ((DWORD *)pOut_buf_cur)[1]=((DWORD *)pSrc)[1];
            pOut_buf_cur+=8;
            }while((pSrc+=8) < pSrc_end);
          if((counter &= 7) < 3)
            {
            if(counter)
              {
              pOut_buf_cur[0]=pSrc[0];
              if(counter > 1)
                pOut_buf_cur[1]=pSrc[1];
              pOut_buf_cur+=counter;
              }
            continue;
            }
          }
        do
          {
          pOut_buf_cur[0]=pSrc[0];
          pOut_buf_cur[1]=pSrc[1];
          pOut_buf_cur[2]=pSrc[2];
          pOut_buf_cur+=3; pSrc+=3;
          }while((int)(counter-=3) > 2);
        if((int)counter > 0)
          {
          pOut_buf_cur[0]=pSrc[0];
          if((int)counter > 1)
            pOut_buf_cur[1]=pSrc[1];
          pOut_buf_cur+=counter;
          }
        }
      }
    }while(!(r->m_final & 1));
  if(decomp_flags & TINFL_FLAG_PARSE_ZLIB_HEADER)
    {
    do{
      if(num_bits < (DWORD)(num_bits & 7))   { do{
        DWORD c; do{
          if(pIn_buf_cur >= pIn_buf_end)   { for( ; ; )   { if(decomp_flags & TINFL_FLAG_HAS_MORE_INPUT)   { do{
            status=TINFL_STATUS_NEEDS_MORE_INPUT; r->m_state=32; goto common_exit; case 32:;
            }while(0);
          if(pIn_buf_cur < pIn_buf_end)   { c=*pIn_buf_cur++; break;   }   } else   { c=0; break;   }   }   } else c=*pIn_buf_cur++;   }while(0);
        bit_buf |= (((tinfl_bit_buf_t)c) << num_bits); num_bits+=8;   }while(num_bits < (DWORD)(num_bits & 7));   } bit_buf >>= (num_bits & 7); num_bits-=(num_bits & 7);   }while(0);
      for(counter=0; counter < 4; ++counter)   { DWORD s; if(num_bits) do{
              if(num_bits < (DWORD)(8))   { do{
                DWORD c; do{
                  if(pIn_buf_cur >= pIn_buf_end)   { for( ; ; )   { if(decomp_flags & TINFL_FLAG_HAS_MORE_INPUT)   { do{
                    status=TINFL_STATUS_NEEDS_MORE_INPUT; r->m_state=41; goto common_exit; case 41:;
                    }while(0);
                  if(pIn_buf_cur < pIn_buf_end)   { c=*pIn_buf_cur++; break;   }   } else   { c=0; break;   }   }   } else c=*pIn_buf_cur++;   }while(0);
                bit_buf |= (((tinfl_bit_buf_t)c) << num_bits); num_bits+=8;   }while(num_bits < (DWORD)(8));   } s=bit_buf & ((1 << (8)) - 1); bit_buf >>= (8); num_bits-=(8);   }while(0);
              else do{
                      if(pIn_buf_cur >= pIn_buf_end)   { for( ; ; )   { if(decomp_flags & TINFL_FLAG_HAS_MORE_INPUT)   { do{
                        status=TINFL_STATUS_NEEDS_MORE_INPUT; r->m_state=42; goto common_exit; case 42:;
                        }while(0);
                      if(pIn_buf_cur < pIn_buf_end)   { s=*pIn_buf_cur++; break;   }   } else   { s=0; break;   }   }   } else s=*pIn_buf_cur++;   }while(0);
                    r->m_z_adler32=(r->m_z_adler32 << 8) | s;   }
    }
  do{
    for( ; ; )   { do{
      status=TINFL_STATUS_DONE; r->m_state=34; goto common_exit; case 34:;
      }while(0);
      }   }while(0);
    }
common_exit:
  r->m_num_bits=num_bits; r->m_bit_buf=bit_buf; r->m_dist=dist; r->m_counter=counter; r->m_num_extra=num_extra; r->m_dist_from_out_buf_start=dist_from_out_buf_start;
  *pIn_buf_size=pIn_buf_cur - pIn_buf_next; *pOut_buf_size=pOut_buf_cur - pOut_buf_next;
  if((decomp_flags & (TINFL_FLAG_PARSE_ZLIB_HEADER | TINFL_FLAG_COMPUTE_ADLER32)) && (status >= 0))
    {
    BYTE *ptr=pOut_buf_next; size_t buf_len=*pOut_buf_size;
    DWORD i,s1=r->m_check_adler32 & 0xffff,s2=r->m_check_adler32 >> 16; size_t block_len=buf_len % 5552;
    while(buf_len)
      {
      for(i=0; i + 7 < block_len; i+=8,ptr+=8)
        {
        s1+=ptr[0],s2+=s1; s1+=ptr[1],s2+=s1; s1+=ptr[2],s2+=s1; s1+=ptr[3],s2+=s1;
        s1+=ptr[4],s2+=s1; s1+=ptr[5],s2+=s1; s1+=ptr[6],s2+=s1; s1+=ptr[7],s2+=s1;
        }
      for( ; i < block_len; ++i) s1+=*ptr++,s2+=s1;
      s1 %= 65521U,s2 %= 65521U; buf_len-=block_len; block_len=5552;
      }
    r->m_check_adler32=(s2 << 16) + s1; if((status == TINFL_STATUS_DONE) && (decomp_flags & TINFL_FLAG_PARSE_ZLIB_HEADER) && (r->m_check_adler32!=r->m_z_adler32)) status=TINFL_STATUS_ADLER32_MISMATCH;
    }
  return status;
  }

int mz_inflate(mz_stream* pStream,int flush)
  {
  inflate_state* pState;
  DWORD n,first_call,decomp_flags=TINFL_FLAG_COMPUTE_ADLER32;
  size_t in_bytes,out_bytes,orig_avail_in;
  tinfl_status status;
  if((!pStream) || (!pStream->state)) return MZ_STREAM_ERROR;
  if(flush == MZ_PARTIAL_FLUSH) flush=MZ_SYNC_FLUSH;
  if((flush) && (flush!=MZ_SYNC_FLUSH) && (flush!=MZ_FINISH)) return MZ_STREAM_ERROR;
  pState=(inflate_state*)pStream->state;
  if(pState->m_window_bits > 0) decomp_flags |= TINFL_FLAG_PARSE_ZLIB_HEADER;
  orig_avail_in=pStream->avail_in;
  first_call=pState->m_first_call; pState->m_first_call=0;
  if(pState->m_last_status < 0) return MZ_DATA_ERROR;
  if(pState->m_has_flushed && (flush!=MZ_FINISH)) return MZ_STREAM_ERROR;
  pState->m_has_flushed |= (flush == MZ_FINISH);
  if((flush == MZ_FINISH) && (first_call))
    {
    decomp_flags |= TINFL_FLAG_USING_NON_WRAPPING_OUTPUT_BUF;
    in_bytes=pStream->avail_in; out_bytes=pStream->avail_out;
    status=tinfl_decompress(&pState->m_decomp,pStream->next_in,&in_bytes,pStream->next_out,pStream->next_out,&out_bytes,decomp_flags);
    pState->m_last_status=status;
    pStream->next_in+=(DWORD)in_bytes; pStream->avail_in-=(DWORD)in_bytes; pStream->total_in+=(DWORD)in_bytes;
    pStream->adler=(&pState->m_decomp)->m_check_adler32;
    pStream->next_out+=(DWORD)out_bytes; pStream->avail_out-=(DWORD)out_bytes; pStream->total_out+=(DWORD)out_bytes;
    if(status < 0)
      return MZ_DATA_ERROR;
    else if(status!=TINFL_STATUS_DONE)
      {
      pState->m_last_status=TINFL_STATUS_FAILED;
      return MZ_BUF_ERROR;
      }
    return MZ_STREAM_END;
    }
  if(flush!=MZ_FINISH) decomp_flags |= TINFL_FLAG_HAS_MORE_INPUT;
  if(pState->m_dict_avail)
    {
    n=min(pState->m_dict_avail,pStream->avail_out);
    memcpy(pStream->next_out,pState->m_dict + pState->m_dict_ofs,n);
    pStream->next_out+=n; pStream->avail_out-=n; pStream->total_out+=n;
    pState->m_dict_avail-=n; pState->m_dict_ofs=(pState->m_dict_ofs + n) & (32768 - 1);
    return ((pState->m_last_status == TINFL_STATUS_DONE) && (!pState->m_dict_avail)) ? MZ_STREAM_END : MZ_OK;
    }
  for( ; ; )
    {
    in_bytes=pStream->avail_in;
    out_bytes=32768 - pState->m_dict_ofs;
    status=tinfl_decompress(&pState->m_decomp,pStream->next_in,&in_bytes,pState->m_dict,pState->m_dict + pState->m_dict_ofs,&out_bytes,decomp_flags);
    pState->m_last_status=status;
    pStream->next_in+=(DWORD)in_bytes; pStream->avail_in-=(DWORD)in_bytes;
    pStream->total_in+=(DWORD)in_bytes; pStream->adler=(&pState->m_decomp)->m_check_adler32;
    pState->m_dict_avail=(DWORD)out_bytes;
    n=min(pState->m_dict_avail,pStream->avail_out);
    memcpy(pStream->next_out,pState->m_dict + pState->m_dict_ofs,n);
    pStream->next_out+=n; pStream->avail_out-=n; pStream->total_out+=n;
    pState->m_dict_avail-=n; pState->m_dict_ofs=(pState->m_dict_ofs + n) & (32768 - 1);
    if(status < 0)
       return MZ_DATA_ERROR;
    else if((status == TINFL_STATUS_NEEDS_MORE_INPUT) && (!orig_avail_in))
      return MZ_BUF_ERROR;
    else if(flush == MZ_FINISH)
      {
       if(status == TINFL_STATUS_DONE)
          return pState->m_dict_avail ? MZ_BUF_ERROR : MZ_STREAM_END;
       else if(!pStream->avail_out)
          return MZ_BUF_ERROR;
      }
    else if((status == TINFL_STATUS_DONE) || (!pStream->avail_in) || (!pStream->avail_out) || (pState->m_dict_avail))
      break;
    }
  return ((status == TINFL_STATUS_DONE) && (!pState->m_dict_avail)) ? MZ_STREAM_END : MZ_OK;
  }

int mz_inflateEnd(mz_stream* pStream)
  {
  if(!pStream)
    return MZ_STREAM_ERROR;
  if(pStream->state)
    {
    pStream->zfree(pStream->opaque,pStream->state);
    pStream->state=NULL;
    }
  return MZ_OK;
  }

int mz_uncompress(unsigned char*pDest,DWORD *pDest_len,unsigned char*pSource,DWORD source_len)
  {
  mz_stream stream;
  int status;
  memset(&stream,0,sizeof(stream));
  if((source_len | *pDest_len) > 0xFFFFFFFFU) return MZ_PARAM_ERROR;
  stream.next_in=pSource;
  stream.avail_in=(DWORD)source_len;
  stream.next_out=pDest;
  stream.avail_out=(DWORD)*pDest_len;
  status=mz_inflateInit(&stream);
  if(status!=MZ_OK)
    return status;
  status=mz_inflate(&stream,MZ_FINISH);
  if(status!=MZ_STREAM_END)
    {
    mz_inflateEnd(&stream);
    return ((status == MZ_BUF_ERROR) && (!stream.avail_in)) ? MZ_DATA_ERROR : status;
    }
  *pDest_len=stream.total_out;
  return mz_inflateEnd(&stream);
  }

char*mz_error(int err)
  {
  static struct   { int m_err; char*m_pDesc;   } s_error_descs[] =
    {
      { MZ_OK,""   },  { MZ_STREAM_END,"stream end"   },  { MZ_NEED_DICT,"need dictionary"   },  { MZ_ERRNO,"file error"   },  { MZ_STREAM_ERROR,"stream error"   },
      { MZ_DATA_ERROR,"data error"   },  { MZ_MEM_ERROR,"out of memory"   },  { MZ_BUF_ERROR,"buf error"   },  { MZ_VERSION_ERROR,"version error"   },  { MZ_PARAM_ERROR,"parameter error"   }
    };
  DWORD i; for(i=0; i < sizeof(s_error_descs) / sizeof(s_error_descs[0]); ++i) if(s_error_descs[i].m_err == err) return s_error_descs[i].m_pDesc;
  return NULL;
  }

void*tinfl_decompress_mem_to_heap(void*pSrc_buf,size_t src_buf_len,size_t *pOut_len,int flags)
  {
  tinfl_decompressor decomp; void*pBuf=NULL,*pNew_buf; size_t src_buf_ofs=0,out_buf_capacity=0;
  *pOut_len=0;
  do{
    (&decomp)->m_state=0;
    }while(0);
  for( ; ; )
    {
    size_t src_buf_size=src_buf_len - src_buf_ofs,dst_buf_size=out_buf_capacity - *pOut_len,new_out_buf_capacity;
    tinfl_status status=tinfl_decompress(&decomp,(BYTE*)pSrc_buf + src_buf_ofs,&src_buf_size,(BYTE*)pBuf,pBuf ? (BYTE*)pBuf + *pOut_len : NULL,&dst_buf_size,
      (flags & ~TINFL_FLAG_HAS_MORE_INPUT) | TINFL_FLAG_USING_NON_WRAPPING_OUTPUT_BUF);
    if((status < 0) || (status == TINFL_STATUS_NEEDS_MORE_INPUT))
      {
      free(pBuf); *pOut_len=0; return NULL;
      }
    src_buf_ofs+=src_buf_size;
    *pOut_len+=dst_buf_size;
    if(status == TINFL_STATUS_DONE) break;
    new_out_buf_capacity=out_buf_capacity * 2; if(new_out_buf_capacity < 128) new_out_buf_capacity=128;
    pNew_buf=realloc(pBuf,new_out_buf_capacity);
    if(!pNew_buf)
      {
      free(pBuf); *pOut_len=0; return NULL;
      }
    pBuf=pNew_buf; out_buf_capacity=new_out_buf_capacity;
    }
  return pBuf;
  }

size_t tinfl_decompress_mem_to_mem(void*pOut_buf,size_t out_buf_len,void*pSrc_buf,size_t src_buf_len,int flags)
  {
  tinfl_decompressor decomp; tinfl_status status; do{
    (&decomp)->m_state=0;
    }while(0);
  status=tinfl_decompress(&decomp,(BYTE*)pSrc_buf,&src_buf_len,(BYTE*)pOut_buf,(BYTE*)pOut_buf,&out_buf_len,(flags & ~TINFL_FLAG_HAS_MORE_INPUT) | TINFL_FLAG_USING_NON_WRAPPING_OUTPUT_BUF);
  return (status!=TINFL_STATUS_DONE) ? ((size_t)(-1)) : out_buf_len;
  }

typedef int (*tinfl_put_buf_func_ptr)(void* pBuf,int len,void*pUser);

int tinfl_decompress_mem_to_callback(void*pIn_buf,size_t *pIn_buf_size,tinfl_put_buf_func_ptr pPut_buf_func,void*pPut_buf_user,int flags)
  {
  int result=0;
  tinfl_decompressor decomp;
  BYTE *pDict=(BYTE*)malloc(32768); size_t in_buf_ofs=0,dict_ofs=0;
  if(!pDict)
    return TINFL_STATUS_FAILED;
  do{
    (&decomp)->m_state=0;
    }while(0);
  for( ; ; )
    {
    size_t in_buf_size=*pIn_buf_size - in_buf_ofs,dst_buf_size=32768 - dict_ofs;
    tinfl_status status=tinfl_decompress(&decomp,(BYTE*)pIn_buf + in_buf_ofs,&in_buf_size,pDict,pDict + dict_ofs,&dst_buf_size,
      (flags & ~(TINFL_FLAG_HAS_MORE_INPUT | TINFL_FLAG_USING_NON_WRAPPING_OUTPUT_BUF)));
    in_buf_ofs+=in_buf_size;
    if((dst_buf_size) && (!(*pPut_buf_func)(pDict + dict_ofs,(int)dst_buf_size,pPut_buf_user)))
      break;
    if(status!=TINFL_STATUS_HAS_MORE_OUTPUT)
      {
      result=(status == TINFL_STATUS_DONE);
      break;
      }
    dict_ofs=(dict_ofs + dst_buf_size) & (32768 - 1);
    }
  free(pDict);
  *pIn_buf_size=in_buf_ofs;
  return result;
  }

int tdefl_compress_mem_to_output(void*pBuf,size_t buf_len,tdefl_put_buf_func_ptr pPut_buf_func,void*pPut_buf_user,int flags)
  {
  tdefl_compressor *pComp; int succeeded; if(((buf_len) && (!pBuf)) || (!pPut_buf_func)) return 0;
  pComp=(tdefl_compressor*)malloc(sizeof(tdefl_compressor)); if(!pComp) return 0;
  succeeded=(tdefl_init(pComp,pPut_buf_func,pPut_buf_user,flags) == TDEFL_STATUS_OKAY);
  succeeded=succeeded && (tdefl_compress_buffer(pComp,pBuf,buf_len,TDEFL_FINISH) == TDEFL_STATUS_DONE);
  free(pComp); return succeeded;
  }

typedef struct
  {
  size_t m_size,m_capacity;
  BYTE *m_pBuf;
  int m_expandable;
  } tdefl_output_buffer;

int tdefl_output_buffer_putter(void*pBuf,int len,void*pUser)
  {
  tdefl_output_buffer *p=(tdefl_output_buffer *)pUser;
  size_t new_size=p->m_size + len;
  if(new_size > p->m_capacity)
    {
    size_t new_capacity=p->m_capacity; BYTE *pNew_buf; if(!p->m_expandable) return 0;
    do{
      new_capacity=max(128U,new_capacity << 1U);   }while(new_size > new_capacity);
    pNew_buf=(BYTE*)realloc(p->m_pBuf,new_capacity); if(!pNew_buf) return 0;
    p->m_pBuf=pNew_buf; p->m_capacity=new_capacity;
    }
  memcpy((BYTE*)p->m_pBuf + p->m_size,pBuf,len); p->m_size=new_size;
  return 1;
  }

void*tdefl_compress_mem_to_heap(void*pSrc_buf,size_t src_buf_len,size_t *pOut_len,int flags)
  {
  tdefl_output_buffer out_buf; memset(&out_buf,0,sizeof(out_buf));
  if(!pOut_len) return 0; else *pOut_len=0;
  out_buf.m_expandable=1;
  if(!tdefl_compress_mem_to_output(pSrc_buf,src_buf_len,tdefl_output_buffer_putter,&out_buf,flags)) return NULL;
  *pOut_len=out_buf.m_size; return out_buf.m_pBuf;
  }

size_t tdefl_compress_mem_to_mem(void*pOut_buf,size_t out_buf_len,void*pSrc_buf,size_t src_buf_len,int flags)
  {
  tdefl_output_buffer out_buf; memset(&out_buf,0,sizeof(out_buf));
  if(!pOut_buf) return 0;
  out_buf.m_pBuf=(BYTE*)pOut_buf; out_buf.m_capacity=out_buf_len;
  if(!tdefl_compress_mem_to_output(pSrc_buf,src_buf_len,tdefl_output_buffer_putter,&out_buf,flags)) return 0;
  return out_buf.m_size;
  }

void mz_zip_array_clear(mz_zip_archive *pZip,mz_zip_array *pArray)
  {
  pZip->m_pFree(pZip->m_pAlloc_opaque,pArray->m_p);
  memset(pArray,0,sizeof(mz_zip_array));
  }

int mz_zip_array_ensure_capacity(mz_zip_archive *pZip,mz_zip_array *pArray,size_t min_new_capacity,DWORD growing)
  {
  void*pNew_p; size_t new_capacity=min_new_capacity; (void)( (!!(pArray->m_element_size)) || (assert("pArray->m_element_size"),0) ); if(pArray->m_capacity >= min_new_capacity) return 1;
  if(growing)   { new_capacity=max(1,pArray->m_capacity); while(new_capacity < min_new_capacity) new_capacity *= 2;   }
  if(NULL == (pNew_p=pZip->m_pRealloc(pZip->m_pAlloc_opaque,pArray->m_p,pArray->m_element_size,new_capacity))) return 0;
  pArray->m_p=pNew_p; pArray->m_capacity=new_capacity;
  return 1;
  }

int mz_zip_array_reserve(mz_zip_archive *pZip,mz_zip_array *pArray,size_t new_capacity,DWORD growing)
  {
  if(new_capacity > pArray->m_capacity)   { if(!mz_zip_array_ensure_capacity(pZip,pArray,new_capacity,growing)) return 0;   }
  return 1;
  }

int mz_zip_array_resize(mz_zip_archive *pZip,mz_zip_array *pArray,size_t new_size,DWORD growing)
  {
  if(new_size > pArray->m_capacity)   { if(!mz_zip_array_ensure_capacity(pZip,pArray,new_size,growing)) return 0;   }
  pArray->m_size=new_size;
  return 1;
  }

int mz_zip_array_ensure_room(mz_zip_archive *pZip,mz_zip_array *pArray,size_t n)
  {
  return mz_zip_array_reserve(pZip,pArray,pArray->m_size + n,1);
  }

int mz_zip_array_push_back(mz_zip_archive *pZip,mz_zip_array *pArray,void*pElements,size_t n)
  {
  size_t orig_size=pArray->m_size; if(!mz_zip_array_resize(pZip,pArray,orig_size + n,1)) return 0;
  memcpy((BYTE*)pArray->m_p + orig_size * pArray->m_element_size,pElements,n * pArray->m_element_size);
  return 1;
  }

time_t mz_zip_dos_to_time_t(int dos_time,int dos_date)
  {
  struct tm tm;
  memset(&tm,0,sizeof(tm)); tm.tm_isdst=-1;
  tm.tm_year=((dos_date >> 9) & 127) + 1980 - 1900; tm.tm_mon=((dos_date >> 5) & 15) - 1; tm.tm_mday=dos_date & 31;
  tm.tm_hour=(dos_time >> 11) & 31; tm.tm_min=(dos_time >> 5) & 63; tm.tm_sec=(dos_time << 1) & 62;
  return mktime(&tm);
  }

void mz_zip_time_to_dos_time(time_t time,WORD *pDOS_time,WORD *pDOS_date)
    {
  struct tm t;
  memcpy(&t,localtime(&time),sizeof(t));
  *pDOS_time=(WORD)(((t.tm_hour) << 11) + ((t.tm_min) << 5) + ((t.tm_sec) >> 1));
  *pDOS_date=(WORD)(((t.tm_year + 1900 - 1980) << 9) + ((t.tm_mon + 1) << 5) + t.tm_mday);
    }

int mz_zip_get_file_modified_time(char*pFilename,WORD *pDOS_time,WORD *pDOS_date)
    {
  FILETIME ft;
  if(!GetFileTime(pFilename,NULL,NULL,&ft))
    return 0;
  FileTimeToDosDateTime(&ft,pDOS_time,pDOS_date);
  return 1;
    }

int mz_zip_set_file_times(char*pFilename,time_t access_time,time_t modified_time)
    {
  struct utimbuf t; t.actime=access_time; t.modtime=modified_time;
  return !utime(pFilename,&t);
    }

int mz_zip_reader_init_internal(mz_zip_archive *pZip,DWORD flags)
    {
  (void)flags;
  if((!pZip) || (pZip->m_pState) || (pZip->m_zip_mode!=MZ_ZIP_MODE_INVALID))
    return 0;
  if(!pZip->m_pAlloc) pZip->m_pAlloc=def_alloc_func;
  if(!pZip->m_pFree) pZip->m_pFree=def_free_func;
  if(!pZip->m_pRealloc) pZip->m_pRealloc=def_realloc_func;
  pZip->m_zip_mode=MZ_ZIP_MODE_READING;
  pZip->m_archive_size=0;
  pZip->m_central_directory_file_ofs=0;
  pZip->m_total_files=0;
  if(NULL == (pZip->m_pState=(mz_zip_internal_state *)pZip->m_pAlloc(pZip->m_pAlloc_opaque,1,sizeof(mz_zip_internal_state))))
    return 0;
  memset(pZip->m_pState,0,sizeof(mz_zip_internal_state));
  (&pZip->m_pState->m_central_dir)->m_element_size=sizeof(BYTE);
  (&pZip->m_pState->m_central_dir_offsets)->m_element_size=sizeof(DWORD);
  (&pZip->m_pState->m_sorted_central_dir_offsets)->m_element_size=sizeof(DWORD);
  return 1;
    }

int mz_zip_reader_filename_less(mz_zip_array *pCentral_dir_array,mz_zip_array *pCentral_dir_offsets,DWORD l_index,DWORD r_index)
  {
  BYTE *pL=&((BYTE *)((pCentral_dir_array)->m_p))[((DWORD *)((pCentral_dir_offsets)->m_p))[l_index]],*pE;
  BYTE *pR=&((BYTE *)((pCentral_dir_array)->m_p))[((DWORD *)((pCentral_dir_offsets)->m_p))[r_index]];
  DWORD l_len=*((WORD *)(pL + MZ_ZIP_CDH_FILENAME_LEN_OFS)),r_len=*((WORD *)(pR + MZ_ZIP_CDH_FILENAME_LEN_OFS));
  BYTE l=0,r=0;
  pL+=MZ_ZIP_CENTRAL_DIR_HEADER_SIZE; pR+=MZ_ZIP_CENTRAL_DIR_HEADER_SIZE;
  pE=pL + min(l_len,r_len);
  while(pL < pE)
    {
    if((l=tolower(*pL))!=(r=tolower(*pR)))
      break;
    pL++; pR++;
    }
  return (pL == pE) ? (l_len < r_len) : (l < r);
  }

void mz_zip_reader_sort_central_dir_offsets_by_filename(mz_zip_archive *pZip)
  {
  mz_zip_internal_state *pState=pZip->m_pState;
  mz_zip_array *pCentral_dir_offsets=&pState->m_central_dir_offsets;
  mz_zip_array *pCentral_dir=&pState->m_central_dir;
  DWORD *pIndices=&((DWORD *)((&pState->m_sorted_central_dir_offsets)->m_p))[0];
  int size=pZip->m_total_files;
  int start=(size - 2) >> 1,end;
  while(start >= 0)
    {
    int child,root=start;
    for( ; ; )
      {
      if((child=(root << 1) + 1) >= size)
        break;
      child+=(((child + 1) < size) && (mz_zip_reader_filename_less(pCentral_dir,pCentral_dir_offsets,pIndices[child],pIndices[child + 1])));
      if(!mz_zip_reader_filename_less(pCentral_dir,pCentral_dir_offsets,pIndices[root],pIndices[child]))
        break;
      do{
        DWORD t=pIndices[root]; pIndices[root]=pIndices[child]; pIndices[child]=t;
        }while(0);
      root=child;
      }
    start--;
    }
  end=size - 1;
  while(end > 0)
    {
    int child,root=0;
    do{
      DWORD t=pIndices[end]; pIndices[end]=pIndices[0]; pIndices[0]=t;
      }while(0);
    for( ; ; )
      {
      if((child=(root << 1) + 1) >= end)
        break;
      child+=(((child + 1) < end) && mz_zip_reader_filename_less(pCentral_dir,pCentral_dir_offsets,pIndices[child],pIndices[child + 1]));
      if(!mz_zip_reader_filename_less(pCentral_dir,pCentral_dir_offsets,pIndices[root],pIndices[child]))
        break;
      do{
        DWORD t=pIndices[root]; pIndices[root]=pIndices[child]; pIndices[child]=t;
        }while(0);
      root=child;
      }
    end--;
    }
  }

int mz_zip_reader_read_central_dir(mz_zip_archive *pZip,DWORD flags)
  {
  DWORD cdir_size,num_this_disk,cdir_disk_index;
  QWORD cdir_ofs;
  __int64 cur_file_ofs;
  BYTE *p;
  DWORD buf_u32[4096 / sizeof(DWORD)]; BYTE *pBuf=(BYTE *)buf_u32;
  int sort_central_dir=((flags & MZ_ZIP_FLAG_DO_NOT_SORT_CENTRAL_DIRECTORY) == 0);
  if(pZip->m_archive_size < MZ_ZIP_END_OF_CENTRAL_DIR_HEADER_SIZE)
    return 0;
  cur_file_ofs=max((__int64)pZip->m_archive_size - (__int64)sizeof(buf_u32),0);
  for( ; ; )
    {
    int i,n=(int)min(sizeof(buf_u32),pZip->m_archive_size - cur_file_ofs);
    if(pZip->m_pRead(pZip->m_pIO_opaque,cur_file_ofs,pBuf,n)!=(DWORD)n)
      return 0;
    for(i=n - 4; i >= 0; --i)
      if(*((DWORD *)(pBuf + i)) == MZ_ZIP_END_OF_CENTRAL_DIR_HEADER_SIG)
        break;
    if(i >= 0)
      {
      cur_file_ofs+=i;
      break;
      }
    if((!cur_file_ofs) || ((pZip->m_archive_size - cur_file_ofs) >= (0xFFFF + MZ_ZIP_END_OF_CENTRAL_DIR_HEADER_SIZE)))
      return 0;
    cur_file_ofs=max(cur_file_ofs - (sizeof(buf_u32) - 3),0);
    }
  if(pZip->m_pRead(pZip->m_pIO_opaque,cur_file_ofs,pBuf,MZ_ZIP_END_OF_CENTRAL_DIR_HEADER_SIZE)!=MZ_ZIP_END_OF_CENTRAL_DIR_HEADER_SIZE)
    return 0;
  if((*((DWORD *)(pBuf + MZ_ZIP_ECDH_SIG_OFS))!=MZ_ZIP_END_OF_CENTRAL_DIR_HEADER_SIG) ||
      ((pZip->m_total_files=*((WORD *)(pBuf + MZ_ZIP_ECDH_CDIR_TOTAL_ENTRIES_OFS)))!=*((WORD *)(pBuf + MZ_ZIP_ECDH_CDIR_NUM_ENTRIES_ON_DISK_OFS))))
    return 0;
  num_this_disk=*((WORD *)(pBuf + MZ_ZIP_ECDH_NUM_THIS_DISK_OFS));
  cdir_disk_index=*((WORD *)(pBuf + MZ_ZIP_ECDH_NUM_DISK_CDIR_OFS));
  if(((num_this_disk | cdir_disk_index)!=0) && ((num_this_disk!=1) || (cdir_disk_index!=1)))
    return 0;
  if((cdir_size=*((DWORD *)(pBuf + MZ_ZIP_ECDH_CDIR_SIZE_OFS))) < pZip->m_total_files * MZ_ZIP_CENTRAL_DIR_HEADER_SIZE)
    return 0;
  cdir_ofs=*((DWORD *)(pBuf + MZ_ZIP_ECDH_CDIR_OFS_OFS));
  if((cdir_ofs + (QWORD)cdir_size) > pZip->m_archive_size)
    return 0;
  pZip->m_central_directory_file_ofs=cdir_ofs;
  if(pZip->m_total_files)
    {
     DWORD i,n;
    if((!mz_zip_array_resize(pZip,&pZip->m_pState->m_central_dir,cdir_size,0)) ||
        (!mz_zip_array_resize(pZip,&pZip->m_pState->m_central_dir_offsets,pZip->m_total_files,0)))
      return 0;
    if(sort_central_dir)
      {
      if(!mz_zip_array_resize(pZip,&pZip->m_pState->m_sorted_central_dir_offsets,pZip->m_total_files,0))
        return 0;
      }
    if(pZip->m_pRead(pZip->m_pIO_opaque,cdir_ofs,pZip->m_pState->m_central_dir.m_p,cdir_size)!=cdir_size)
      return 0;
    p=(BYTE *)pZip->m_pState->m_central_dir.m_p;
    for(n=cdir_size,i=0; i < pZip->m_total_files; ++i)
      {
      DWORD total_header_size,comp_size,decomp_size,disk_index;
      if((n < MZ_ZIP_CENTRAL_DIR_HEADER_SIZE) || (*((DWORD *)(p))!=MZ_ZIP_CENTRAL_DIR_HEADER_SIG))
        return 0;
      ((DWORD *)((&pZip->m_pState->m_central_dir_offsets)->m_p))[i]=(DWORD)(p - (BYTE *)pZip->m_pState->m_central_dir.m_p);
      if(sort_central_dir)
        ((DWORD *)((&pZip->m_pState->m_sorted_central_dir_offsets)->m_p))[i]=i;
      comp_size=*((DWORD *)(p + MZ_ZIP_CDH_COMPRESSED_SIZE_OFS));
      decomp_size=*((DWORD *)(p + MZ_ZIP_CDH_DECOMPRESSED_SIZE_OFS));
      if(((!*((DWORD *)(p + MZ_ZIP_CDH_METHOD_OFS))) && (decomp_size!=comp_size)) || (decomp_size && !comp_size) || (decomp_size == 0xFFFFFFFF) || (comp_size == 0xFFFFFFFF))
        return 0;
      disk_index=*((WORD *)(p + MZ_ZIP_CDH_DISK_START_OFS));
      if((disk_index!=num_this_disk) && (disk_index!=1))
        return 0;
      if(((QWORD)*((DWORD *)(p + MZ_ZIP_CDH_LOCAL_HEADER_OFS)) + MZ_ZIP_LOCAL_DIR_HEADER_SIZE + comp_size) > pZip->m_archive_size)
        return 0;
      if((total_header_size=MZ_ZIP_CENTRAL_DIR_HEADER_SIZE + *((WORD *)(p + MZ_ZIP_CDH_FILENAME_LEN_OFS)) + *((WORD *)(p + MZ_ZIP_CDH_EXTRA_LEN_OFS)) + *((WORD *)(p + MZ_ZIP_CDH_COMMENT_LEN_OFS))) > n)
        return 0;
      n-=total_header_size; p+=total_header_size;
      }
    }
  if(sort_central_dir)
    mz_zip_reader_sort_central_dir_offsets_by_filename(pZip);
  return 1;
  }

int mz_zip_reader_end(mz_zip_archive *pZip)
  {
  if((!pZip) || (!pZip->m_pState) || (!pZip->m_pAlloc) || (!pZip->m_pFree) || (pZip->m_zip_mode!=MZ_ZIP_MODE_READING))
    return 0;
  if(pZip->m_pState)
    {
    mz_zip_internal_state *pState=pZip->m_pState; pZip->m_pState=NULL;
    mz_zip_array_clear(pZip,&pState->m_central_dir);
    mz_zip_array_clear(pZip,&pState->m_central_dir_offsets);
    mz_zip_array_clear(pZip,&pState->m_sorted_central_dir_offsets);
    if(pState->m_pFile)
      {
      fclose(pState->m_pFile);
      pState->m_pFile=NULL;
      }
    pZip->m_pFree(pZip->m_pAlloc_opaque,pState);
    }
  pZip->m_zip_mode=MZ_ZIP_MODE_INVALID;
  return 1;
  }

int mz_zip_reader_init(mz_zip_archive *pZip,QWORD size,DWORD flags)
  {
  if((!pZip) || (!pZip->m_pRead))
    return 0;
  if(!mz_zip_reader_init_internal(pZip,flags))
    return 0;
  pZip->m_archive_size=size;
  if(!mz_zip_reader_read_central_dir(pZip,flags))
    {
    mz_zip_reader_end(pZip);
    return 0;
    }
  return 1;
  }

size_t mz_zip_mem_read_func(void*pOpaque,QWORD file_ofs,void*pBuf,size_t n)
  {
  mz_zip_archive *pZip=(mz_zip_archive *)pOpaque;
  size_t s=(file_ofs >= pZip->m_archive_size) ? 0 : (size_t)min(pZip->m_archive_size - file_ofs,n);
  memcpy(pBuf,(BYTE *)pZip->m_pState->m_pMem + file_ofs,s);
  return s;
  }

int mz_zip_reader_init_mem(mz_zip_archive *pZip,void*pMem,size_t size,DWORD flags)
  {
  if(!mz_zip_reader_init_internal(pZip,flags))
    return 0;
  pZip->m_archive_size=size;
  pZip->m_pRead=mz_zip_mem_read_func;
  pZip->m_pIO_opaque=pZip;
  pZip->m_pState->m_pMem=pMem;
  pZip->m_pState->m_mem_size=size;
  if(!mz_zip_reader_read_central_dir(pZip,flags))
    {
    mz_zip_reader_end(pZip);
    return 0;
    }
  return 1;
  }

size_t mz_zip_file_read_func(void*pOpaque,QWORD file_ofs,void*pBuf,size_t n)
  {
  mz_zip_archive *pZip=(mz_zip_archive *)pOpaque;
  __int64 cur_ofs=ftell(pZip->m_pState->m_pFile);
  if(((__int64)file_ofs < 0) ||
    (((cur_ofs!=(__int64)file_ofs)) &&
    (fseek(pZip->m_pState->m_pFile,(long)((__int64)file_ofs),0))))
    return 0;
  return fread(pBuf,1,n,pZip->m_pState->m_pFile);
  }

int mz_zip_reader_init_file(mz_zip_archive *pZip,char*pFilename,DWORD flags)
  {
  QWORD file_size;
  FILE *pFile;
  pFile=fopen(pFilename,"rb");
  if(!pFile)
    return 0;
  if(fseek(pFile,0,2))
    {
    fclose(pFile);
    return 0;
    }
  file_size=ftell(pFile);
  if(!mz_zip_reader_init_internal(pZip,flags))
    {
    fclose(pFile);
    return 0;
    }
  pZip->m_pRead=mz_zip_file_read_func;
  pZip->m_pIO_opaque=pZip;
  pZip->m_pState->m_pFile=pFile;
  pZip->m_archive_size=file_size;
  if(!mz_zip_reader_read_central_dir(pZip,flags))
    {
    mz_zip_reader_end(pZip);
    return 0;
    }
  return 1;
  }

DWORD mz_zip_reader_get_num_files(mz_zip_archive *pZip)
  {
  return pZip ? pZip->m_total_files : 0;
  }

BYTE *mz_zip_reader_get_cdh(mz_zip_archive *pZip,DWORD file_index)
  {
  if((!pZip) || (!pZip->m_pState) || (file_index >= pZip->m_total_files) || (pZip->m_zip_mode!=MZ_ZIP_MODE_READING))
    return NULL;
  return &((BYTE *)((&pZip->m_pState->m_central_dir)->m_p))[((DWORD *)((&pZip->m_pState->m_central_dir_offsets)->m_p))[file_index]];
  }

int mz_zip_reader_is_file_encrypted(mz_zip_archive *pZip,DWORD file_index)
  {
  DWORD m_bit_flag;
  BYTE *p=mz_zip_reader_get_cdh(pZip,file_index);
  if(!p)
    return 0;
  m_bit_flag=*((WORD *)(p + MZ_ZIP_CDH_BIT_FLAG_OFS));
  return (m_bit_flag & 1);
  }

int mz_zip_reader_is_file_a_directory(mz_zip_archive *pZip,DWORD file_index)
  {
  DWORD filename_len,external_attr;
  BYTE *p=mz_zip_reader_get_cdh(pZip,file_index);
  if(!p)
    return 0;
  filename_len=*((WORD *)(p + MZ_ZIP_CDH_FILENAME_LEN_OFS));
  if(filename_len)
    {
    if(*(p + MZ_ZIP_CENTRAL_DIR_HEADER_SIZE + filename_len - 1) == '/')
      return 1;
    }
  external_attr=*((DWORD *)(p + MZ_ZIP_CDH_EXTERNAL_ATTR_OFS));
  if((external_attr & 0x10)!=0)
    return 1;
  return 0;
  }

int mz_zip_reader_file_stat(mz_zip_archive *pZip,DWORD file_index,mz_zip_archive_file_stat *pStat)
  {
  DWORD n;
  BYTE *p=mz_zip_reader_get_cdh(pZip,file_index);
  if((!p) || (!pStat))
    return 0;
  pStat->m_file_index=file_index;
  pStat->m_central_dir_ofs=((DWORD *)((&pZip->m_pState->m_central_dir_offsets)->m_p))[file_index];
  pStat->m_version_made_by=*((WORD *)(p + MZ_ZIP_CDH_VERSION_MADE_BY_OFS));
  pStat->m_version_needed=*((WORD *)(p + MZ_ZIP_CDH_VERSION_NEEDED_OFS));
  pStat->m_bit_flag=*((WORD *)(p + MZ_ZIP_CDH_BIT_FLAG_OFS));
  pStat->m_method=*((WORD *)(p + MZ_ZIP_CDH_METHOD_OFS));
  pStat->m_time=mz_zip_dos_to_time_t(*((WORD *)(p + MZ_ZIP_CDH_FILE_TIME_OFS)),*((WORD *)(p + MZ_ZIP_CDH_FILE_DATE_OFS)));
  pStat->m_crc32=*((DWORD *)(p + MZ_ZIP_CDH_CRC32_OFS));
  pStat->m_comp_size=*((DWORD *)(p + MZ_ZIP_CDH_COMPRESSED_SIZE_OFS));
  pStat->m_uncomp_size=*((DWORD *)(p + MZ_ZIP_CDH_DECOMPRESSED_SIZE_OFS));
  pStat->m_internal_attr=*((WORD *)(p + MZ_ZIP_CDH_INTERNAL_ATTR_OFS));
  pStat->m_external_attr=*((DWORD *)(p + MZ_ZIP_CDH_EXTERNAL_ATTR_OFS));
  pStat->m_local_header_ofs=*((DWORD *)(p + MZ_ZIP_CDH_LOCAL_HEADER_OFS));
  n=*((WORD *)(p + MZ_ZIP_CDH_FILENAME_LEN_OFS)); n=min(n,MZ_ZIP_MAX_ARCHIVE_FILENAME_SIZE - 1);
  memcpy(pStat->m_filename,p + MZ_ZIP_CENTRAL_DIR_HEADER_SIZE,n); pStat->m_filename[n]='\0';
  n=*((WORD *)(p + MZ_ZIP_CDH_COMMENT_LEN_OFS)); n=min(n,MZ_ZIP_MAX_ARCHIVE_FILE_COMMENT_SIZE - 1);
  pStat->m_comment_size=n;
  memcpy(pStat->m_comment,p + MZ_ZIP_CENTRAL_DIR_HEADER_SIZE + *((WORD *)(p + MZ_ZIP_CDH_FILENAME_LEN_OFS)) + *((WORD *)(p + MZ_ZIP_CDH_EXTRA_LEN_OFS)),n); pStat->m_comment[n]='\0';
  return 1;
  }

DWORD mz_zip_reader_get_filename(mz_zip_archive *pZip,DWORD file_index,char*pFilename,DWORD filename_buf_size)
  {
  DWORD n;
  BYTE *p=mz_zip_reader_get_cdh(pZip,file_index);
  if(!p)   { if(filename_buf_size) pFilename[0]='\0'; return 0;   }
  n=*((WORD *)(p + MZ_ZIP_CDH_FILENAME_LEN_OFS));
  if(filename_buf_size)
    {
    n=min(n,filename_buf_size - 1);
    memcpy(pFilename,p + MZ_ZIP_CENTRAL_DIR_HEADER_SIZE,n);
    pFilename[n]='\0';
    }
  return n + 1;
  }

int mz_zip_reader_string_equal(char*pA,char*pB,DWORD len,DWORD flags)
  {
  DWORD i;
  if(flags & MZ_ZIP_FLAG_CASE_SENSITIVE)
    return 0 == memcmp(pA,pB,len);
  for(i=0; i < len; ++i)
    if(tolower(pA[i])!=tolower(pB[i]))
      return 0;
  return 1;
  }

int mz_zip_reader_filename_compare(mz_zip_array *pCentral_dir_array,mz_zip_array *pCentral_dir_offsets,DWORD l_index,char*pR,DWORD r_len)
  {
  BYTE *pL=&((BYTE *)((pCentral_dir_array)->m_p))[((DWORD *)((pCentral_dir_offsets)->m_p))[l_index]],*pE;
  DWORD l_len=*((WORD *)(pL + MZ_ZIP_CDH_FILENAME_LEN_OFS));
  BYTE l=0,r=0;
  pL+=MZ_ZIP_CENTRAL_DIR_HEADER_SIZE;
  pE=pL + min(l_len,r_len);
  while(pL < pE)
    {
    if((l=tolower(*pL))!=(r=tolower(*pR)))
      break;
    pL++; pR++;
    }
  return (pL == pE) ? (int)(l_len - r_len) : (l - r);
  }

int mz_zip_reader_locate_file_binary_search(mz_zip_archive *pZip,char*pFilename)
  {
  mz_zip_internal_state *pState=pZip->m_pState;
  mz_zip_array *pCentral_dir_offsets=&pState->m_central_dir_offsets;
  mz_zip_array *pCentral_dir=&pState->m_central_dir;
  DWORD *pIndices=&((DWORD *)((&pState->m_sorted_central_dir_offsets)->m_p))[0];
  int size=pZip->m_total_files;
  DWORD filename_len=(DWORD)strlen(pFilename);
  int l=0,h=size - 1;
  while(l <= h)
    {
    int m=(l + h) >> 1,file_index=pIndices[m],comp=mz_zip_reader_filename_compare(pCentral_dir,pCentral_dir_offsets,file_index,pFilename,filename_len);
    if(!comp)
      return file_index;
    else if(comp < 0)
      l=m + 1;
    else
      h=m - 1;
    }
  return -1;
  }

int mz_zip_reader_locate_file(mz_zip_archive *pZip,char*pName,char*pComment,DWORD flags)
  {
  DWORD file_index; size_t name_len,comment_len;
  if((!pZip) || (!pZip->m_pState) || (!pName) || (pZip->m_zip_mode!=MZ_ZIP_MODE_READING))
    return -1;
  if(((flags & (MZ_ZIP_FLAG_IGNORE_PATH | MZ_ZIP_FLAG_CASE_SENSITIVE)) == 0) && (!pComment) && (pZip->m_pState->m_sorted_central_dir_offsets.m_size))
    return mz_zip_reader_locate_file_binary_search(pZip,pName);
  name_len=strlen(pName); if(name_len > 0xFFFF) return -1;
  comment_len=pComment ? strlen(pComment) : 0; if(comment_len > 0xFFFF) return -1;
  for(file_index=0; file_index < pZip->m_total_files; file_index++)
    {
    BYTE *pHeader=&((BYTE *)((&pZip->m_pState->m_central_dir)->m_p))[((DWORD *)((&pZip->m_pState->m_central_dir_offsets)->m_p))[file_index]];
    DWORD filename_len=*((WORD *)(pHeader + MZ_ZIP_CDH_FILENAME_LEN_OFS));
    char*pFilename=(char*)pHeader + MZ_ZIP_CENTRAL_DIR_HEADER_SIZE;
    if(filename_len < name_len)
      continue;
    if(comment_len)
      {
      DWORD file_extra_len=*((WORD *)(pHeader + MZ_ZIP_CDH_EXTRA_LEN_OFS)),file_comment_len=*((WORD *)(pHeader + MZ_ZIP_CDH_COMMENT_LEN_OFS));
      char*pFile_comment=pFilename + filename_len + file_extra_len;
      if((file_comment_len!=comment_len) || (!mz_zip_reader_string_equal(pComment,pFile_comment,file_comment_len,flags)))
        continue;
      }
    if((flags & MZ_ZIP_FLAG_IGNORE_PATH) && (filename_len))
      {
      int ofs=filename_len - 1;
      do
        {
        if((pFilename[ofs] == '/') || (pFilename[ofs] == '\\') || (pFilename[ofs] == ':'))
          break;
        }while(--ofs >= 0);
      ofs++;
      pFilename+=ofs; filename_len-=ofs;
      }
    if((filename_len == name_len) && (mz_zip_reader_string_equal(pName,pFilename,filename_len,flags)))
      return file_index;
    }
  return -1;
  }

int mz_zip_reader_extract_to_mem_no_alloc(mz_zip_archive *pZip,DWORD file_index,void*pBuf,size_t buf_size,DWORD flags,void*pUser_read_buf,size_t user_read_buf_size)
  {
  int status=TINFL_STATUS_DONE;
  QWORD needed_size,cur_file_ofs,comp_remaining,out_buf_ofs=0,read_buf_size,read_buf_ofs=0,read_buf_avail;
  mz_zip_archive_file_stat file_stat;
  void*pRead_buf;
  DWORD local_header_u32[(MZ_ZIP_LOCAL_DIR_HEADER_SIZE + sizeof(DWORD) - 1) / sizeof(DWORD)]; BYTE *pLocal_header=(BYTE *)local_header_u32;
  tinfl_decompressor inflator;
  if((buf_size) && (!pBuf))
    return 0;
  if(!mz_zip_reader_file_stat(pZip,file_index,&file_stat))
    return 0;
  if(!file_stat.m_comp_size)
    return 1;
  if(mz_zip_reader_is_file_a_directory(pZip,file_index))
    return 1;
  if(file_stat.m_bit_flag & (1 | 32))
    return 0;
  if((!(flags & MZ_ZIP_FLAG_COMPRESSED_DATA)) && (file_stat.m_method!=0) && (file_stat.m_method!=8))
    return 0;
  needed_size=(flags & MZ_ZIP_FLAG_COMPRESSED_DATA) ? file_stat.m_comp_size : file_stat.m_uncomp_size;
  if(buf_size < needed_size)
    return 0;
  cur_file_ofs=file_stat.m_local_header_ofs;
  if(pZip->m_pRead(pZip->m_pIO_opaque,cur_file_ofs,pLocal_header,MZ_ZIP_LOCAL_DIR_HEADER_SIZE)!=MZ_ZIP_LOCAL_DIR_HEADER_SIZE)
    return 0;
  if(*((DWORD *)(pLocal_header))!=MZ_ZIP_LOCAL_DIR_HEADER_SIG)
    return 0;
  cur_file_ofs+=MZ_ZIP_LOCAL_DIR_HEADER_SIZE + *((WORD *)(pLocal_header + MZ_ZIP_LDH_FILENAME_LEN_OFS)) + *((WORD *)(pLocal_header + MZ_ZIP_LDH_EXTRA_LEN_OFS));
  if((cur_file_ofs + file_stat.m_comp_size) > pZip->m_archive_size)
    return 0;
  if((flags & MZ_ZIP_FLAG_COMPRESSED_DATA) || (!file_stat.m_method))
    {
    if(pZip->m_pRead(pZip->m_pIO_opaque,cur_file_ofs,pBuf,(size_t)needed_size)!=needed_size)
      return 0;
    return ((flags & MZ_ZIP_FLAG_COMPRESSED_DATA)!=0) || (mz_crc32((0),(BYTE *)pBuf,(size_t)file_stat.m_uncomp_size) == file_stat.m_crc32);
    }
  do{
    (&inflator)->m_state=0;
    }while(0);
  if(pZip->m_pState->m_pMem)
    {
    pRead_buf=(BYTE *)pZip->m_pState->m_pMem + cur_file_ofs;
    read_buf_size=read_buf_avail=file_stat.m_comp_size;
    comp_remaining=0;
    }
  else if(pUser_read_buf)
    {
    if(!user_read_buf_size)
      return 0;
    pRead_buf=(BYTE *)pUser_read_buf;
    read_buf_size=user_read_buf_size;
    read_buf_avail=0;
    comp_remaining=file_stat.m_comp_size;
    }
  else
    {
    read_buf_size=min(file_stat.m_comp_size,MZ_ZIP_MAX_IO_BUF_SIZE);
    if(((0,sizeof(size_t) == sizeof(DWORD))) && (read_buf_size > 0x7FFFFFFF))
      return 0;
    if(NULL == (pRead_buf=pZip->m_pAlloc(pZip->m_pAlloc_opaque,1,(size_t)read_buf_size)))
      return 0;
    read_buf_avail=0;
    comp_remaining=file_stat.m_comp_size;
    }
  do
    {
    size_t in_buf_size,out_buf_size=(size_t)(file_stat.m_uncomp_size - out_buf_ofs);
    if((!read_buf_avail) && (!pZip->m_pState->m_pMem))
      {
      read_buf_avail=min(read_buf_size,comp_remaining);
      if(pZip->m_pRead(pZip->m_pIO_opaque,cur_file_ofs,pRead_buf,(size_t)read_buf_avail)!=read_buf_avail)
        {
        status=TINFL_STATUS_FAILED;
        break;
        }
      cur_file_ofs+=read_buf_avail;
      comp_remaining-=read_buf_avail;
      read_buf_ofs=0;
      }
    in_buf_size=(size_t)read_buf_avail;
    status=tinfl_decompress(&inflator,(BYTE *)pRead_buf + read_buf_ofs,&in_buf_size,(BYTE *)pBuf,(BYTE *)pBuf + out_buf_ofs,&out_buf_size,TINFL_FLAG_USING_NON_WRAPPING_OUTPUT_BUF | (comp_remaining ? TINFL_FLAG_HAS_MORE_INPUT : 0));
    read_buf_avail-=in_buf_size;
    read_buf_ofs+=in_buf_size;
    out_buf_ofs+=out_buf_size;
    }while(status == TINFL_STATUS_NEEDS_MORE_INPUT);
  if(status == TINFL_STATUS_DONE)
    {
    if((out_buf_ofs!=file_stat.m_uncomp_size) || (mz_crc32((0),(BYTE *)pBuf,(size_t)file_stat.m_uncomp_size)!=file_stat.m_crc32))
      status=TINFL_STATUS_FAILED;
    }
  if((!pZip->m_pState->m_pMem) && (!pUser_read_buf))
    pZip->m_pFree(pZip->m_pAlloc_opaque,pRead_buf);
  return status == TINFL_STATUS_DONE;
  }

int mz_zip_reader_extract_file_to_mem_no_alloc(mz_zip_archive *pZip,char*pFilename,void*pBuf,size_t buf_size,DWORD flags,void*pUser_read_buf,size_t user_read_buf_size)
  {
  int file_index=mz_zip_reader_locate_file(pZip,pFilename,NULL,flags);
  if(file_index < 0)
    return 0;
  return mz_zip_reader_extract_to_mem_no_alloc(pZip,file_index,pBuf,buf_size,flags,pUser_read_buf,user_read_buf_size);
  }

int mz_zip_reader_extract_to_mem(mz_zip_archive *pZip,DWORD file_index,void*pBuf,size_t buf_size,DWORD flags)
  {
  return mz_zip_reader_extract_to_mem_no_alloc(pZip,file_index,pBuf,buf_size,flags,NULL,0);
  }

int mz_zip_reader_extract_file_to_mem(mz_zip_archive *pZip,char*pFilename,void*pBuf,size_t buf_size,DWORD flags)
  {
  return mz_zip_reader_extract_file_to_mem_no_alloc(pZip,pFilename,pBuf,buf_size,flags,NULL,0);
  }

void*mz_zip_reader_extract_to_heap(mz_zip_archive *pZip,DWORD file_index,size_t *pSize,DWORD flags)
  {
  QWORD comp_size,uncomp_size,alloc_size;
  BYTE *p=mz_zip_reader_get_cdh(pZip,file_index);
  void*pBuf;
  if(pSize)
    *pSize=0;
  if(!p)
    return NULL;
  comp_size=*((DWORD *)(p + MZ_ZIP_CDH_COMPRESSED_SIZE_OFS));
  uncomp_size=*((DWORD *)(p + MZ_ZIP_CDH_DECOMPRESSED_SIZE_OFS));
  alloc_size=(flags & MZ_ZIP_FLAG_COMPRESSED_DATA) ? comp_size : uncomp_size;
  if(((0,sizeof(size_t) == sizeof(DWORD))) && (alloc_size > 0x7FFFFFFF))
    return NULL;
  if(NULL == (pBuf=pZip->m_pAlloc(pZip->m_pAlloc_opaque,1,(size_t)alloc_size)))
    return NULL;
  if(!mz_zip_reader_extract_to_mem(pZip,file_index,pBuf,(size_t)alloc_size,flags))
    {
    pZip->m_pFree(pZip->m_pAlloc_opaque,pBuf);
    return NULL;
    }
  if(pSize) *pSize=(size_t)alloc_size;
  return pBuf;
  }

void*mz_zip_reader_extract_file_to_heap(mz_zip_archive *pZip,char*pFilename,size_t *pSize,DWORD flags)
  {
  int file_index=mz_zip_reader_locate_file(pZip,pFilename,NULL,flags);
  if(file_index < 0)
    {
    if(pSize) *pSize=0;
    return 0;
    }
  return mz_zip_reader_extract_to_heap(pZip,file_index,pSize,flags);
  }

int mz_zip_reader_extract_to_callback(mz_zip_archive *pZip,DWORD file_index,mz_file_write_func pCallback,void*pOpaque,DWORD flags)
  {
  int status=TINFL_STATUS_DONE; DWORD file_crc32=(0);
  QWORD read_buf_size,read_buf_ofs=0,read_buf_avail,comp_remaining,out_buf_ofs=0,cur_file_ofs;
  mz_zip_archive_file_stat file_stat;
  void*pRead_buf=NULL; void*pWrite_buf=NULL;
  DWORD local_header_u32[(MZ_ZIP_LOCAL_DIR_HEADER_SIZE + sizeof(DWORD) - 1) / sizeof(DWORD)]; BYTE *pLocal_header=(BYTE *)local_header_u32;
  if(!mz_zip_reader_file_stat(pZip,file_index,&file_stat))
    return 0;
  if(!file_stat.m_comp_size)
    return 1;
  if(mz_zip_reader_is_file_a_directory(pZip,file_index))
    return 1;
  if(file_stat.m_bit_flag & (1 | 32))
    return 0;
  if((!(flags & MZ_ZIP_FLAG_COMPRESSED_DATA)) && (file_stat.m_method!=0) && (file_stat.m_method!=8))
    return 0;
  cur_file_ofs=file_stat.m_local_header_ofs;
  if(pZip->m_pRead(pZip->m_pIO_opaque,cur_file_ofs,pLocal_header,MZ_ZIP_LOCAL_DIR_HEADER_SIZE)!=MZ_ZIP_LOCAL_DIR_HEADER_SIZE)
    return 0;
  if(*((DWORD *)(pLocal_header))!=MZ_ZIP_LOCAL_DIR_HEADER_SIG)
    return 0;
  cur_file_ofs+=MZ_ZIP_LOCAL_DIR_HEADER_SIZE + *((WORD *)(pLocal_header + MZ_ZIP_LDH_FILENAME_LEN_OFS)) + *((WORD *)(pLocal_header + MZ_ZIP_LDH_EXTRA_LEN_OFS));
  if((cur_file_ofs + file_stat.m_comp_size) > pZip->m_archive_size)
    return 0;
  if(pZip->m_pState->m_pMem)
    {
    pRead_buf=(BYTE *)pZip->m_pState->m_pMem + cur_file_ofs;
    read_buf_size=read_buf_avail=file_stat.m_comp_size;
    comp_remaining=0;
    }
  else
    {
    read_buf_size=min(file_stat.m_comp_size,MZ_ZIP_MAX_IO_BUF_SIZE);
    if(NULL == (pRead_buf=pZip->m_pAlloc(pZip->m_pAlloc_opaque,1,(size_t)read_buf_size)))
      return 0;
    read_buf_avail=0;
    comp_remaining=file_stat.m_comp_size;
    }
  if((flags & MZ_ZIP_FLAG_COMPRESSED_DATA) || (!file_stat.m_method))
    {
    if(pZip->m_pState->m_pMem)
      {
      if(((0,sizeof(size_t) == sizeof(DWORD))) && (file_stat.m_comp_size > 0xFFFFFFFF))
        return 0;
      if(pCallback(pOpaque,out_buf_ofs,pRead_buf,(size_t)file_stat.m_comp_size)!=file_stat.m_comp_size)
        status=TINFL_STATUS_FAILED;
      else if(!(flags & MZ_ZIP_FLAG_COMPRESSED_DATA))
        file_crc32=(DWORD)mz_crc32(file_crc32,(BYTE *)pRead_buf,(size_t)file_stat.m_comp_size);
      cur_file_ofs+=file_stat.m_comp_size;
      out_buf_ofs+=file_stat.m_comp_size;
      comp_remaining=0;
      }
    else
      {
      while(comp_remaining)
        {
        read_buf_avail=min(read_buf_size,comp_remaining);
        if(pZip->m_pRead(pZip->m_pIO_opaque,cur_file_ofs,pRead_buf,(size_t)read_buf_avail)!=read_buf_avail)
          {
          status=TINFL_STATUS_FAILED;
          break;
          }
        if(!(flags & MZ_ZIP_FLAG_COMPRESSED_DATA))
          file_crc32=(DWORD)mz_crc32(file_crc32,(BYTE *)pRead_buf,(size_t)read_buf_avail);
        if(pCallback(pOpaque,out_buf_ofs,pRead_buf,(size_t)read_buf_avail)!=read_buf_avail)
          {
          status=TINFL_STATUS_FAILED;
          break;
          }
        cur_file_ofs+=read_buf_avail;
        out_buf_ofs+=read_buf_avail;
        comp_remaining-=read_buf_avail;
        }
      }
    }
  else
    {
    tinfl_decompressor inflator;
    do{
      (&inflator)->m_state=0;
      }while(0);
    if(NULL == (pWrite_buf=pZip->m_pAlloc(pZip->m_pAlloc_opaque,1,32768)))
      status=TINFL_STATUS_FAILED;
    else
      {
      do
        {
        BYTE *pWrite_buf_cur=(BYTE *)pWrite_buf + (out_buf_ofs & (32768 - 1));
        size_t in_buf_size,out_buf_size=32768 - (out_buf_ofs & (32768 - 1));
        if((!read_buf_avail) && (!pZip->m_pState->m_pMem))
          {
          read_buf_avail=min(read_buf_size,comp_remaining);
          if(pZip->m_pRead(pZip->m_pIO_opaque,cur_file_ofs,pRead_buf,(size_t)read_buf_avail)!=read_buf_avail)
            {
            status=TINFL_STATUS_FAILED;
            break;
            }
          cur_file_ofs+=read_buf_avail;
          comp_remaining-=read_buf_avail;
          read_buf_ofs=0;
          }
        in_buf_size=(size_t)read_buf_avail;
        status=tinfl_decompress(&inflator,(BYTE *)pRead_buf + read_buf_ofs,&in_buf_size,(BYTE *)pWrite_buf,pWrite_buf_cur,&out_buf_size,comp_remaining ? TINFL_FLAG_HAS_MORE_INPUT : 0);
        read_buf_avail-=in_buf_size;
        read_buf_ofs+=in_buf_size;
        if(out_buf_size)
          {
          if(pCallback(pOpaque,out_buf_ofs,pWrite_buf_cur,out_buf_size)!=out_buf_size)
            {
            status=TINFL_STATUS_FAILED;
            break;
            }
          file_crc32=(DWORD)mz_crc32(file_crc32,pWrite_buf_cur,out_buf_size);
          if((out_buf_ofs+=out_buf_size) > file_stat.m_uncomp_size)
            {
            status=TINFL_STATUS_FAILED;
            break;
            }
          }
        }while((status == TINFL_STATUS_NEEDS_MORE_INPUT) || (status == TINFL_STATUS_HAS_MORE_OUTPUT));
      }
    }
  if((status == TINFL_STATUS_DONE) && (!(flags & MZ_ZIP_FLAG_COMPRESSED_DATA)))
    {
    if((out_buf_ofs!=file_stat.m_uncomp_size) || (file_crc32!=file_stat.m_crc32))
      status=TINFL_STATUS_FAILED;
    }
  if(!pZip->m_pState->m_pMem)
    pZip->m_pFree(pZip->m_pAlloc_opaque,pRead_buf);
  if(pWrite_buf)
    pZip->m_pFree(pZip->m_pAlloc_opaque,pWrite_buf);
  return status == TINFL_STATUS_DONE;
  }

int mz_zip_reader_extract_file_to_callback(mz_zip_archive *pZip,char*pFilename,mz_file_write_func pCallback,void*pOpaque,DWORD flags)
  {
  int file_index=mz_zip_reader_locate_file(pZip,pFilename,NULL,flags);
  if(file_index < 0)
    return 0;
  return mz_zip_reader_extract_to_callback(pZip,file_index,pCallback,pOpaque,flags);
  }

size_t mz_zip_file_write_callback(void*pOpaque,QWORD ofs,void*pBuf,size_t n)
  {
  (void)ofs; return fwrite(pBuf,1,n,(FILE*)pOpaque);
  }

void mz_write_le16(BYTE *p,WORD v)   { p[0]=(BYTE)v; p[1]=(BYTE)(v >> 8);   }

void mz_write_le32(BYTE *p,DWORD v)   { p[0]=(BYTE)v; p[1]=(BYTE)(v >> 8); p[2]=(BYTE)(v >> 16); p[3]=(BYTE)(v >> 24);   }

int mz_zip_writer_init(mz_zip_archive *pZip,QWORD existing_size)
  {
  if((!pZip) || (pZip->m_pState) || (!pZip->m_pWrite) || (pZip->m_zip_mode!=MZ_ZIP_MODE_INVALID))
    return 0;
  if(pZip->m_file_offset_alignment)
    {
    if(pZip->m_file_offset_alignment & (pZip->m_file_offset_alignment - 1))
      return 0;
    }
  if(!pZip->m_pAlloc) pZip->m_pAlloc=def_alloc_func;
  if(!pZip->m_pFree) pZip->m_pFree=def_free_func;
  if(!pZip->m_pRealloc) pZip->m_pRealloc=def_realloc_func;
  pZip->m_zip_mode=MZ_ZIP_MODE_WRITING;
  pZip->m_archive_size=existing_size;
  pZip->m_central_directory_file_ofs=0;
  pZip->m_total_files=0;
  if(NULL == (pZip->m_pState=(mz_zip_internal_state *)pZip->m_pAlloc(pZip->m_pAlloc_opaque,1,sizeof(mz_zip_internal_state))))
    return 0;
  memset(pZip->m_pState,0,sizeof(mz_zip_internal_state));
  (&pZip->m_pState->m_central_dir)->m_element_size=sizeof(BYTE);
  (&pZip->m_pState->m_central_dir_offsets)->m_element_size=sizeof(DWORD);
  (&pZip->m_pState->m_sorted_central_dir_offsets)->m_element_size=sizeof(DWORD);
  return 1;
  }

size_t mz_zip_heap_write_func(void*pOpaque,QWORD file_ofs,void*pBuf,size_t n)
  {
  mz_zip_archive *pZip=(mz_zip_archive *)pOpaque;
  mz_zip_internal_state *pState=pZip->m_pState;
  QWORD new_size=max(file_ofs + n,pState->m_mem_size);
  if((!n) || ((0,sizeof(size_t) == sizeof(DWORD)) && (new_size > 0x7FFFFFFF)))
    return 0;
  if(new_size > pState->m_mem_capacity)
    {
    void*pNew_block;
    size_t new_capacity=max(64,pState->m_mem_capacity); while(new_capacity < new_size) new_capacity *= 2;
    if(NULL == (pNew_block=pZip->m_pRealloc(pZip->m_pAlloc_opaque,pState->m_pMem,1,new_capacity)))
      return 0;
    pState->m_pMem=pNew_block; pState->m_mem_capacity=new_capacity;
    }
  memcpy((BYTE *)pState->m_pMem + file_ofs,pBuf,n);
  pState->m_mem_size=(size_t)new_size;
  return n;
  }

int mz_zip_writer_end(mz_zip_archive *pZip)
  {
  mz_zip_internal_state *pState;
  int status=1;
  if((!pZip) || (!pZip->m_pState) || (!pZip->m_pAlloc) || (!pZip->m_pFree) || ((pZip->m_zip_mode!=MZ_ZIP_MODE_WRITING) && (pZip->m_zip_mode!=MZ_ZIP_MODE_WRITING_HAS_BEEN_FINALIZED)))
    return 0;
  pState=pZip->m_pState;
  pZip->m_pState=NULL;
  mz_zip_array_clear(pZip,&pState->m_central_dir);
  mz_zip_array_clear(pZip,&pState->m_central_dir_offsets);
  mz_zip_array_clear(pZip,&pState->m_sorted_central_dir_offsets);
  if(pState->m_pFile)
    {
    fclose(pState->m_pFile);
    pState->m_pFile=NULL;
    }
  if((pZip->m_pWrite == mz_zip_heap_write_func) && (pState->m_pMem))
    {
    pZip->m_pFree(pZip->m_pAlloc_opaque,pState->m_pMem);
    pState->m_pMem=NULL;
    }
  pZip->m_pFree(pZip->m_pAlloc_opaque,pState);
  pZip->m_zip_mode=MZ_ZIP_MODE_INVALID;
  return status;
  }

int mz_zip_writer_init_heap(mz_zip_archive *pZip,size_t size_to_reserve_at_beginning,size_t initial_allocation_size)
  {
  pZip->m_pWrite=mz_zip_heap_write_func;
  pZip->m_pIO_opaque=pZip;
  if(!mz_zip_writer_init(pZip,size_to_reserve_at_beginning))
    return 0;
  if(0!=(initial_allocation_size=max(initial_allocation_size,size_to_reserve_at_beginning)))
    {
    if(NULL == (pZip->m_pState->m_pMem=pZip->m_pAlloc(pZip->m_pAlloc_opaque,1,initial_allocation_size)))
      {
      mz_zip_writer_end(pZip);
      return 0;
      }
    pZip->m_pState->m_mem_capacity=initial_allocation_size;
    }
  return 1;
  }

size_t mz_zip_file_write_func(void*pOpaque,QWORD file_ofs,void*pBuf,size_t n)
  {
  mz_zip_archive *pZip=(mz_zip_archive *)pOpaque;
  __int64 cur_ofs=ftell(pZip->m_pState->m_pFile);
  if(((__int64)file_ofs < 0) ||
    (((cur_ofs!=(__int64)file_ofs)) &&
    (fseek(pZip->m_pState->m_pFile,(long)((__int64)file_ofs),0))))
    return 0;
  return fwrite(pBuf,1,n,pZip->m_pState->m_pFile);
  }

int mz_zip_writer_init_file(mz_zip_archive *pZip,char*pFilename,QWORD size_to_reserve_at_beginning)
  {
  FILE *pFile;
  pZip->m_pWrite=mz_zip_file_write_func;
  pZip->m_pIO_opaque=pZip;
  if(!mz_zip_writer_init(pZip,size_to_reserve_at_beginning))
    return 0;
  if(NULL == (pFile=fopen(pFilename,"wb")))
    {
    mz_zip_writer_end(pZip);
    return 0;
    }
  pZip->m_pState->m_pFile=pFile;
  if(size_to_reserve_at_beginning)
    {
    QWORD cur_ofs=0; char buf[4096]; memset(&buf,0,sizeof(buf));
    do
      {
      size_t n=(size_t)min(sizeof(buf),size_to_reserve_at_beginning);
      if(pZip->m_pWrite(pZip->m_pIO_opaque,cur_ofs,buf,n)!=n)
        {
        mz_zip_writer_end(pZip);
        return 0;
        }
      cur_ofs+=n; size_to_reserve_at_beginning-=n;
      }while(size_to_reserve_at_beginning);
    }
  return 1;
  }

int mz_zip_writer_init_from_reader(mz_zip_archive *pZip,char*pFilename)
  {
  mz_zip_internal_state *pState;
  if((!pZip) || (!pZip->m_pState) || (pZip->m_zip_mode!=MZ_ZIP_MODE_READING))
    return 0;
  if((pZip->m_total_files == 0xFFFF) || ((pZip->m_archive_size + MZ_ZIP_CENTRAL_DIR_HEADER_SIZE + MZ_ZIP_LOCAL_DIR_HEADER_SIZE) > 0xFFFFFFFF))
    return 0;
  pState=pZip->m_pState;
  if(pState->m_pFile)
    {
    if(pZip->m_pIO_opaque!=pZip)
      return 0;
    if(!pFilename)
      return 0;
    pZip->m_pWrite=mz_zip_file_write_func;
    if(NULL == (pState->m_pFile=freopen(pFilename,"r+b",pState->m_pFile)))
      {
      mz_zip_reader_end(pZip);
      return 0;
      }
    }
  else if(pState->m_pMem)
    {
    if(pZip->m_pIO_opaque!=pZip)
      return 0;
    pState->m_mem_capacity=pState->m_mem_size;
    pZip->m_pWrite=mz_zip_heap_write_func;
    }
  else if(!pZip->m_pWrite)
    return 0;
  pZip->m_archive_size=pZip->m_central_directory_file_ofs;
  pZip->m_zip_mode=MZ_ZIP_MODE_WRITING;
  pZip->m_central_directory_file_ofs=0;
  return 1;
  }

int mz_zip_writer_validate_archive_name(char*pArchive_name)
  {
  if(*pArchive_name == '/')
    return 0;
  while(*pArchive_name)
    {
    if((*pArchive_name == '\\') || (*pArchive_name == ':'))
      return 0;
    pArchive_name++;
    }
  return 1;
  }

DWORD mz_zip_writer_compute_padding_needed_for_file_alignment(mz_zip_archive *pZip)
  {
  DWORD n;
  if(!pZip->m_file_offset_alignment)
    return 0;
  n=(DWORD)(pZip->m_archive_size & (pZip->m_file_offset_alignment - 1));
  return (pZip->m_file_offset_alignment - n) & (pZip->m_file_offset_alignment - 1);
  }

int mz_zip_writer_write_zeros(mz_zip_archive *pZip,QWORD cur_file_ofs,DWORD n)
  {
  char buf[4096];
  memset(buf,0,min(sizeof(buf),n));
  while(n)
    {
    DWORD s=min(sizeof(buf),n);
    if(pZip->m_pWrite(pZip->m_pIO_opaque,cur_file_ofs,buf,s)!=s)
      return 0;
    cur_file_ofs+=s; n-=s;
    }
  return 1;
  }

typedef struct
  {
  mz_zip_archive *m_pZip;
  QWORD m_cur_archive_file_ofs;
  QWORD m_comp_size;
  } mz_zip_writer_add_state;

int mz_zip_writer_add_put_buf_callback(void* pBuf,int len,void*pUser)
  {
  mz_zip_writer_add_state *pState=(mz_zip_writer_add_state *)pUser;
  if((int)pState->m_pZip->m_pWrite(pState->m_pZip->m_pIO_opaque,pState->m_cur_archive_file_ofs,pBuf,len)!=len)
    return 0;
  pState->m_cur_archive_file_ofs+=len;
  pState->m_comp_size+=len;
  return 1;
  }

int mz_zip_writer_create_local_dir_header(mz_zip_archive *pZip,BYTE *pDst,WORD filename_size,WORD extra_size,QWORD uncomp_size,QWORD comp_size,DWORD uncomp_crc32,WORD method,WORD bit_flags,WORD dos_time,WORD dos_date)
  {
  (void)pZip;
  memset(pDst,0,MZ_ZIP_LOCAL_DIR_HEADER_SIZE);
  mz_write_le32((BYTE *)(pDst + MZ_ZIP_LDH_SIG_OFS),(DWORD)(MZ_ZIP_LOCAL_DIR_HEADER_SIG));
  mz_write_le16((BYTE *)(pDst + MZ_ZIP_LDH_VERSION_NEEDED_OFS),(WORD)(method ? 20 : 0));
  mz_write_le16((BYTE *)(pDst + MZ_ZIP_LDH_BIT_FLAG_OFS),(WORD)(bit_flags));
  mz_write_le16((BYTE *)(pDst + MZ_ZIP_LDH_METHOD_OFS),(WORD)(method));
  mz_write_le16((BYTE *)(pDst + MZ_ZIP_LDH_FILE_TIME_OFS),(WORD)(dos_time));
  mz_write_le16((BYTE *)(pDst + MZ_ZIP_LDH_FILE_DATE_OFS),(WORD)(dos_date));
  mz_write_le32((BYTE *)(pDst + MZ_ZIP_LDH_CRC32_OFS),(DWORD)(uncomp_crc32));
  mz_write_le32((BYTE *)(pDst + MZ_ZIP_LDH_COMPRESSED_SIZE_OFS),(DWORD)(comp_size));
  mz_write_le32((BYTE *)(pDst + MZ_ZIP_LDH_DECOMPRESSED_SIZE_OFS),(DWORD)(uncomp_size));
  mz_write_le16((BYTE *)(pDst + MZ_ZIP_LDH_FILENAME_LEN_OFS),(WORD)(filename_size));
  mz_write_le16((BYTE *)(pDst + MZ_ZIP_LDH_EXTRA_LEN_OFS),(WORD)(extra_size));
  return 1;
  }

int mz_zip_writer_create_central_dir_header(mz_zip_archive *pZip,BYTE *pDst,WORD filename_size,WORD extra_size,WORD comment_size,QWORD uncomp_size,QWORD comp_size,DWORD uncomp_crc32,WORD method,WORD bit_flags,WORD dos_time,WORD dos_date,QWORD local_header_ofs,DWORD ext_attributes)
  {
  (void)pZip;
  memset(pDst,0,MZ_ZIP_CENTRAL_DIR_HEADER_SIZE);
  mz_write_le32((BYTE *)(pDst + MZ_ZIP_CDH_SIG_OFS),(DWORD)(MZ_ZIP_CENTRAL_DIR_HEADER_SIG));
  mz_write_le16((BYTE *)(pDst + MZ_ZIP_CDH_VERSION_NEEDED_OFS),(WORD)(method ? 20 : 0));
  mz_write_le16((BYTE *)(pDst + MZ_ZIP_CDH_BIT_FLAG_OFS),(WORD)(bit_flags));
  mz_write_le16((BYTE *)(pDst + MZ_ZIP_CDH_METHOD_OFS),(WORD)(method));
  mz_write_le16((BYTE *)(pDst + MZ_ZIP_CDH_FILE_TIME_OFS),(WORD)(dos_time));
  mz_write_le16((BYTE *)(pDst + MZ_ZIP_CDH_FILE_DATE_OFS),(WORD)(dos_date));
  mz_write_le32((BYTE *)(pDst + MZ_ZIP_CDH_CRC32_OFS),(DWORD)(uncomp_crc32));
  mz_write_le32((BYTE *)(pDst + MZ_ZIP_CDH_COMPRESSED_SIZE_OFS),(DWORD)(comp_size));
  mz_write_le32((BYTE *)(pDst + MZ_ZIP_CDH_DECOMPRESSED_SIZE_OFS),(DWORD)(uncomp_size));
  mz_write_le16((BYTE *)(pDst + MZ_ZIP_CDH_FILENAME_LEN_OFS),(WORD)(filename_size));
  mz_write_le16((BYTE *)(pDst + MZ_ZIP_CDH_EXTRA_LEN_OFS),(WORD)(extra_size));
  mz_write_le16((BYTE *)(pDst + MZ_ZIP_CDH_COMMENT_LEN_OFS),(WORD)(comment_size));
  mz_write_le32((BYTE *)(pDst + MZ_ZIP_CDH_EXTERNAL_ATTR_OFS),(DWORD)(ext_attributes));
  mz_write_le32((BYTE *)(pDst + MZ_ZIP_CDH_LOCAL_HEADER_OFS),(DWORD)(local_header_ofs));
  return 1;
  }

int mz_zip_writer_add_to_central_dir(mz_zip_archive *pZip,char*pFilename,WORD filename_size,void*pExtra,WORD extra_size,void*pComment,WORD comment_size,QWORD uncomp_size,QWORD comp_size,DWORD uncomp_crc32,WORD method,WORD bit_flags,WORD dos_time,WORD dos_date,QWORD local_header_ofs,DWORD ext_attributes)
  {
  mz_zip_internal_state *pState=pZip->m_pState;
  DWORD central_dir_ofs=(DWORD)pState->m_central_dir.m_size;
  size_t orig_central_dir_size=pState->m_central_dir.m_size;
  BYTE central_dir_header[MZ_ZIP_CENTRAL_DIR_HEADER_SIZE];
  if((local_header_ofs > 0xFFFFFFFF) || (((QWORD)pState->m_central_dir.m_size + MZ_ZIP_CENTRAL_DIR_HEADER_SIZE + filename_size + extra_size + comment_size) > 0xFFFFFFFF))
    return 0;
  if(!mz_zip_writer_create_central_dir_header(pZip,central_dir_header,filename_size,extra_size,comment_size,uncomp_size,comp_size,uncomp_crc32,method,bit_flags,dos_time,dos_date,local_header_ofs,ext_attributes))
    return 0;
  if((!mz_zip_array_push_back(pZip,&pState->m_central_dir,central_dir_header,MZ_ZIP_CENTRAL_DIR_HEADER_SIZE)) ||
      (!mz_zip_array_push_back(pZip,&pState->m_central_dir,pFilename,filename_size)) ||
      (!mz_zip_array_push_back(pZip,&pState->m_central_dir,pExtra,extra_size)) ||
      (!mz_zip_array_push_back(pZip,&pState->m_central_dir,pComment,comment_size)) ||
      (!mz_zip_array_push_back(pZip,&pState->m_central_dir_offsets,&central_dir_ofs,1)))
    {
    mz_zip_array_resize(pZip,&pState->m_central_dir,orig_central_dir_size,0);
    return 0;
    }
  return 1;
  }

int mz_zip_writer_add_file(mz_zip_archive *pZip,char*pArchive_name,char*pSrc_filename,void*pComment,WORD comment_size,DWORD level_and_flags)
  {
  DWORD uncomp_crc32=(0),level,num_alignment_padding_bytes;
  WORD method=0,dos_time=0,dos_date=0,ext_attributes=0;
  QWORD local_dir_header_ofs=pZip->m_archive_size,cur_archive_file_ofs=pZip->m_archive_size,uncomp_size=0,comp_size=0;
  size_t archive_name_size;
  BYTE local_dir_header[MZ_ZIP_LOCAL_DIR_HEADER_SIZE];
  FILE *pSrc_file=NULL;
  if((int)level_and_flags < 0)
    level_and_flags=MZ_DEFAULT_LEVEL;
  level=level_and_flags & 0xF;
  if((!pZip) || (!pZip->m_pState) || (pZip->m_zip_mode!=MZ_ZIP_MODE_WRITING) || (!pArchive_name) || ((comment_size) && (!pComment)) || (level > MZ_UBER_COMPRESSION))
    return 0;
  if(level_and_flags & MZ_ZIP_FLAG_COMPRESSED_DATA)
    return 0;
  if(!mz_zip_writer_validate_archive_name(pArchive_name))
    return 0;
  archive_name_size=strlen(pArchive_name);
  if(archive_name_size > 0xFFFF)
    return 0;
  num_alignment_padding_bytes=mz_zip_writer_compute_padding_needed_for_file_alignment(pZip);
  if((pZip->m_total_files == 0xFFFF) || ((pZip->m_archive_size + num_alignment_padding_bytes + MZ_ZIP_LOCAL_DIR_HEADER_SIZE + MZ_ZIP_CENTRAL_DIR_HEADER_SIZE + comment_size + archive_name_size) > 0xFFFFFFFF))
    return 0;
  if(!mz_zip_get_file_modified_time(pSrc_filename,&dos_time,&dos_date))
    return 0;
  pSrc_file=fopen(pSrc_filename,"rb");
  if(!pSrc_file)
    return 0;
  fseek(pSrc_file,0,2);
  uncomp_size=ftell(pSrc_file);
  fseek(pSrc_file,0,0);
  if(uncomp_size > 0xFFFFFFFF)
    {
    fclose(pSrc_file);
    return 0;
    }
  if(uncomp_size <= 3)
    level=0;
  if(!mz_zip_writer_write_zeros(pZip,cur_archive_file_ofs,num_alignment_padding_bytes + sizeof(local_dir_header)))
    {
    fclose(pSrc_file);
    return 0;
    }
  local_dir_header_ofs+=num_alignment_padding_bytes;
  if(pZip->m_file_offset_alignment)   { (void)( (!!((local_dir_header_ofs & (pZip->m_file_offset_alignment - 1)) == 0)) || (assert("(local_dir_header_ofs & (pZip->m_file_offset_alignment - 1)) == 0"),0) );   }
  cur_archive_file_ofs+=num_alignment_padding_bytes + sizeof(local_dir_header);
  memset(&local_dir_header,0,sizeof(local_dir_header));
  if(pZip->m_pWrite(pZip->m_pIO_opaque,cur_archive_file_ofs,pArchive_name,archive_name_size)!=archive_name_size)
    {
    fclose(pSrc_file);
    return 0;
    }
  cur_archive_file_ofs+=archive_name_size;
  if(uncomp_size)
    {
    QWORD uncomp_remaining=uncomp_size;
    void*pRead_buf=pZip->m_pAlloc(pZip->m_pAlloc_opaque,1,MZ_ZIP_MAX_IO_BUF_SIZE);
    if(!pRead_buf)
      {
      fclose(pSrc_file);
      return 0;
      }
    if(!level)
      {
      while(uncomp_remaining)
        {
        DWORD n=(DWORD)min(MZ_ZIP_MAX_IO_BUF_SIZE,uncomp_remaining);
        if((fread(pRead_buf,1,n,pSrc_file)!=n) || (pZip->m_pWrite(pZip->m_pIO_opaque,cur_archive_file_ofs,pRead_buf,n)!=n))
          {
          pZip->m_pFree(pZip->m_pAlloc_opaque,pRead_buf);
          fclose(pSrc_file);
          return 0;
          }
        uncomp_crc32=(DWORD)mz_crc32(uncomp_crc32,(BYTE *)pRead_buf,n);
        uncomp_remaining-=n;
        cur_archive_file_ofs+=n;
        }
      comp_size=uncomp_size;
      }
    else
      {
      int result=0;
      mz_zip_writer_add_state state;
      tdefl_compressor *pComp=(tdefl_compressor *)pZip->m_pAlloc(pZip->m_pAlloc_opaque,1,sizeof(tdefl_compressor));
      if(!pComp)
        {
        pZip->m_pFree(pZip->m_pAlloc_opaque,pRead_buf);
        fclose(pSrc_file);
        return 0;
        }
      state.m_pZip=pZip;
      state.m_cur_archive_file_ofs=cur_archive_file_ofs;
      state.m_comp_size=0;
      if(tdefl_init(pComp,mz_zip_writer_add_put_buf_callback,&state,tdefl_create_comp_flags_from_zip_params(level,-15,MZ_DEFAULT_STRATEGY))!=TDEFL_STATUS_OKAY)
        {
        pZip->m_pFree(pZip->m_pAlloc_opaque,pComp);
        pZip->m_pFree(pZip->m_pAlloc_opaque,pRead_buf);
        fclose(pSrc_file);
        return 0;
        }
      for( ; ; )
        {
        size_t in_buf_size=(DWORD)min(uncomp_remaining,MZ_ZIP_MAX_IO_BUF_SIZE);
        tdefl_status status;
        if(fread(pRead_buf,1,in_buf_size,pSrc_file)!=in_buf_size)
          break;
        uncomp_crc32=(DWORD)mz_crc32(uncomp_crc32,(BYTE *)pRead_buf,in_buf_size);
        uncomp_remaining-=in_buf_size;
        status=tdefl_compress_buffer(pComp,pRead_buf,in_buf_size,uncomp_remaining ? TDEFL_NO_FLUSH : TDEFL_FINISH);
        if(status == TDEFL_STATUS_DONE)
          {
          result=1;
          break;
          }
        else if(status!=TDEFL_STATUS_OKAY)
          break;
        }
      pZip->m_pFree(pZip->m_pAlloc_opaque,pComp);
      if(!result)
        {
        pZip->m_pFree(pZip->m_pAlloc_opaque,pRead_buf);
        fclose(pSrc_file);
        return 0;
        }
      comp_size=state.m_comp_size;
      cur_archive_file_ofs=state.m_cur_archive_file_ofs;
      method=8;
      }
    pZip->m_pFree(pZip->m_pAlloc_opaque,pRead_buf);
    }
  fclose(pSrc_file); pSrc_file=NULL;
  if((comp_size > 0xFFFFFFFF) || (cur_archive_file_ofs > 0xFFFFFFFF))
    return 0;
  if(!mz_zip_writer_create_local_dir_header(pZip,local_dir_header,(WORD)archive_name_size,0,uncomp_size,comp_size,uncomp_crc32,method,0,dos_time,dos_date))
    return 0;
  if(pZip->m_pWrite(pZip->m_pIO_opaque,local_dir_header_ofs,local_dir_header,sizeof(local_dir_header))!=sizeof(local_dir_header))
    return 0;
  if(!mz_zip_writer_add_to_central_dir(pZip,pArchive_name,(WORD)archive_name_size,NULL,0,pComment,comment_size,uncomp_size,comp_size,uncomp_crc32,method,0,dos_time,dos_date,local_dir_header_ofs,ext_attributes))
    return 0;
  pZip->m_total_files++;
  pZip->m_archive_size=cur_archive_file_ofs;
  return 1;
  }

int mz_zip_writer_add_mem_ex(mz_zip_archive *pZip,char*pArchive_name,void*pBuf,size_t buf_size,void*pComment,WORD comment_size,DWORD level_and_flags,QWORD uncomp_size,DWORD uncomp_crc32)
  {
  WORD method=0,dos_time=0,dos_date=0;
  DWORD level,ext_attributes=0,num_alignment_padding_bytes;
  QWORD local_dir_header_ofs=pZip->m_archive_size,cur_archive_file_ofs=pZip->m_archive_size,comp_size=0;
  size_t archive_name_size;
  BYTE local_dir_header[MZ_ZIP_LOCAL_DIR_HEADER_SIZE];
  tdefl_compressor *pComp=NULL;
  int store_data_uncompressed;
  mz_zip_internal_state *pState;
  if((int)level_and_flags < 0)
    level_and_flags=MZ_DEFAULT_LEVEL;
  level=level_and_flags & 0xF;
  store_data_uncompressed=((!level) || (level_and_flags & MZ_ZIP_FLAG_COMPRESSED_DATA));
  if((!pZip) || (!pZip->m_pState) || (pZip->m_zip_mode!=MZ_ZIP_MODE_WRITING) || ((buf_size) && (!pBuf)) || (!pArchive_name) || ((comment_size) && (!pComment)) || (pZip->m_total_files == 0xFFFF) || (level > MZ_UBER_COMPRESSION))
    return 0;
  pState=pZip->m_pState;
  if((!(level_and_flags & MZ_ZIP_FLAG_COMPRESSED_DATA)) && (uncomp_size))
    return 0;
  if((buf_size > 0xFFFFFFFF) || (uncomp_size > 0xFFFFFFFF))
    return 0;
  if(!mz_zip_writer_validate_archive_name(pArchive_name))
    return 0;
  mz_zip_time_to_dos_time(this_file.time,&dos_time,&dos_date);
  archive_name_size=strlen(pArchive_name);
  if(archive_name_size > 0xFFFF)
    return 0;
  num_alignment_padding_bytes=mz_zip_writer_compute_padding_needed_for_file_alignment(pZip);
  if((pZip->m_total_files == 0xFFFF) || ((pZip->m_archive_size + num_alignment_padding_bytes + MZ_ZIP_LOCAL_DIR_HEADER_SIZE + MZ_ZIP_CENTRAL_DIR_HEADER_SIZE + comment_size + archive_name_size) > 0xFFFFFFFF))
    return 0;
  if((archive_name_size) && (pArchive_name[archive_name_size - 1] == '/'))
    {
    ext_attributes |= 0x10;
    if((buf_size) || (uncomp_size))
      return 0;
    }
  if((!mz_zip_array_ensure_room(pZip,&pState->m_central_dir,MZ_ZIP_CENTRAL_DIR_HEADER_SIZE + archive_name_size + comment_size)) || (!mz_zip_array_ensure_room(pZip,&pState->m_central_dir_offsets,1)))
    return 0;
  if((!store_data_uncompressed) && (buf_size))
    {
    if(NULL == (pComp=(tdefl_compressor *)pZip->m_pAlloc(pZip->m_pAlloc_opaque,1,sizeof(tdefl_compressor))))
      return 0;
    }
  if(!mz_zip_writer_write_zeros(pZip,cur_archive_file_ofs,num_alignment_padding_bytes + sizeof(local_dir_header)))
    {
    pZip->m_pFree(pZip->m_pAlloc_opaque,pComp);
    return 0;
    }
  local_dir_header_ofs+=num_alignment_padding_bytes;
  if(pZip->m_file_offset_alignment)   { (void)( (!!((local_dir_header_ofs & (pZip->m_file_offset_alignment - 1)) == 0)) || (assert("(local_dir_header_ofs & (pZip->m_file_offset_alignment - 1)) == 0"),0) );   }
  cur_archive_file_ofs+=num_alignment_padding_bytes + sizeof(local_dir_header);
  memset(&local_dir_header,0,sizeof(local_dir_header));
  if(pZip->m_pWrite(pZip->m_pIO_opaque,cur_archive_file_ofs,pArchive_name,archive_name_size)!=archive_name_size)
    {
    pZip->m_pFree(pZip->m_pAlloc_opaque,pComp);
    return 0;
    }
  cur_archive_file_ofs+=archive_name_size;
  if(!(level_and_flags & MZ_ZIP_FLAG_COMPRESSED_DATA))
    {
    uncomp_crc32=(DWORD)mz_crc32((0),(BYTE*)pBuf,buf_size);
    uncomp_size=buf_size;
    if(uncomp_size <= 3)
      {
      level=0;
      store_data_uncompressed=1;
      }
    }
  if(store_data_uncompressed)
    {
    if(pZip->m_pWrite(pZip->m_pIO_opaque,cur_archive_file_ofs,pBuf,buf_size)!=buf_size)
      {
      pZip->m_pFree(pZip->m_pAlloc_opaque,pComp);
      return 0;
      }
    cur_archive_file_ofs+=buf_size;
    comp_size=buf_size;
    if(level_and_flags & MZ_ZIP_FLAG_COMPRESSED_DATA)
      method=8;
    }
  else if(buf_size)
    {
    mz_zip_writer_add_state state;
    state.m_pZip=pZip;
    state.m_cur_archive_file_ofs=cur_archive_file_ofs;
    state.m_comp_size=0;
    if((tdefl_init(pComp,mz_zip_writer_add_put_buf_callback,&state,tdefl_create_comp_flags_from_zip_params(level,-15,MZ_DEFAULT_STRATEGY))!=TDEFL_STATUS_OKAY) ||
        (tdefl_compress_buffer(pComp,pBuf,buf_size,TDEFL_FINISH)!=TDEFL_STATUS_DONE))
      {
      pZip->m_pFree(pZip->m_pAlloc_opaque,pComp);
      return 0;
      }
    comp_size=state.m_comp_size;
    cur_archive_file_ofs=state.m_cur_archive_file_ofs;
    method=8;
    }
  pZip->m_pFree(pZip->m_pAlloc_opaque,pComp);
  pComp=NULL;
  if((comp_size > 0xFFFFFFFF) || (cur_archive_file_ofs > 0xFFFFFFFF))
    return 0;
  if(!mz_zip_writer_create_local_dir_header(pZip,local_dir_header,(WORD)archive_name_size,0,uncomp_size,comp_size,uncomp_crc32,method,0,dos_time,dos_date))
    return 0;
  if(pZip->m_pWrite(pZip->m_pIO_opaque,local_dir_header_ofs,local_dir_header,sizeof(local_dir_header))!=sizeof(local_dir_header))
    return 0;
  if(!mz_zip_writer_add_to_central_dir(pZip,pArchive_name,(WORD)archive_name_size,NULL,0,pComment,comment_size,uncomp_size,comp_size,uncomp_crc32,method,0,dos_time,dos_date,local_dir_header_ofs,ext_attributes))
    return 0;
  pZip->m_total_files++;
  pZip->m_archive_size=cur_archive_file_ofs;
  this_file.comp_size=(size_t)comp_size;
  this_file.uncomp_size=(size_t)uncomp_size;
  return 1;
  }

int mz_zip_writer_add_mem(mz_zip_archive *pZip,char*pArchive_name,void*pBuf,size_t buf_size,DWORD level_and_flags)
  {
  return mz_zip_writer_add_mem_ex(pZip,pArchive_name,pBuf,buf_size,NULL,0,level_and_flags,0,0);
  }

int mz_zip_writer_add_from_zip_reader(mz_zip_archive *pZip,mz_zip_archive *pSource_zip,DWORD file_index)
  {
  DWORD n,bit_flags,num_alignment_padding_bytes;
  QWORD comp_bytes_remaining,local_dir_header_ofs;
  QWORD cur_src_file_ofs,cur_dst_file_ofs;
  DWORD local_header_u32[(MZ_ZIP_LOCAL_DIR_HEADER_SIZE + sizeof(DWORD) - 1) / sizeof(DWORD)]; BYTE *pLocal_header=(BYTE *)local_header_u32;
  BYTE central_header[MZ_ZIP_CENTRAL_DIR_HEADER_SIZE];
  size_t orig_central_dir_size;
  mz_zip_internal_state *pState;
  void*pBuf; BYTE *pSrc_central_header;
  if((!pZip) || (!pZip->m_pState) || (pZip->m_zip_mode!=MZ_ZIP_MODE_WRITING))
    return 0;
  if(NULL == (pSrc_central_header=mz_zip_reader_get_cdh(pSource_zip,file_index)))
    return 0;
  pState=pZip->m_pState;
  num_alignment_padding_bytes=mz_zip_writer_compute_padding_needed_for_file_alignment(pZip);
  if((pZip->m_total_files == 0xFFFF) || ((pZip->m_archive_size + num_alignment_padding_bytes + MZ_ZIP_LOCAL_DIR_HEADER_SIZE + MZ_ZIP_CENTRAL_DIR_HEADER_SIZE) > 0xFFFFFFFF))
    return 0;
  cur_src_file_ofs=*((DWORD *)(pSrc_central_header + MZ_ZIP_CDH_LOCAL_HEADER_OFS));
  cur_dst_file_ofs=pZip->m_archive_size;
  if(pSource_zip->m_pRead(pSource_zip->m_pIO_opaque,cur_src_file_ofs,pLocal_header,MZ_ZIP_LOCAL_DIR_HEADER_SIZE)!=MZ_ZIP_LOCAL_DIR_HEADER_SIZE)
    return 0;
  if(*((DWORD *)(pLocal_header))!=MZ_ZIP_LOCAL_DIR_HEADER_SIG)
    return 0;
  cur_src_file_ofs+=MZ_ZIP_LOCAL_DIR_HEADER_SIZE;
  if(!mz_zip_writer_write_zeros(pZip,cur_dst_file_ofs,num_alignment_padding_bytes))
    return 0;
  cur_dst_file_ofs+=num_alignment_padding_bytes;
  local_dir_header_ofs=cur_dst_file_ofs;
  if(pZip->m_file_offset_alignment)   { (void)( (!!((local_dir_header_ofs & (pZip->m_file_offset_alignment - 1)) == 0)) || (assert("(local_dir_header_ofs & (pZip->m_file_offset_alignment - 1)) == 0"),0) );   }
  if(pZip->m_pWrite(pZip->m_pIO_opaque,cur_dst_file_ofs,pLocal_header,MZ_ZIP_LOCAL_DIR_HEADER_SIZE)!=MZ_ZIP_LOCAL_DIR_HEADER_SIZE)
    return 0;
  cur_dst_file_ofs+=MZ_ZIP_LOCAL_DIR_HEADER_SIZE;
  n=*((WORD *)(pLocal_header + MZ_ZIP_LDH_FILENAME_LEN_OFS)) + *((WORD *)(pLocal_header + MZ_ZIP_LDH_EXTRA_LEN_OFS));
  comp_bytes_remaining=n + *((DWORD *)(pSrc_central_header + MZ_ZIP_CDH_COMPRESSED_SIZE_OFS));
  if(NULL == (pBuf=pZip->m_pAlloc(pZip->m_pAlloc_opaque,1,(size_t)max(sizeof(DWORD) * 4,min(MZ_ZIP_MAX_IO_BUF_SIZE,comp_bytes_remaining)))))
    return 0;
  while(comp_bytes_remaining)
    {
    n=(DWORD)min(MZ_ZIP_MAX_IO_BUF_SIZE,comp_bytes_remaining);
    if(pSource_zip->m_pRead(pSource_zip->m_pIO_opaque,cur_src_file_ofs,pBuf,n)!=n)
      {
      pZip->m_pFree(pZip->m_pAlloc_opaque,pBuf);
      return 0;
      }
    cur_src_file_ofs+=n;
    if(pZip->m_pWrite(pZip->m_pIO_opaque,cur_dst_file_ofs,pBuf,n)!=n)
      {
      pZip->m_pFree(pZip->m_pAlloc_opaque,pBuf);
      return 0;
      }
    cur_dst_file_ofs+=n;
    comp_bytes_remaining-=n;
    }
  bit_flags=*((WORD *)(pLocal_header + MZ_ZIP_LDH_BIT_FLAG_OFS));
  if(bit_flags & 8)
    {
    if(pSource_zip->m_pRead(pSource_zip->m_pIO_opaque,cur_src_file_ofs,pBuf,sizeof(DWORD) * 4)!=sizeof(DWORD) * 4)
      {
      pZip->m_pFree(pZip->m_pAlloc_opaque,pBuf);
      return 0;
      }
    n=sizeof(DWORD) * ((*((DWORD *)(pBuf)) == 0x08074b50) ? 4 : 3);
    if(pZip->m_pWrite(pZip->m_pIO_opaque,cur_dst_file_ofs,pBuf,n)!=n)
      {
      pZip->m_pFree(pZip->m_pAlloc_opaque,pBuf);
      return 0;
      }
    cur_src_file_ofs+=n;
    cur_dst_file_ofs+=n;
    }
  pZip->m_pFree(pZip->m_pAlloc_opaque,pBuf);
  if(cur_dst_file_ofs > 0xFFFFFFFF)
    return 0;
  orig_central_dir_size=pState->m_central_dir.m_size;
  memcpy(central_header,pSrc_central_header,MZ_ZIP_CENTRAL_DIR_HEADER_SIZE);
  mz_write_le32((BYTE *)(central_header + MZ_ZIP_CDH_LOCAL_HEADER_OFS),(DWORD)(local_dir_header_ofs));
  if(!mz_zip_array_push_back(pZip,&pState->m_central_dir,central_header,MZ_ZIP_CENTRAL_DIR_HEADER_SIZE))
    return 0;
  n=*((WORD *)(pSrc_central_header + MZ_ZIP_CDH_FILENAME_LEN_OFS)) + *((WORD *)(pSrc_central_header + MZ_ZIP_CDH_EXTRA_LEN_OFS)) + *((WORD *)(pSrc_central_header + MZ_ZIP_CDH_COMMENT_LEN_OFS));
  if(!mz_zip_array_push_back(pZip,&pState->m_central_dir,pSrc_central_header + MZ_ZIP_CENTRAL_DIR_HEADER_SIZE,n))
    {
    mz_zip_array_resize(pZip,&pState->m_central_dir,orig_central_dir_size,0);
    return 0;
    }
  if(pState->m_central_dir.m_size > 0xFFFFFFFF)
    return 0;
  n=(DWORD)orig_central_dir_size;
  if(!mz_zip_array_push_back(pZip,&pState->m_central_dir_offsets,&n,1))
    {
    mz_zip_array_resize(pZip,&pState->m_central_dir,orig_central_dir_size,0);
    return 0;
    }
  pZip->m_total_files++;
  pZip->m_archive_size=cur_dst_file_ofs;
  return 1;
  }

int mz_zip_writer_finalize_archive(mz_zip_archive *pZip)
  {
  mz_zip_internal_state *pState;
  QWORD central_dir_ofs,central_dir_size;
  BYTE hdr[MZ_ZIP_END_OF_CENTRAL_DIR_HEADER_SIZE];
  if((!pZip) || (!pZip->m_pState) || (pZip->m_zip_mode!=MZ_ZIP_MODE_WRITING))
    return 0;
  pState=pZip->m_pState;
  if((pZip->m_total_files > 0xFFFF) || ((pZip->m_archive_size + pState->m_central_dir.m_size + MZ_ZIP_END_OF_CENTRAL_DIR_HEADER_SIZE) > 0xFFFFFFFF))
    return 0;
  central_dir_ofs=0;
  central_dir_size=0;
  if(pZip->m_total_files)
    {
    central_dir_ofs=pZip->m_archive_size;
    central_dir_size=pState->m_central_dir.m_size;
    pZip->m_central_directory_file_ofs=central_dir_ofs;
    if(pZip->m_pWrite(pZip->m_pIO_opaque,central_dir_ofs,pState->m_central_dir.m_p,(size_t)central_dir_size)!=central_dir_size)
      return 0;
    pZip->m_archive_size+=central_dir_size;
    }
  memset(&hdr,0,sizeof(hdr));
  mz_write_le32((BYTE *)(hdr + MZ_ZIP_ECDH_SIG_OFS),(DWORD)(MZ_ZIP_END_OF_CENTRAL_DIR_HEADER_SIG));
  mz_write_le16((BYTE *)(hdr + MZ_ZIP_ECDH_CDIR_NUM_ENTRIES_ON_DISK_OFS),(WORD)(pZip->m_total_files));
  mz_write_le16((BYTE *)(hdr + MZ_ZIP_ECDH_CDIR_TOTAL_ENTRIES_OFS),(WORD)(pZip->m_total_files));
  mz_write_le32((BYTE *)(hdr + MZ_ZIP_ECDH_CDIR_SIZE_OFS),(DWORD)(central_dir_size));
  mz_write_le32((BYTE *)(hdr + MZ_ZIP_ECDH_CDIR_OFS_OFS),(DWORD)(central_dir_ofs));
  if(pZip->m_pWrite(pZip->m_pIO_opaque,pZip->m_archive_size,hdr,sizeof(hdr))!=sizeof(hdr))
    return 0;
  if((pState->m_pFile) && (fflush(pState->m_pFile) == (-1)))
    return 0;
  pZip->m_archive_size+=sizeof(hdr);
  pZip->m_zip_mode=MZ_ZIP_MODE_WRITING_HAS_BEEN_FINALIZED;
  return 1;
  }

int mz_zip_writer_finalize_heap_archive(mz_zip_archive *pZip,void**pBuf,size_t *pSize)
  {
  if((!pZip) || (!pZip->m_pState) || (!pBuf) || (!pSize))
    return 0;
  if(pZip->m_pWrite!=mz_zip_heap_write_func)
    return 0;
  if(!mz_zip_writer_finalize_archive(pZip))
    return 0;
  *pBuf=pZip->m_pState->m_pMem;
  *pSize=pZip->m_pState->m_mem_size;
  pZip->m_pState->m_pMem=NULL;
  pZip->m_pState->m_mem_size=pZip->m_pState->m_mem_capacity=0;
  return 1;
  }

int mz_zip_add_mem_to_archive_file_in_place(char*pZip_filename,char*pArchive_name,void*pBuf,size_t buf_size,void*pComment,WORD comment_size,DWORD level_and_flags)
  {
  int status,created_new_archive=0;
  mz_zip_archive zip_archive;
  struct _stat file_stat;
  memset(&zip_archive,0,sizeof(zip_archive));
  if((int)level_and_flags < 0)
     level_and_flags=MZ_DEFAULT_LEVEL;
  if((!pZip_filename) || (!pArchive_name) || ((buf_size) && (!pBuf)) || ((comment_size) && (!pComment)) || ((level_and_flags & 0xF) > MZ_UBER_COMPRESSION))
    return 0;
  if(!mz_zip_writer_validate_archive_name(pArchive_name))
    return 0;
  if(_stat(pZip_filename,&file_stat)!=0)
    {
    if(!mz_zip_writer_init_file(&zip_archive,pZip_filename,0))
      return 0;
    created_new_archive=1;
    }
  else
    {
    if(!mz_zip_reader_init_file(&zip_archive,pZip_filename,level_and_flags | MZ_ZIP_FLAG_DO_NOT_SORT_CENTRAL_DIRECTORY))
      return 0;
    if(!mz_zip_writer_init_from_reader(&zip_archive,pZip_filename))
      {
      mz_zip_reader_end(&zip_archive);
      return 0;
      }
    }
  status=mz_zip_writer_add_mem_ex(&zip_archive,pArchive_name,pBuf,buf_size,pComment,comment_size,level_and_flags,0,0);
  if(!mz_zip_writer_finalize_archive(&zip_archive))
    status=0;
  if(!mz_zip_writer_end(&zip_archive))
    status=0;
  if((!status) && (created_new_archive))
    {
    int ignoredStatus=remove(pZip_filename);
    (void)ignoredStatus;
    }
  return status;
  }

void*mz_zip_extract_archive_file_to_heap(char*pZip_filename,char*pArchive_name,size_t *pSize,DWORD flags)
  {
  int file_index;
  mz_zip_archive zip_archive;
  void*p=NULL;
  if(pSize)
    *pSize=0;
  if((!pZip_filename) || (!pArchive_name))
    return NULL;
  memset(&zip_archive,0,sizeof(zip_archive));
  if(!mz_zip_reader_init_file(&zip_archive,pZip_filename,flags | MZ_ZIP_FLAG_DO_NOT_SORT_CENTRAL_DIRECTORY))
    return NULL;
  if((file_index=mz_zip_reader_locate_file(&zip_archive,pArchive_name,NULL,flags)) >= 0)
    p=mz_zip_reader_extract_to_heap(&zip_archive,file_index,pSize,flags);
  mz_zip_reader_end(&zip_archive);
  return p;
  }

typedef struct _ZIP_FILE_DATA{
  char path[MAX_PATH];  /* path name */
  char name[MAX_PATH];  /* file name */
  int index;            /* original index in archive (before sort) */
  int depth;            /* folder depth of file */
  QWORD size;           /* size of file */
  FILETIME time;        /* file date:time stamp */
  }ZIP_FILE_DATA;

typedef struct _ZIP_FIND_DATA{
  char path[MAX_PATH];  /* path match string */
  char match[MAX_PATH]; /* file match string */
  int files;            /* number of files in archive (not including folders) */
  int index;            /* index of pending comparison */
  int depth;            /* folder depth of match string */
  ZIP_FILE_DATA*zFile;  /* pointer to list of files */
  mz_zip_archive*zip;   /* pointer to archive */
  }ZIP_FIND_DATA;

int CompareZipFileNames(const void**v1,const void**v2)
  {
  int cmp;
  ZIP_FILE_DATA*z1,*z2;
  z1=(ZIP_FILE_DATA*)v1;
  z2=(ZIP_FILE_DATA*)v2;
  cmp=_stricmp(z1->path,z2->path);
  if(cmp)
    return(cmp);
  return(_stricmp(z1->name,z2->name));
  }

FILETIME ZipTimeToFileTime(time_t zt)
  {
  struct{WORD date,time;}dos;
  static FILETIME ft;
  mz_zip_time_to_dos_time(zt,&dos.time,&dos.date);
  DosDateTimeToFileTime(dos.date,dos.time,&ft);
  return(ft);
  }

char _toupper_(char c)
  {
  if(c<'a')
    return(c);
  if(c>'z')
    return(c);
  return(c-' ');
  }

int _is_match_(char*fname,char*wild)
  {
  char*cp,*mp;
  if(fname==NULL||wild==NULL)
    return(0);
  if(strcmp(wild,"*")==0)
    return(1);
  while((*fname)&&(*wild!='*'))
    {
    if((_toupper_(*wild)!=_toupper_(*fname))&&(*wild!='?'))
      return(0);
    wild++;
    fname++;
    }
  while(*fname)
    {
    if(*wild=='*')
      {
      if(!*++wild)
        return(1);
      mp=wild;
      cp=fname+1;
      }
    else if((_toupper_(*wild)==_toupper_(*fname))||(*wild=='?'))
      {
      wild++;
      fname++;
      }
    else
      {
      wild=mp;
      fname=cp++;
      }
    }
  while(*wild=='*')
    wild++;
  return(!*wild);
  }

BOOL IsFileMatch(char*fname,char*wild)
  {
  char*ptr;
  char fn[MAX_PATH],fe[MAX_PATH],wn[MAX_PATH],we[MAX_PATH];
  if(fname==NULL)
    return(FALSE);
  if(strlen(fname)==0)
    return(FALSE);
  if(wild==NULL)
    return(FALSE);
  if(strlen(wild)==0)
    return(FALSE);
  strcpy(fn,fname);
  if((ptr=strchr(fn,'.'))!=NULL)
    {
    strcpy(fe,ptr+1);
    *ptr=0;
    }
  else
    fe[0]=0;
  strcpy(wn,wild);
  if((ptr=strchr(wn,'.'))!=NULL)
    {
    strcpy(we,ptr+1);
    *ptr=0;
    }
  else
    we[0]=0;
  if(!_is_match_(fn,wn))
    return(FALSE);
  if(fe[0]==0)
    {
    if(we[0]==0)
      return(TRUE);
    if(strcmp(we,"*")==0)
      return(TRUE);
    }
  if(we[0]==0)
    return(FALSE);
  if(_is_match_(fe,we))
    return(TRUE);
  return(FALSE);
  }

HANDLE FindFirstZip(mz_zip_archive*zip,char*lpFileName,WIN32_FIND_DATAA*lpFindData)
  {
  char*ptr;
  int i,m,n;
  ZIP_FIND_DATA*zFind;
  mz_zip_archive_file_stat stat;

/* clear find structure */

  memset(lpFindData,0,sizeof(WIN32_FIND_DATAA));

/* check for bad zip file pointer */

  if(zip==NULL)
    return(INVALID_HANDLE_VALUE);

/* get number of files and folders in archive */

  m=(int)mz_zip_reader_get_num_files(zip);
  if(m<1)
    return(INVALID_HANDLE_VALUE);

/* count the number of files excluding folders */

  for(n=i=0;i<m;i++)
    if(!mz_zip_reader_is_file_a_directory(zip,i))
      n++;
  if(n<1)
    return(INVALID_HANDLE_VALUE);

/* allocate memory for find structure */

  if((zFind=calloc(1,sizeof(ZIP_FIND_DATA)))==NULL)
    return(INVALID_HANDLE_VALUE);

/* allocate memory for list of files */

  zFind->files=n;
  if((zFind->zFile=calloc(n,sizeof(ZIP_FILE_DATA)))==NULL)
    {
    free(zFind);
    return(INVALID_HANDLE_VALUE);
    }

/* get file data into list */

  for(n=i=0;i<m;i++)
    {
    if(mz_zip_reader_is_file_a_directory(zip,i))
      continue;
    if(!mz_zip_reader_file_stat(zip,i,&stat))
      {
      free(zFind->zFile);
      free(zFind);
      return(INVALID_HANDLE_VALUE);
      }
    strcpy(zFind->zFile[n].name,stat.m_filename);
    for(ptr=zFind->zFile[n].name;*ptr;ptr++)
      {
      if(*ptr=='/')
        {
        *ptr='\\';
        zFind->zFile[n].depth+=1;
        }
      }
    zFind->zFile[n].size=stat.m_uncomp_size;
    zFind->zFile[n].time=ZipTimeToFileTime(stat.m_time);
    zFind->zFile[n].index=i;
    n++;
    }

/* split names into path and file */

  for(i=0;i<n;i++)
    {
    if(strchr(zFind->zFile[i].name,'\\'))
      {
      strcpy(zFind->zFile[i].path,zFind->zFile[i].name);
      ptr=strrchr(zFind->zFile[i].path,'\\');
      *ptr=0;
      ptr=strrchr(zFind->zFile[i].name,'\\');
      strcpy(zFind->zFile[i].name,ptr+1);
      }
    }

/* sort list by names */

  qsort((void*)zFind->zFile,n,sizeof(ZIP_FILE_DATA),CompareZipFileNames);

/* save match and folder depth */

  strcpy(zFind->match,lpFileName);
  for(ptr=zFind->match;*ptr;ptr++)
    {
    if(*ptr=='/')
      {
      *ptr='\\';
      zFind->depth+=1;
      }
    else if(*ptr=='\\')
      zFind->depth+=1;
    }
  if(zFind->depth)
    {
    strcpy(zFind->path,zFind->match);
    ptr=strrchr(zFind->path,'\\');
    *ptr=0;
    ptr=strrchr(zFind->match,'\\');
    strcpy(zFind->match,ptr+1);
    }

/* return matching file if applicable */

  for(i=0;i<n;i++)
    {
    if(zFind->zFile[i].depth==zFind->depth)
      {
      if(_stricmp(zFind->zFile[i].path,zFind->path)==0)
        {
        if(IsFileMatch(zFind->zFile[i].name,zFind->match))
          {
          strcpy(lpFindData->cFileName,zFind->zFile[i].name);
          lpFindData->nFileSizeHigh=(DWORD)(zFind->zFile[i].size>>32);
          lpFindData->nFileSizeLow =(DWORD) zFind->zFile[i].size     ;
          lpFindData->ftCreationTime  =
          lpFindData->ftLastAccessTime=
          lpFindData->ftLastWriteTime =zFind->zFile[i].time;
          lpFindData->dwFileAttributes=FILE_ATTRIBUTE_NORMAL;
          zFind->index=i+1;
          return((HANDLE)zFind);
          }
        }
      }
    else if(zFind->zFile[i].depth==zFind->depth+1)
      {
      if(zFind->depth==0)
        {
        if(IsFileMatch(zFind->zFile[i].path,zFind->match))
          {
          strcpy(lpFindData->cFileName,zFind->zFile[i].path);
          lpFindData->dwFileAttributes=FILE_ATTRIBUTE_DIRECTORY;
          zFind->index=i+1;
          return((HANDLE)zFind);
          }
        }
      else
        {
        char name[MAX_PATH],path[MAX_PATH];
        strcpy(path,zFind->zFile[i].path);
        ptr=strrchr(path,'\\');
        *ptr=0;
        strcpy(name,zFind->zFile[i].path);
        ptr=strrchr(name,'\\');
        *ptr=0;
        ptr=strrchr(name,'\\');
        strcpy(name,ptr+1);
        if(_stricmp(path,zFind->path)==0&&IsFileMatch(name,zFind->match))
          {
          strcpy(lpFindData->cFileName,name);
          lpFindData->dwFileAttributes=FILE_ATTRIBUTE_DIRECTORY;
          zFind->index=i+1;
          return((HANDLE)zFind);
          }
        }
      }
    }

  return(INVALID_HANDLE_VALUE);
  }

BOOL FindNextZip(HANDLE hFind,WIN32_FIND_DATAA*lpFindData)
  {
  char*ptr;
  int i;
  ZIP_FIND_DATA*zFind;
  if(hFind==NULL||hFind==INVALID_HANDLE_VALUE)
    return(FALSE);
  zFind=(ZIP_FIND_DATA*)hFind;
  for(i=zFind->index;i<zFind->files;i++)
    {
    if(zFind->zFile[i].depth==zFind->depth)
      {
      if(_stricmp(zFind->zFile[i].path,zFind->path)==0)
        {
        if(IsFileMatch(zFind->zFile[i].name,zFind->match))
          {
          strcpy(lpFindData->cFileName,zFind->zFile[i].name);
          lpFindData->nFileSizeHigh=(DWORD)(zFind->zFile[i].size>>32);
          lpFindData->nFileSizeLow =(DWORD) zFind->zFile[i].size     ;
          lpFindData->ftCreationTime  =
          lpFindData->ftLastAccessTime=
          lpFindData->ftLastWriteTime =zFind->zFile[i].time;
          lpFindData->dwFileAttributes=FILE_ATTRIBUTE_NORMAL;
          zFind->index=i+1;
          return(TRUE);
          }
        }
      }
    else if(zFind->zFile[i].depth==zFind->depth+1)
      {
      if(_stricmp(zFind->zFile[i-1].path,zFind->zFile[i].path))
        {
        if(zFind->depth==0)
          {
          if(IsFileMatch(zFind->zFile[i].path,zFind->match))
            {
            strcpy(lpFindData->cFileName,zFind->zFile[i].path);
            lpFindData->dwFileAttributes=FILE_ATTRIBUTE_DIRECTORY;
            zFind->index=i+1;
            return(TRUE);
            }
          }
        else
          {
          char name[MAX_PATH],path[MAX_PATH];
          strcpy(path,zFind->zFile[i].path);
          ptr=strrchr(path,'\\');
          *ptr=0;
          ptr=strrchr(zFind->zFile[i].path,'\\');
          strcpy(name,ptr+1);
          if(_stricmp(path,zFind->path)==0&&IsFileMatch(name,zFind->match))
            {
            strcpy(lpFindData->cFileName,name);
            lpFindData->dwFileAttributes=FILE_ATTRIBUTE_DIRECTORY;
            zFind->index=i+1;
            return(TRUE);
            }
          }
        }
      }
    }
  return(FALSE);
  }

BOOL FindCloseZip(HANDLE hFind)
  {
  ZIP_FIND_DATA*zFind;
  if(hFind==NULL||hFind==INVALID_HANDLE_VALUE)
    return(FALSE);
  zFind=(ZIP_FIND_DATA*)hFind;
  free(zFind->zFile);
  free(zFind);
  return(TRUE);
  }

char*GetNameZip(HANDLE hFind,WIN32_FIND_DATAA*lpFindData)
  {
  static char fname[MAX_PATH];
  ZIP_FIND_DATA*zFind;
  zFind=(ZIP_FIND_DATA*)hFind;
  if(strlen(zFind->path))
    {
    strcpy(fname,zFind->path);
    strcat(fname,"\\");
    strcat(fname,lpFindData->cFileName);
    }
  else
    strcpy(fname,lpFindData->cFileName);
  return(fname);
  }

ZILE*zopen(mz_zip_archive*zip,char*name,char*access)
  {
  char c,d,*na,*nf,*ptr;
  int i,n;
  void*bufr;
  size_t uncomp_size;
  mz_zip_archive_file_stat stat;
  ZILE*zp;
  if(strcmp(access,"rb")&&strcmp(access,"rt"))
    return(NULL);
  n=(int)mz_zip_reader_get_num_files(zip);
  for(i=0;i<n;i++)
    {
    if(mz_zip_reader_is_file_a_directory(zip,i))
      continue;
    if(!mz_zip_reader_file_stat(zip,i,&stat))
      return(NULL);
    na=name;
    nf=stat.m_filename;
    if(*na==':')
      {
      na++;
      if((ptr=strrchr(nf,'\\'))!=NULL)
        nf=ptr+1;
      if((ptr=strrchr(nf,'/'))!=NULL)
        nf=ptr+1;
      }
    while((*na)||(*nf))
      {
      c=_toupper_(*na++);
      if(c=='\\')
        c='/';
      d=_toupper_(*nf++);
      if(c!=d)
        break;
      }
    if((*na)||(*nf))
      continue;
    if((bufr=mz_zip_reader_extract_file_to_heap(zip,stat.m_filename,&uncomp_size,0))==NULL)
      return(NULL);
    if(uncomp_size!=stat.m_uncomp_size)
      return(NULL);
    if((zp=calloc(1,sizeof(ZILE)))==NULL)
      return(NULL);
    strcpy(zp->name,name);
    zp->data.size=(DWORD)uncomp_size;
    zp->data.data=bufr;
    zp->time=stat.m_time;
    return(zp);
    }
  return(NULL);
  }

DWORD zsize(ZILE*zp)
  {
  return(zp->data.size);
  }

time_t ztime(ZILE*zp)
  {
  return(zp->time);
  }

void zrewind(ZILE*zp)
  {
  zp->position=0;
  }

int zseek(ZILE*zp,long offset,int origin)
  {
  if(origin==SEEK_SET)
    {
    if(offset<0)
      return(1);
    if((DWORD)offset<=zp->data.size)
      {
      zp->position=offset;
      return(0);
      }
    return(1);
    }
  if(origin==SEEK_CUR)
    {
    if(offset>0)
      {
      if(zp->position+(DWORD)offset<=zp->data.size)
        {
        zp->position=zp->position+(DWORD)offset;
        return(0);
        }
      return(1);
      }
    if(offset<0)
      {
      if((DWORD)labs(offset)<=zp->position)
        {
        zp->position-=(DWORD)labs(offset);
        return(0);
        }
      return(1);
      }
    return(0);
    }
  if(origin==SEEK_END)
    {
    if(offset>0)
      return(1);
    if(offset<0)
      {
      if((DWORD)labs(offset)<=zp->data.size)
        {
        zp->position=zp->data.size-(DWORD)labs(offset);
        return(0);
        }
      return(1);
      }
    zp->position=zp->data.size;
    return(0);
    }
  return(1);
  }

long ztell(ZILE*zp)
  {
  return((long)zp->position);
  }

size_t zread(void*bufr,size_t size,size_t count,ZILE*zp)
  {
  size_t l,n;
  l=size*count;
  if(zp->position+l>zp->data.size)
    n=zp->data.size-zp->position;
  else
    n=l;
  memcpy(bufr,zp->data.data+zp->position,n);
  zp->position+=(DWORD)n;
  return(n/size);
  }

size_t zwrite(void*bufr,size_t size,size_t count,ZILE*zp)
  {
  return(0);
  }

char*zgets(char*bufr,int n,ZILE*zp)
  {
  char c;
  int i,j;
  if(zp==NULL)
    return(NULL);
  bufr[0]=0;
  for(i=j=0;i<n-1;i++)
    {
    if(zread(&c,1,1,zp)!=1)
      return(NULL);
    if(c==13)
      continue;
    bufr[j++]=c;
    bufr[j]=0;
    if(c==10)
      return(bufr);
    }
  return(NULL);
  }

int zclose(ZILE*zp)
  {
  if(zp==NULL)
    return(1);
  free(zp->data.data);
  memset(zp,0,sizeof(ZILE));
  return(0);
  }
  